/*
 * -----------------------------------------------------------------
 * AperiodicParameters.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/** 
  This release parameter object characterizes a schedulable object that may
  become active at any time.

  When a reference to a {@link AperiodicParameters} object is given
  as a parameter to a constructor the {@link AperiodicParameters} object becomes bound
  to the object being created. Changes to the values in the {@link AperiodicParameters}
  object affect the constructed object. If given to more than one constructor
  then changes to the values in the {@link AperiodicParameters} object affect <em>all</em>
  of the associated objects. Note that this is a one-to-many relationship and
  <em>not</em> a many-to-many.

  <p><b>Caution:</b> This class is explicitly unsafe in multithreaded
  situations when it is being changed.  No synchronization is done.  It
  is assumed that users of this class who are mutating instances will be
  doing their own synchronization at a higher level.
  */

public class AperiodicParameters extends ReleaseParameters
{
    /** Create an {@link AperiodicParameters} object.
     *
     *  @param cost            Processing time per invocation.
     *                         On implementations which can measure the amount of time
     *                         a schedulable object is executed, this value is the maximum amount of time
     *                         a schedulable object receives.
     *                         On implementations which cannot measure execution time, this value is used
     *                         as a hint to the feasibility algorithm. On such systems it is not possible
     *                         to determine when any particular object exceeds cost.
     *                         Equivalent to <code>RelativeTime(0,0)</code> if null.
     *
     *  @param  deadline       The latest permissible completion time measured from the
     *                         release time of the associated invocation of the schedulable object.
     *                         Not used in feasibility analysis for minimum implementation.
     *                         If null, the deadline will be 
     *                         <code>RelativeTime(Long.MAX_VALUE,999999)</code>.
     *
     *  @param overrunHandler  This handler is invoked if an invocation of the schedulable
     *                         object exceeds cost.  Not required for minimum implementation.
     *                         If null, nothing happens on the overrun condition.
     *
     *  @param missHandler     This handler is invoked if the <code>run()</code> method of the schedulable
     *                         object is still executing after the deadline has passed.
     *                         Although minimum implementations do not consider deadlines in
     *                         feasibility calculations, they must recognize variable deadlines
     *                         and invoke the miss handler as appropriate.
     *                         If null, nothing happens on the miss deadline condition.
     */
    public AperiodicParameters(RelativeTime cost,
			       RelativeTime deadline,
			       AsyncEventHandler overrunHandler,
			       AsyncEventHandler missHandler)
    {
	this.cost = cost;
	this.deadline = deadline;
	this.overrunHandler = overrunHandler;
	this.missHandler = missHandler;
    }
}
