/*
 * -----------------------------------------------------------------
 * AsynchronouslyInterruptedException.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


package javax.realtime;

/** 
 An special exception that is thrown in response to an attempt to asynchronously
 transfer the locus of control of a {@link RealtimeThread}.
 
 <p>When a method is declared with <code>AsynchronouslyInterruptedException</code>
 in its <code>throws</code> clause the platform is expected to asynchronously throw this
 exception if <code>RealtimeThread.interrupt()</code> is called while the
 method is executing, or if such an interrupt is pending any
 time control returns to the method.  The interrupt is <em>not</em> thrown
 while any methods it invokes are executing, unless they are, in turn,
 declared to throw the exception.  This is intended to allow long-running
 computations to be terminated without the overhead or
 latency of polling with {@link Thread#interrupted}.
 
 <p>The <code>throws AsynchronouslyInterruptedException</code> clause
 is a marker on a stack frame which allows a method to be
 statically marked as asynchronously interruptible.  Only methods that are
 marked this way can be interrupted.  
 
 <p>When <code>Thread.interrupt()</code>, {@link RealtimeThread#interrupt}, 
 or <code>this.fire()</code> is called,
 the <code>AsynchronouslyInterruptedException</code> is compared against any currently
 pending <code>AsynchronouslyInterruptedException</code>
 on the thread.  If there is none, or if the depth of the <code>AsynchronouslyInterruptedException</code>
 is less than the currently pending <code>AsynchronouslyInterruptedException</code> -- i.e., it is targeted
 at a less deeply nested method call -- it becomes the currently
 pending interrupt.  Otherwise, it is discarded.
 
 <p>If the current method is interruptible, the exception is thrown on the thread.
 Otherwise, it just remains pending until control returns to an interruptible
 method, at which point the <code>AsynchronouslyInterruptedException</code> is thrown.  When an interrupt is
 caught, the caller should invoke the <code>happened()</code> method 
 on the <code>AsynchronouslyInterruptedException</code> in which it is 
 interested to see if it matches the pending <code>AsynchronouslyInterruptedException</code>.
 If so, the pending <code>AsynchronouslyInterruptedException</code> is cleared from the thread.  Otherwise,
 it will continue to propagate outward.
 
 <p><code>Thread.interrupt()</code> and <code>RealtimeThread.interrupt()</code>
 generate a system available generic <code>AsynchronouslyInterruptedException</code>
 which will always propagate outward through interruptible methods until the 
 generic <code>AsynchronouslyInterruptedException</code> is identified and stopped.
 Other sources (e.g., <code>this.fire()</code> and {@link Timed})
 will generate
 a specific instance of <code>AsynchronouslyInterruptedException</code> which 
 applications can identify and thus limit propogation.
 */

public class AsynchronouslyInterruptedException extends InterruptedException {

    protected RealtimeThread target;

    protected int priority;

    protected boolean enabled;
    protected boolean doInterruptibleInProcess;

    private static AsynchronouslyInterruptedException genericAie = null;
    private static Object lock = new Object();
    

  /**
    Return the system generic <code>AsynchronouslyInterruptedException</code>, which is
    generated when <code>RealtimeThread.interrupt()</code> is invoked.
   */

    public static AsynchronouslyInterruptedException getGeneric() {
	synchronized(lock){
		if(genericAie==null){
		    try{
			genericAie = 
				(AsynchronouslyInterruptedException)(ImmortalMemory.instance()).newInstance(Class.forName("javax.realtime.AsynchronouslyInterruptedException"));

			}
			catch(Exception e){
				e.printStackTrace();
				return null;
			}
		}
	}
	
 
	return genericAie;

    }

    /**
      Enable the throwing of this exception.  This is valid only within a
      call to doInterruptible().  Otherwise it returns false and does nothing.
      @return True if this is enabled otherwise returns false.
     */

    public synchronized boolean enable() {
	Thread currentThread = Thread.currentThread();
	if(doInterruptibleInProcess == false || target!=currentThread)
		return false;
	
	enabled = true;

	boolean pending = target.getPending();

	if(pending==true){
		target.setPendingFalse();
		postAIE(target);
	}

	return true;
    }
    /**
      Defer the throwing of this exception. If <code>interrupt()</code> is
      called when this exception is disabled, the exception is put in
      pending state.  The exception will be thrown if this exception is
      subsequently enabled. This is valid only within a
      call to <code>doInterruptible()</code>.  Otherwise it returns false and does nothing.
      @return True if this is disabled otherwise returns false.
     */

    public synchronized boolean disable() {
	Thread currentThread = Thread.currentThread();
	if(doInterruptibleInProcess == false || target!=currentThread)
		return false;
	enabled = false;
	return true;
    }
    /**
      Query the enabled status of this exception.
      @return True if this is enabled otherwise returns false.
     */

    public boolean isEnabled() {
	return enabled;
    }

    /**
      Create an instance of <code>AsynchronouslyInterruptedException</code>.
     */

    public AsynchronouslyInterruptedException () {
	priority = 0;
	doInterruptibleInProcess = false;
	enabled = false;
    }

    /**
      Make this exception the current exception if <code>doInterruptible()</code> has
      been invoked and not completed.
      @return True if this was fired. If there is no current invocation
      of <code>doInterruptible()</code>, then false is returned with no other effect.
      False is also returned if there is already a current <code>doInterruptible()</code> 
      or if <code>disable()</code> has been called.
     */
    public synchronized boolean fire() {
	if(enabled && target.isAlive()){
	    synchronized(target) {
		AsynchronouslyInterruptedException pendingAIE = getPendingAIE(target);
		if(pendingAIE!=null){
			if(pendingAIE.priority < priority){
			     return false;
			}
		}
		disable();
		postAIE(target);
		return true;
		
	    }

	}
	return false;
    }

    /**
      Execute the <code>run()</code> method of the given {@link Interruptible}.
      This method may be on the stack in exactly one {@link RealtimeThread}.
      An attempt to invoke this method in a thread while it is on the
      stack of another or the same thread will cause an
      immediate return with a value of false.
      @param code An instance of an {@link Interruptible} whose <code>run()</code> method will be
      called.
      @return True if the method call completed normally. Returns
      false if another call to <code>doInterruptible</code> has not completed.
     */

    public boolean doInterruptible(Interruptible logic) {
	if(logic==null)
	    return true;
	
	synchronized(this){
		if (doInterruptibleInProcess == true)
		    return false;
		markStack();
		doInterruptibleInProcess = true;
	}

	try {
	    markStack();
	    helper(logic);
	} catch(AsynchronouslyInterruptedException e) {
		try{
			logic.interruptAction(e);
		}
		finally{
			happened(true);
		}
	}
	finally{

		unmarkStack();
		doInterruptibleInProcess = false;

	}
	return true;
    }

    private void helper(Interruptible r)
	throws AsynchronouslyInterruptedException
    {
	enable();
	r.run(this);
	disable();
    }

    /**
      Used with an instance of this exception to see if the current exception
      is this exception. 
      @param propagate Propagate the exception if true and this exception
      is not the current one. If false, then the state of this is set to 
      nonpending (i.e., it will stop propagating).
      @return True if this is the current exception.  Returns false if
      this is not the current exception.
     */
    public boolean happened(boolean propagate) {
	RealtimeThread target = (RealtimeThread)Thread.currentThread();
	synchronized(target) {
	    AsynchronouslyInterruptedException pending = getPendingAIE(target);
	    if(pending==null){
		return false;
	    }
	    if(pending == this){
		makeNonPending(target);//target.currentInterrupt = null;//make current non pending.
		return true;
	    }
	    if(propagate){
		propagate();	//Propagate the pending exception in this thread.
	    }
	    else{
		makeNonPending(target);
 		//makes pening nonpending//
	    }
	    return false;
	}
    }

    boolean isDoInterruptibleInProcess(){
	return doInterruptibleInProcess;
    }

    private void markStack(){
	target = (RealtimeThread)(Thread.currentThread());	 
	priority = getStackDepth();
    }

    private void unmarkStack(){
	priority = 0;
    }

    /**
      Cause the pending exception to continue up the stack.
     */
    public static void propagate(){

	RealtimeThread rtt =(RealtimeThread)Thread.currentThread();
	AsynchronouslyInterruptedException pending =  getPendingAIE(rtt);

	if(pending!=null){
		try{
			throw pending;
		}
		catch(AsynchronouslyInterruptedException aie){

			return;
		}
	}
    }
   
    /*blindly sets aie as pending. Does not perform precidence checking.
    */
    private native boolean postAIE(RealtimeThread target);
    
    //Returns current threads stack depth//
    private native int getStackDepth();
    
    //This returns null if there is no current aie
    private static native AsynchronouslyInterruptedException getPendingAIE(RealtimeThread target);
    
    //sets pendingAI flag to zero and pendingAIE ICell to null.
    private native void makeNonPending(RealtimeThread target);

}

