/*
 * -----------------------------------------------------------------
 * BoundAsyncEventHandler.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
  A bound asynchronous event handler is an asynchronous event handler that
  is permanently bound to a thread.  Bound asynchronous event handlers are
  meant for use in situations where the added timeliness is worth the
  overhead of binding the handler to a thread.
  */
public abstract class BoundAsyncEventHandler extends AsyncEventHandler
{
  

  // here's the big difference, instead of (potentially) generating
  // a new thread evey time the event happens, we start the thread
  // from the get-go, and it blocks until there's something to do.
    /**
      Create a handler whose parameters are
      inherited from the current thread, if it is a {@link RealtimeThread},
      or null otherwise.
      */
  public BoundAsyncEventHandler()
  {
    // When the AsyncEventHandler's constructor is run, the HandlerThread
    // that gets created get the "this" from the BoundAsyncEventHandler 
    // and as such gets our "run" method.
    super();
    HandlerThread.setDaemon(true);
    HandlerThread.start();
  }

    /**
      Create a handler with the specified {@link ReleaseParameters} and
      {@link MemoryParameters}.
      @param scheduling A {@link SchedulingParameters} object which will be associated with the
                constructed instance of this. If null this will be assigned the reference to the
                {@link SchedulingParameters} of the current thread.
      @param release The {@link ReleaseParameters} object for this.  A value of null will
                construct this without a {@link ReleaseParameters} object.
      @param memory The {@link MemoryParameters} object for this.  A value of null will
                construct this without a {@link MemoryParameters} object.
      @param area The {@link MemoryArea} for this. Must be a reference to 
               a {@link ScopedMemory} or {@link ImmortalMemory} object if <code>noheap</code> is true.
      @param nonheap A flag meaning, when true, that this will have characteristics
                identical to a {@link NoHeapRealtimeThread}.
      @param group A {@link ProcessingGroupParameters} object to which this will be associated.  If
               null this will not be associated with any processing group.
      @param logic The {@link Runnable} object whose <code>run</code>
               is executed by handleAsyncEvent.      
      */
  public BoundAsyncEventHandler(SchedulingParameters scheduling,
				ReleaseParameters release,
				MemoryParameters memory,
				MemoryArea area,
				ProcessingGroupParameters group,
				boolean nonheap,
				Runnable logic)
  {
    super(scheduling, release, memory, area, group, nonheap, logic);
    //HandlerThread.setDaemon(false);
    HandlerThread.setDaemon(true);
    HandlerThread.start();
  }
  

  // I am concered about race conditions with this blockIfNotReady
  // system with the way we unblock the HandlerThread.

  protected synchronized void handle()
  {
    getAndIncrementPendingFireCount();
    notify();
  }

  // this is the code that run runs.
  void runHandler()
  {
    while(blockIfNotReady()){
      handleAsyncEvent();
      getAndDecrementPendingFireCount();
    }
  }   
  
  // Atomically (wrt to this class) block if fireCount = 0 
  private synchronized boolean blockIfNotReady()
  {
    boolean ready = false;
    // does this really need to be a while statement?
    while(!ready){
      synchronized(countLock){
	if(fireCount<1) ready=false;
	else ready=true;
      }
      if(!ready){
	try{
	  wait();
	}
	// if we're interrupted manually, we assume the user
	// knows what he/she's doing.
	catch(Exception e){}
      }
    }
    return(true);
  }

}
