/*
 * -----------------------------------------------------------------
 * MemoryArea.java : The abstract base class of all classes dealing
 *      the representations of allocatable memory areas.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

import java.lang.reflect.Constructor;

/**
 * <code>MemoryArea</code> is the abstract base class of all classes
 * dealing with the representations of allocatable memory areas, 
 * including the immortal memory area, physical memory and scoped memory
 * areas.
 *
 * @author Pratik Solanki
 */
public abstract class MemoryArea
{
	long memorySize;
    Runnable attachedRunnable = null;
	
	/* 
     * This is javax.realtime.Finalizer used to implement finalization
     * (not used anymore. Finalization is done in native code in VM).
	 */
//	Finalizer list = null;
	
    /**
     *  
     * @param sizeInBytes The size of <code>MemoryArea</code> to allocate,
     *            in bytes.
     */
    protected MemoryArea(long sizeInBytes)
    {
		memorySize = sizeInBytes;
    }

    
    /**
     *  
     * @param size A <code>SizeEstimator</code> object which indicates the
     * amount of memory required by this <code>MemoryArea</code>. 
     */
    protected MemoryArea(SizeEstimator size)
    {
        this(size.getEstimate());
    }
 
    /**
     *  
     * @param sizeInBytes The size of <code>MemoryArea</code> to allocate,
     *            in bytes.
     * @param logic The <code>run()</code> method of this object will be 
     * called whenever {@link #enter()} is called.
     */
    protected MemoryArea(long sizeInBytes, Runnable logic)
    {
		memorySize = sizeInBytes;
        attachedRunnable = logic;
    }

    
    /**
     *  
     * @param size A <code>SizeEstimator</code> object which indicates the
     * amount of memory required by this <code>MemoryArea</code>. 
     * @param logic The <code>run()</code> method of this object will be 
     * called whenever {@link #enter()} is called.
     */
    protected MemoryArea(SizeEstimator size, Runnable logic)
    {
        this(size.getEstimate(), logic);
    }

    /**
     * Associate this memory area to the current real-time thread for 
     * the duration of the execution of the <code>run()</code> method 
     * of the <code>java.lang.Runnable</code> passed at constructin time. 
     * During this bound period of execution, all objects are allocated
     * from the memory area until another one takes effect, or the  
     * <code>enter()</code> method is exited. A runtime exception is thrown
     * if this method is called from thread other than a 
     * {@link RealtimeThread} or {@link NoHeapRealtimeThread}.
     *
     * @throws IllegalArgumentException Thrown if no <code>Runnable</code>
     * was passed in the constructor.
     * @throws ScopedCycleException If entering this ScopedMemory would 
     * violate the single parent rule.
     */
    public void enter()
    {
        if(this.attachedRunnable != null)
            enter(attachedRunnable);
        else
            throw new IllegalArgumentException("MemoryArea did not have a Runnable attached");
    }
    
    /**
     * Associate this memory area to the current real-time thread for 
     * the duration of the execution of the <code>run()</code> method 
     * of the given <code>java.lang.Runnable</code>. During this bound 
     * period of execution, all objects are allocated from the memory 
     * area until another one takes effect, or the <code>enter()</code>
     * method is exited. A runtime exception is thrown if this method 
     * is called from thread other than a {@link RealtimeThread} or 
     * {@link NoHeapRealtimeThread}.
     *
     * @param logic The Runnable object whose <code>run()</code> method 
     *              should be invoked.
     * @throws ScopedCycleException If entering this ScopedMemory would 
     * violate the single parent rule.

     */
    public void enter(Runnable logic)
    {
        RealtimeThread currentThread = null;
        try
        {
            currentThread = RealtimeThread.currentRealtimeThread();
        }
        catch(ClassCastException e)
        {
            throw new IllegalStateException("MemoryArea.enter() should be called from a RealtimeThread");
        }
     
        synchronized(this)
        {
            currentThread.pushMemoryArea(this);
        }
        
        try
        {
            logic.run();
        }
        finally
        {
            synchronized(this)
            {
                currentThread.popMemoryArea();
            }
        }
     }

    /**
     * Returns the <code>MemoryArea</code> in which the given 
     * object is located.
     * 
     * @return The <code>MemoryArea</code> of the object.
     */
    public static MemoryArea getMemoryArea(Object object)
    {
		return getMemoryAreaForObject(object); 
    }

    /**
     * An exact count, in bytes, of the all of the memory currently 
     * used by the system for the allocated objects.
     *
     * @return The amount of memory consumed in bytes.
     */
    public long memoryConsumed()
    {
		return getMemoryConsumed0();
    }

    /**
     * An approximation to the total amount of memory currently 
     * available for future allocated objects, measured in bytes.
     *
     * @return The amount of remaining memory in bytes
     */
    public long memoryRemaining()
    {
		return getMemoryRemaining0();
    }


    /**
     * Allocate an array of T in this memory area.
     *
     * @param type   The class of the elements of the new array.
     * @param number The number of elements in the new array.
     * @return A new array of class type, of number elements.
     * @throws IllegalAccessException The class or initializer is inaccessible.
     * @throws InstantiationException The array cannot be instantiated.
     * @throws OutOfMemoryError       Space in the memory area is exhausted.
     */
    public synchronized Object newArray(Class type, int number)
		throws IllegalAccessException, InstantiationException
    {
        return newArray0(type, number);
    }
  
    /**
     * Allocate an object in this memory area.
     *
     * @param type The class of which to create a new instance.
     * @return A new instance of class <code>type</code>.
     * @throws IllegalAccessException The class or initializer is inaccessible.
     * @throws InstantiationException The specified class object could not be
     *         instantiated. Possible causes are: it is an interface, it is abstract,
     *         it is an array, or an exception was thrown by the constructor.
     * @throws OutOfMemoryError Space in the memory area is exhausted.
     */
    public synchronized Object newInstance(Class type)
		throws IllegalAccessException, InstantiationException
    {
        return newInstance0(type);
    }

    /**
     * Allocate an object in this memory area.
     *
     * @param type The class of which to create a new instance.
     *
     * @return A new instance of class <code>type</code>.
     *
     * @throws IllegalAccessException The class or initializer is inaccessible.
     * @throws InstantiationException The specified class object could not be
     *         instantiated. Possible causes are: it is an interface, it is abstract,
     *         it is an array, or an exception was thrown by the constructor.
     * @throws OutOfMemoryError Space in the memory area is exhausted.
     */
    public synchronized Object newInstance(Constructor c, Object[] args)
		throws IllegalAccessException, InstantiationException
    {
        return newInstance1(c, args);
    }



    /**
     * Query the size of the memory area. The returned value is the current size.
     * Current size may be larger than initial size for those areas that are allowed
     * to grow.
     *
     * @return The size of the memory area in bytes.
     */
    public long size()
    {
		return memorySize;
    }

    /**
     * Execute the run method from the <code>logic</code> parameter using this 
     * memory area as the current allocation context. If the memory area is a 
     * scoped memory type, this method behaves as if it had moved the
     * allocation context up the scope stackto the occurance of the memory area.
     * If the memory area is heap or immortal memory, this method behaves as if the 
     * run method were running in that memory type with an empty scope stack.
     * 
     * @param logic The runnable object whose <code>run()</code> method should 
     * be executed.
     *
     * @throws IllegalStateException A non-realtime thread attempted to enter the 
     * memory area. 
     * @throws InaccessibleAreaException The memory area is not in the thread's
     * scope stack.
     */
    public void executeInArea(Runnable logic) throws InaccessibleAreaException
    {
        executeInArea0(logic);
    }

    
	native void runAllFinalizers();
    native MemoryArea getCurrentMemoryArea();
    native void setCurrentMemoryArea(MemoryArea memory);
	
	
	void registerMemoryArea(long base, long size)
    {
      /* Highly dependent code. If the enum in native code changes, 
       * this has to change too!
       */
      if(this instanceof HeapMemory)
        registerMemoryArea0(base, size, 1);
      else if((this instanceof ImmortalMemory) || (this instanceof ImmortalPhysicalMemory))
        registerMemoryArea0(base, size, 2);
      else if(this instanceof ScopedMemory)
        registerMemoryArea0(base, size, 3);
      else
        System.out.println("Trying to register an unknown memory area " + this);
      return;
    }
    
	native void registerMemoryArea0(long base, long size, int memType);
	native static MemoryArea getMemoryAreaForObject(Object obj);
	private native Object newInstance0(Class type);
	private native Object newInstance1(Constructor c, Object[] args);
	private native Object newArray0(Class type, int number);
    private native void executeInArea0(Runnable logic);
	private native long getMemoryConsumed0();
	private native long getMemoryRemaining0();
	private native void freeMemory();
	
	protected void finalize() throws Throwable
	{
		super.finalize();
		runAllFinalizers();
		freeMemory();	
	}
}
