/*
 * -----------------------------------------------------------------
 * NoHeapRealtimeThread.java : This class is a specialized version of RealtimeThread.
 * It can preempt any impelemnted garbage collector and is not allowed to allocate or
 * reference any object allocated in the heap.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/** 
  A <code>NoHeapRealtimeThread</code> is a specialized form of {@link RealtimeThread}. 
  Because an instance of <code>NoHeapRealtimeThread</code> may immediately preempt any 
  implemented garbage collector logic contained in its <code>run()</code> 
  is never allowed to allocate or reference any object allocated in the heap nor
  it is even allowed to manipulate the references to objects in the heap.
  

  For example, if a and b are objects in immortal memory, b.p is reference to an object 
  on the heap, and a.p is type compatible with b.p, then
  a  <code>NoHeapRealtimeThread</code> is <em>not</em> allowed to execute anyting like the following:<p>
  <code>
   a.p = b.p; 
   b.p = null;
  </code>
  
  <p>Thus, it is always safe for a <code>NoHeapRealtimeThread</code> to interrupt the garbage
  collector at any time, without waiting for the end of the garbage collection
  cycle or a defined preemption point. Due to these restrictions,
  a <code>NoHeapRealtimeThread</code> object must be placed in a memory area such that
  thread logic may unexceptionally access instance variables and such that
  Java methods on {@link java.lang.Thread} (e.g., enumerate and join) complete normally
  except where execution would cause access violations.
  (Implementation hint: They could be allocated in {@link ImmortalMemory}.)
  The constructors of <code>NoHeapRealtimeThread</code> require a reference to {@link ScopedMemory} 
  or {@link ImmortalMemory}.
  
  When the thread is started, all
  execution occurs in the scope of the given memory area. 
  Thus, all memory allocation performed with the "new" operator is taken
  from this given area. 

 */

public class NoHeapRealtimeThread extends RealtimeThread
{
     /**
      Create a <code>NoHeapRealtimeThread</code>.
      @param scheduling A {@link SchedulingParameters} object that will 
                be associated with this. A null value means this will not
		have an associated {@link SchedulingParameters} object.
      @param area A {@link MemoryArea} object.  Must be a {@link ScopedMemory} or {@link ImmortalMemory} type.
                A null value causes an <code>IllegalArgumentException</code> to be thrown.
      */
   
    public NoHeapRealtimeThread(SchedulingParameters sp, MemoryArea ma) throws IllegalArgumentException
	{
        this(sp, null, ma);
	    /*super(sp, null, null, ma, null);
	    if ( ma == null || ((!(ma instanceof ScopedMemory)) && (!(ma instanceof ImmortalMemory)) 
		 && (!(ma instanceof ImmortalPhysicalMemory))) )
		throw new IllegalArgumentException("Illegal Memory Area Passed") ;*/

	}

    
    /**
      Create a <code>NoHeapRealtimeThread</code>.
      @param scheduling A {@link SchedulingParameters} object that will 
                be associated with this. A null value means this will not
                have an associated {@link SchedulingParameters} object.
      @param release A {@link ReleaseParameters} object that will be associated 
                with this. A null value means this will not
                have an associated {@link ReleaseParameters} object.
      @param area A {@link MemoryArea} object.  Must be a {@link ScopedMemory} or {@link ImmortalMemory} type.
                      A null value causes an <code>IllegalArgumentException</code> to be thrown.

      */
    public NoHeapRealtimeThread(SchedulingParameters sp, ReleaseParameters rp, MemoryArea ma) throws IllegalArgumentException
	{
        this(sp, rp, null, ma, null, null);
	    /*super(sp, rp, null, ma, null, null);

	    if (ma == null || ( (!(ma instanceof ScopedMemory)) && (!(ma instanceof ImmortalMemory)) 
		 && (!(ma instanceof ImmortalPhysicalMemory))) )
		throw new IllegalArgumentException("Illegal Memory Area Passed") ;*/

	}
     
     /** Create a <code>NoHeapRealtimeThread</code>.
      @param scheduling A {@link SchedulingParameters} object that will
                be associated with this. A null value means this will not
                have an associated {@link SchedulingParameters} object.
      @param release A {@link ReleaseParameters} object that will 
                be associated with this. A null value means this will not
                have an associated {@link ReleaseParameters} object.
      @param memory A {@link MemoryParameters} object that will be associated 
                with this. A null value means this will not
                have a {@link MemoryParameters} object.
      @param area A {@link MemoryArea} object.  Must be a {@link ScopedMemory} 
                or {@link ImmortalMemory} type.
                A null value causes an <code>IllegalArgumentException</code> to be thrown.
      @param group A {@link ProcessingGroupParameters} object
                that will be associated with this. A null value means
		this will not have an associated {@link ProcessingGroupParameters} object.
      @param logic A Runnable whose <code>run()</code> method will be executed for this.
     */
    public NoHeapRealtimeThread(SchedulingParameters sp, 
				ReleaseParameters rp,
				MemoryParameters mp,
				MemoryArea ma,
				ProcessingGroupParameters group,
				java.lang.Runnable logic) throws IllegalArgumentException
    {
	super(sp, rp, mp, ma, group, logic);

	if ( ma == null || ((!(ma instanceof ScopedMemory)) && (!(ma instanceof ImmortalMemory)) 
	     && (!(ma instanceof ImmortalPhysicalMemory))) )
	{
//	    System.out.println("NRT throwing Illegal Memory Area exception");
	    throw new IllegalArgumentException("Illegal Memory Area Passed") ;
	}
	
    }



//  	protected void finalize()
//  	{
//  	}

    /**
     * Checks if the NoHeapRealtimeThread is startable and starts it 
     * if it is. Checks that the parameters associated with this NHRT
     * object are not allocated in heap. Also checks if 
     * <code>this</code> object is allocated in heap. If any of them 
     * are allocated, <code>start()</code> throws a {@link MemoryAccessError}
     * 
     * @throws MemoryAccessError If any of the parameters or 
     *          <code>this</code> is allocated on heap.
     */
    public void start()
    {
        Object param = null;
        HeapMemory heap = HeapMemory.instance();
 
        if(heap.equals(MemoryArea.getMemoryArea(this)))
        {
            throw new MemoryAccessError("Tried starting an NHRT which was allocated in heap");
        }
    
        if((param = getMemoryParameters()) != null)
        {
            if(heap.equals(MemoryArea.getMemoryArea(param)))
            {
                throw new MemoryAccessError("Tried starting an NHRT with MemoryParameters allocated on heap");
            }
        }
 
        if((param = getProcessingGroupParameters()) != null)
        {
            if(heap.equals(MemoryArea.getMemoryArea(param)))
            {
                throw new MemoryAccessError("Tried starting an NHRT with ProcessingGroupParameters allocated on heap");
            }
        }
        
        if((param = getReleaseParameters()) != null)
        {
            if(heap.equals(MemoryArea.getMemoryArea(param)))
            {
                throw new MemoryAccessError("Tried starting an NHRT with ReleaseParameters allocated on heap");
            }
        }
        
        if((param = getSchedulingParameters()) != null)
        {
            if(heap.equals(MemoryArea.getMemoryArea(param)))
            {
                throw new MemoryAccessError("Tried starting an NHRT with SchedulingParameters allocated on heap");
            }
        }
        
        super.start();
    }

}
