/*
 * -----------------------------------------------------------------
 *  POSIXSignalHandler.java : This class is used to handle POSIX signals.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;
/** 
  Use instances of {@link AsyncEvent} to handle POSIX signals. Usage:
  <pre>POSIXSignalHandler.addHandler(SIGINT, intHandler);</pre>
  This class is required to be implemented only if the underlying
  operating system supports POSIX signals. 
  */

public final class POSIXSignalHandler
{
	/* All the signals defined in POSIX */
	
    /** Hangup. */
    public static final int SIGHUP = 1;
    /** Interrupt (rubout). */
    public static final int SIGINT = 2;
    /** Quit (ASCII FS). */
    public static final int SIGQUIT = 3;
    /** Illegal instruction (not reset when caught). */
    public static final int SIGILL = 4;
    /** Trace trap (not reset when caught). */
    public static final int SIGTRAP = 5;
    /** IOT instruction. */
    public static final int SIGIOT = 6;
    /** Used by abort, replace SIGIOT in the future. */
    public static final int SIGABRT = 6;
    /** EMT instruction. */
    public static final int SIGEMT = 7;
    /** Floating point exception. */
    public static final int SIGFPE = 8;
    /** Kill (cannot be caught or ignored). */
    public static final int SIGKILL = 9;
    /** Bus error. */
    public static final int SIGBUS = 10;
    /** Segmentation violation. */
    public static final int SIGSEGV = 11;
    /** Bad argument to system call. */
    public static final int SIGSYS = 12;
    /** Write on a pipe with no one to read it. */
    public static final int SIGPIPE = 13;
    /** Alarm clock. */
    public static final int SIGALRM = 14;
    /** Software termination signal from kill. */
    public static final int SIGTERM = 15;
    /** User defined signal	= 1. */
    public static final int SIGUSR1 = 16;
    /** User defined signal	= 2. */
    public static final int SIGUSR2 = 17;
    /** Child status change. */
    public static final int SIGCLD = 18;
    /** Child status change alias (POSIX). */
    public static final int SIGCHLD = 18;
    /** Power-fail restart. */
    public static final int SIGPWR = 19;
    /** Window size change. */
    public static final int SIGWINCH = 20;
    /** Urgent socket condition. */
    public static final int SIGURG = 21;
    /** Pollable event occured. */
    public static final int SIGPOLL = 22;
    /** Socket I/O possible (SIGPOLL alias). */
    public static final int SIGIO = SIGPOLL;
    /** Stop (cannot be caught or ignored). */
    public static final int SIGSTOP = 23;
    /** User stop requested from tty. */
    public static final int SIGTSTP = 24;
    /** Stopped process has been continued. */
    public static final int SIGCONT = 25;
    /** Background tty read attempted. */
    public static final int SIGTTIN = 26;
    /** Background tty write attempted. */
    public static final int SIGTTOU = 27;
    /** Virtual timer expired. */
    public static final int SIGVTALRM = 28;
    /** Profiling timer expired. */
    public static final int SIGPROF = 29;
    /** Exceeded cpu limit. */
    public static final int SIGXCPU = 30;
    /** Exceeded file size limit. */
    public static final int SIGXFSZ = 31;
    /** Process's lwps are blocked. */
    public static final int SIGWAITING = 32;
    /** Special signal used by thread library. */
    public static final int SIGLWP = 33;
    /** Special signal used by CPR. */
    public static final int SIGFREEZE = 34;
    /** Special signal used by CPR. */
    public static final int SIGTHAW = 35;
    /** Thread cancellation signal used by libthread. */
    public static final int SIGCANCEL = 36;
    /** Resource lost (e.g., record-lock lost). */
    public static final int SIGLOST = 37;

	private static final int total_signal = 37;

	private static AsyncEvent[] aeArr = new AsyncEvent[total_signal+1];
	static {
		for(int i=1;i<= total_signal ;i++)
			aeArr[i] = new AsyncEvent();
	}
    
    /**
      Add the given {@link AsyncEventHandler} to the list of handlers of the {@link AsyncEvent} of the given signal.
      @param signal One of the POSIX signals from this  (e.g., <code>this.SIGLOST</code>).
      @param handler An {@link AsyncEventHandler} which will be scheduled when the given signal occurs.
     */
	public static synchronized void addHandler(int signal, 
										AsyncEventHandler handler)
	{
		if(signal < 1 || signal > total_signal)
			return;
		aeArr[signal].addHandler(handler);
	}
	    
    /**
      Remove the given {@link AsyncEventHandler} to the list of handlers 
      of the {@link AsyncEvent} of the given signal.
      @param signal One of the POSIX signals from this (e.g., <code>this.SIGLOST</code>).
      @param handler An {@link AsyncEventHandler} which will be scheduled when the given signal occurs.
     */
	public static synchronized void removeHandler(int signal, 
										AsyncEventHandler handler)
	{
		if(signal < 1 || signal > total_signal)
			return;
		aeArr[signal].removeHandler(handler);
	}
    
    /**
      Set the given {@link AsyncEventHandler} as the handler of the {@link AsyncEvent} of the given signal.
      @param signal One of the POSIX signals from this  (e.g., <code>this.SIGLOST</code>).
      @param handler An {@link AsyncEventHandler} which will be scheduled when the given signal occurs.
             If h is null then no handler will be associated with this 
	     (i.e., remove all handlers).
     */
	public static synchronized void setHandler(int signal, 
										AsyncEventHandler handler)
	{
		if(signal < 1 || signal > total_signal)
			return;
		aeArr[signal].setHandler(handler);
	}
	
}
