/*
 * -----------------------------------------------------------------
 * PeriodicParameters.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;
import java.util.*;

/**
  This release parameter indicates that the {@link RealtimeThread#waitForNextPeriod} method on the
  associated {@link Schedulable} object will be unblocked at the start of each period.

  When a reference to a <code>PeriodicParameters</code> object is given
  as a parameter to a constructor the <code>PeriodicParameters</code> object becomes bound
  to the object being created. Changes to the values in the <code>PeriodicParameters</code>
  object affect the constructed object. If given to more than one constructor
  then changes to the values in the <code>PeriodicParameters</code> object affect <em>all</em>
  of the associated objects. Note that this is a one-to-many relationship and
  <em>not</em> a many-to-many.

  <p><b>Caution:</b> This class is explicitly unsafe in multithreaded
  situations when it is being changed.  No synchronization is done.  It
  is assumed that users of this class who are mutating instances will be
  doing their own synchronization at a higher level.
  */

public class PeriodicParameters extends ReleaseParameters
{

    protected HighResolutionTime start;
    protected RelativeTime period;

    /** Create a <code>PeriodicParameters</code> object.
     *
     *  @param start           Time at which the first period begins.
     *                         If a {@link RelativeTime}, this time is relative to
     *                         the first time the schedulable object becomes schedulable
     *                         <em>(schedulable time)</em>
     *                         (e.g., when <code>start()</code> is called on a thread).
     *                         If an {@link AbsoluteTime} and it is before the
     *                         schedulable time, start is equivalent to the schedulable time.
     *
     *  @param period          The period is the interval between successive
     *                         unblocks of {@link RealtimeThread#waitForNextPeriod}.
     *                         Must be greater than zero when entering feasibility analysis.
     *
     *  @param cost            Processing time per period.
     *                         On implementations which can measure the amount of time
     *                         a schedulable object is executed, this value is the maximum
     *                         amount of time a schedulable object receives per period.
     *                         On implementations which cannot measure execution time, this value is used
     *                         as a hint to the feasibility algorithm.  On such systems it is not possible
     *                         to determine when
     *                         any particular object exceeds or will exceed cost time units in a period.
     *                         Equivalent to <code>RelativeTime(0,0)</code> if null.
     *
     *  @param  deadline       The latest permissible completion time measured from the
     *                         release time of the associated invocation of the schedulable object.
     *                         For a minimum implementation for purposes of feasibility analysis,
     *                         the deadline is equal to the period.
     *                         Other implementations may use this parameter
     *                         to compute execution eligibility.
     *                         If null, deadline will equal the period.
     *
     *  @param overrunHandler  This handler is invoked if an invocation of the schedulable
     *                         object exceeds cost in the given period.
     *                         Not required for minimum implementation.
     *                         If null, nothing happens on the overrun condition.
     *
     *  @param missHandler     This handler is invoked if the <code>run()</code> method of the schedulable
     *                         object is still executing after the deadline has passed.
     *                         Although minimum implementations do not consider deadlines in
     *                         feasibility calculations, they must recognize variable deadlines
     *                         and invoke the miss handler as appropriate.
     *                         If null, nothing happens on the miss deadline condition.
     */
    public PeriodicParameters(HighResolutionTime start,
			      RelativeTime period,
			      RelativeTime cost,
			      RelativeTime deadline,
			      AsyncEventHandler overrunHandler,
			      AsyncEventHandler missHandler
			      )
    {
	if (start == null)
	    this.start = new AbsoluteTime(0,0);
	else
	    this.start = start;
	this.period = period;
	this.cost = cost;
	if (deadline == null)
	    this.deadline = period;
	else
	    this.deadline = deadline;

	this.overrunHandler = overrunHandler;
	this.missHandler = missHandler;
    }


    public boolean setIfFeasible(RelativeTime period, RelativeTime cost, RelativeTime deadline)
    {
	boolean retval = true;
	Schedulable s;
	RelativeTime oldCost, oldDeadline, oldPeriod;

	oldPeriod = getPeriod();
	oldCost = getCost();
	oldDeadline = getDeadline();

	setPeriod(period);
	setCost(cost);
	setDeadline(deadline);

	// get globalLock to lock all schedulers
	
	synchronized(Scheduler.class)
	    {
		for (int i=0; i< schedSet.size() ; i++)
		    {
			s = (Schedulable) schedSet.elementAt(i);
			if (!s.setReleaseParametersIfFeasible(this))
			    {
				// undo the changes

				setPeriod(oldPeriod);
				setCost(oldCost);
				setDeadline(oldDeadline);
				for (int j = i -1; i> 0 ; i--)
				    {
					s = (Schedulable) schedSet.elementAt(i);
					s.setReleaseParametersIfFeasible(this);
				    }
				retval = false;
				break;
			    }
		    }
	    }
	return retval;
    }



  
  /** 
    Get the start time.
    */
    public HighResolutionTime getStart()
    {
	return start;
    }
    
    /** 
      Set the start time.
     @param start Time at which the first period begins.
     */
    public void setStart(HighResolutionTime s)
    {
	if (s == null)
	    start = new AbsoluteTime(0,0);
	else
	    start = s;
    }
    
    /** Get the period.
     */
    public RelativeTime getPeriod()
    {
	return period;
    }
    
    /** 
      Set the period.
      @param period The period is the interval between successive 
                    unblocks of {@link RealtimeThread#waitForNextPeriod}.
		    Also used in the feasibility analysis and admission control algorithms.
     */
    public void setPeriod(RelativeTime p)
    {
	period = p;
    }
}
