package javax.realtime;

import java.util.*;

/**
  The <code>PhysicalMemoryManager</code> is available for use by the various physical
  memory accessor objects ({@link VTPhysicalMemory}, {@link LTPhysicalMemory}, 
    {@link ImmortalPhysicalMemory}, {@link RawMemoryAccess}, and 
    {@link RawMemoryFloatAccess})
    to create objects of the correct type that
  are bound to areas of physical memory with the appropriate
  characteristics -- or with appropriate accessor behavior.  Examples of
  characteristics that might be specified are: DMA memory, accessors
  with byte swapping, etc.
  <p>  
  The base implementation will provide a <code>PhysicalMemoryManager</code> and
  a set of 
  {@link PhysicalMemoryTypeFilter} classes that correctly identify memory classes that
  are standard for the (OS, JVM, and processor) platform.
  <p>
  OEMs may provide {@link PhysicalMemoryTypeFilter} classes
  that allow additional characteristics of memory devices
  to be specified.
  <p>
  Memory attributes that are configured may not be compatible with one another.
  For instance, copy-back cache enable may be incompatible with execute-only.  In this
  case, the implementation of memory filters may detect conflicts and throw a 
  {@link MemoryTypeConflictException}, but since filters are not part of the normative
  RTSJ, this exception is at best advisory.
*/

public final class PhysicalMemoryManager 
{
    // This field will be initialized by native method
    // NOTE: If you change this field in any way, you should make
    // corresponding changes in PhysicalMemoryManager.c
    private static long pagesize;
    
    //  Utility class for internal use only
    private static class Filter 
    {
        Object typeName;
        PhysicalMemoryTypeFilter filter;
        byte[] bitMap;
    };
   
    /** 
     *  Preserve the memory type filters that have been registered with the this 
     *  manager. 
     */
    static private Filter filterArray[];
 
    // Utility class for internal use only
    private static class AddressList
    {
        long base;
        long size;
        long virtualAddress;
        AddressList next;
    };
    
    /**
     * A list of all base/size pairs that are in use
     */
    private static AddressList addrInUse = null;
    private static Object addrLock = new Object();


    //
    // Memory types:
    // what other types are there? are they system-dependent?
    // are there any common ones should be specified here?
    //
    /**
      Specify this to identify DMA memory.
     */
    public static final String DMA = "dma";
    /**
      Specify this to identify shared memory.
     */
    public static final String SHARED = "shared";
    /**
      Specify this to identify aligned memory.
     */
    public static final String ALIGNED = "aligned";
    /**
      Specify this if byte swapping should be used.
     */
    public static final String BYTESWAP = "byteswap";


    // Highly Linux dependent values! 
    // To be used as attributes for mmap
    static final int PROT_READ  = 1;     // Page can be read.  
    static final int PROT_WRITE = 2;     // Page can be written.  
    static final int PROT_EXEC  = 4;     // Page can be executed.  
    static final int PROT_NONE  = 0;     // Page can not be accessed.  

    static final int MAP_SHARED = 1;        // Share changes.  
    static final int MAP_PRIVATE= 2;        // Changes are private.  
    static final int MAP_FIXED  = 16;       // Interpret addr exactly.  
    
        
    //  Initialize filterArray to the default size, or the override size found
    //  in System.properties.
    //  Then install the default set of filters in filterArray.
    static 
    {
        final int MAXFILTERS = 10;
        //int maxFilters = Integer.getInteger("javax.realtime.MAXFILTERS", MAXFILTERS).intValue();
        int maxFilters = MAXFILTERS;
        //  filterArray must be immor
        filterArray = new Filter[maxFilters];
        init();
/*        for(int i=0; i<maxFilters; i++)
        {
            filterArray[i] = new Filter();
        }*/

        //  Install the default set of filters.

        //  The filters should be in immortal memory.
        /*PhysicalMemoryTypeFilter f = new DefaultPhysicalMemoryFilter();
        try 
        {
            registerFilter(DMA, f);
            registerFilter(SHARED, f);
            registerFilter(ALIGNED, f);
            registerFilter(BYTESWAP, f);
        } 
        catch (DuplicateFilterException e)
        {
        }*/
    }
    
    /** Inaccessible constructor to prevent ordinary construction of memory
     *  factories.
     */
    private PhysicalMemoryManager()
    {
    }

    /** 
     * If the memory at the specified physical address and size is of the specified 
     * type, unused, and accessible map it to a virtual address (if necessary)
     * and return that address.
     *
     * @param memoryType  An object or array of objects that expresses the desired
     * type of memory.
     * @param base The starting address in physical memory to which the caller 
     * wants access
     * @param size The size of the memory area to which the caller wants access
     *
     * @return The virtual address that was assigned to the requested memory.
     *
     * @throws MemoryTypeConflictException The specified memory cannot be the 
     * requested type
     * @throws MemoryInUseException The specified memory is already in use
     */
    static synchronized long create(Object memoryType, long base, long size) 
                        throws MemoryTypeConflictException, MemoryInUseException
    {
                            
        //  Check with the filters for a match between the specified base and type.
        //  Map the memory..checking with the filters for mapping rules.
        //  Be sure to check with the security subsystem.
        //  Return the memory's virtual address.

        RealtimeSystem.getSecurityManager().checkAccessPhysicalRange(base, size);
        
        //boolean isDummy = false;
        long virtualAddress = -1;
 
        // Make size a multiple of pagesize i.e 4096
        if((size % 4096) != 0)
        {
            int factor = (int) size/4096;
            size = (factor+1)*4096;
        }
            
       
        // Check if the memory is already in use
        // This method will throw a MemoryInUseException
        addrCheckIfInUse(base, size);
        
        Vector filters = new Vector();
        // Check if memoryType is an array
        if(memoryType instanceof Object[])
        {
            int i=0;
            Object[] array = (Object[]) memoryType;
            //if(array instanceof DummyFilter[])
            //    isDummy = true;
            int arraySize = array.length;
            while(i<arraySize)
            {
                int index = findFilter(array[i]);
                if(index != -1)
                    filters.add(filterArray[index]);
                i++;
            }
        }
        else
        {
            //if(memoryType instanceof DummyFilter)
            //    isDummy = true;
            int index = findFilter(memoryType);
            if(index != -1)
                filters.add(filterArray[index]);
        }

        if(filters.size() == 0)
        {
            throw new MemoryTypeConflictException("Requested type of memory could not be found");
        }

        // Now go through the list of filters to see if the
        // base/size is acceptable to all
        // Also keep a note of the flags and protection bits
        int prot = 0;
        int flags = 0;
        
        for(int i=0; i<filters.size(); i++)
        {
            PhysicalMemoryTypeFilter f = ((Filter) filters.elementAt(i)).filter;

            long found = f.find(base, size);
            if(found != base)
                throw new MemoryTypeConflictException("Requested base/size pair cannot be the given type (Physical address failed)");
            
            prot = prot | f.getVMAttributes();
            flags = flags | f.getVMFlags();
            
            /*Object[] attr = f.getAttributes();
            if(attr instanceof Integer[])
            {
                Integer[] attributes = (Integer[]) attr;
                if(attributes.length != 2)
                    throw new RuntimeException("Filter " + f + " did not give the right number of attributes to Manager");
                
                prot = prot | attributes[0].intValue();
                flags = flags | attributes[1].intValue();
            }
            else
                throw new RuntimeException("Filter " + f + " did not give the right attributes to Manager");*/
        }

        //Now do the mmap() if this isn't a dummy filter
        //if(isDummy)
        //{
        //    virtualAddress = base;
        //}
        //else
        //{
            virtualAddress = mmap(base, size, prot, flags);
        //}

        // Now we need to check if the virtual address is okay
        for(int i=0; i<filters.size(); i++)
        {
            //PhysicalMemoryTypeFilter f = (PhysicalMemoryTypeFilter) filters.elementAt(i);
            PhysicalMemoryTypeFilter f = ((Filter) filters.elementAt(i)).filter;
            
            long found = f.vFind(virtualAddress, size);
            if(found != virtualAddress)
            {
                //if(!isDummy)
                    unmap(base, size);
                throw new MemoryTypeConflictException("Requested base/size pair cannot be the given type (Virtual address failed)");
            }
        }
        
        // Add to InUse list
        addrAddToList(base, size, virtualAddress);
        
        return virtualAddress;
    }

    /** 
     * If the memory of the specified type and size is: available, and accessible,
     * map it to a virtual address (if necessary) and return that address.
     *
     * @param memoryType  An object or array of objects that expresses the desired
     * type of memory.
     * @param size The size of the memory area to which the caller wants access
     *
     * @return The virtual address that was assigned to the requested memory.
     *
     * @throws OutOfMemoryError Thrown if the requested memory cannot be found.
     * @throws MemoryTypeConflictException The specified memory cannot be the 
     * requested type
     */
    static synchronized long create(Object memoryType, long size)
                        throws MemoryTypeConflictException 
    {
        //  Use the registered filters and a map of free memory that is maintained
        //  somewhere to find the right size extent of memory.
        //  The spec says nothing about how this should be found (ie., first fit,
        //  best fit, buddy...)
        //  Map the memory..checking with the filters for mapping rules.
        //  Be sure to check with the security subsystem.
        //  Return the memory's virtual address.
        RealtimeSystem.getSecurityManager().checkAccessPhysical();
        
        //boolean isDummy = false;
        long virtualAddress = -1;
        
        // Make size a multiple of pagesize i.e 4096
        if((size % 4096) != 0)
        {
            int factor = (int) size/4096;
            size = (factor+1)*4096;
        }
            
        Vector filters = new Vector();
        // Check if memoryType is an array
        if(memoryType instanceof Object[])
        {
            int i=0;
            Object[] array = (Object[]) memoryType;
            //if(array instanceof DummyFilter[])
            //    isDummy = true;
            int arraySize = array.length;
            while(i<arraySize)
            {
                int index = findFilter(array[i]);
                if(index != -1)
                    filters.add(filterArray[index]);
                i++;
            }
        }
        else
        {
            //if(memoryType instanceof DummyFilter)
            //    isDummy = true;
            int index = findFilter(memoryType);
            if(index != -1)
                filters.add(filterArray[index]);
        }

        // Now go through the list of filters to see if the
        // base/size is acceptable to all
        // Also keep a not of the flags and protection bits
        int prot = 0;
        int flags = 0;
        
        boolean acceptableToAll = false;
        int arraySize = filters.size();
        long tryAddr = -1;
        int position;
       
        //System.out.println("PMM:Going into loop. arraysize = " + arraySize);
        // This loop gives a chance to each filter to "suggest" a physical memory address
        // and then 
        for(position=0; position<arraySize; position++)
        {
            PhysicalMemoryTypeFilter f = ((Filter) filters.elementAt(position)).filter;
            
            // Find an address 
            tryAddr = findFreeSlot(f, size);
            //System.out.println("PMM: position = " + position + " findFreeSlot returned " + tryAddr);
            if(tryAddr == -1)
                continue;
           
            // Get the protection and flag attributes
            prot = f.getVMAttributes();
            flags = f.getVMFlags();
                
            // Now check if that address is acceptable to everyone else
            int traverser = (position+1)%arraySize;
            while(traverser != position)
            {
                PhysicalMemoryTypeFilter ft = (PhysicalMemoryTypeFilter) filters.elementAt(traverser);
                long addr2 = ft.find(tryAddr, size);
                if(addr2 != tryAddr)
                {
                    // break out of while loop. Someone didn't like our address.
                    break;
                }
                prot = prot | ft.getVMAttributes();
                flags = flags | ft.getVMFlags();
 
                traverser = (traverser+1)%arraySize;
            }
            if(traverser == position)
            {
                // We found it!
                acceptableToAll = true;
                break;
            }
        }
        if(!acceptableToAll)
        {
            throw new MemoryTypeConflictException("Requested memory cannot be the given type (Physical address failed)");
        }
        
        //Now do the mmap() if this isn't a dummy filter
        //if(isDummy)
        //{
        //    virtualAddress = tryAddr;
        //}
        //else
        //{
            virtualAddress = mmap(tryAddr, size, prot, flags);
        //}

        // Now we need to check if the virtual address is okay
        for(int i=0; i<filters.size(); i++)
        {
            //PhysicalMemoryTypeFilter f = (PhysicalMemoryTypeFilter) filters.elementAt(i);
            PhysicalMemoryTypeFilter f = ((Filter) filters.elementAt(position)).filter;
            
            long found = f.vFind(virtualAddress, size);
            if(found != virtualAddress)
            {
                //if(!isDummy)
                    unmap(tryAddr, size);
                throw new MemoryTypeConflictException("Requested memory cannot be the given type (Virtual address failed)");
            }
        }
       
        addrAddToList(tryAddr, size, virtualAddress);

        return virtualAddress;
    }

    /**
     * Get the physical address corresponding to a virtual address.
     * <p>
     * Note that the physical memory backing a range of virtual memory may not be
     * contiguous. This method only gives the physical address corresponding to a
     * single virtual address, not a range of memory.
     *
     * @param addr Physical address
     *
     * @return Virtual address
     * @throws IllegalArgumentException The virtual address does not refer to a mapped
     * address.
     */
     static long getPhysical(long addr)
     {
        return addr;
     }

    /** Free a range of the physical address space and remove the mapping into the
     *  application's virtual address space.
     *  This will only work for memory that has been allocated through the 
     *  <code>PhysicalMemoryManager</code>.
     *
     *  @param address the base of the memory to be unmapped. The implementation
     *  will use the same size it used to map the memory starting at that address.
     *
     *  @throws IllegalArgumentException No memory has been mapped starting at that virtual address
     */
    static void free(long virtualAddress, long size)
    {
       // addrRemoveFromListAndUnmap(address, size);
        synchronized(addrLock)
        {
            AddressList tr = addrInUse;
            AddressList next = addrInUse.next;
            if(tr.virtualAddress == virtualAddress)
            {
                addrInUse = tr.next;
                return;
            }

            while(next != null)
            {
                if(next.virtualAddress == virtualAddress)
                {
                    // Call munmap first
                    unmap(next.base, next.size);

                    //now remove from list
                    tr.next = next.next;
                    return;
                }
                tr = next;
                next = next.next;
            }
            
            // Since we reached here, we didn't find the base address. Throw RuntimeException
            
            throw new RuntimeException("Unable to remove base/size node from list");
        }
    }
    
    /** Register a memory type filter with the physical memory manager.
     *  
     *  @param name The type of memory handled by this filter
     *  @param filter The filter object
     *
     *  @throws DuplicateFilterException A filter for this type of memory
     *  already exists
     *  @throws RuntimeException The system is configured for a bounded
     *  number of filters.  This filter exceeds the bound.
     *  @throws IllegalArgumentException The name parameter must not be an array
     *      of objects.
     *  @throws IllegalArgumentException The name and filter must both be in
     *  immortal memory.
     */
    public static final void registerFilter(Object name, PhysicalMemoryTypeFilter filter)
            throws DuplicateFilterException, IllegalArgumentException 
    {
 
       RealtimeSystem.getSecurityManager().checkSetFilter();
        if(name instanceof Object [])
            throw new IllegalArgumentException("The name argument of addFilter must not be an array.");

        //  Add test for the memory areas of name and filter.
        synchronized(filterArray)
        {
            if(findFilter(name) >= 0)
                throw new DuplicateFilterException();
            
            int n = findEmpty();
            if(n >= filterArray.length || n < 0) 
                throw new RuntimeException("System is not configured for more than " + 
                    filterArray.length + " types of memory.");
            
            filterArray[n] = new Filter(); 
            filterArray[n].typeName = name;
            filterArray[n].filter = filter;
            //System.out.println("PMM:Added filter " + filter);
        }

    }
    
    /** Remove the identified filter from the set of registered filters.
     *  @param name The identifying object for this memory attribute.
     */
    public static final void removeFilter(Object name)
    {

//        checkSetFilter();
        //  If the caller passes an array of names, remove all the associated filters.
        if(name instanceof Object [])
        {
            Object [] nameArray = (Object [])name;
            for(int i = 0; i < nameArray.length; ++i)
            {
                try 
                {
                    removeFilter(nameArray[i]);
                } 
                catch (RuntimeException e)
                {
                    //  Silently ignore runtime exceptions if the name parameter is
                    //  an array.
                }
            }
        }
        else
        {
            synchronized(filterArray)
            {
                int n = findFilter(name);
                if(n >=0)
                {
                    //  Do we need to do anything before we lose it?
                    filterArray[n] = null;
                    //System.out.println("Removed filter");
                }
                else
                    throw new RuntimeException("Attempting to remove a filter that is not registered");
            }
        }

    }

    
    ////////////////////////////////////////////////////////////////////////////////
    //  Methods to handle/manage devices that can be removed while the system     //
    //  runs.                                                                     //
    ////////////////////////////////////////////////////////////////////////////////
        
    /** Is the specified range of memory removable? 
     * 
     *  @param base The starting address in physical memory
     *  @param size The size of the memory area
     *
     *  @return true if any part of the specified range can be removed
     */
    public static boolean isRemovable(long address, long size)
    {
        //  This will operate in conjunction with filter objects that 
        //  contain memory range properties.
        //  If any filter says that any part of the range is removable, then it is.
        return false;
    }

    /** Is any part of the specified range of memory presently removed?
     *  This method is used for devices that lie in the memory address space and can
     *  be removed while the system is running. (Such as PC cards)
     *  @param base The starting address in physical memory
     *  @param size The size of the memory area
     *  @return true if any part of the specified range is currently not usable
     */
    public static boolean isRemoved(long address, long size)
    {
        //  We do not want to store state at this level (it would have to use 
        //  immortal memory), so ask the OS/hardware for the state of the region.
        return false;
    }
    
    /** Register the specified AEH to run when any memory in the range is 
     *  removed from the system.
     *  If the specified range of physical memory contains multiple different
     *  types of removable memory, the aeh will be registered with any one of them.
     *
     *  If the size or the base is less than 0, remove all "remove" references to the aeh.
     *  @param base The starting address in physical memory
     *  @param size The size of the memory area
     *  @param aeh  Register this aeh.
     *  @throws IllegalArgumentException if the specified 
     *  range contains no removable memory.
     */     
    public static void onRemoval(long base, long size, AsyncEventHandler aeh)
    {
        /*
        int n = findFilterByAddress(base, size);
        */
    }

    /** Register the specified {@link AsyncEventHandler} to run when any memory in the range is 
     *  added to the system.
     *  If the specified range of physical memory contains multiple different
     *  types of removable memory, the AEH will be registered with any one of them.
     *
     *  If the size or the base is less than 0, unregister all "remove" references to the AEH.
     *  @param base The starting address in physical memory
     *  @param size The size of the memory area
     *  @param aeh  Register this AEH.
     *  @throws IllegalArgumentException If the specified 
     *  range contains no removable memory.
     */     
    public static void onInsertion(long base, long size, AsyncEventHandler aeh)
    {
    }
    
    //////////////////////////////
    //  Private support methods //
    //////////////////////////////
    
    /** Search for a filter for the specified memory type.
     *
     *  @param name The "name" of the memory type
     *  @return The index of the filter in the filter array or -1 if it
     *      is not found.
     */
    private static int findFilter(Object name)
    {
        //System.out.println("PMM: Searching for filter " + name);
        if(name == null) 
            return -1;
        synchronized(filterArray)
        {
            for(int i = 0; i < filterArray.length; ++i)
            {
                if(filterArray[i] != null && name.equals(filterArray[i].typeName))
                    return i;
            }
        }

        return -1;
    }
    
    /** Find an empty entry in the filter array.
     *  @returns The index of an empty entry in filterArray or -1 if all entries
     *      are full.
     */
    private static int findEmpty()
    {

        synchronized(filterArray)
        {
            for(int i = 0; i < filterArray.length; ++i)
            {
                if(filterArray[i] == null) 
                    return i;
            }
        }

        return -1;
     }
     
    /** Find any filter that has memory in the specified range.
     *  @param base The beginning of the physical address range 
     *  @param size The size of the range in bytes
     *  @returns The index of the filter, or -1 if none is found
     */
    private static int findFilterByAddress(long base, long size)
    {

        for(int i = 0; i < filterArray.length; ++i)
        {
            if(filterArray[i] != null)
            {
                if(filterArray[i].filter.contains(base, size))
                    return i;
            }
        }

        return -1;
    }

    /** Find any filter for removable memory that has memory in the specified range.
     *  @param base The beginning of the physical address range 
     *  @param size The size of the range in bytes
     *  @returns The index of the filter, or -1 if none is found
     */
    private static int findRemovableFilterByAddress(long base, long size)
    {

        for(int i = 0; i < filterArray.length; ++i)
        {
            if(filterArray[i] != null)
            {
                if(filterArray[i].filter.isRemovable() &&
                        filterArray[i].filter.contains(base, size))
                    return i;
            }
        }

        return -1;
    }

    
   /**
     * Check if the base/size pair has already been mmaped. If so throw a 
     * MemoryInUseException.
     * 
     * @throws MemoryInUseException If the base/size pair was already mmaped.
     */
    private static void addrCheckIfInUse(long base, long size) throws MemoryInUseException
    {
        AddressList traverser = addrInUse;
        while(traverser != null)
        {
            // Okay this might seem like a complicated if condition. Let me explain.
            //
            // The first part of the OR condition checks if the base/size pair
            // memory region lies completely below the base address of the current node.
            //
            // Second part of the condition checks if the base/size pair lies completely 
            // above current nodes maximum address (traverser.base+traverser.size)
            //
            // If either of these conditions is true, we're fine. The NOT at the 
            // beginning indicates that if both these conditions evaluated to false
            // then the base/size pair conflicts and we should raise an exception
            if(!(((base+size)  <= traverser.base   )  ||
                  (base        >= (traverser.base + traverser.size))))
            {
               
                //Oops. This conflicts. Thrown a MemoryInUseException
                throw new MemoryInUseException("Base/size pair conlicts with an already mapped memory region");
            }
            traverser = traverser.next;
        }

        // Nothing conflicted, return and continue.
        return;
    }

    /**
     * Adds the base/size pair to the InUse list. The list is unsorted.
     *
     */
    private static void addrAddToList(long base, long size, long virtualAddress)
    {
        synchronized(addrLock)
        {
            AddressList addr = new AddressList();
            addr.base = base;
            addr.size = size;
            addr.virtualAddress = virtualAddress;
            addr.next = addrInUse;
            addrInUse = addr;
        }
    }
    
    /**
     * Removes the base/size pair from the InUse list. The list is unsorted.
     *
     */
/*    private static void addrRemoveFromListAndUnmap(long virtualAddress, long size)
    {
        synchronized(addrLock)
        {
            AddressList tr = addrInUse;
            AddressList next = addrInUse.next;
            if(tr.virtualAddress == virtualAddress)
            {
                addrInUse = tr.next;
                return;
            }

            while(next != null)
            {
                if(next.virtualAddress == virtualAddress)
                {
                    // Call munmap first
                    unmap(next.base, next.size);

                    //now remove from list
                    tr.next = next.next;
                    return;
                }
                tr = next;
                next = next.next;
            }
            
            // Since we reached here, we didn't find the base address. Throw RuntimeException
            
            throw new RuntimeException("Unable to remove base/size node from list");
        }
    }*/

    private static long findFreeSlot(PhysicalMemoryTypeFilter f, long size)
    {
        long addr = f.find(0, size);
        if(addr == -1)
        {
            //System.out.println("findFreeSlot: returning -1 at the beginning");
            return -1;
        }
        do 
        {
            AddressList traverser = addrInUse;
            while(traverser != null)
            {
                // Okay this might seem like a complicated if condition. Let me explain.
                //
                // The first part of the OR condition checks if the base/size pair
                // memory region lies completely below the base address of the current node.
                //
                // Second part of the condition checks if the base/size pair lies completely 
                // above current nodes maximum address (traverser.base+traverser.size)
                //
                // If either of these conditions is true, we're fine. The NOT at the 
                // beginning indicates that if both these conditions evaluated to false
                // then the base/size pair conflicts and we should raise an exception
                if(!(( addr        < traverser.base &&
                      (addr+size)  < traverser.base   )  ||
                    (( addr        > (traverser.base + traverser.size) &&
                       (addr+size) > (traverser.base + traverser.size)))))
                {
                    // conflict. break out.
                    break;
                }
                traverser = traverser.next;
            }

            if(traverser != null)
            {
                // Conflict. I have to find a new one
                // Calculate the next base address after this base/size pair
                addr = f.find(traverser.base+traverser.size, size);
            }
        }while(addr == -1);

        //System.out.println("*** PMM:findFreeSlot() returning " + addr);        

        return addr;
       
    }

    private static long mmap(long base, long size, int prot, int flags)
    {
        return mmap0(base, size, prot, flags);
    }

    private static void unmap(long base, long size)
    {
        munmap0(base, size);
        return;
    }
    
    private static native void init();
    private static native long mmap0(long base, long size, int prot, int flags);
    private static native void munmap0(long base, long size);
}
