package javax.realtime;

public interface PhysicalMemoryTypeFilter 
{
    /** Search for memory of the right type.
     *   @param base start searching here
     *   @param size look for at least this much memory
     *   @return the address where memory was found or -1 if it was not found.
     */
    public long find(long base, long size);
    
    /** Search for virtual memory of the right type.
     *      This is important for systems where attributes are associated with 
     *      particular ranges of virtual memory.
     *   @param base start searching here
     *   @param size look for at least this much memory
     *   @return the address where memory was found or -1 if it was not found.
     */
    public long vFind(long base, long size);
    
    /** If configuration is required for memory to fit the attribute of this object,
     *  do the configuration here.
     *  @param base address of the beginning of the physical memory region.
     *  @param vBase address of the beginning of the virtual memory region.
     *  @param size size of the memory region
     *  @exception IllegalArgumentException if the base and size do not fall into
     *  this type of memory
     */
    public void initialize(long base, long vBase, long size);
    
    /** If this type of memory is removable, return true.
     * @return true if this type of memory is removeable.
     */
    public boolean isRemovable();
    
    /** Does the specified range of memory contain any of this type?
     *  @param base physical address of the beginning of the memory region
     *  @param size size of the memory region
     *  @return true if the specified range contains ANY of this type of memory.
     */
    public boolean contains(long base, long size);
    
    /** Arrange for the specified AEH to be called if any memory in the specified
     *  range is removed.
     *  @param physical base address of the beginning of the memory region
     *  @param size size of the memory region
     *  @param aeh Run this if any memory in the specified range is removed.
     *  @exception IllegalArgumentException if the base and size do not fall into
     *  this type of memory
     */
    public void onRemoval(long base, long size, AsyncEventHandler aeh);
    
    /** Arrange for the specified AEH to be called if any memory in the specified
     *  range is inserted.
     *  @param physical base address of the beginning of the memory region
     *  @param size size of the memory region
     *  @param aeh Run this if any memory in the specified range is inserted.
     *  @exception IllegalArgumentException if the base and size do not fall into
     *  this type of memory
     */
    public void onInsertion(long base, long size, AsyncEventHandler aeh);     
    
    /** Is all of the specified range of physical memory present in the system?
     *  (If any of it has been removed, return false.)
     *  @param physical base address of the beginning of the memory region
     *  @param size size of the memory region
     *  @param aeh Run this if any memory in the specified range is inserted.
     *  @exception IllegalArgumentException if the base and size do not fall into
     *  this type of memory
     */
    public boolean isPresent(long base, long size);

    //public Object[] getAttributes();
    public int getVMAttributes();

    public int getVMFlags();
        
}
    
