/*
 * -----------------------------------------------------------------
 * RawMemoryFloatAccess.java : This class holds the accessor methods 
 * for accessing a raw memory area by float and double types.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
  This class holds the accessor methods for accessing a raw memory area
  by float and double types. Implementations are required to implement
  this class if and only if the underlying Java Virtual Machine supports
  floating point data types.
  
  <p> Many of the constructors and methods in this class throw 
  {@link OffsetOutOfBoundsException}.  This exception means that the 
  value given in the offset parameter is either negative or outside the
  memory area.
  
  <p> Many of the constructors and methods in this class throw 
  {@link SizeOutOfBoundsException}.  This exception means that the value
  given in the size parameter is either negative, larger than an allowable
  range, or would cause an accessor method to access an address outside of
  the memory area.
                              
  @author Pratik Solanki
 */
public class RawMemoryFloatAccess extends RawMemoryAccess
{
  private final int SIZEOF_FLOAT = 4;
  private final int SIZEOF_DOUBLE = 8;
  
  /* 
    Constructor reserved for use by the memory object factory.
   */
                 
  public RawMemoryFloatAccess(Object type, long size)
      throws SecurityException, OffsetOutOfBoundsException, 
            SizeOutOfBoundsException,UnsupportedPhysicalMemoryException,
            MemoryTypeConflictException 
  {
    super(type, size);
  }
 
  /* 
    Constructor reserved for use by the memory object factory.
   */
  public RawMemoryFloatAccess(Object type, long base, long size)
      throws SecurityException, OffsetOutOfBoundsException, 
            SizeOutOfBoundsException,UnsupportedPhysicalMemoryException,
            MemoryTypeConflictException 
  {
    super(type, base, size);
  }
  
  /**
    Create a <code>RawMemoryFloatAccess</code> object.
    
    @param type An Object representing the type of memory required 
                (e.g., <em>dma, shared</em>) - used to define the base 
                address and control the mapping
    @param size The size of the area in bytes.
    
    @throws SecurityException The application doesn't have
                permissions to access physical memory or the given type
                of memory.
    @throws OffsetOutOfBoundsException The address is invalid.
    @throws SizeOutOfBoundsException The size is negative or
                extends into an invalid range of memory.
    @throws UnsupportedPhysicalMemoryException Thrown if the underlying 
                hardware does not support the given type.
   */
  /*public static RawMemoryFloatAccess createFloatAccess(Object type, long size)
    throws SecurityException, OffsetOutOfBoundsException, 
    SizeOutOfBoundsException, UnsupportedPhysicalMemoryException
  {
    return createFloatAccess(type, factory.getTypedMemoryBase(type, size), size);
  }*/
  
  /**
    Create a <code>RawMemoryFloatAccess</code> object.
    
    @param type An Object representing the type of memory required 
                (e.g., <em>dma, shared</em>)
    @param base The physical memory address of the area.
    @param size The size of the area in bytes.
    
    @throws SecurityException The application doesn't have
            permissions to access physical memory or the given range of memory.
    @throws OffsetOutOfBoundsException The address is invalid.
    @throws SizeOutOfBoundsException The size is negative or
            extends into an invalid range of memory.
    @throws UnsupportedPhysicalMemoryException Thrown if the underlying hardware does not
            support the given type.

   */
  /*public static RawMemoryFloatAccess createFloatAccess(Object type, long base, long size)
    throws SecurityException, OffsetOutOfBoundsException, 
    SizeOutOfBoundsException, UnsupportedPhysicalMemoryException
  {
    return new RawMemoryFloatAccess(base, size);
  }*/
 

  /** 
    Get the double at the given <code>offset</code>.
    
    @throws OffsetOutOfBoundsException The address is invalid.
    @throws SizeOutOfBoundsException The size is negative or
            extends into an invalid range of memory.
   */
  public double getDouble(long offset)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    standardChecks(offset, SIZEOF_DOUBLE);
    return getDouble0(virtAddress+offset);
  }
 
  /** 
    Get <code>number</code> <code>double</code> values starting at the
    given <code>offset</code> in this, and assigns them into the 
    <code>double</code> array starting at position <code>low</code>.
    
    @throws OffsetOutOfBoundsException The address is invalid.
    @throws SizeOutOfBoundsException The size is negative or
            extends into an invalid range of memory.
   */
                
  public void getDoubles(long offset, double[] doubles, int low, int number)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    arrayChecks(offset, SIZEOF_DOUBLE, doubles.length, low, number);
    getDoubles0(virtAddress+offset, doubles, low, number);
  }
  
  /** 
    Get the float at the given <code>offset</code>.
    
    @throws OffsetOutOfBoundsException The address is invalid.
    @throws SizeOutOfBoundsException The size is negative or
            extends into an invalid range of memory.
   */         
  public float getFloat(long offset)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    standardChecks(offset, SIZEOF_FLOAT);
    return getFloat0(virtAddress+offset);
  }
  
  /** 
    Get <code>number</code> <code>float</code> values starting at the
    given <code>offset</code> in this and assign them into the 
    <code>byte</code> array starting at position <code>low</code>.
    
    @throws OffsetOutOfBoundsException The address is invalid.
    @throws SizeOutOfBoundsException The size is negative or
            extends into an invalid range of memory.
   */
                   
  public void getFloats(long offset, float[] floats, int low, int number)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {  
    arrayChecks(offset, SIZEOF_FLOAT, floats.length, low, number);
    getFloats0(virtAddress+offset, floats, low, number);
  }
  
  /** 
    Set the float at the given <code>offset</code>.
    
    @throws OffsetOutOfBoundsException The address is invalid.
    @throws SizeOutOfBoundsException The size is negative or
            extends into an invalid range of memory.
   */         
  public void setFloat(long offset, float value)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    standardChecks(offset, SIZEOF_FLOAT);
    setFloat0(virtAddress+offset, value);
  }
  
  /** 
    Set <code>number</code> <code>float</code> values starting at the 
    given <code>offset</code> in this from the <code>byte</code> array 
    starting at position <code>low</code>.
    
    @throws OffsetOutOfBoundsException The address is invalid.
    @throws SizeOutOfBoundsException The size is negative or
            extends into an invalid range of memory.
   */                      
  public void setFloats(long offset, float[] floats, int low, int number)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    arrayChecks(offset, SIZEOF_FLOAT, floats.length, low, number);
    setFloats0(virtAddress+offset, floats, low, number);
  }
  
  /** 
    Set the double at the given <code>offset</code>.
    
    @throws OffsetOutOfBoundsException The address is invalid.
    @throws SizeOutOfBoundsException The size is negative or
            extends into an invalid range of memory.
   */         
  public void setDouble(long offset, double value)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    standardChecks(offset, SIZEOF_DOUBLE);
    setDouble0(virtAddress+offset, value);
  }
  
  /** 
    Get <code>number</code> <code>double</code> values starting at the
    given <code>offset</code> in this, and assigns them into the 
    <code>double</code> array starting at position <code>low</code>.
    
    @throws OffsetOutOfBoundsException The address is invalid.
    @throws SizeOutOfBoundsException The size is negative or
            extends into an invalid range of memory.
   */
                   
  public void setDoubles(long offset, double[] doubles, int low, int number)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    arrayChecks(offset, SIZEOF_DOUBLE, doubles.length, low, number);
    setDoubles0(virtAddress+offset, doubles, low, number);
  }

  protected void finalize()
  {
      super.finalize();
  }
  
  private native double getDouble0(long address);
  private native void setDouble0(long address, double value);

  private native float getFloat0(long address);
  private native void setFloat0(long address, float value);

  private native void getDoubles0(long address, double[] doubles, int low, int number);
  private native void setDoubles0(long address, double[] doubles, int low, int number);

  private native void getFloats0(long address, float[] floats, int low, int number);
  private native void setFloats0(long address, float[] floats, int low, int number);
}
