/*
 * -----------------------------------------------------------------
 * RealtimeThread.java : RealtimeThread is a special java thread 
 * that gives the user finer control over its execution.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

import java.io.*;

/**
  <code>RealtimeThread</code> extends {@link java.lang.Thread} and includes classes and methods to
  get and set parameter objects, manage the execution of those threads
  with a {@link ReleaseParameters} type of {@link PeriodicParameters}, and waiting.
  <p>
  A <code>RealtimeThread</code>object must be placed in a memory area such that
  thread logic may unexceptionally access instance variables and such that
  Java methods on {@link Thread} (e.g., enumerate and join) complete normally
  except where such execution would cause access violations.
  (Implementation hint: They could be allocated in {@link HeapMemory}.)
*/
public class RealtimeThread extends Thread implements Schedulable
{
    private boolean blockOnWaitForNextPeriod = false;
    int RTPriority = 0;
    Scheduler scheduler;
    SchedulingParameters scheduling = null;
    ReleaseParameters releaseParameters = null;
    MemoryParameters memoryParameters = null;
    MemoryArea memoryArea = getCurrentMemoryArea();//null;
    ProcessingGroupParameters groupParameters = null;
    java.lang.Runnable logic = null;

    private boolean pending = false;

    private long startOfPeriodMicros = 0;

    private boolean debug = true;
    

    /*
    static {
	System.loadLibrary("RealtimeThread");
    }
    */

    /* Initial memory area index value */
    private int initialMemoryAreaIndex = 0;
    
    /* Variables for memory allocation constraints. */
    int immortalAllocated = 0;
    int memAreaAllocated = 0;
    int heapAllocated = 0;
    long startTime = 0;

    private native void sleepMicrosUntil0(long time, int num);

    private native int waitForNextPeriod0(long periodMicros);

    private native boolean canChangeScheduler0();

    private native void setNativeRealtimePriority(int priority);

    private native long startNextPeriodMicros0();

    private static native void yield0();

    private void setRTPriority(int priority)
    {
      RTPriority = priority;
    }

    /**
      Return current thread's real-time priority.
      */
    private int getRTPriority()
    {
      return RTPriority;
    }

  /************** ISHIN ***********/
    private native void setDeadline (long millis, long nanos);

    /**
      Return the deadline in milli seconds.
      */
    private long getDeadlineMillis()
    {
	try  {
	  return releaseParameters.deadline.getMilliseconds();
	} catch (Exception e) {}
	return 0;
    }

    /**
      Return thread's deadline in nano seconds.
      */
    private long getDeadlineNanos()
    {
	try  {
	  return releaseParameters.deadline.getNanoseconds();
	} catch (Exception e) {}
	return 0;
    }

    private static int getIntScheduler()
    {
	return Scheduler.getScheduler();
    }
  /************** ISHIN ***********/


    /**
      Return the period in milli seconds.
      */
    private long getPeriodMillis()
    {
	try
	    {
	      if (releaseParameters instanceof PeriodicParameters)
		return ((PeriodicParameters) releaseParameters).period.getMilliseconds();
	      if (groupParameters != null)
		  return groupParameters.getPeriod().getMilliseconds();
	      else
		  {
		      if (releaseParameters instanceof SporadicParameters)
			  return ((SporadicParameters) releaseParameters).getMinimumInterarrival().getMilliseconds();
		      else
			  return 0;
		  }
	    }
	catch (Exception e)
	    {
	    }
	return 0;
    }

    /**
      Return thread's period in nano seconds.
      */
    private long getPeriodNanos()
    {
	try
	    {
	      if (releaseParameters instanceof PeriodicParameters)
		return ((PeriodicParameters) releaseParameters).period.getNanoseconds();
	      if (groupParameters != null)
		  return groupParameters.getPeriod().getNanoseconds();
	      else
		  {
		      if (releaseParameters instanceof SporadicParameters)
			  return ((SporadicParameters) releaseParameters).getMinimumInterarrival().getNanoseconds();
		      else
			  return 0;
		  }
	    }
	catch (Exception e)
	    {
	    }
	return 0;
    }

    private AbsoluteTime absoluteStart = null;

    /**
       Returns the milliseconds of the start time. If start time
       is a relative time it will convert it first to absolute.
     */
    private long getStartMillis()
    {
	try
	    {
		if (releaseParameters instanceof PeriodicParameters)
		    {
			PeriodicParameters pp = (PeriodicParameters) releaseParameters;
			//System.out.println("START MILLIS "+((PeriodicParameters) releaseParameters).start.getMilliseconds());
		       
			if (pp.start instanceof RelativeTime)
			{
			    if (absoluteStart == null)
				{
				    absoluteStart = new AbsoluteTime(new java.util.Date(System.currentTimeMillis()));
				    absoluteStart = ((RelativeTime)pp.start).absolute(null,absoluteStart);
				}
			}
			else
			{
			    absoluteStart = (AbsoluteTime)pp.start;
			}
			return absoluteStart.getMilliseconds();
		    }
		if (groupParameters != null)
		    {
			//System.out.println("START MILLIS "+groupParameters.getStart().getMilliseconds());
			HighResolutionTime ht = groupParameters.getStart();

			if (ht instanceof RelativeTime)
			{
			    if (absoluteStart == null)
				{
				    absoluteStart = new AbsoluteTime(new java.util.Date(System.currentTimeMillis()));
				    absoluteStart = ht.absolute(null,absoluteStart);
				}
			}
			else
			{
			    absoluteStart = (AbsoluteTime) ht;
			}
			return absoluteStart.getMilliseconds();
		    }
		else
		    return 0;
	    }
	catch (Exception e)
	    {
		e.printStackTrace();
	    }
	return 0;
    }

    private int getStartNanos()
    {
	try
	    {
		if (releaseParameters instanceof PeriodicParameters)
		    {
			PeriodicParameters pp = (PeriodicParameters) releaseParameters;
			if (pp.start instanceof RelativeTime)
			{
			    if (absoluteStart == null)
				{
				    absoluteStart = new AbsoluteTime(new java.util.Date(System.currentTimeMillis()));
				    absoluteStart = ((RelativeTime)pp.start).absolute(null,absoluteStart);
				}
			}
			else
			{
			    absoluteStart = (AbsoluteTime) pp.start;
			}
			
			return absoluteStart.getNanoseconds();
		    }
		if (groupParameters != null)
		    {
			HighResolutionTime ht = groupParameters.getStart();

			if (ht instanceof RelativeTime)
			{
			    if (absoluteStart == null)
				{
				    absoluteStart = new AbsoluteTime(new java.util.Date(System.currentTimeMillis()));
				    absoluteStart = ht.absolute(null,absoluteStart);
				}
			}
			else
			{
			    absoluteStart = (AbsoluteTime) ht;
			}
			return absoluteStart.getNanoseconds();
		    }
		else
		    return 0;
	    }
	catch (Exception e)
	    {
		e.printStackTrace();
	    }
	return 0;
    }
    
    /** 
      Create a real-time thread. All parameter values are null.  The default values for null parameter 
      objects are dependent on the value of the default {@link Scheduler} at the time the thread is created.
   */
    public RealtimeThread()
    {
	at = new AbsoluteTime();
	SchedulingParameters sp;
	scheduling = null;

	try
	    {
		sp = currentRealtimeThread().getSchedulingParameters();
		if (sp instanceof ImportanceParameters)
		    {
			sp = new ImportanceParameters(((ImportanceParameters)sp).getPriority(),
						      ((ImportanceParameters)sp).getImportance());
		    }
		else
		    {
			sp = new PriorityParameters(((PriorityParameters)sp).getPriority());
		    }
	    }
	catch (Exception e)
	    {
		/* Create default priority parameter */
		sp = new PriorityParameters(PriorityScheduler.instance().getNormPriority());
	    }

	setSchedulingParameters(sp);

	scheduler = Scheduler.getDefaultScheduler();
	startTime = System.currentTimeMillis();
	arrival = new Object();
	MonitorControl.setMonitorControl(arrival, new PriorityInheritance());
    }
    
    /** 
      Create a real-time thread with the given {@link SchedulingParameters}.
      @param scheduling The {@link SchedulingParameters} associated with this (and possibly other <code>RealtimeThread</code>).
    */
    public RealtimeThread(SchedulingParameters scheduling)
    {
	at = new AbsoluteTime();
	scheduler = Scheduler.getDefaultScheduler();

	setSchedulingParameters(scheduling);

	if (scheduling == null)
	    {
		SchedulingParameters sp;

		try
		    {
			sp = currentRealtimeThread().getSchedulingParameters();
			if (sp instanceof ImportanceParameters)
			    {
				sp = new ImportanceParameters(((ImportanceParameters)sp).getPriority(),
							      ((ImportanceParameters)sp).getImportance());
			    }
			else
			    {
				sp = new PriorityParameters(((PriorityParameters)sp).getPriority());
			    }
		    }
		catch (Exception e)
		    {
			/* Create default priority parameter */
			sp = new PriorityParameters(PriorityScheduler.instance().getNormPriority());
		    }
		setSchedulingParameters(sp);
	    }

	arrival = new Object();
	MonitorControl.setMonitorControl(arrival, new PriorityInheritance());
    startTime = System.currentTimeMillis();
    }
    
    /** 
      Create a real-time thread with the given {@link SchedulingParameters} and {@link ReleaseParameters}.
      @param scheduling The {@link SchedulingParameters} associated with this (and possibly other <code>RealtimeThread</code>).
      @param release The {@link ReleaseParameters} associated with this (and possibly other <code>RealtimeThread</code>).
    */
    public RealtimeThread(SchedulingParameters scheduling,
			  ReleaseParameters release)
    {
	at = new AbsoluteTime();
	scheduler = Scheduler.getDefaultScheduler();
	// change sched = ... to setSchedulingParamters(*)
	setSchedulingParameters(scheduling);

	/************* ISHIN **************/
	if (scheduling == null)
	/*
	if (scheduling != null) 
	    {
		if (scheduling instanceof PriorityParameters) {
		    setRTPriority(((PriorityParameters)scheduling).getPriority());
		}
	    }
	else
	*/
	/************* ISHIN **************/
	    {
		SchedulingParameters sp;
		try
		    {
			sp = currentRealtimeThread().getSchedulingParameters();
			if (sp instanceof ImportanceParameters)
			    {
				sp = new ImportanceParameters(((ImportanceParameters)sp).getPriority(),
							      ((ImportanceParameters)sp).getImportance());
			    }
			else
			    {
				sp = new PriorityParameters(((PriorityParameters)sp).getPriority());
			    }
		    }
		catch (Exception e)
		    {
			/* Create default priority parameter */
			sp = new PriorityParameters(PriorityScheduler.instance().getNormPriority());
		    }

		setSchedulingParameters(sp);
	    }


	this.releaseParameters= release;
	arrival = new Object();
	MonitorControl.setMonitorControl(arrival, new PriorityInheritance());

	if ((release != null) 
	    && (release instanceof PeriodicParameters) 
	    && ( ((PeriodicParameters)release).getStart() != null) )
	    {
		startOfPeriodMicros = ((PeriodicParameters) release).getStart().getMilliseconds() * 1000 ;
		startOfPeriodMicros += ((PeriodicParameters) release).getStart().getNanoseconds() / 1000 ;
	    }
	else
	    {
		startOfPeriodMicros = System.currentTimeMillis() * 1000;
	    }

    startTime = System.currentTimeMillis();
    }
    
    /** 
      Create a real-time thread with the given characteristics and a {@link Runnable}.
      @param scheduling The {@link SchedulingParameters} associated with this (and possibly other <code>RealtimeThread</code>).
      @param release The {@link ReleaseParameters} associated with this (and possibly other <code>RealtimeThread</code> ).
      @param memory The {@link MemoryParameters} associated with this (and possibly other <code>RealtimeThread</code> ).
      @param area The {@link MemoryArea} associated with this.
      @param group The {@link ProcessingGroupParameters} associated with this 
                    (and possibly other <code>RealtimeThread</code> ).
    */
    public RealtimeThread(SchedulingParameters scheduling,
			  ReleaseParameters release,
			  MemoryParameters memory, MemoryArea area,
			  ProcessingGroupParameters group,
			  java.lang.Runnable logic)
    {
	super(logic);
	at = new AbsoluteTime();
	scheduler = Scheduler.getDefaultScheduler();
	setSchedulingParameters(scheduling);
	if (scheduling == null)
	    {
		SchedulingParameters sp;
		try
		    {
			sp = currentRealtimeThread().getSchedulingParameters();
			if (sp instanceof ImportanceParameters)
			    {
				sp = new ImportanceParameters(((ImportanceParameters)sp).getPriority(),
							      ((ImportanceParameters)sp).getImportance());
			    }
			else
			    {
				sp = new PriorityParameters(((PriorityParameters)sp).getPriority());
			    }
		    }
		catch (Exception e)
		    {
			/* Create default priority parameter */
			sp = new PriorityParameters(PriorityScheduler.instance().getNormPriority());
		    }
		setSchedulingParameters(sp);
	    }

	//this.releaseParameters = release;
	setReleaseParameters(release);

	//this.memoryParameters = memory;
	setMemoryParameters(memory);

	arrival = new Object();
	MonitorControl.setMonitorControl(arrival, new PriorityInheritance());

	if ((release != null) &&
	    (release instanceof PeriodicParameters) && 
	    ( ((PeriodicParameters)release).getStart() != null) )
	    {
		startOfPeriodMicros = ((PeriodicParameters) release).getStart().getMilliseconds() * 1000 ;
		startOfPeriodMicros += ((PeriodicParameters) release).getStart().getNanoseconds() / 1000 ;
	    }
	else if((group != null) &&
		( group.getStart() != null) )
	    {
		startOfPeriodMicros = group.getStart().getMilliseconds() * 1000 ;
		startOfPeriodMicros += group.getStart().getNanoseconds() / 1000 ;		
	    } 
	else
	    {
		startOfPeriodMicros = System.currentTimeMillis() * 1000;
	    }


    if(this.memoryParameters != null)
        memoryParameters.associateThread(this);
	this.memoryArea = area;
	this.groupParameters = group;
	this.logic = logic;
	//incrementScopedMemoryRC();
    startTime = System.currentTimeMillis();
    }

    /* This particular constructor is not in the RTSJ standar
     * so I am commenting it out.

    public RealtimeThread(SchedulingParameters scheduling,
			  ReleaseParameters release,
			  MemoryParameters memory, MemoryArea area,
			  ProcessingGroupParameters group)
    {
	scheduler = Scheduler.getDefaultScheduler();
	setSchedulingParameters(scheduling);
	//this.releaseParameters = release;
	setReleaseParameters(release);

	//this.memoryParameters = memory;
	setMemoryParameters(memory);

	this.memoryArea = area;
	this.groupParameters = group;
	this.logic = logic;
	arrival = new Lock();
	MonitorControl.setMonitorControl(arrival, new PriorityInheritance());
	//System.out.println("calling incrementScopedMemory()");
	//incrementScopedMemoryRC();
    startTime = System.currentTimeMillis();
    }

    */


 
    /** 
      Inform the scheduler and cooperating facilities that this thread's 
      feasibility parameters should be considered in 
      feasibility analysis until further notified.
      */
    public boolean addToFeasibility()
    {
	return scheduler.addToFeasibility(this);
    }
  
  /**
    This will throw a <code>ClassCastException</code> if the current thread is not a
    <code>RealtimeThread</code>.
   */
    public static RealtimeThread currentRealtimeThread() throws ClassCastException
    {
	if (!(Thread.currentThread() instanceof RealtimeThread))
	    throw new ClassCastException();
	
	return (RealtimeThread) Thread.currentThread();
    }
    
    private int descheduleBlockings = 0;

    /**
      Stop unblocking {@link RealtimeThread#waitForNextPeriod} for a periodic schedulable object.
      If this does not have a type of {@link PeriodicParameters} as it {@link ReleaseParameters} nothing happens.
     */
    public synchronized void deschedulePeriodic()
    {
	if ((getReleaseParameters() == null) ||
	    !(getReleaseParameters() instanceof PeriodicParameters))
	    return;
	
	blockOnWaitForNextPeriod = true;
	descheduleBlockings ++;
    }
    
    /** Get the current {@link MemoryArea}.
     *	@return	The current memory area in which allocations occur.
     */
    public MemoryArea getMemoryArea()
    {
	return memoryArea;
    }
  
    /** 
      * Return a reference to the {@link MemoryParameters} object.
      */
    public MemoryParameters getMemoryParameters()
    {
	return memoryParameters;
    }
    
    /** 
      Return a reference to the {@link ProcessingGroupParameters} object.
     */
    public ProcessingGroupParameters getProcessingGroupParameters()
    {
	return groupParameters;
    }
    
    /** 
      Returns a reference to the {@link ReleaseParameters} object.
     */
    public ReleaseParameters getReleaseParameters()
    {
	return releaseParameters;
    }
   
   /** 
      Get the scheduler for this thread.
     */
    public Scheduler getScheduler()
    {
	return scheduler;
    }
    
    /** 
      Return a reference to the {@link SchedulingParameters} object.
     */
    public SchedulingParameters getSchedulingParameters()
    {
	return scheduling;
    }
    
    /**
      Set the state of the generic {@link AsynchronouslyInterruptedException} to pending.
      */
    public synchronized void interrupt()
    {
	 AsynchronouslyInterruptedException genericAIE = AsynchronouslyInterruptedException.getGeneric();
  
	 if(!genericAIE.isDoInterruptibleInProcess()){
	     postAIE(genericAIE);
	     /* ROBBINS SEZ:
	      * we additionally interrupt with the original interrupt to wake
	      * threads from sleep, wait 
	      */
	     super.interrupt();
	     
	     return;
 	 }
	 if(genericAIE.isEnabled()){
	     postAIE(genericAIE);
	     super.interrupt();		 
	 }
	 else{
	     pending = true;
	 }
    }
     
     /** 
      Inform the scheduler and cooperating facilities that this thread's 
      feasibility parameters should not be considered in 
      feasibility analysis until further notified.
      */
  public boolean removeFromFeasibility()
  {
    if (scheduler != null)
      return scheduler.removeFromFeasibility(this);
    return true;
  }
    
    /** 
      Begin unblocking {@link RealtimeThread#waitForNextPeriod} for a periodic thread.
      Typically used when a periodic schedulable object is in an overrun condition.
      The scheduler should recompute the schedule and perform admission
      control. 
      If this does not have a type of {@link PeriodicParameters} as it {@link ReleaseParameters} nothing happens.
      */
    public synchronized void schedulePeriodic()
    {
	if ((getReleaseParameters() == null) ||
	    !(getReleaseParameters() instanceof PeriodicParameters))
	    return;
	
	blockOnWaitForNextPeriod = false;
	descheduleBlockings --;
	this.notifyAll();
    }
    
    /** 
      Set the reference to the {@link MemoryParameters} object.
     */
    public void setMemoryParameters(MemoryParameters parameters) throws IllegalThreadStateException
    {
	if (scheduler != null)
	    {
		if (scheduler.inFeasibility(this))
		    {
			throw (new IllegalThreadStateException());
		    }
	    }

        if(memoryParameters != null)
            memoryParameters.deassociateThread(this);
    	this.memoryParameters = parameters;
        if(memoryParameters != null)
            memoryParameters.associateThread(this);
    }
    
    /** 
      Set the reference to the {@link ProcessingGroupParameters} object.
     */
    public void setProcessingGroupParameters(ProcessingGroupParameters parameters)
    {
	if (scheduler != null)
	    {
		if (scheduler.inFeasibility(this))
		    {
			throw (new IllegalThreadStateException());
		    }
	    }

	if (this.groupParameters != null)
	    {
		this.groupParameters.unbind(this);
	    }

	this.groupParameters = parameters;

	if (this.groupParameters != null)
	    {
		this.groupParameters.bind(this);
	    }
    }
    

    private long getSkippedPeriods()
    {
	/* Adjust the start of this period */
	long periodMicros = (getPeriodMillis() * 1000 ) + 
	    (getPeriodNanos() / 1000);
	
	long uElapsed = (System.currentTimeMillis() * 1000) - startOfPeriodMicros;
	long numberOfPeriodsToAdd = (long) Math.floor( (double)uElapsed / (double)periodMicros);
	return numberOfPeriodsToAdd;
    }

    private void resetStartOfPeriod()
    {
	/* Adjust the start of this period */
	long periodMicros = (getPeriodMillis() * 1000 ) + 
	    (getPeriodNanos() / 1000);
	
//  	long uElapsed = (System.currentTimeMillis() * 1000) - startOfPeriodMicros;
//  	long numberOfPeriodsToAdd = (long) Math.floor( (double)uElapsed / (double)periodMicros);

	long numberOfPeriodsToAdd = getSkippedPeriods();

	// current period
	numberOfPeriodsToAdd++;
	
	startOfPeriodMicros += numberOfPeriodsToAdd * periodMicros;

	/*
	if (debug == true)
	    System.out.println("Adjusting startOfPeriod("+
			   startOfPeriodMicros+") Now("+
			   System.currentTimeMillis()+") CurrentDeadlineMicros("+
			   currentDeadlineMicros()+")");
	*/
    }


    /**
      Set the reference to the {@link ReleaseParameters} object.
     */
    public void setReleaseParameters(ReleaseParameters parameters) throws IllegalThreadStateException
    {
	if (scheduler != null)
	    {
		if (scheduler.inFeasibility(this))
		    {
			throw (new IllegalThreadStateException());
		    }
	    }

	if (this.releaseParameters != null)
	    {
		this.releaseParameters.unbind(this);
	    }

	this.releaseParameters = parameters;

	if (this.releaseParameters != null)
	    {
		this.releaseParameters.bind(this);
		setDeadline(parameters.deadline.getMilliseconds(),
		            parameters.deadline.getNanoseconds());
	    }

	if (parameters instanceof PeriodicParameters)
	  {
	      // reset the deadlineMissedCounter
	      missedDeadlineCounter = 0;
	      resetDeadlineTimer();
	      resetStartOfPeriod();
	  }
    }

    /**
      Set the scheduler.
      This is a reference to the scheduler that will manage the execution of this thread.
      @exception IllegalThreadStateException Thrown when
                 <code>((Thread.isAlive() && Not Blocked) == true).</code>
                 (Where blocked means waiting in
		 <code>Thread.wait(), Thread.join(),</code> or <code>Thread.sleep()</code>)
      */
    public void setScheduler(Scheduler scheduler, SchedulingParameters scheduling, ReleaseParameters release, MemoryParameters memoryParameters, ProcessingGroupParameters processingGroup) throws IllegalThreadStateException
    {
	/* remove from the previous scheduler */
	if (this.scheduler != null)
	    synchronized(this.scheduler)
		{
		    this.scheduler.removeFromFeasibility((Schedulable)this);
		}

	setSchedulingParameters(scheduling);
	setReleaseParameters(release);
	setProcessingGroupParameters(processingGroup);
	this.scheduler = scheduler;
	addToFeasibility();
    }

    public void setScheduler(Scheduler scheduler) throws IllegalThreadStateException
    {
	if (isAlive() && !canChangeScheduler0())
	    throw (new IllegalThreadStateException());

	this.scheduler = scheduler;
    }

    /**
      Set the reference to the {@link SchedulingParameters} object.
     */
    public void setSchedulingParameters(SchedulingParameters scheduling) throws IllegalThreadStateException
    {
	if (isAlive() && !canChangeScheduler0())
	    throw (new IllegalThreadStateException());

//  	if (scheduler != null)
//  	    {
//  		if (scheduler.inFeasibility(this))
//  		    {
//  			throw (new IllegalThreadStateException());
//  		    }
//  	    }

	if (scheduling == null)
	    return;

	if (this.scheduling != null)
	    {
		((PriorityParameters)this.scheduling).unbind(this);
	    }

	this.scheduling = scheduling;
	if (this.scheduling != null)
	    {
		((PriorityParameters)this.scheduling).bind(this);
	    }


	/* propagate if thread already running */

	if (scheduling instanceof PriorityParameters)
	    {
		PriorityParameters p = (PriorityParameters) scheduling;
		int rtPriority = (p.getPriority() - PriorityScheduler.instance().getMinPriority()) + 1;
		if (rtPriority >= 0)
		    {
			RTPriority = rtPriority;

		    }
		if (isAlive())
		    {
			setNativeRealtimePriority(rtPriority);
		    }
	    }
    }

    /**
      An accurate timer with nanosecond granularity.  The actual
      resolution available for the clock must be queried from somewhere
      else. The time base is the given {@link Clock}.
      The sleep time may be relative or absolute. If relative, then the
      calling thread is blocked for the amount of time given by the parameter.
      If absolute, then the calling thread is blocked until the indicated
      point in time.  If the given absolute time is before the current time,
      the call to sleep returns immediately.
      */

    private AbsoluteTime at;
    private RelativeTime rt = new RelativeTime();

    public static void sleep(Clock clock, HighResolutionTime time) throws InterruptedException
    {
	long millis;
	int nanos;
	RealtimeThread crt = null;
	
	try
	    {
		crt = currentRealtimeThread();
	    }
	catch (Exception e)
	    {
	    }
	
	clock.getTime(crt.at);

	
	if (time instanceof AbsoluteTime)
	    {
		crt.rt = ((AbsoluteTime)time).subtract(crt.at, crt.rt);
		millis = crt.rt.getMilliseconds();
		nanos = crt.rt.getNanoseconds();
		//millis = time.getMilliseconds() - crt.at.getMilliseconds();
		//nanos = time.getNanoseconds() - crt.at.getNanoseconds();
	    }
	else
	    {
		millis = time.getMilliseconds();
		nanos = time.getNanoseconds();
	    }

	synchronized(crt)
	    {
		crt.wait(millis, nanos);
	    }

    }

    /**
      An accurate timer with nanosecond granularity.  The actual
      resolution available for the clock must be queried from somewhere
      else. The time base is the default {@link Clock}.
      The sleep time may be relative or absolute. If relative, then the
      calling thread is blocked for the amount of time given by the parameter.
      If absolute, then the calling thread is blocked until the indicated
      point in time.  If the given absolute time is before the current time,
      the call to sleep returns immediately.
      */
    public static void sleep(HighResolutionTime time) throws InterruptedException
    {
	sleep(Clock.getRealtimeClock(),time);
	return;

    }

    private long currentDeadlineMicros()
    {

	long tmp = startOfPeriodMicros;
	tmp += (releaseParameters.getDeadline().getMilliseconds() * 1000) +
	    (releaseParameters.getDeadline().getNanoseconds() / 1000);
	
	return tmp;
    }
   
    /**
       Returns true if, after changing the parameters the task set was still feasible
    */
    
    public boolean setIfFeasible( ReleaseParameters release,
				     MemoryParameters memory,
				    ProcessingGroupParameters group)
    {
	if (scheduler != null)
	    {
		return scheduler.setIfFeasible(this, release, memory, group);
	    }

	return false;
    }
    
    public boolean setIfFeasible(ReleaseParameters release,
				    ProcessingGroupParameters group)
    {
	return  setIfFeasible(release, memoryParameters, group);
    }

    private boolean setIfFeasible(ProcessingGroupParameters group)
    {
	return setIfFeasible(releaseParameters, memoryParameters, group);
    }

    public boolean setIfFeasible(ReleaseParameters release,
				    MemoryParameters memory)
    {
	return setIfFeasible(release, memory, groupParameters);
    }

    private boolean setIfFeasible(ReleaseParameters release)
    {
	return setIfFeasible(release, memoryParameters, groupParameters);
    }

    private boolean setIfFeasible(MemoryParameters memory)
    {
	return setIfFeasible(releaseParameters, memory, groupParameters);
    }

    /**
       add the the feasibility of the already set scheduler if the resulting feasibility set is
       schedulable. If successful return true, if not return false. If there is not assigned 
       scheduler it will return false
    */

    public boolean addIfFeasible()
    {
	if (scheduler != null)
	    {
		synchronized(scheduler)
		    {
			scheduler.addToFeasibility(this);
			if (!scheduler.isFeasible())
			    {
				scheduler.removeFromFeasibility(this);
				return false;
			    }
			return true;
		    }
	    }
	return false;
    }

    private class DeadlineMissPeriodicTimer extends PeriodicTimer
    {
	public RelativeTime period;
	private boolean firstTime = true;
	private AbsoluteTime starting;

	public DeadlineMissPeriodicTimer(HighResolutionTime start,
					 RelativeTime interval,
					 AsyncEventHandler handler)
	{
	    super(start, interval, handler);
	    starting = start.absolute(Clock.getRealtimeClock(),starting);
	}
	public void fire()
	{
	    if (firstTime)
		{
		    setInterval(period);
		    firstTime = false;
		}
	    super.fire();
	}
    }

    private DeadlineMissPeriodicTimer deadlineTimer = null;

    private void stopDeadlineTimer()
    {
	if (deadlineTimer != null)
	    {
		// first disable 
		deadlineTimer.disable();
		
		// then free the reference so it can be collected;
		deadlineTimer = null;
	    }
    }

    private void resetDeadlineTimer()
    {
	if (isAlive())
	    stopDeadlineTimer();

	if (getReleaseParameters() != null)
	    {
		AsyncEventHandler hdlr = getReleaseParameters().getDeadlineMissHandler();
		if ((hdlr != null) && isAlive())
		    {
			// new deadline timer start it
			startDeadlineTimer();
		    }
	    }
    }

    private void startDeadlineTimer()
    {
	if (getReleaseParameters() != null)
	    {
		AsyncEventHandler hdlr = getReleaseParameters().getDeadlineMissHandler();
		if (hdlr != null)
		    {
			RelativeTime startingTime;
			long startMillis = getStartMillis();
			long now = System.currentTimeMillis();
			deadlineTimer = 
			    new DeadlineMissPeriodicTimer(new RelativeTime(getReleaseParameters().getDeadline().getMilliseconds(), 
									   getReleaseParameters().getDeadline().getNanoseconds()),
							  new RelativeTime(getPeriodMillis(), (int) getPeriodNanos()),
							  hdlr);


			deadlineTimer.period = new RelativeTime(getPeriodMillis(), (int) getPeriodNanos());

			deadlineTimer.start();
			deadlineTimer.enable();
			synchronized(deadlineTimer)
			  {
			    if (!deadlineTimer.started)
			      try
				{
				  deadlineTimer.wait();
				}
			      catch(Exception e)
				{
				}
			  }			
		    }
	    }
    }


    private int missedDeadlineCounter = 0;

    private void updateDeadlineMissedCounter()
    {
	long skippedPeriods = getSkippedPeriods();
	missedDeadlineCounter += skippedPeriods;

	long nextDeadline = startOfPeriodMicros + skippedPeriods + 
	    (releaseParameters.getDeadline().getMilliseconds() * 1000) +
	    (releaseParameters.getDeadline().getNanoseconds() / 1000);

	long nowMicros = System.currentTimeMillis() * 1000;

	if (nextDeadline <= nowMicros)
	    missedDeadlineCounter++;
    }



   /**
      Used by threads that have a reference to a {@link ReleaseParameters} 
      type of {@link PeriodicParameters}
      to block until the start of each period.  Periods start at either
      the start time in {@link PeriodicParameters} or when <code>this.start()</code> is called. This
      method will block until the start of the next period unless the thread
      is in either an overrun or deadline miss condition.
      
      If both overrun and miss handlers are null and the thread has overrun its cost or 
      missed a deadline {@link RealtimeThread#waitForNextPeriod} will immediately 
      return false once per overrun or deadline miss.  It will then again block until
      the start of the next period 
      (unless, of course, the thread has overrun or missed again).
 
      If either the overrun or deadline miss handlers are not null 
      and the thread is in either an overrun or deadline miss condition
      {@link RealtimeThread#waitForNextPeriod} will block until the handler
      corrects the situation (possibly by calling {@link RealtimeThread#schedulePeriodic}).

      {@link RealtimeThread#waitForNextPeriod} throws <code>IllegalThreadStateException</code> if
      this does not have a reference to a {@link ReleaseParameters} type of {@link PeriodicParameters}.
      @return True when the thread is not in an overrun or deadline miss condition and
              unblocks at the start of the next period.
      */


    public boolean waitForNextPeriod() throws IllegalThreadStateException
    {
	int ret;
	boolean retValue = true;

	if (debug == true)
	    System.out.println ("                     [RealtimeThread] waitForNextPeriod() is invoked.");

     	if ((releaseParameters == null) ||
	    !(releaseParameters instanceof PeriodicParameters))
	    throw new IllegalThreadStateException("RealtimeThread not periodic");

	/* Check for deadline condition */
	long currentMicros = System.currentTimeMillis() * 1000;

  	if (deadlineTimer != null)
  	    deadlineTimer.disable();


	if (currentMicros > currentDeadlineMicros())
	    {
		/* call the deadline missed handler */
		try
		    {
			AsyncEventHandler hdlr = getReleaseParameters().getDeadlineMissHandler();
			if (hdlr != null)
			    {
				synchronized(this)
				    {
					long uElapsed  = currentMicros - startOfPeriodMicros;
					long lostDeadlines = getSkippedPeriods();
					lostDeadlines += Math.floor( (((double)uElapsed) - 
								     ( (double)lostDeadlines * 
								       (double)(getPeriodMillis()* 1000 + 
										getPeriodNanos() / 1000))) / 
								     (double)((releaseParameters.getDeadline().getMilliseconds() * 1000) +
									      (releaseParameters.getDeadline().getNanoseconds() / 1000))
								     );
					
					descheduleBlockings += lostDeadlines;
					if (lostDeadlines > 0)
					    retValue = false;
					else
					    retValue = true;

				    }
				//retValue = true;
			    }
			else
			    {
				updateDeadlineMissedCounter();
				resetStartOfPeriod();
				missedDeadlineCounter--;
				if (deadlineTimer != null)
				    deadlineTimer.enable();
				return false;
			    }
		    }
		catch (Exception e)
		    {
			e.printStackTrace();
		    }
	    }
	else if (missedDeadlineCounter > 0)
	    {
		missedDeadlineCounter--;
		if (deadlineTimer != null)
		    deadlineTimer.enable();
		return false;
	    }
	
	synchronized(this)
	    {
		
		while (descheduleBlockings > 0)
		    {
			try
			    {
				if (deadlineTimer != null)
				    deadlineTimer.enable();
				this.wait();
			    }
			catch (Exception e)
			    {
			    }
		    }
	    }
	resetStartOfPeriod();
	
	int tnum =  -1;
	try{ tnum = Integer.parseInt(getName()); } catch (Exception e) {}
	
	sleepMicrosUntil0(startOfPeriodMicros, tnum); //nextPeriodMicros, tnum);
	
  	if (deadlineTimer != null)
  	    deadlineTimer.enable();
	

	return retValue;
    }

    private Object arrival;
    private long previousArrival=0;
    private int arrivalCount = 0;

    /* Dio - this was not made part of the standard */
    /* ISHIN -- */
    public boolean waitForArrival()
    {
	synchronized(arrival)
	    {
		if (previousArrival == 0)
		    previousArrival = System.currentTimeMillis();
		if (arrivalCount == 0)
		    {
			try
			    {
				arrival.wait();
			    }
			catch(Exception e)
			    {
				return false;
			    }
		    }
		arrivalCount --;
	    }
	if (releaseParameters != null &&
	    releaseParameters instanceof SporadicParameters)
	    {
		long interArrival = System.currentTimeMillis() - previousArrival;
		long minInterarrival = ((SporadicParameters)releaseParameters).getMinimumInterarrival().getMilliseconds();
		if ( interArrival < minInterarrival)
		    {
			try
			    {
				Thread.sleep(minInterarrival - interArrival);
			    }
			catch(Exception e)
			    {
			    }
		    }
		previousArrival = System.currentTimeMillis();
	    }
	return true;
    }

    public void  notifyArrival()
    {
	synchronized(arrival)
	    {
		arrivalCount ++;
		arrival.notify();
	    }
    }

    /* ISHIN -- */

/*    void incrementScopedMemoryRC()
    {
        if(memoryArea != null && memoryArea instanceof ScopedMemory)
        {
            ((ScopedMemory)memoryArea).incrementShared();
        }
    }

	void decrementScopedMemoryRC()
	{
		if(memoryArea != null && memoryArea instanceof ScopedMemory)
		{
			((ScopedMemory)memoryArea).decrementShared();
		}
	}*/

    /**
      Enter into current memory area set and allocate memory from it.
      */
	void enterMemoryArea()
	{
		if(memoryArea != null)
        {
            int depth = getMemoryAreaStackDepth();
         //   System.out.println("*** entermemoryArea() - depth = " + depth);
            MemoryArea area = getOuterMemoryArea(depth-1);
         //   System.out.println("*** entermemoryArea() - outerMem(depth-1) = " + area);
         //   System.out.println("*** entermemoryArea() - this.memoryArea = " + memoryArea);
            if(memoryArea.equals(area) == false)
            {
                synchronized(memoryArea)
                {
//                    try
//                    {
			            pushMemoryArea(memoryArea);
                        if(memoryArea instanceof ScopedMemory)
                            ((ScopedMemory)memoryArea).incrementReferenceCount();
//                    }
//                    catch(ScopedCycleException se)
//                    {
//                        se.printStackTrace();
//                    }

                }
            }
        }
	}

    /**
      Exit from the current memory area to the previous one.
      */
	void exitMemoryArea()
	{
        MemoryArea popped;
        //int depth = getMemoryAreaStackDepth();
        //popped = getOuterMemoryArea(depth-1); 
        // pop all the memory areas
        
        // System.out.println("exitMemoryArea called!");
        popped = getCurrentMemoryArea();
        do
        {
            // System.out.println("Popping memory area (initial)" + popped);
            synchronized(popped)
            {
                // System.out.println("Popping memory area " + popped);
	            if(popped instanceof ScopedMemory)
                {
                    ScopedMemory sm = (ScopedMemory) popped;
                    // System.out.println("\t Ref count of scope is " + sm.getReferenceCount());
                    sm.decrementReferenceCount();
                    // System.out.println("\t After decrementing ref count");
                }
                // System.out.println("Popping memory area - 2");
		        popMemoryArea();
                // System.out.println("Popping memory area - 3");
                popped = getCurrentMemoryArea();
                // System.out.println("Popping memory area - 4");
            }
        }while(popped != null);
//        System.out.println("exitMemoryArea() finished");
	}

    /*
     * Checks if this realtime thread is allowed to allocate numBytes
     * from memArea. Returns false if the MemoryParameters object 
     * associated with this RealtimeThread allows the allocation.
     *
     * Synchronized to make it atomic.
     *
     * This function is called from the VM whenever an allocation 
     * request is made.
     * This fn MUST only be called from
     *  CVMgcAllocNewInstance and CVMgcAllocNewArray
     */
    private synchronized boolean allowAllocation(MemoryArea memArea, int numBytes)
    {
        boolean result = false;
        if(memoryParameters == null)
            return true;
        
        // New algo.
        // First add numBytes to all memory area fields that apply
        if(memArea == null || memArea instanceof HeapMemory)
        {
            heapAllocated+=numBytes;
        }

        // I need to have this condition because instanceof will return 
        // true for null, i.e. (null instanceof ImmortalMemory) = true.
        if(memArea != null)
        {
            if(memArea instanceof ImmortalMemory || 
                memArea instanceof ImmortalPhysicalMemory)
            {
                immortalAllocated+=numBytes;
            }

            if(memArea.equals(this.memoryArea))
            {
                memAreaAllocated+=numBytes;
            }
        }
        

        // Now check if we are exceeding any bounds.
        // First heap
        long limit;
        if((limit=memoryParameters.getAllocationRate()) >= 0)
        {
            long currentTime = System.currentTimeMillis();
            float expectedAllocRate = heapAllocated/(currentTime-startTime);
            if(expectedAllocRate > limit)
            {
                return false;
            }
        }

        // Now check immortal
        if((limit=memoryParameters.getMaxImmortal()) >= 0)
        {
            if(immortalAllocated > limit)
                return false;
        }
        
        // Now check memory area
        if((limit=memoryParameters.getMaxMemoryArea()) >= 0)
        {
            if(memAreaAllocated > limit)
                return false;
        }
        
        // If false never got returned above then all checks passed
        // and I should return true.
        return true;
        
    }

	void setPendingFalse(){
		pending = false;
	}
	boolean getPending(){
		return pending;
	}
  
    //This is an oldmethod that is no longer used. getOuterMemoryArea(int)
    // is used instead
	static MemoryArea getOuterMemoryArea(MemoryArea area)
    {
        Thread current = Thread.currentThread();
        MemoryArea outer = null;
        if(current instanceof RealtimeThread)
        {
            outer = ((RealtimeThread) current).getOuterMemoryArea0(area);
        }
        if(outer == null)
            return ((RealtimeThread)current).memoryArea;

        return outer;
    }
    
    public static MemoryArea getCurrentMemoryArea()
    {
        return getCurrentMemoryArea0();
    }
    
    private native static MemoryArea getCurrentMemoryArea0();
    
    public void start()
    {
        /* This is for enabling stack inheritance.
         * Increment all scoped memories reference 
         * count
         */
        MemoryArea area = MemoryArea.getMemoryArea(this);
        int depth=0;
        MemoryArea[] stack = null;
 
        if(area instanceof ScopedMemory)
        {
            depth = getMemoryAreaStackDepth();
            stack = new MemoryArea[depth];
            getMemoryAreaStack(stack);
            for(int i=0; i<depth; i++)
            {
                if(stack[i] instanceof ScopedMemory)
                {
                    ((ScopedMemory)stack[i]).incrementReferenceCount();
                }
            }
        }
        try
        {
            if((this.memoryArea instanceof ScopedMemory) && 
                    (stack != null) &&
                    (this.memoryArea == stack[depth-1]))
            {
                // Its a ScopedMemory and its the same as last one one stack
                // Hence it wil not be pushed on stack and initial index = current depth-1
                this.initialMemoryAreaIndex = depth-1;
            }
            else
            {
                // Its different. Hence it will be pushed. initial = depth
                this.initialMemoryAreaIndex = depth;
            }
	    /*********** ISHIN ***********/
            super.start();            
	    //	    super.start(RTPriority);
	    /*********** ISHIN ***********/
        }
        catch(Throwable t)
        {
            // An exception was thrown. Undo all the ref count 
            // increments
            if(area instanceof ScopedMemory && stack != null)
            {
                for(int i=0; i<depth; i++)
                {
                    if(stack[i] instanceof ScopedMemory)
                    {
                        ((ScopedMemory)stack[i]).decrementReferenceCount();
                    }
                }
            }
        }
    }

    /**
       Set the {@link ProcessingGroupParameters}  of this schedulable object only if the resulting
       task set is feasible.
       @param groupParameters The {@link ProcessingGroupParameters} object.
    */
    
    public boolean setProcessingGroupParametersIfFeasible(ProcessingGroupParameters groupParameters)
    {
	if (groupParameters ==null)
	    return false;

	return setIfFeasible(groupParameters);
    }
    
    /** 
	Set the {@link SchedulingParameters} of this scheduable object only if the resulting task
	set is feasible.
	@param  scheduling The {@link SchedulingParameters} object. If null nothing happens. 
    */
    public boolean setSchedulingParametersIfFeasible(SchedulingParameters scheduling)
    {
	if (scheduling == null)
	    return false;

	if (scheduler == null)
	    return false;

	synchronized(scheduler)
	    {
		SchedulingParameters tmpsched = this.scheduling;
		removeFromFeasibility();
		setSchedulingParameters(scheduling);
		if (!addIfFeasible())
		    {
			setSchedulingParameters(tmpsched);
			addToFeasibility();
			return false;
		    }
	    }
	return true;
    }

    /**
       Set the {@link ReleaseParameters} for this schedulable object only if the resulting
       task set is feasible.
       @param  release The {@link ReleaseParameters} object. If null nothing happens. 
    */
    public boolean setReleaseParametersIfFeasible(ReleaseParameters release)
    {
	if (release == null)
	    return false ;

	return setIfFeasible(release);
    }
    
    
  /** 
    Set the {@link MemoryParameters} of this schedulable object if the resulting task set is feasible. 
    @param memory The {@link MemoryParameters} object. If null nothing happens.
    */
    public boolean setMemoryParametersIfFeasible(MemoryParameters memParam)
    {
	if (memParam == null)
	    return false;

	return setIfFeasible(memParam);
    }


    /** 
     * Get the size of the stack of {@link MemoryArea} instances to 
     * which this <code>RealtimeThread</code> has access. 
     *
     * @return The size of the stack of {@link MemoryArea} instances.
     */
    static int getMemoryAreaStackSize()
    {
        return getMemoryAreaStackDepth0();
    }
    
    /** 
     * Get the size of the stack of {@link MemoryArea} instances to 
     * which this <code>RealtimeThread</code> has access. 
     *
     * @return The size of the stack of {@link MemoryArea} instances.
     */
    public static int getMemoryAreaStackDepth()
    {
        return getMemoryAreaStackDepth0();
    }
   
   
    /** 
     * Get the instance of {@link MemoryArea} in the memory area stack 
     * at the index given. If the given index does not exist in the 
     * memory area scope stack then null is returned.
     * 
     * @param index The offset into the memory area stack.
     *
     * @return The instance of {@link MemoryArea} at index or <code>null</code>
     * if the given value is does not correspond to a position 
     * in the stack.
     * 
     */
    public static MemoryArea getOuterMemoryArea(int index)
    {
        if(index < 0)
        {
            return null;
        }
        try
        {
            RealtimeThread.currentRealtimeThread();
            return getOuterMemoryArea1(index);
        }
        catch(ClassCastException ce)
        {
            // We were called from a non RealtimeThread
            // Fall through and return null
        }
        return null;
    }

   
    /** 
     * Memory area stacks include inherited stacks from parent threads. 
     * The inital memory area for the current 
     * <code>RealtimeThread</code> is the memory area given as a parameter to
     * the constructor. This method returns the position in the memory areai
     * stack of that initial memory area.
     * 
     * @return The index into the memory area stack of the inital memory area
     * of the current <code>RealtimeThread</code>
     */
    public static int getInitialMemoryAreaIndex()
    {
        try
        {
            RealtimeThread currentThread = RealtimeThread.currentRealtimeThread();
            return currentThread.getInitialMemoryAreaIndexInternal();        
        }
        catch(ClassCastException ce)
        {
            //We were called from a non RealtimeThread
            return -1;
        }
    }

    int getInitialMemoryAreaIndexInternal()
    {
        return this.initialMemoryAreaIndex;
    }
    
    private static native int getMemoryAreaStackDepth0();
    private static native MemoryArea getOuterMemoryArea1(int index);
    static native void getMemoryAreaStack(MemoryArea[] stack);

    native void pushMemoryArea(MemoryArea area);
    native MemoryArea popMemoryArea();

    //This is an old method that I have to get rid of. Its no longer used.
    // getOuterMemoryArea1 is used instead.
    native MemoryArea getOuterMemoryArea0(MemoryArea area);
    native void postAIE(AsynchronouslyInterruptedException genericAIE);

}

