/*
 * -----------------------------------------------------------------
 * Schedulable.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
  Handlers and other objects can be run by a {@link Scheduler} if they
  provide a <code>run()</code> method and the methods defined below.  The
  {@link Scheduler} uses this information to create a suitable context
  to execute the <code>run()</code> method.
 */

public interface Schedulable extends java.lang.Runnable
{
  
  /** 
      Inform the {@link Scheduler} and cooperating facilities that this thread's 
      feasibility parameters should be considered in 
      feasibility analysis until further notified.
      */
    public boolean addToFeasibility();
    
    public boolean addIfFeasible();

    public boolean setIfFeasible(ReleaseParameters release, 
				    MemoryParameters memory);
  
    public boolean setIfFeasible(ReleaseParameters release, 
				    MemoryParameters memory,
				    ProcessingGroupParameters group);

    public boolean setIfFeasible(ReleaseParameters release,
				    ProcessingGroupParameters group);
  /** 
    Return the {@link MemoryParameters} of this schedulable object. 
    */
    public MemoryParameters getMemoryParameters();
  
  /**
    Return the {@link ReleaseParameters} of this schedulable object.
    */
    public ReleaseParameters getReleaseParameters();
    
    /** 
      Return the {@link Scheduler} for this schedulable object.
     */
    public Scheduler getScheduler();
  
  /** 
    Return the {@link SchedulingParameters} of this schedulable object. 
    */
    public SchedulingParameters getSchedulingParameters();

  /**
    Return the {@link ProcessingGroupParameters} of this schedulable object.
    */
    public ProcessingGroupParameters getProcessingGroupParameters();
  
  /** 
      Inform the {@link Scheduler} and cooperating facilities that this 
      thread's feasibility parameters should not be considered in 
      feasibility analysis until further notified.
      */
    public boolean removeFromFeasibility();
  
  /** 
    Set the {@link MemoryParameters} of this schedulable object. 
    @param memory The {@link MemoryParameters} object. If null nothing happens.
    */
    public void setMemoryParameters(MemoryParameters memory);
  
  /**
    Set the {@link ReleaseParameters} for this schedulable object.
    @param  release The {@link ReleaseParameters} object. If null nothing happens. 
   */
    public void setReleaseParameters(ReleaseParameters release);
  

  /** 
    Set the {@link Scheduler} for this schedulable object.
    @param scheduler The {@link Scheduler} object. If null nothing happens.
    */
    public void setScheduler(Scheduler scheduler) throws IllegalThreadStateException;

    
  /** 
    Set the {@link Scheduler} for this schedulable object.
    @param scheduler The {@link Scheduler} object. If null nothing happens.
    */
    public void setScheduler(Scheduler scheduler, SchedulingParameters scheduling, ReleaseParameters release, MemoryParameters memoryParameters, ProcessingGroupParameters processingGroup) throws IllegalThreadStateException;
    
    /** 
    Set the {@link SchedulingParameters} of this scheduable object.
    @param  scheduling The {@link SchedulingParameters} object. If null nothing happens. 
    */
    public void setSchedulingParameters(SchedulingParameters scheduling);

    /**
      Set the {@link ProcessingGroupParameters}  of this schedulable object.
      @param groupParameters The {@link ProcessingGroupParameters} object. If null nothing happens.
      */
    public void setProcessingGroupParameters(ProcessingGroupParameters groupParameters);

    /**
      Set the {@link ProcessingGroupParameters}  of this schedulable object only if the resulting
      task set is feasible.
      @param groupParameters The {@link ProcessingGroupParameters} object.
      */

    public boolean setProcessingGroupParametersIfFeasible(ProcessingGroupParameters groupParameters);

    /** 
    Set the {@link SchedulingParameters} of this scheduable object only if the resulting task
    set is feasible.
    @param  scheduling The {@link SchedulingParameters} object. If null nothing happens. 
    */
    public boolean setSchedulingParametersIfFeasible(SchedulingParameters scheduling);

    /**
       Set the {@link ReleaseParameters} for this schedulable object only if the resulting
       task set is feasible.
       @param  release The {@link ReleaseParameters} object. If null nothing happens. 
    */
    public boolean setReleaseParametersIfFeasible(ReleaseParameters release);

  /** 
    Set the {@link MemoryParameters} of this schedulable object. 
    @param memory The {@link MemoryParameters} object. If null nothing happens.
    */
    public boolean setMemoryParametersIfFeasible(MemoryParameters memParam);
    
}


