/*
 * -----------------------------------------------------------------
 * ScopedMemory.java : The abstract base class dealing with 
 * representations of memory spaces with limited lifetime.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;


/** 
 * <code>ScopedMemory</code> is the abstract base class of all classes dealing with
 * representations of memory spaces  with a limited lifetime.  The
 * <code>ScopedMemory</code> area is valid as long as there are real-time threads
 * with access to it.  A reference is created for each accessor when
 * either a real-time thread is created with the <code>ScopedMemory</code> object
 * as its memory area, or a real-time thread runs the {@link ScopedMemory#enter}
 * method for the memory area.  When the last reference to the object
 * is removed, by exiting the thread or exiting the <code>enter()</code> method,
 * finalizers are run for all objects in the memory area, and the area
 * is emptied.
 * <p>
 * A <code>ScopedMemory</code> area is a connection to a particular region of
 * memory and reflects the current status of it.  The object does not
 * necessarily contain direct references to the region of memory
 * that is implementation dependent.
 * <p> 
 * When a <code>ScopedMemory</code> area is instantiated, the object itself is
 * allocated from the current memory allocation scheme in use, but the
 * memory space that object represents is not.  Typically, the
 * memory for a <code>ScopedMemory</code> area might be allocated using native
 * method implementations that make appropriate use of <code>malloc()</code> and
 * <code>free()</code> or similar routines to manipulate memory.
 * <p> 
 * The <code>enter()</code> method of <code>ScopedMemory</code> is the mechanism used to activate
 * a new memory scope. Entry into the scope is done by calling the
 * method:
 *
 *        <p><code>public void enter(Runnable r)</code>
 * <p>
 * Where <em>r</em> is a Runnable object whose <code>run()</code> method represents
 * the entry point to the code that will run in the new scope. Exit
 * from the scope occurs when the <code>r.run()</code> completes.
 * Allocations of objects within <code>r.run()</code> are done with the <code>ScopedMemory</code>
 * area. When <code>r.run()</code> is complete, the scoped memory area is no
 * longer active. Its reference count will be decremented and if it
 * is zero all of the objects in the memory area finalized and collected.
 * <p>
 * Objects allocated from a <code>ScopedMemory</code> area have a unique lifetime.
 * They cease to exist on exiting a {@link ScopedMemory#enter} method or upon
 * exiting the last real-time thread referencing the area, regardless
 * of any references that may exist to the object.  Thus, to maintain
  *the safety of Java and avoid dangling references, a very restrictive
 * set of rules apply to <code>ScopedMemory</code> area objects:
 * <ol>
 * <li>A reference to an object in <code>ScopedMemory</code> can never be stored in an Object 
 *     allocated in the Java heap.</li>
 * <li>A reference to an object in <code>ScopedMemory</code> can never be stored in an Object 
 *     allocated in {@link ImmortalMemory}.</li>
 * <li>A reference to an object in <code>ScopedMemory</code> can only be stored in Objects allocated in the same
 *     <code>ScopedMemory</code> area, or into a -- more inner -- <code>ScopedMemory</code> area nested by the
 *      use of its <code>enter()</code> method.</li>
 * <li>References to immortal or heap objects <em>may</em> be stored into an object 
 *     allocated in a <code>ScopedMemory</code> area.</li>
 * </ol>
 *
 * @author Pratik Solanki
 */
public abstract class ScopedMemory extends MemoryArea
{

    private int referenceCount = 0;
    //Object portal;

    /** 
     * This is for printing out a readable <code>toString()</code>. 
     * Assigns a unique ID to each <code>ScopedMemory</code> object. 
     */
    private static int idTracker = 1;
 
    /** 
     * This is for printing out a readable <code>toString()</code>. 
     * Assigns a unique ID to each <code>ScopedMemory</code> object. 
     */
    private int myID = -1;
    
    /**
     * Lock for use by <code>join()</code> and its variants.
     */
    private Object lock = new Object();

    private Runnable attachedRunnable = null;

    /**
     * Create a new <code>ScopedMemory</code> area with a particular size.
     *
     * @param size The size of the new <code>ScopedMemory</code> area
     *             in bytes. If size is less than or equal to zero
     *             nothing happens.
     */
    public ScopedMemory(long size)
    {
        this(size, null);
    }

    public ScopedMemory(long size, Runnable r)
    {
        super(size);
        attachedRunnable = r;
        myID = idTracker++;
    }

    public ScopedMemory(SizeEstimator size)
    {
        this(size.getEstimate());
    }

    public ScopedMemory(SizeEstimator size, Runnable r)
    {
        this(size.getEstimate(), r);
    }

    public void enter()
    {
        if(attachedRunnable != null)
            enter(attachedRunnable);
    }
    
    /**
     * Associate this <code>ScopedMemory</code> area to the current 
     * realtime thread for the duration of the execution of the 
     * <code>run()</code> method of the given {@link Runnable}.  
     * During this bound period of execution, all objects are allocated 
     * from the <code>ScopedMemory</code> area until another
     * one takes effect, or the <code>enter()</code> method is exited. 
     * A runtime exception is thrown if this method is called from a 
     * thread other than a {@link RealtimeThread} or 
     * {@link NoHeapRealtimeThread}.
     *
     * @param logic The runnable object which contains the code to execute.
     */
    public void enter(Runnable logic)
    {
        RealtimeThread currentThread = this.getCurrentThread();
        Throwable exception = null;
        
        pushAndIncrementRC(currentThread);
        
        try
        {
            runRunnable(logic);
        }
        catch(Throwable t)
        {
            exception = t;
        }
        popAndDecrementRC(currentThread);

        if(exception != null)
        {
            if(exception instanceof RuntimeException)
                throw (RuntimeException) exception;
            else if(exception instanceof Error)
                throw (Error) exception;
            else
                System.out.println("PANIC: ScopedMemory.enter() - The Runnable threw an exception that wasn't an Error or a RuntimeException: " + exception);
        }

    }

    private void pushAndIncrementRC(RealtimeThread currentThread)
    {
        synchronized(this)
        {
            currentThread.pushMemoryArea(this);
            incrementReferenceCount();
        }
    }
    
    private void popAndDecrementRC(RealtimeThread currentThread)
    {
        synchronized(this)
        {
            decrementReferenceCount();
            currentThread.popMemoryArea();
        }
    }

    private void runRunnable(Runnable logic)
    {
        try
        {
            logic.run();
        }
        catch(Throwable t)
        {
            MemoryArea mem = MemoryArea.getMemoryArea(t);
            if(mem.equals(this))
            {
                System.out.println("Trying to propagate the following exception beyond my scope. The VM will now throw a ThrowBoundaryError");
                t.printStackTrace();
                ThrowBoundaryError e = null;
                try
                {
                    e = (ThrowBoundaryError) ImmortalMemory.instance().newInstance(Class.forName("javax.realtime.ThrowBoundaryError")); 
                }
                catch(ClassNotFoundException cnfe)
                {
                    System.err.println("ThrowBoundaryError class not found");
                }
                catch(IllegalAccessException iae)
                {
                    System.err.println("IllegalAccessException while trying to throw ThrowBoundaryError");
                }
                catch(InstantiationException ie)
                {
                    System.err.println("InstantiationException while trying to throw ThrowBoundaryError");
                }
                if(e != null) 
                    throw e;
                else
                    throw new OutOfMemoryError();
            }
            
            if(t instanceof RuntimeException)
                throw (RuntimeException)t;
            if(t instanceof Error)
                throw (Error)t;
        }
    }

    /** 
     * Get the maximum size this memory area can attain. If this is
     * a fixed size memorea area, the returned value will be equal 
     * to the initial size.
     *
     * @return The maximum size attainable.
     */
    public long getMaximumSize()
    {
        return memorySize;
    }

    /**
     * Return a reference to the portal object in this instance of
     * <code>ScopedMemory</code>. For a more detailed explanation of portals
     * see {@link ScopedMemory#setPortal(java.lang.Object)}
     *
     * @return The portal object or null if there is no portal object.
     */
    public synchronized Object getPortal()
    {
        return getPortal0();
    }


    /**
     * Set the argument to the portal object in the memory area
     * represented by this instance of <code>ScopedMemory</code>.
     * <p>
     * A portal can serve as a means of interthread communication and they are
     * used primarily when threads need to share an object that is allocated in a
     * <code>ScopedMemory</code>. The portal object for a ScopedMemory
     * must be allocated in the same <code>ScopedMemory</code>. Thus the
     * following condition has to evaluate to <code>true</code> for the 
     * portal to be set
     * <p>
     * <code>this.equals(MemoryArea.getMemoryArea(object))</code>
     *
     * @param object The object which will become the portal for this.
     *               If null the previous portal object remains the
     *               portal object for this or if there was no previous
     *               portal object then there is still no portal object
     *               for this.
     */
    public synchronized void setPortal(Object object)
    {
        if(object == null)
            return;
        
        if(MemoryArea.getMemoryArea(object).equals(this))
            setPortal0(object);
        else
            throw new IllegalArgumentException("Portal object not allocated in this ScopedMemory");
    }

    void incrementReferenceCount()
    {
        synchronized(lock)
        {
            referenceCount++;
        }
    }
    
    void decrementReferenceCount()
    {
        synchronized(lock)
        {
            referenceCount--;
            if(referenceCount == 0)
            {
                runAllFinalizers();
                reclaimMemoryAllocated();
                lock.notify();
            }
        }
    }

    /**
     * Returns the reference count of this <code>ScopedMemory</code>. 
     * The reference count is an indication of the number of threads 
     * that may have access to this scope.
     *
     * @return The reference count of this <code>ScopedMemory</code>.
     */
    public int getReferenceCount()
    {
        return referenceCount;
    }

    /**
     * Returns a user-friendly representation of this <code>ScopedMemory</code>.
     * @return The string representation
     */
    public String toString()
    {
        return new String("ScopedMemory#" + myID);
    }

    /**
     * Wait at most until the time designated by the <code>time</code> parameter for 
     * the reference count of this <code>ScopedMemory</code> to go down to zero.
     *
     * @param time If this time is an absolute time, the wait is bounded by that point in time.
     * If the time is a relative time (or a member of the <code>RationalTime</code>
     * subclass of <code>RelativeTime</code> the wait is bounded by a the specified
     * interval from some time between the time <code>join</code> is called and the
     * time it starts waiting for the reference count to reach zero.
     *
     * @exception InterruptedException if another thread interrupts this thread while it is waiting.
     */
    public void join(HighResolutionTime time) throws InterruptedException
    {
        this.getCurrentThread();
        if(time == null)
            throw new IllegalArgumentException("The HighResolutionTime parameter passed was null");
        synchronized(lock)
        {
            if(referenceCount == 0)
                return;
            HighResolutionTime.waitForObject(lock, time);
        }
    }
 
    /**
     * Wait until the reference count of this <code>ScopedMemory</code> goes down to zero.
     *
     * @exception InterruptedException If another thread interrupts this
     * thread while it is waiting.
     */
    public void join() throws InterruptedException
    {
       join(new RelativeTime(0,0)); 
    }

    /**
     * Combine <code>join();enter();</code> such that no <code>enter</code> from
     * another thread can intervene between the two method invocations. The resulting method
     * will wait for the reference count on this <code>ScopedMemory</code> to reach zero,
     * then enter the <code>ScopedMemory</code> and execute the <code>run</code> method from
     * <code>logic</code> passed in the constructor. If no Runnable was passed, the method 
     * returns immediately.
     *
     * @exception InterruptedException If another thread interrupts this thread while it is waiting.
     * @exception ScopedCycleException If entering this <code>ScopedMemory</code> would
     * violate the single parent rule.
     */
    public void joinAndEnter() throws InterruptedException, ScopedCycleException
    {
        if(attachedRunnable != null)
            joinAndEnter(attachedRunnable);
    }

    
    /**
     * Combine <code>join(time);enter();</code> such that no <code>enter</code> from
     * another thread can intervene between the two method invocations. The resulting method
     * will wait for the reference count on this <code>ScopedMemory</code> to reach zero, or
     * for the current time to reach the designated time,
     * then enter the <code>ScopedMemory</code> and execute the <code>run</code> method from
     * <code>Runnable</code> object passed at constructin time. If no <code>Runnable</code> was 
     * passed then this method returns immediately.
     *
     * @param time The time that bounds the wait.
     *
     * @exception InterruptedException if another thread interrupts this thread while it is waiting.
     * @exception ScopedCycleException If entering this <code>ScopedMemory</code> would
     * violate the single parent rule.
     */
    public void joinAndEnter(HighResolutionTime time) throws InterruptedException, ScopedCycleException
    {
        if(attachedRunnable != null)
            joinAndEnter(attachedRunnable, time);
    }

    
    /**
     * Combine <code>join();enter(logic);</code> such that no <code>enter</code> from
     * another thread can intervene between the two method invocations. The resulting method
     * will wait for the reference count on this <code>ScopedMemory</code> to reach zero,
     * then enter the <code>ScopedMemory</code> and execute the <code>run</code> method from
     * <code>logic</code>
     *
     * @param logic The {@link Runnable} object which contains the code to execute.
     *
     * @exception InterruptedException If another thread interrupts this thread while it is waiting.
     * @exception ScopedCycleException If entering this <code>ScopedMemory</code> would
     * violate the single parent rule.
     */
    public void joinAndEnter(Runnable logic) throws InterruptedException,
                                        ScopedCycleException
    {
        joinAndEnter(logic, new RelativeTime(0,0));
    }

 

    /**
     * Combine <code>join(time);enter(logic);</code> such that no <code>enter</code> from
     * another thread can intervene between the two method invocations. The resulting method
     * will wait for the reference count on this <code>ScopedMemory</code> to reach zero, or
     * for the current time to reach the designated time,
     * then enter the <code>ScopedMemory</code> and execute the <code>run</code> method from
     * <code>logic</code>.
     *
     * @param logic The {@link Runnable} object which contains the code to execute.
     * @param time The time that bounds the wait.
     *
     * @exception InterruptedException if another thread interrupts this thread while it is waiting.
     * @exception ScopedCycleException If entering this <code>ScopedMemory</code> would
     * violate the single parent rule.
     */
    public void joinAndEnter(Runnable logic, HighResolutionTime time) 
                    throws InterruptedException, ScopedCycleException
    {
        RealtimeThread currentThread = this.getCurrentThread();
       
        if(time == null)
            throw new IllegalArgumentException("The HighResolutionTime parameter passed was null");
        synchronized(lock)
        {
            if(referenceCount > 0)
                HighResolutionTime.waitForObject(lock, time);

            pushAndIncrementRC(currentThread);
        }

        try
        {
            runRunnable(logic);
        }
        finally
        {
            popAndDecrementRC(currentThread);
        }
    }

    /**
     * Utility function which returns the current thread if it is a 
     * {@link RealtimeThread}.
     *
     * @throws MemoryAccessError If the current thread is not a 
     *          <code>RealtimeThread</code>.
     */
    private RealtimeThread getCurrentThread() 
    {
        try
        {
            return RealtimeThread.currentRealtimeThread();
        }
        catch(ClassCastException c)
        {
            throw new MemoryAccessError("Current thread is not a RealtimeThread");
        }
    }
 
    native void initScoped(long size);
    native void reclaimMemoryAllocated();
    private native Object getPortal0();
    private native void setPortal0(Object portal);

    protected void finalize() throws Throwable
    {
        if(referenceCount > 0)
        {
            System.err.println("PANIC: My reference count is > 0 and finalize has been called: " + this);
            return;
        }
        super.finalize();
    }

}

