/*
 * -----------------------------------------------------------------
 * Timed.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


package javax.realtime;

/**
  Create a scope in a {@link RealtimeThread} for which <code>interrupt()</code> will be called at
  the expiration of a timer.  This timer will begin measuring time 
  at some point between the time <code>doInterruptible()</code> 
  is invoked and the time the <code>run()</code> method of
  the <code>Interruptible</code> object is invoked. Each call of <code>doInterruptible()</code>
  on an instance of <code>Timed</code> will restart the timer for the amount of time
  given in the constructor or the most recent invocation of <code>resetTime()</code>.
  All memory use of <code>Timed</code> occurs during construction or the first invocation
  of <code>doInterruptible()</code>. Subsequent invokes of <code>doInterruptible()</code> do not
  allocate memory.

  <p>Usage: <code>new Timed(T).doInterruptible(interruptible);</code> */

public class Timed extends AsynchronouslyInterruptedException{

	OneShotTimer ost;
	boolean isTicking = false;
	HighResolutionTime hrt;
	InterruptibleLogicAdapter logicAdapter = new InterruptibleLogicAdapter();
	Object lock = new Object();

    /** 
      Create an instance of <code>Timed</code> with a timer set to timeout.  If the time is in the past the
      {@link AsynchronouslyInterruptedException} mechanism is immediately activated.
      @param time The interval of time between the invocation of
      <code>doInterruptible()</code> and when <code>interrupt()</code> is
      called on <code>currentRealtimeThread()</code>. If null the {@link IllegalArgumentException} is thrown.
     */
	public Timed(HighResolutionTime time) throws IllegalArgumentException{
		if(time==null)
			throw new IllegalArgumentException("HighResolutionTime is null");

		hrt = time;
		AsyncEventHandler handler = new AsyncEventHandler(){
			public void handleAsyncEvent(){
				if(target==null)
					return;
				/*Not sure, whether to use interrupt or fire..
				  According to ATC examples it looks like we should use fire()
				  and According to comments in Timed class it looks like we should use interrupt
				  -Kalpesh.
				*/

				//target.interrupt();
				if(isTicking)
					fire();
			}

		};

		ost = new OneShotTimer(time,null,handler);	
		ost.start();	
	}
	

    /**
      Execute a timeout method.
      Starts the timer and executes the <code>run()</code> method of the 
      given {@link Interruptible} object.
      @param logic   Implements an {@link Interruptible} <code>run()</code> method.  If null nothing happens.
      */
	public boolean doInterruptible(Interruptible logic){
		synchronized(lock){

			//if(isTicking){
			//	resetTime(hrt);
			//}
		
		//	ost.start();
			ost.reschedule(hrt);
			
			isTicking = true;
		}
		logicAdapter.setLogic(logic);
		boolean b = super.doInterruptible(logicAdapter);//logic Adapter to stop timer if this aie is interrupted by "higher-priority" aie - Kalpesh.
		isTicking = false;
		return b;
	}
	
    /** 
      To reschedule the timeout for the next invocation of <code>doInterruptible()</code>.
      @param time  This can be an absolute time or a relative time.  If null the
      timeout is not changed.
      */
	public void resetTime(HighResolutionTime time){
		hrt = time;		
		ost.reschedule(time);
	}



/* This class is used as part of the solution to defuse  timer related to Timed object if doInterruptible is interrupted due to
   interrupt or fire
*/
class InterruptibleLogicAdapter implements Interruptible{
	
	Interruptible logic;
	public void setLogic(Interruptible logic){
		this.logic = logic;
	}

	public void run(AsynchronouslyInterruptedException aie) throws AsynchronouslyInterruptedException{

			logic.run(aie);
	}

	public void interruptAction(AsynchronouslyInterruptedException aie){

		isTicking = false;
		logic.interruptAction(aie);

	}

}


}


