/*
 * -----------------------------------------------------------------
 * Timer.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


package javax.realtime;

/**
 * A Timer is a timed event that measures time relative to a given Clock. This
 * class defines basic functionality available to all timers. Applications will generally
 * use either PeriodicTimer to create an event that is fired repeatedly at regular
 * intervals, or OneShotTimer for an event that just fires once at a specific time. A
 * timer is always based on a Clock, which provides the basic facilities of something
 * that ticks along following some time line (real-time, cpu-time, user-time, simulation-time,
 * etc.). All timers are created disabled and do nothing until start() is called.
 */


public abstract class Timer extends AsyncEvent{

	protected Clock clock;
	protected HighResolutionTime time;
	protected boolean isEnabled = true;
	protected RealtimeThread timerThread;
	private Object objlock = new Object();
	private Object restartLock = new Object();

	protected Object waitLock = new Object();
	private boolean isRunning = false;
        private boolean firsttime = false;
        private Timer selfReference ;
        protected boolean started = false;

	boolean rescheduled = false;

        AbsoluteTime at = new AbsoluteTime();        

	/** Create a timer that fires at time t, according to Clock c and is handled by the
	* specified handler
	*
	* @param  t       The time to fire the event, Will be converted to absolute time.
	* @param  c       The clock on which to base this time. If null, the system realtime
	*                 clock is used.
	* @param  handler The default handler to use for this event. If null, no handler is
	*                 associated with it and nothing will happen when this event
	*                 fires until a handler is provided
	*/
	protected Timer(HighResolutionTime t, Clock c ,AsyncEventHandler handler){
		clock = c;
		if(clock==null){
			clock = Clock.getRealtimeClock();
		}
		
		time = t;
		addHandler(handler);

		//Is this thread should be a realtime thread!//
		selfReference = this;
		timerThread = new RealtimeThread(new PriorityParameters(PriorityScheduler.MAX_PRIORITY)){

			public void run(){
			  synchronized(selfReference)
			    {
			      started = true;
			      selfReference.notifyAll();
			    }
				try{
					synchronized(objlock){
						firsttime = true;
					}

					while(true){
						synchronized(restartLock){
							synchronized(waitLock){

								isRunning = true;
								Init();
								isRunning = false;
							}
							restartLock.wait();
						}
					}
				}
				catch(InterruptedException ie){
					ie.printStackTrace();
				}
			}
		};
		timerThread.setDaemon(true);
	}



	/**
	* overrides: public ReleaseParameters createReleaseParameters() in class AsyncEvent
	* @return an instance of ReleaseParameters
	*/
	public ReleaseParameters createReleaseParameters(){
	    return new AperiodicParameters(null, null, null, null);
	}

	/** Disable this timer, preventing it from firing. It may subsequently be re-enabled.
	* If the timer is disabled when its fire time occurs then it will not fire.
	* However, a disabled timer continues to count while it is disabled and if it is
	* subsequently reabled before its fire time occures and is enabled when its fire time
	* occurs it will fire.
	*/	
	public void disable(){
		synchronized(objlock){
			isEnabled = false;
		}
	}

	/**
	* Re-enable this timer after it has been disabled.
	*/
	public void enable(){
		synchronized(objlock){

			isEnabled = true;
		}

	}

    public void destroy(){
    }
    
	/**
	* Return the Clock that this timer is based on
	* @return  clock   The clock of this timer based on
	*/
	public Clock getClock(){

		return clock; 
	}

	/**
	*  Get the time at which this event will fire
	*  @return an AbsoluteTime object representing the absolute time at which this will fire.
	*/
	public AbsoluteTime getFireTime(){
		return time.absolute(clock,null);
	}

	/**
	*  Change the scheduled time for this event. can take either absolute or relative times.
	*  @param t  the time to reschedule for this event firing
	*            if t is null, the previous fire time is still the time at which this will fire.
	*/
	public void reschedule(HighResolutionTime time){
	
		if(time==null)
			return;
		synchronized(waitLock){
			this.time = time;
			if(isRunning){

				rescheduled = true;
				waitLock.notifyAll();//If timerThread is waiting on old time then wakeup it and start it again using new time.
						  
			}
		}		
	}
	
	/**
	* A Timer starts measuring time from when it is started
	*/
	public void start(){
		
		if(!firsttime)
		  {
			timerThread.start();
		  }
		else{	
			synchronized(restartLock){
				restartLock.notifyAll();
			}
		}

	}

	protected abstract void Init();

}


