/*
 * -----------------------------------------------------------------
 * mem_registry.c : The crux of the memory management system of Real-
 * time java.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


#include "javavm/include/directmem.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/globalroots.h"
#include "javavm/include/gc_common.h"

#include <stdio.h>
#include <malloc.h>
#include "javavm/include/mem_registry.h"

static CVMObjectICell jtHeapMemoryTempObject;

/* global variables */
CVMObjectICell *jtHeapMemoryArea; 

/* local-to-the-file variables */
static MemoryAreaInfo *heapInfo = NULL;
static MemoryAreaInfo *headLinkedList = NULL;
static MemoryAreaInfo *tailLinkedList = NULL;
static CVMSysMutex singletonLock;
static CVMBool addGlobalRef;
static MemoryAreaInfo* saved;

/* Locking and unlocking functions */

/* Forward declaration of private methods */
static void gcsafeLockAll(CVMExecEnv* ee);
static void gcunsafeLockAll(CVMExecEnv* ee);
static void unlockAll(CVMExecEnv* ee);
static MemoryAreaInfo* gcsafeGetNewMemoryAreaInfo(CVMExecEnv *ee);
static void freeMemoryAreaInfo(CVMExecEnv *ee, MemoryAreaInfo *info);
static MemoryAreaInfo* gcunsafeLookupNoLock(CVMExecEnv *ee, CVMObjectICell *nodeicell);
static CVMBool gcunsafeAddMemoryAreaInfoNoLock(CVMExecEnv *ee, MemoryAreaInfo *info);
static MemoryAreaInfo* gcunsafeRemoveMemoryAreaNoLock(CVMExecEnv *ee, CVMObjectICell *memoryAreaicell);
static CVMBool gcunsafeCheckMemoryAreaAssignmentNoLock(CVMExecEnv *ee, 
                                          CVMObjectICell *dstMemoryArea, 
                                          CVMObjectICell *srcMemoryArea,
                                          CVMObject *srcObject);
static CVMObjectICell* gcunsafeGetMemoryAreaForObjectNoLock(CVMExecEnv *ee, CVMObject *ref);
static void gcsafeThrowOutOfMemoryError(CVMExecEnv *ee);
static MemoryAreaInfo* gcunsafeGetMemoryAreaInfoForObjectNoLock(CVMExecEnv *ee, CVMObject *ref);

extern char* CVMgcimplGetHeapBaseAddress(CVMExecEnv* ee);

#define GrabLockAndRun(ee_, action_) \
    { \
        gcunsafeLockAll(ee_); \
        action_; \
        unlockAll(ee_); \
    }

#define CheckAndExecuteGcUnsafe(ee_, action_) \
    { \
        if(CVMD_isgcUnsafe(ee_)) \
        { \
            action_; \
        } \
        else \
        { \
            CVMD_gcUnsafeExec(ee_, action_); \
        } \
    } \
                        
#define GrabLockAndExecuteGcUnsafe(ee_, action_) \
    { \
       CheckAndExecuteGcUnsafe(ee_, GrabLockAndRun(ee_, action_)); \
    }
               
               
/* ***  Public methods *** */
int memInitialised = 0;
/* Initialises the mutex */
void initialiseMemoryRegistry(CVMExecEnv *ee)
{
    if(memInitialised == 0)
    {
        CVMsysMutexInit(&singletonLock, "Memory Registry Lock", 28);
        jtHeapMemoryTempObject.ref_DONT_ACCESS_DIRECTLY = NULL;
        jtHeapMemoryArea = &jtHeapMemoryTempObject;
        memInitialised = 1;
    }
}

/* 
 * Register a new memory area.
 * gc-unsafe method 
 */
CVMBool gcunsafeRegisterMemoryArea(CVMExecEnv *ee, MemoryAreaInfo *info, CVMObjectICell *memoryArea,
                                    CVMJavaLong base, CVMJavaLong size, int memType)
{
  CVMBool result;
  CVMID_icellAssignDirect(ee, info->memoryArea, memoryArea);
  info->memBase = (char *) ((int) base);
  info->currentPointer = info->memBase;
  info->memSize = size;
  info->memAllocated = 0;
  info->memRemaining = size;

  info->memType = memType;
#ifdef CVM_DEBUG_MEMORY
  printf("****Registering memory area at %x; size = %u\n", base, size);
#endif
 
  
  GrabLockAndRun(ee, { 
          result = gcunsafeAddMemoryAreaInfoNoLock(ee, info);
          });

  return result;
}

/* 
 * Register a new memory area.
 * gc-safe method.
 */
CVMBool CVMJTregistryRegisterMemoryArea(CVMExecEnv *ee, CVMObjectICell *memoryArea,
										CVMJavaLong base, CVMJavaLong size, int memType)
{
    CVMBool result;
    MemoryAreaInfo *info = gcsafeGetNewMemoryAreaInfo(ee);
    if(info == NULL)
    {
        // We ran out of memory. Throw OutOfMemoryError
        gcsafeThrowOutOfMemoryError(ee);
        return CVM_FALSE;
    }
    
    CheckAndExecuteGcUnsafe(ee, {
                      result = gcunsafeRegisterMemoryArea(ee, info, memoryArea, base, size, memType);
                    });
    
    return result;
}

/* 
 * gc-unsafe method to unregister the memory area 
 */
/*CVMBool gcunsafeUnregisterMemoryArea(CVMExecEnv *ee, CVMObjectICell *memoryArea)
{
    CVMBool result;
    MemoryAreaInfo *node;
    
    GrabLockAndRun(ee, {
             node = gcunsafeRemoveMemoryAreaNoLock(ee, memoryArea);
            });
    
    return result;
}*/

/* 
 * gc-safe method to unregister memory area 
 */
CVMBool CVMJTregistryUnregisterMemoryArea(CVMExecEnv *ee, 
										  CVMObjectICell *memoryArea)
{
    MemoryAreaInfo *node;
    
    GrabLockAndExecuteGcUnsafe(ee, {
                  node = gcunsafeRemoveMemoryAreaNoLock(ee, memoryArea);
                });
    if(node)
    {
        freeMemoryAreaInfo(ee, node);
        return CVM_TRUE;
    }
    return CVM_FALSE;
}


/*
 * Reclaim the memory area i.e. set currentPointer to base address.
 * Does NOT free the memory, just starts reusing from the base address.
 * 
 * Also frees all the global roots associated with this memory area.
 * These global roots are actually references to heap memory objects 
 * that were stored in this memory area.
 */
void CVMJTregistryReclaimMemory(CVMExecEnv *ee, CVMObjectICell *memoryArea)
{
    GlobalRefList *s, *t;
    CVMdebugPrintf((" ...... Reclaim memory started\n"));
    GrabLockAndExecuteGcUnsafe(ee, 
            { 
                MemoryAreaInfo *info = gcunsafeLookupNoLock(ee, memoryArea); 
                info->currentPointer = info->memBase;
                info->memRemaining = info->memSize;
                info->memAllocated = 0;
                t = info->reflistHead;
                info->reflistHead = NULL;
                memset(info->memBase, 0, info->memSize);
           });
    
    CVMdebugPrintf((" ...... Reclaim memory  - 2\n"));
    while(t != NULL)
    {
        CVMID_freeGlobalRoot(ee, t->ref);
        s = t->next;
        free(t);
        t = s;
    };
    CVMdebugPrintf((" ...... Reclaim memory finished\n"));
}     

/*
 * This function should only be called from CVMgcAllocnewInstance which is 
 * in gc-unsafe mode.
 * 
 * This function returns CVM_FALSE if the system runs out of memory.
 * 
 * gc-unsafe function.
 */
CVMBool CVMJTregistryAddFinalizableObject(CVMExecEnv *ee, CVMObject *newInstance)
{
	MemoryAreaInfo *info;
	FinalizableObject  *newNode;
	
	newNode = malloc(sizeof(*newNode));
    if(newNode == NULL)
        return CVM_FALSE;

	GrabLockAndRun(ee, {
			info = gcunsafeLookupNoLock(ee, CVMmemoryAreaICell(ee));
			newNode->reference = newInstance;
			newNode->next = info->finalizeListHead;
			info->finalizeListHead = newNode;
		});

	return CVM_TRUE;
}

/*
 * Invokes finalizer methods on all finalizable objects allocated in 
 * MemoryArea 'mem'.
 *
 * This is a gc-safe method.
 */
void CVMJTregistryInvokeAllFinalizers(CVMExecEnv *ee, CVMObjectICell *mem)
{
	jclass cls;
	jmethodID mid;
	JNIEnv *env;
	MemoryAreaInfo *info;
	
	FinalizableObject *list_head, *tempRef;
	CVMObjectICell temp ;
	
	env = CVMexecEnv2JniEnv(ee);

	GrabLockAndExecuteGcUnsafe(ee, { 
				info = gcunsafeLookupNoLock(ee,mem);
			});
	
    // The following should execute in 'locked' + gc-safe mode
    // but I am not doing that right now because it could cause other 
	// allocations to block. I need to have a finer granularity of locks 
	// to avoid this.
	// 
	// -Pratik.
	
	if(!info)
    	return ;
	
	list_head = info->finalizeListHead;

    while(list_head)
    {
       temp.ref_DONT_ACCESS_DIRECTLY = list_head->reference;
       cls = (*env)->GetObjectClass(env,&temp);
       if(!cls)
            return ;
       mid = (*env)->GetMethodID(env, cls,"finalize","()V");
      
	  /* This check shouldn't be here, coz we have only those nodes, who
       * have finalizer() method 
	   */
      if(!mid)
           continue ;

      (*env)->CallVoidMethod(env, &temp, mid);
      tempRef = list_head;
      list_head = list_head->next;
      free(tempRef);
   }
   GrabLockAndExecuteGcUnsafe(ee, {
			info->finalizeListHead = NULL;
			});
}



/*
 * A new object of size 'size' is to be allocated from memory area 
 * 'memoryAreaicell'. Allocate the space and return a pointer to it.
 * or NULL if I can't satisfy the request.
 *
 * This function HAS to be called from gcunsafe code.
 */
CVMObject* CVMJTregistryInvokeNewInstance(CVMExecEnv *ee, CVMObjectICell *memoryAreaicell, CVMUint32 size)
{
  MemoryAreaInfo *node;
  CVMObject *allocatedObj;
  
  CVMassert(CVMD_isgcUnsafe(ee));
  
  GrabLockAndRun(ee, {  
      node = gcunsafeLookupNoLock(ee, memoryAreaicell);
  
      CVMassert(node != NULL);
  
      if(size>node->memRemaining)
      {
    	printf("mem_registry.c::InvokeNewInstance() - Went out of memory\n");
	    allocatedObj = NULL;
      }
      else
      {
        allocatedObj = (CVMObject*) node->currentPointer;
        node->currentPointer += size;
        node->memAllocated += size;
        node->memRemaining -= size;
  
      }
    });
  
#ifdef CVM_DEBUG_MEMORY
  printf("****Allocated %u bytes in %x\n", size, node->memBase);
#endif
  return allocatedObj;
}

/*
 * Returns a reference to the MemoryArea object in which the object to 
 * which the reference is passed is allocated.
 * i.e. Will return the MemoryArea where obj->ref_DONT_ACCESS_DIRECTLY 
 * is allocated.
 *
 * This method is gc-safe.
 */
CVMObjectICell* CVMJTregistryGetMemoryAreaForObject(CVMExecEnv *ee, CVMObjectICell* obj)
{
  CVMObject *ref;
  CVMObjectICell *returnValue;
  GrabLockAndExecuteGcUnsafe(ee, {
        	    ref = CVMID_icellDirect(ee, obj);
                returnValue = gcunsafeGetMemoryAreaForObjectNoLock(ee, ref);
        });
  return returnValue;
}

MemoryAreaInfo* CVMJTregistryGetMemoryAreaInfoForObject(CVMExecEnv *ee, CVMObjectICell *obj)
{
  CVMObject *ref;
  MemoryAreaInfo *returnValue;
  GrabLockAndExecuteGcUnsafe(ee, {
        	    ref = CVMID_icellDirect(ee, obj);
                returnValue = gcunsafeGetMemoryAreaInfoForObjectNoLock(ee, ref);
        });
  return returnValue;
}

CVMBool gcunsafeAllowLocalVarAssignment(CVMExecEnv *ee, CVMObject *directObj)
{
    CVMObjectICell *dstReferenceMemoryArea, *srcObjectMemoryArea;
    CVMBool result;
    MemoryAreaInfo *savedInfo = NULL;
    
    // Special case. If directObj is NULL, then allow access
    if(directObj == NULL)
        return CVM_TRUE;

    GrabLockAndRun(ee, {        
        addGlobalRef = CVM_FALSE;
        dstReferenceMemoryArea = CVMmemoryAreaICell(ee); //CVMregistryUnsafeGetMemoryAreaForObject(ee, );
        srcObjectMemoryArea = gcunsafeGetMemoryAreaForObjectNoLock(ee, directObj);
  
        //Sanity check. In case the memoryAreaICell value didn't get set.
        if(dstReferenceMemoryArea == NULL || CVMID_icellIsNull(dstReferenceMemoryArea))
            dstReferenceMemoryArea = jtHeapMemoryArea;
        
        if(srcObjectMemoryArea == NULL) // This object is at an address I know nothing about!
            result = CVM_TRUE;
        else
            result = gcunsafeCheckMemoryAreaAssignmentNoLock(ee, dstReferenceMemoryArea, srcObjectMemoryArea, directObj);
        // I need to save this because I will be relinquishing the lock after this.
        if(addGlobalRef)
            savedInfo = saved;
        });

    // We tried storing a reference to a heap memory object in our memory areas.
    // Create a new global ref to directobj so that GC can find this object.
    // 
    // A local ref would do because I am trying to do an astore here but 
    // since I can't figure out hwo exactly to create a local root, I am
    // going to create a global root.
    // - Pratik.
    if(savedInfo)
    {
        CVMObjectICell *heapRef;
        CVMD_gcSafeExec(ee, {
                heapRef = CVMID_getGlobalRoot(ee);
            });

        // back in gc-unsafe mode
        CVMID_icellSetDirect(ee, heapRef, directObj);
                
        GrabLockAndRun(ee, {
                GlobalRefList *newNode = (GlobalRefList *) malloc(sizeof(GlobalRefList));
                if(newNode)
                {
                    newNode->ref = heapRef;
                    newNode->next = savedInfo->reflistHead;
                    savedInfo->reflistHead = newNode;
                }
                else
                {
                    printf("PANIC: malloc() failed\n");
                }
            });
    }
    
    return result;
}


/* 
 * This function checks if directObj can be assigned to a local variable.
 * This function MUST be executed in gc-unsafe mode.
 */
CVMBool CVMJTrcAllowLocalVarAssignment(CVMExecEnv *ee, CVMObject *directObj)
{
    CVMBool result;
    //CheckAndExecuteGcUnsafe(ee,
    //        {
    CVMassert(CVMD_isgcUnsafe(ee));
                result = gcunsafeAllowLocalVarAssignment(ee, directObj);  
    //        });
    return result;
}

/* 
 * This function checks if reference assignment is allowed.
 *
 * More precisely,
 * -> check if a reference to srcObject can be stored in dstObject
 *
 * This function MUST be called in gc-unsafe mode.
 */
CVMBool CVMJTrcAllowReferenceAssignment(CVMExecEnv *ee, CVMObject *dstObj, CVMObject *srcObj)
{
    CVMObjectICell *srcArea, *dstArea;
    CVMBool result;
    MemoryAreaInfo *savedInfo = NULL;
    
    CVMassert(CVMD_isgcUnsafe(ee));
    //GrabLockAndExecuteGcUnsafe(ee, {
    GrabLockAndRun(ee, {
                // Sanity check. See if the source object is NULL. If so, return true
                if(srcObj == NULL)
                    result = CVM_TRUE;
                else
                {
                    addGlobalRef = CVM_FALSE;
                    srcArea = gcunsafeGetMemoryAreaForObjectNoLock(ee, srcObj);
                    dstArea = gcunsafeGetMemoryAreaForObjectNoLock(ee, dstObj);
                    if(srcArea == NULL || dstArea == NULL)
                        result = CVM_TRUE;
                    else
                        result = gcunsafeCheckMemoryAreaAssignmentNoLock(ee, dstArea, srcArea, srcObj);
                   
                    // I need to save this because I will be relinquishing the lock after this.
                    if(addGlobalRef)
                        savedInfo = saved;
                }
            });

    // We tried storing a reference to a heap memory object in our memory areas.
    // Create a new global ref to directobj so that GC can find this object.
    // 
    // - Pratik.
    if(savedInfo)
    {
        CVMObjectICell *heapRef;
        CVMD_gcSafeExec(ee, {
                heapRef = CVMID_getGlobalRoot(ee);
            });

        // back in gc-unsafe mode
        CVMID_icellSetDirect(ee, heapRef, srcObj);

        GrabLockAndRun(ee, {
                GlobalRefList *newNode = (GlobalRefList *) malloc(sizeof(GlobalRefList));
                if(newNode)
                {
                    newNode->ref = heapRef;
                    newNode->next = savedInfo->reflistHead;
                    savedInfo->reflistHead = newNode;
                }
                else
                {
                    printf("PANIC: malloc() failed\n");
                }
            });
    }
    
 
    return result;
}

/*
 * Create the MemoryAreaInfo for Heap and add it to the list.
 */
void CVMJTregistryCreateHeapMemoryAreaInfo(CVMExecEnv* ee, CVMObjectICell* theHeap)
{
    if(heapInfo == NULL)
    {
        heapInfo = gcsafeGetNewMemoryAreaInfo(ee);
        if(heapInfo == NULL)
        {
            // We ran out of memory. Throw OutOfMemoryError
            gcsafeThrowOutOfMemoryError(ee);
            return;
        }
        heapInfo->memType = MEMTYPE_HEAP;
        heapInfo->memBase = CVMgcimplGetHeapBaseAddress(ee);
        heapInfo->memSize = CVMgcimplTotalMemory(ee);
        heapInfo->currentPointer = heapInfo->memBase + heapInfo->memSize;
        heapInfo->memAllocated = heapInfo->memSize;
        heapInfo->memRemaining = 0;

        CVMID_icellAssign(ee, heapInfo->memoryArea, theHeap);
        GrabLockAndExecuteGcUnsafe(ee, {
                gcunsafeAddMemoryAreaInfoNoLock(ee, heapInfo);
            });
    }
}

/*
 * Find the MemoryAreaInfo for the passed memoryAreaicell
 */
MemoryAreaInfo* CVMJTregistryLookupMemoryArea(CVMExecEnv *ee, CVMObjectICell *memoryAreaicell)
{
   MemoryAreaInfo *info; 

   GrabLockAndExecuteGcUnsafe(ee, {
            info = gcunsafeLookupNoLock(ee, memoryAreaicell);
         });

   return info;
}

/* Returns NULL if the system runs out of memory */
PerThreadMemoryInfo* CVMJTregistryCreatePTMI(CVMExecEnv *ee)
{
  PerThreadMemoryInfo *ptmi = (PerThreadMemoryInfo *)malloc(sizeof(PerThreadMemoryInfo));
  if(ptmi == NULL)
      return NULL;
  ptmi->upInStack = NULL;
  ptmi->downInStack = NULL;
  ptmi->info = NULL;
  return ptmi;
}

void CVMJTregistryDeletePTMI(PerThreadMemoryInfo *info)
{
  free(info);
}

/* gc-unsafe method call */
CVMBool CVMJTtreeAddEdge(CVMExecEnv *ee, MemoryAreaInfo *child, 
                            MemoryAreaInfo *parent, int refCount)
{
    CVMBool result = CVM_FALSE;
    CVMassert(CVMD_isgcUnsafe(ee));
    CVMassert(child->memType == MEMTYPE_SCOPED);
    CVMassert(parent==NULL || parent->memType == MEMTYPE_SCOPED);

/*    CVMdebugPrintf(("Adding a new tree edge\n"));
    CVMdebugPrintf(("child->parent = %x\n", child->parent));
    CVMdebugPrintf(("parent = %x\n", parent));
    CVMdebugPrintf(("refcount = %d\n", refCount));*/
    
    GrabLockAndRun(ee, 
        {
            /* Adding edge is allowed if
             * 1. child-parent relationship already exists
             * 2. If the child has no parent yet and its ref count is 0
             */
            if(child->parent == parent || 
                (child->parent == NULL && refCount == 0))
            {
                child->parent = parent;
                result = CVM_TRUE;
            }
            else
                result = CVM_FALSE;
                
        });
    
    return result;
}

/* gc-unsafe function */
CVMBool CVMJTtreeRemoveEdge(CVMExecEnv* ee, MemoryAreaInfo *child, MemoryAreaInfo *parent)
{
    CVMassert(CVMD_isgcUnsafe(ee));
    CVMassert(child->memType == MEMTYPE_SCOPED);
    
    GrabLockAndRun(ee,
        {
            child->parent = NULL;
        });
    return CVM_TRUE;
}

int CVMJTregistryHandleReference(void *arg)
{
/*  struct arg_t
  {
	CVMExecEnv *ee;
	CVMObject *ref;
  };
  
  struct arg_t *data = (struct arg_t*) arg;
  */
  // FIXME:
  // over here I should do some intelligent things like check if
  // data->ref actually points to a memory location in my memory 
  // area. But instead I just return without doing anything
  // - Pratik.
  return 1;
}

/*
 * Returns the amount of free memory for the MemoryArea object 
 * referenced by memoryAreaicell
 */
CVMJavaLong CVMJTregistryGetMemoryRemaining(CVMExecEnv *ee, CVMObjectICell *memoryAreaicell)
{
  MemoryAreaInfo *node;
  int memRemaining;
  CVMBool res;

  GrabLockAndExecuteGcUnsafe(ee, 
          {
  
           	  //Is the memory area heap?
              JT_gcunsafeIsSame(ee, memoryAreaicell, jtHeapMemoryArea, res)

              if(res)
           	  {
           	    memRemaining = (CVMJavaLong) CVMgcimplFreeMemory(ee);
           	  }
           	  else
           	  {
                node = gcunsafeLookupNoLock(ee, memoryAreaicell);
                if(node == NULL)
                {
           	        printf("ERROR: CVMJTregistryGetMemoryRemaining - Could not identify memory area\n");
           	        memRemaining = 0;
                }
                else
                    memRemaining = node->memRemaining;
           	  }
          });

  return memRemaining;
}
 

CVMJavaLong CVMJTregistryGetMemoryAllocated(CVMExecEnv *ee, CVMObjectICell *memoryAreaicell)
{
  MemoryAreaInfo *node;
  int memAllocated;
  CVMBool res;

  GrabLockAndExecuteGcUnsafe(ee, 
          {
  
           	  //Is the memory area heap?
              JT_gcunsafeIsSame(ee, memoryAreaicell, jtHeapMemoryArea, res)

              if(res)
           	  {
           	    memAllocated = (CVMJavaLong) (CVMgcimplTotalMemory(ee) - CVMgcimplFreeMemory(ee));
           	  }
           	  else
           	  {
                node = gcunsafeLookupNoLock(ee, memoryAreaicell);
                if(node == NULL)
                {
           	        printf("ERROR: CVMJTregistryGetMemoryAllocated - Could not identify memory area\n");
           	        memAllocated = 0;
                }
                else
                    memAllocated = node->memAllocated;
           	  }
          });

  return memAllocated;
}


/* gc-safe function */
void CVMJTregistryRegisterImmortal(CVMExecEnv *ee, CVMObjectICell *immortalMemory,
									CVMJavaLong base, CVMJavaLong size, int immortalObjSize)
{
    CVMJTregistryRegisterMemoryArea(ee, immortalMemory, base, size, MEMTYPE_IMMORTAL);
    GrabLockAndRun(ee, {
                // Make use of the fact that immortal memory got added at the tail
                CVMBool res;
                CVMID_icellSameObject(ee, tailLinkedList->memoryArea, immortalMemory, res);
                CVMassert(res);
                tailLinkedList->currentPointer += immortalObjSize; 
                tailLinkedList->memAllocated += immortalObjSize;
                tailLinkedList->memRemaining -= immortalObjSize;
            });
}

/* *** Private (static) functions *** */

static void gcsafeLockAll(CVMExecEnv* ee)
{
    CVMassert(CVMD_isgcSafe(ee));
    CVMsysMutexLock(ee, &singletonLock);
}

static void gcunsafeLockAll(CVMExecEnv* ee)
{
#ifdef CVM_DEBUG_MEMORY
    CVMconsolePrintf("Locking memory lock\n");
#endif
    if(!CVMsysMutexTryLock(ee, &singletonLock))
    {
        CVMD_gcSafeExec(ee, 
                {
                  gcsafeLockAll(ee);
                });
    }
}

static void unlockAll(CVMExecEnv* ee)
{
#ifdef CVM_DEBUG_MEMORY
    CVMconsolePrintf("Unlocking memory lock\n");
#endif
    CVMsysMutexUnlock(ee, &singletonLock);
}

/* gc-safe method 
 * Returns NULL if the system runs out of memory
 */

static MemoryAreaInfo* gcsafeGetNewMemoryAreaInfo(CVMExecEnv *ee)
{
  MemoryAreaInfo *info = (MemoryAreaInfo*) malloc(sizeof(MemoryAreaInfo));
  if(info == NULL)
      return NULL;
  info->memoryArea = CVMID_getWeakGlobalRoot(ee);
  info->memBase = NULL;
  info->currentPointer = NULL;
  info->memSize = 0;
  info->memAllocated = 0;
  info->memRemaining = 0;

  info->portal = NULL;

  info->memType = MEMTYPE_UNKNOWN;
  info->parent = NULL;
  //info->child = NULL;
  info->next = NULL;
  info->prev = NULL;
  info->finalizeListHead = NULL;
  info->reflistHead = NULL;

  return info;
}

/* Free the memory area info.
 * NOTE that this does not free the memory region. Maybe I SHOULD be doing that too.
 */
static void freeMemoryAreaInfo(CVMExecEnv *ee, MemoryAreaInfo *info)
{
  //Finalizers on all the objects should have been run before
  
  //free the CVMObjectICell first
  CVMID_freeWeakGlobalRoot(ee, info->memoryArea);

  free(info);
}


//static MemoryAreaInfo* lookupSimpleLinkedList(CVMExecEnv *ee, CVMObjectICell *nodeicell)
static MemoryAreaInfo* gcunsafeLookupNoLock(CVMExecEnv *ee, CVMObjectICell *nodeicell)
{
  MemoryAreaInfo *temp = headLinkedList;
//  CVMdebugPrintf(("gcunsafeLookupNoLock - nodeicell = %x\t", nodeicell));
//  CVMdebugPrintf(("looking for %O\n", CVMID_icellDirect(ee, nodeicell)));
  while(temp != NULL)
	{
	  CVMBool res = CVM_FALSE;
	  CVMObjectICell *tempicell = temp->memoryArea;
	  if(CVMD_isgcSafe(ee))
		{
		  CVMID_icellSameObject(ee, tempicell, nodeicell, res);
		}
	  else
		{
		  JT_gcunsafeIsSame(ee, tempicell, nodeicell, res)
		}
	  if(res == CVM_TRUE)
		return temp;
	  temp = temp->next;
	}
  return NULL;
}

//static CVMBool addToSimpleLinkedList(CVMExecEnv *ee, MemoryAreaInfo *info)
static CVMBool gcunsafeAddMemoryAreaInfoNoLock(CVMExecEnv *ee, MemoryAreaInfo *info)
{
  MemoryAreaInfo *nullinfo = gcunsafeLookupNoLock(ee, info->memoryArea);
  if(nullinfo != NULL)
	{
	  printf("mem_registry.c::addToSimpleLinkedList - memoryarea already present!\n");
	  return CVM_FALSE;
	}

  if(headLinkedList == NULL)
  {
	headLinkedList = tailLinkedList = info;
	return CVM_TRUE;
  }

  tailLinkedList->next = info;
  info->prev = tailLinkedList;
  tailLinkedList = info;
  return CVM_TRUE;
}

/*
 * Removes the MemoryAreaInfo associated with the MemoryArea object 
 * referenced by memoryAreaicell.
 *
 * Does NOT grab the lock and expects to be running in gcunsafe mode.
 */
static MemoryAreaInfo* gcunsafeRemoveMemoryAreaNoLock(CVMExecEnv *ee, CVMObjectICell *memoryAreaicell)
{
  MemoryAreaInfo *info = gcunsafeLookupNoLock(ee, memoryAreaicell);
  if(info == NULL)
	return CVM_FALSE;

  if(info->prev != NULL)
	info->prev->next = info->next;
  else
	headLinkedList = info->next;
  
  if(info->next != NULL)
	info->next->prev = info->prev;
  else
	tailLinkedList = info->prev;

//  freeMemoryAreaInfo(ee, info);
//  return CVM_TRUE;
    return info;
}

/*
 * Checks if reference assignment between source and destination memory
 * areas is allowed according to the assignmnet rules in the spec.
 *
 * Expects to be in gc-unsafe mode and does not grab any locks.
 */
static CVMBool gcunsafeCheckMemoryAreaAssignmentNoLock(CVMExecEnv *ee, 
                                          CVMObjectICell *dstMemoryArea, 
                                          CVMObjectICell *srcMemoryArea,
                                          CVMObject *srcObject)
{
    MemoryAreaInfo *srcInfo, *dstInfo;

    srcInfo = gcunsafeLookupNoLock(ee, srcMemoryArea);
    dstInfo = gcunsafeLookupNoLock(ee, dstMemoryArea);
    
    if(srcInfo == NULL)
    {
#ifdef CVM_DEBUG
//        printf("PANIC: gcunsafeCheckMemoryAreaAssignmentNoLock: Unable to recognize source memory area\n");
#endif
        return CVM_TRUE;
    }
  
    if(dstInfo == NULL)
    {
#ifdef CVM_DEBUG
  //      printf("PANIC: gcunsafeCheckMemoryAreaAssignmentNoLock: Unable to recognize destination memory area\n");
#endif
        return CVM_TRUE;
    }
  
    // Allow access for Heap memory
    // (pg 59, 1st column is all Yes)
    if(srcInfo->memType == MEMTYPE_HEAP)
    {
        //Disallow access for NoHeaprealtimeThreads
        if(CVMisNoHeapRealtimeThread(ee))
        {
            return CVM_FALSE;
        }
       
        // This is the extra logic to ensure that if Immortal/Scoped 
        // are holding references to Heap then those objects don't get
        // garbage collected until the Scoped memory is reclaimed (never,
        // in the case of Immortal.
        if(dstInfo->memType == MEMTYPE_IMMORTAL ||
           dstInfo->memType == MEMTYPE_SCOPED)
        {
            addGlobalRef = CVM_TRUE;
            saved = dstInfo;
        }
        
        return CVM_TRUE;
    }
  
    // Allow access fo immortal 
    // (pg 59, 2nd column is all Yes)
    if(srcInfo->memType == MEMTYPE_IMMORTAL)
        return CVM_TRUE;

    if(srcInfo->memType == MEMTYPE_SCOPED)
    {
        // pg 59, row 1, column 3
        // Heap cannot hold a reference to Scoped
        if(dstInfo->memType == MEMTYPE_HEAP)
        {
#ifdef CVM_DEBUG
            printf("gcunsafeCheckMemoryAreaAssignmentNoLock: Heap cannot hold a reference to Scoped.\n");
#endif
            return CVM_FALSE;
        }
    
        // pg 59, row 2, column 3
        // Immortal cannot hold a reference to Scoped
        if(dstInfo->memType == MEMTYPE_IMMORTAL)
        {
#ifdef CVM_DEBUG
            printf("gcunsafeCheckMemoryAreaAssignmentNoLock: Immortal cannot hold a reference to Scoped.\n");
#endif
            return CVM_FALSE;
        }

        // pg 59, row 3, column 3
        // This is the big fun part. Scoped memory restrictions.
        // A ScopedMemory can ONLY hold a reference to an object
        // in itself, an outer scope or a shared scope.
        //
        // As of 12th March 2001, I don't know what a shared scope is.
        if(dstInfo->memType == MEMTYPE_SCOPED)
        {
            // Initialise to the current memory 'stack' element.
            PerThreadMemoryInfo *traverser = ee->perThreadStack;
      
            // Shortcut check for same scope
            //if(srcInfo == dstInfo)
            //    return CVM_TRUE;
      
            // Now, in the current threads's PerThreadMemoryInfo structure, 
            // go all the way up to the destination scope
            while(traverser != NULL && traverser->info != dstInfo)
            {
                traverser = traverser->upInStack;
            }
      
            // Error! I can't find the destination scope in current thread.
            // Return CVM_FLASE. This will trigger the  interpreter to throw 
            // an exception.
            if(traverser == NULL)
            {
//                printf("PANIC: Unable to find the destination scoped memory in the current threads memory stack!\n");
                return CVM_FALSE;
            }

            // Now go up to see if the source is above me (or the same as me)
            while(traverser != NULL && traverser->info != srcInfo)
            {
                traverser = traverser->upInStack;
            }
            // Illegal assignment! Source is not above me. The only case in 
            // which this would be valid is if it were a 'shared' scope, which
            // as of now, I don't anything about it. (i.e. I don't know what a 
            // shared scope is).
            if(traverser == NULL)
            {
#ifdef CVM_DEBUG
                printf("gcunsafeCheckMemoryAreaAssignmentNoLock: Illegal scoped memory assignment.\n");
#endif
                return CVM_FALSE;
            }
            
            // We found the source scope above the destination (or maybe it was
            // the same as destination).
            return CVM_TRUE;
        } // end 'if(dstInfo->memType == MEMTYPE_SCOPED)'
    
        // Unable to figure out the destination memory area.
        printf("PANIC: gcunsafeCheckMemoryAreaAssignmentNoLock - This shouln't happen. I can't figure out the destination memType field.\n");
        return CVM_FALSE;
    } // end 'if(srcInfo->memType == MEMTYPE_SCOPED)'
  
    // Unable to figure out the source memory area.
    printf("PANIC: gcunsafeCheckMemoryAreaAssignmentNoLock - This shouln't happen. I can't figure out the source memType field.\n");
    return CVM_FALSE;
}

/*
 * Returns a reference to the MemoryArea object in which 'ref' is 
 * allocated.
 *
 * Expects to be running in gc-unsafe mode and does not grab any locks.
 * 
 */
static CVMObjectICell* gcunsafeGetMemoryAreaForObjectNoLock(CVMExecEnv *ee, CVMObject *ref)
{
  MemoryAreaInfo *traverser = gcunsafeGetMemoryAreaInfoForObjectNoLock(ee, ref);
  if(traverser != NULL)
      return traverser->memoryArea;

  return NULL;
}

static MemoryAreaInfo* gcunsafeGetMemoryAreaInfoForObjectNoLock(CVMExecEnv *ee, CVMObject *ref)
{
  MemoryAreaInfo *traverser;
 // CVMBool isHeap;
  
  traverser	= headLinkedList;
  while(traverser != NULL)
  {
    if((char*)ref >= traverser->memBase && (char*)ref < traverser->currentPointer)
        break;
    traverser = traverser->next;
  };
  
  if(traverser == NULL)
  {
      return NULL;//jtHeapMemoryArea;
  }
  
  return traverser;
}

/* gc-safe method for throwing OutOfMemoryError */
static void gcsafeThrowOutOfMemoryError(CVMExecEnv *ee)
{
}


// ScopedMemory get/set Portal functions.

// getPortal - creates a new local ref for the portal object and returns
// Has to be called from gc-safe code.
CVMObjectICell* CVMJTsmGetPortal(CVMExecEnv* ee, CVMObjectICell* scopedMemory)
{
    JNIEnv *env = CVMexecEnv2JniEnv(ee);
    CVMObjectICell tempICell;
    GrabLockAndExecuteGcUnsafe(ee, { 
        MemoryAreaInfo *scopedInfo = gcunsafeLookupNoLock(ee, scopedMemory);
        CVMID_icellSetDirect(ee, &tempICell, scopedInfo->portal);
        });

    //Now create a new local ref and return
    return (*env)->NewLocalRef(env, &tempICell);
}

void CVMJTsmSetPortal(CVMExecEnv* ee, CVMObjectICell* scopedMemory, CVMObjectICell* portalObject)
{
    GrabLockAndExecuteGcUnsafe(ee, { 
        MemoryAreaInfo *scopedInfo = gcunsafeLookupNoLock(ee, scopedMemory);
        if(portalObject != NULL)
        {
            scopedInfo->portal = CVMID_icellDirect(ee, portalObject);
        }
        else
        {
            scopedInfo->portal = NULL;
        }
    });

    return;
}

/* *** Methods not needed *** */

/* This is commented out right now but we might want to use this 
 * macro instead of the function
 * 
#define scanAndWalkRefsInMemoryArea(ee, gcOpts, callback, data, info) \
{ \
  char *traverser = info->memBase; \
  while(traverser < info->currentptr) \
  { \
	CVMUint32 objsize = CVMobjectSize((CVMObject*)traverser); \
	CVMMobjectWalkRefsWithSpecialHandling(ee, gcOpts, CVMobjectGetClassWord((CVMObject*)traverser), \
		{}, callback, NULL); \
	traverser += objsize; \
  } \
}
*/

/* I dont think there is any use of the following 3 functions */
static void scanAndWalkRefsInMemoryArea(CVMExecEnv *ee, CVMGCOptions *gcOpts, 
				CVMRefCallbackFunc callback, void *data, MemoryAreaInfo *info)
{
  char *traverser = info->memBase;
  while(traverser < info->currentPointer)
  {
	CVMUint32 objsize = CVMobjectSize((CVMObject*)traverser);
	CVMobjectWalkRefsWithSpecialHandling(ee, gcOpts, (CVMObject*)traverser, 
		CVMobjectGetClassWord((CVMObject*)traverser), {}, callback, NULL);
	traverser += objsize;
  }
}


static void scanAndWalkRefsInLinkedList(CVMExecEnv *ee, CVMGCOptions *gcOpts, 
				CVMRefCallbackFunc callback, void *data)
{
  MemoryAreaInfo *temp = headLinkedList;
  while(temp != NULL)
  {
	scanAndWalkRefsInMemoryArea(ee, gcOpts, callback, data, temp);
	temp = temp->next;
  }
}

int CVMJTregistryScanAndWalkAllRoots(void *p)
{
  struct arg_t
  {
	CVMExecEnv *ee;
	CVMGCOptions *gcOpts;
	CVMRefCallbackFunc callback;
	void *data;
  };

  struct arg_t* a = (struct arg_t*) p;

  
  scanAndWalkRefsInLinkedList(a->ee, a->gcOpts, a->callback, a->data);
  return 1;
}

/* This function checks, wheather an object is in heap or not,
 * and accordingly returns boolean.
 * This function is used to check, wheather scope has reference to 
 * heap or not. 
 * opcode: aload, aload<_N>, aaload
 */
CVMBool CVMJTisHeapReference(CVMExecEnv *ee, CVMObject *obj)
{
	if(heapInfo == NULL)
		return CVM_TRUE;

	if((char*)obj >= heapInfo->memBase && (char*)obj < heapInfo->currentPointer)
		return CVM_TRUE;
	else
		return CVM_FALSE;
}

void CVMJTdumpImmortalMemory()
{
    int size;
    CVMClassBlock *cb;
//    CVMExecEnv *ee = CVMgetEE();
    char *t = headLinkedList->memBase;
    char *end = headLinkedList->currentPointer;
    while(t <= end)
    {
        cb = ((CVMObject*)t)->hdr.clas;
        size = CVMcbInstanceSize(cb);
        CVMconsolePrintf("Addr: %x Size: %d Class: %C\n", t, size, cb);
        t = t+size;
    }
}
/* ***** */

