/*
 * -----------------------------------------------------------------
 * PhysicalMemoryManager.c: Implementation of PhysicalMemoryManager.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


#include "javax_realtime_PhysicalMemoryManager.h"
#include <unistd.h>
#include <sys/mman.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

extern int errno;

// This is the file descriptor for /dev/mem
//
// This is a global variable. I might need to put a 
// lock around it. - Pratik.
static int memFileFD = -1;



/*
 * Class:     PhysicalMemoryManager
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_javax_realtime_PhysicalMemoryManager_init
  (JNIEnv *env, jclass clazz)
{
	// get the pagesize and save it in PhysicalMemoryManager static object
	long pagesize = 0;
	jfieldID pagesizeID;

	memFileFD = open("/dev/mem", O_RDWR);
	if(memFileFD == -1)
    {
        jclass securityClazz;
#ifdef CVM_DEBUG
        printf("PhysicalmemoryManager::init() - Error while opening /dev/mem.\nAre you superuser?\n");
#endif
        securityClazz = (*env)->FindClass(env, "java/lang/SecurityException");
	    (*env)->ThrowNew(env, securityClazz, "Cannot access physical memory. Are you superuser?");
        return;
    }
	
	pagesize = sysconf(_SC_PAGESIZE);
	if(pagesize == -1)
	{
		// Error! I could not get the pagesize.
		// Throw a RuntimeException
		jclass runtimeExceptionClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
		(*env)->ThrowNew(env, runtimeExceptionClazz, "Unable to query system page size");
		return;
	}
	
	// now save it in PhysicalMemoryManager
    pagesizeID = (*env)->GetStaticFieldID(env, clazz, "pagesize", "J");
	if(pagesizeID == NULL)
	{
		jclass runtimeExceptionClazz = NULL;
		// If an exception was thrown,clear itr and throw a RuntimeException.
		if((*env)->ExceptionCheck(env))
			(*env)->ExceptionClear(env);
		
		runtimeExceptionClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
		(*env)->ThrowNew(env, runtimeExceptionClazz, "Unable to query system page size");
		return;

	}
	(*env)->SetStaticLongField(env, clazz, pagesizeID, (jlong) pagesize);
	return;
}


/*
 * Class:     PhysicalMemoryManager
 * Method:    mmap0
 * Signature: (JJII)J
 */
JNIEXPORT jlong JNICALL Java_javax_realtime_PhysicalMemoryManager_mmap0
  (JNIEnv *env, jclass clazz, jlong base, jlong size, jint prot, jint flags)
{
    void* mappedAddr = NULL;
    mappedAddr = mmap(NULL, size, (int)prot, (int)flags, memFileFD, base);
    if(((int)mappedAddr) == -1)
    {
        switch(errno)
        {
            // EBADF  fd is not a valid file descriptor (and MAP_ANONYMOUS was not set).
            case EBADF:
                {
                    // For the moment just throwing a RuntimeException
                    jclass runtimeClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
                    (*env)->ThrowNew(env, runtimeClazz, "An error occurred when mmap() was called - EBADF");
                    break;
                }
            // EACCES MAP_PRIVATE was asked, but fd is not open for  reading.   Or  MAP_SHARED  was
            // asked and PROT_WRITE is set, fd is not open for writing.
            case EACCES:
                {        
                    // For the moment just throwing a RuntimeException
                    jclass runtimeClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
                    (*env)->ThrowNew(env, runtimeClazz, "An error occurred when mmap() was called - EACCES");
                    break;
                }

            // EINVAL We  don't  like start or length or offset.  (E.g., they are too large, or not
            // aligned on a PAGESIZE boundary.)
            case EINVAL:
                {        
                    // For the moment just throwing a RuntimeException
                    jclass runtimeClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
                    (*env)->ThrowNew(env, runtimeClazz, "An error occurred when mmap() was called - EINVAL");
                    break;
                }

            // ETXTBUSY
            // MAP_DENYWRITE was set but the object specified by fd is open for writing.
            /*case ETXTBUSY:
                {        
                    // For the moment just throwing a RuntimeException
                    jclass runtimeClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
                    (*env)->ThrowNew(env, runtimeClazz, "An error occurred when mmap() was called - ETXTBUSY");
                    break;
                }
            */
            // EAGAIN The file has been locked, or too much memory has been locked.
            case EAGAIN:
                {        
                    // For the moment just throwing a RuntimeException
                    jclass runtimeClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
                    (*env)->ThrowNew(env, runtimeClazz, "An error occurred when mmap() was called - EAGAIN");
                    break;
                }

            // ENOMEM No memory is available.
            case ENOMEM:
                {        
                    // For the moment just throwing a RuntimeException
                    jclass runtimeClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
                    (*env)->ThrowNew(env, runtimeClazz, "An error occurred when mmap() was called - ENOMEM");
                    break;
                }
        }
    }
    return (jlong) ((int)mappedAddr);
}


/*
 * Class:     PhysicalMemoryManager
 * Method:    munmap0
 * Signature: (JJ)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_PhysicalMemoryManager_munmap0
  (JNIEnv *env, jclass clazz, jlong base, jlong size)
{
    int ret;
    ret = munmap( (void*)((int)base), size);
    if(ret != 0)
    {
        switch(errno)
        {
            // EACCES MAP_PRIVATE was asked, but fd is not open for  reading.   Or  MAP_SHARED  was
            // asked and PROT_WRITE is set, fd is not open for writing.
            /*case EACCES:
                {        
                    // For the moment just throwing a RuntimeException
                    jclass runtimeClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
                    (*env)->ThrowNew(env, runtimeClazz, "An error occurred when munmap() was called - EACCES");
                    break;
                }*/

            // EINVAL We  don't  like start or length or offset.  (E.g., they are too large, or not
            // aligned on a PAGESIZE boundary.)
            case EINVAL:
                {        
                    // For the moment just throwing a RuntimeException
                    jclass runtimeClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
                    (*env)->ThrowNew(env, runtimeClazz, "An error occurred when munmap() was called - EINVAL");
                    break;
                }

            /*case EAGAIN:
                {        
                    // For the moment just throwing a RuntimeException
                    jclass runtimeClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
                    (*env)->ThrowNew(env, runtimeClazz, "An error occurred when munmap() was called - EAGAIN");
                    break;
                }*/

            default:
                {
                    
                    jclass runtimeClazz = (*env)->FindClass(env, "java/lang/RuntimeException");
                    perror("munmap() - unknown error: ");
                    (*env)->ThrowNew(env, runtimeClazz, "An unknown error occurred when munmap() was called");
                    break;
                }
        }
    }
    return;
}

