/*
 * @(#)classes.h	1.197 00/11/08
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*
 * This file includes the definitions of interpeter's class file related
 * data structures: CVMClassBlock, CVMMethodBlock, and CVMFieldBlock.
 * Constant pool related data structures are in constantpool.h.
 */

#ifndef _INCLUDED_CLASSES_H
#define _INCLUDED_CLASSES_H

#include "javavm/include/defs.h"
#include "javavm/include/typeid.h"
#include "javavm/include/constantpool.h"

/*
 * CVMGCBitMap - this class is used to specify a bit map for a java
 * object. If a bit is set then the field at that offset is an object
 * reference which must be scanned. If the object has more than 30
 * fields then the CVMGCBitMap points to a bitmap. See gc_common.h for
 * details.
 */

union CVMGCBitMap {
    CVMUint32          map; /* low bit set indicates pointer to BigGCBitMap */
    CVMBigGCBitMap*    bigmap;
};

struct CVMBigGCBitMap {
    CVMUint16         maplen;
    CVMUint32         map[1];
};

/*
 * CVMClassBlock - The main data structure for storing information about
 * a java class. Unlike some other VM's, this data type is not an instance
 * of java/lang/Class. The java/lang/Class instance (stored in the
 * javaInstance field, is simply a 1 field java class that points
 * back to its CVMClassBlock.
 *
 * WARNING: if you change anything in this sturcture then you need to
 * fix up CVM_INIT_CLASSBLOCK.
 */
struct CVMClassBlock {
    /* bit map of object fields containing references */
    CVMGCBitMap     gcMapX;

    CVMClassTypeID  classNameX; /* see typeid.h */

    /* If the class has reached the SUPERCLASS_LOADED state, then
     * superclassX will contain an CVMClassBlock*. Otheriwse it contains
     * the typeID of the superclass.
     */
    union {
	CVMClassBlock*  superclassCb;     /* after loading superclass */
	CVMClassTypeID  superclassTypeID; /* before loading superclass */
	/*
	 * Number of mirnada methods for this class.
	 * Only valid after the class has been added to the list of
	 * classes to free. See staticsX.freeClassLink.
	 */
	CVMUint8        mirandaMethodCountX; /* during class unloading */
    } superclassX;

    union {
	CVMConstantPool* constantpoolX;
	CVMArrayInfo*    arrayInfoX;
    } cpX;

    CVMInterfaces*   interfacesX; /* null if no interfaces supported */
    CVMMethodArray*  methodsX;    /* array of class methods */
    CVMFieldArray*   fieldsX;     /* array of class fields */

    union {
	/* The statics area should have 64 bit statics grouped at the
	 * beginning so they'll all be 64 bit aligned 
	 */
	CVMJavaVal32*    statics;
	/*
	 * The freeClassLink field is used to link the class into
	 * CVMglobals.freeClassList when the class unloading code
	 * detects that the class is no longer referenced.
	 */
	CVMClassBlock*   freeClassLink;
    } staticsX;

    CVMUint16        constantpoolCountX;/* # entries in constantpool */
    CVMUint16        methodCountX;	/* # entries in methods array */
    CVMUint16        fieldCountX;       /* # entries in fields array */
    CVMUint16        methodTableCountX; /* # of entries in methodTable array */
					/* but see also CVMcbBasicTypeCode   */
    
    CVMUint16        instanceSizeX;     /* instance size in bytes */
    CVMUint8	     accessFlagsX;      /* access using CVMcbIs macro */
    CVMUint8         runtimeFlagsX;     /* For non ROM classes only.
					   ROM classes keep runtime flags
					   in CVMROMClassesRuntimeFlags[] */
    CVMInt16         romIndexX;    	/* Index into CVMROMClassesRuntimeFlags
					   > 0 <-> in ROM and has <clinit>
					   = 0 <-> in ROM and has no <clinit> 
					   < 0 <-> not in ROM :
					    -1 <-> check for <clinit>
					    -2 <-> don't check for <clinit> */
    CVMUint16        numStaticRefsX;    /* # of statics that are ref types,
					   but only for dynamic classes */
    CVMClassICell*   javaInstanceX;      /* java.lang.Class instance */
#ifdef CVM_CLASSLOADING
    CVMClassLoaderICell*  classLoaderX;  /* java.lang.ClassLoader    */
    CVMObjectICell*  protectionDomainX;  /* java.security.ProtectionDomain */
#endif

    union {
	/* If this class is currently having it's <clinit> method run, then
	 * clinitEE refers to the ee of the thread that is running the
	 * clinit method. We need to handle ROM classes differently since
	 * we can't modify the clinitEE of a ROM class at runtime.
	 */
	CVMExecEnv** eePtr;    /* for ROM classes */
	CVMExecEnv*  ee;       /* for RAM classes */
	/*
	 * Pointer to the classTable slot that points to this class.
	 * Only valid after the class has been added to the list of
	 * classes to free. See staticsX.freeClassLink. Note that this
	 * never conflicts with the use of the ee or eePtr fields since
	 * they are only used while running the <clinit>, which can't
	 * possibly happen if the class is about to be unloaded.
	 */
	CVMClassBlock** classTableSlotPtr;
    } clinitEEX;

    /* All checked exceptions for all methods in the class are grouped
       into one array to save memory. Each method will have an offset
       into the checkedExceptions array. The first entry at the offset
       is the number of checked exceptions for the method, followed by
       all checked exceptions for the method. If two method have the
       same checked exceptions, they may both have the same offset into
       the checkedExceptions array (saving even more memory). Index 0
       of checkedExcpetions always contains 0. If a method specifies
       offset 0 then it has no checked exceptions. If no class methods
       have checked exceptions, then checkedExcpetions will be NULL.
       */
    CVMUint16*	     checkedExceptionsX;

#ifdef CVM_DEBUG_CLASSINFO
    char*           sourceFileNameX;  /* source file name */
#endif
    CVMMethodBlock** methodTablePtrX;

    /* Pointer to inner classes information. NULL if there is none. */
    /* %comment cjp004 */
    CVMInnerClassesInfo* innerClassesInfoX;
};

#ifdef CVM_CLASSLOADING
#define CVMcbClassLoader(cb)	    ((cb)->classLoaderX)
#define CVMcbProtectionDomain(cb)   ((cb)->protectionDomainX)
#else
#define CVMcbClassLoader(cb)	    ((CVMClassLoaderICell*)NULL)
#define CVMcbProtectionDomain(cb)   ((CVMObjectICell*)NULL)
#endif

#define CVMcbAccessFlags(cb)        ((cb)->accessFlagsX)
#define CVMcbArrayInfo(cb)          ((cb)->cpX.arrayInfoX)
#define CVMcbCheckedExceptions(cb)  ((cb)->checkedExceptionsX)
#define CVMcbClassTableSlotPtr(cb)  ((cb)->clinitEEX.classTableSlotPtr)
#define CVMcbClassName(cb)	    ((cb)->classNameX)
#define CVMcbClinitEE(cb)           (*(CVMcbIsInROM(cb) 	\
				       ? (cb)->clinitEEX.eePtr	\
				       : &(cb)->clinitEEX.ee))
#define CVMcbConstantPool(cb)	    ((cb)->cpX.constantpoolX)
#define CVMcbConstantPoolCount(cb)   ((cb)->constantpoolCountX)
#define CVMcbFieldCount(cb) 	    ((cb)->fieldCountX)
#define CVMcbFreeClassLink(cb)      ((cb)->staticsX.freeClassLink)
#define CVMcbGcMap(cb) 	            ((cb)->gcMapX)
#define CVMcbImplementsCount(cb)					\
    (CVMcbInterfaces(cb) == NULL					\
     ? 0								\
     : (CVMcbInterfaces(cb) == CVMcbInterfaces(CVMcbSuperclass(cb))	\
	? 0 /* CVMcbInterfaces was inherited from the parent */		\
	: CVMcbInterfaces(cb)->implementsCountX))
#define CVMcbInstanceSize(cb) 	    ((cb)->instanceSizeX) 
#define CVMcbInterfacecb(cb,idx)    (CVMcbInterfaces(cb)->itable[idx]. \
				     interfaceCb)
#define CVMcbInterfaceMethodTableIndices(cb, idx)			\
     				    (CVMcbInterfaces(cb)->itable[idx].	\
				     intfInfoX.methodTableIndicesX)
#define CVMcbInterfaceMethodTableIndex(cb, idx, methodIdx)		      \
     				    (CVMcbInterfaceMethodTableIndices(cb, idx)\
				     [methodIdx])
#define CVMcbInterfaceTypeID(cb,idx) \
                                    (CVMcbInterfaces(cb)->itable[idx]. \
				     intfInfoX.interfaceTypeIDX)
#define CVMcbInterfaceCount(cb)     (CVMcbInterfaces(cb) == NULL	     \
				     ? 0 				     \
				     : CVMcbInterfaces(cb)->interfaceCountX)
#define CVMcbInterfaces(cb)         ((cb)->interfacesX)
#define CVMcbIsInROM(cb) 	    (CVMcbRomIndex(cb) >= 0)
#define CVMcbJavaInstance(cb)       ((cb)->javaInstanceX)
#define CVMcbMethodCount(cb) 	    ((cb)->methodCountX)
#define CVMcbMethodTableCount(cb)   ((cb)->methodTableCountX)
#define CVMcbMethodTablePtr(cb)     ((cb)->methodTablePtrX)
#define CVMcbMirandaMethodCount(cb) ((cb)->superclassX.mirandaMethodCountX)
#define CVMcbNumStaticRefs(cb) 	    ((cb)->numStaticRefsX)
#define CVMcbRomIndex(cb) 	    ((cb)->romIndexX)
#define CVMcbStatics(cb)            ((cb)->staticsX.statics)
#define CVMcbSuperclass(cb)  	    ((cb)->superclassX.superclassCb)
#define CVMcbSuperclassTypeID(cb)   ((cb)->superclassX.superclassTypeID)
#define CVMcbSourceFileName(cb)     ((cb)->sourceFileNameX)

#define CVMcbMethodSlot(cb, idx) \
    (&CVMcbMethods(cb)->ranges[(idx) >> 8].mb[(idx) & 0xff])
#define CVMcbFieldSlot(cb, idx)  \
    (&CVMcbFields(cb)->ranges[(idx) >> 8].fb[(idx) & 0xff])

#define CVMcbMethodTableSlot(cb, idx)    \
    (CVMcbMethodTablePtr(cb)[idx])
#define CVMobjMethodTableSlot(obj, idx)  \
    (CVMcbMethodTableSlot(CVMobjectGetClass(obj), idx))

/*
 * This is what CVMobjMethodTableSlot() used to look like before we made
 * the methodTable a ptr field rather than putting the entire methodTable
 * at the end of the cb (and leaving arrays without any methodTable).
 */
#if 0
#define CVMobjMethodTableSlot(obj, idx)					\
    (CVMisArrayObject(obj) ?						\
     CVMcbMethodTableSlot(CVMsystemClass(java_lang_Object), idx) :	\
     CVMcbMethodTableSlot(CVMobjectGetClass(obj), idx))
#endif

/* Only the classloader and the above macros should use these macros. */
#define CVMcbMethods(cb)            ((cb)->methodsX)
#define CVMcbFields(cb)             ((cb)->fieldsX)

/* We need "setter" macros for these 2 fields because the existing access
 * macros for them cannot be used as lvals.
 */
#define CVMcbSetInterfaceCount(cb, count)  \
    CVMcbInterfaces(cb)->interfaceCountX = count;
#define CVMcbSetImplementsCount(cb, count) \
    CVMcbInterfaces(cb)->implementsCountX = count;

/*
 * Runtime flags for classes:
 *
 * Dynamically created CVMClassBlocks store runtime flags in the
 * cb->runtimeFlagsX field. However, romized classes can't do this
 * because they are read-only. For this reason romized classes use a
 * writeable array called CVMROMClassesRuntimeFlags for storing runtime flags.
 *
 * In a number of places we cast away the constness of a romized
 * CVMClassBlock in order to store a pointer to the CVMClassBlock
 * in a field that is also used to point to non-const CVMClassBlocks.
 * This is safe because an invariant of the VM is that we won't write
 * into any CVMClassBlock that is const. This is easy to ensure because
 * the only time we ever write into an CVMClassBlock (after it is
 * initialized) is to update the rutime flags.
 *
 * Romized CVMClassBlocks use cb->romIndex to index into the
 * CVMROMClassesRuntimeFlags writeable array of flags. Dynamically created
 * CVMClassBlocks access the cb->runtimeFlagsX field directly.
 * 
 * For now, only romized classes with static initializers have
 * an entry in CVMROMClassesRuntimeFlags. All others use entry #0 because it is
 * assumed that they will not need to dynamically set any flags.
 */

extern CVMUint8 CVMROMClassesRuntimeFlags[];
#define CVM_CLASS_ERROR	              0x01 /* CVMclassLoadSuperclasses() or
					    * <clinit> caused an error. Look
					    * at SUPERCLASS_LOADED flag to
					    * determine which. */
#define CVM_CLASS_SUPERCLASS_LOADED   0x02 /* superclass and interfaces have
					    * been loaded */
#define CVM_CLASS_LINKED              0x04 /* interfaces, methods and fields
					    * have ben preprared. */
#define CVM_CLASS_INITIALIZED         0x08 /* static initializers have been
					    * run. */
#define CVM_CLASS_VERIFIED            0x10 /* class is verified. */
#define CVM_CLASS_GC_SCANNED          0x20 /* class has been scanned by GC */
#define CVM_CLASS_INHERITED_INTERFACES 0x40/* dynamically loaded class did not
					    * declare at least one interface,
					    * so its CVMcbInterfaces belongs to
					    * its superclass and should not be
					    * freed when the class is unloaded
					    */
#define CVM_CLASS_FIELDS_PREPARED     0x80 /* fields have been prepared. */

#define CVMcbRuntimeFlags(cb) (*(CVMcbIsInROM(cb) ?   \
    &CVMROMClassesRuntimeFlags[CVMcbRomIndex(cb)] : &(cb)->runtimeFlagsX))
#define CVMcbCheckRuntimeFlag(cb, flag)               \
    ((CVMcbRuntimeFlags(cb) & CVM_CLASS_##flag) != 0)
#define CVMcbCheckROMRuntimeFlag(cb, flag)            \
    (CVMassert(CVMcbIsInROM((cb))), 		      \
     (CVMROMClassesRuntimeFlags[CVMcbRomIndex((cb))] & CVM_CLASS_##flag) != 0)
/* We must become unsafe to avoid a race condition with gc setting or
 * clearing the SCANNED bit. */
#define CVMcbSetRuntimeFlag(cb, ee_, flag) {			\
    CVMD_gcUnsafeExec(ee_, {				\
        CVMcbRuntimeFlags(cb) |= CVM_CLASS_##flag;	\
    });							\
}
#define CVMcbClearRuntimeFlag(cb, flag)               \
    CVMcbRuntimeFlags(cb) &= ~CVM_CLASS_##flag;

#define CVMcbGcScanned(cb) \
    ((((cb)->runtimeFlagsX) & CVM_CLASS_GC_SCANNED) != 0)

#define CVMcbSetGcScanned(cb) \
    ((cb)->runtimeFlagsX) |= CVM_CLASS_GC_SCANNED

#define CVMcbClearGcScanned(cb) \
    ((cb)->runtimeFlagsX) &= ~CVM_CLASS_GC_SCANNED


#define CVMcbSetAccessFlag(cb, flag)                 \
    CVMcbAccessFlags(cb) |= CVM_CLASS_ACC_##flag;

/*
 * There are exactly nine primitive classes.
 * For each of them, the methodTableCount field contains a type code
 * that is used to table-lookup more detailed type information. See
 * the data structures declared in basictypes.h.
 */
#define CVMcbBasicTypeCode(cb)	((CVMBasicType) CVMcbMethodTableCount(cb))

/*
 * CVMInterfaces - Contains information about all interfaces that a class
 * or interface implements.
 *
 * For cb's corresponding to Java classes, it contains entries for all of
 * the interfaces the class implements, directly or indirectly; it is a
 * flattened version of the interface hierarchy. 
 *
 * For class blocks corresponding to Java interfaces, the interface array 
 * contains the same information, as well as an entry for this interface.
 * The fact that interface cb's contain entries in the interface table 
 * for themselves as well as a flattened version of the interface
 * hierarchy, makes class preparation easier because all that needs to be
 * done to create a class's interface list is concatenate those of all of 
 * the interfaces it implements. This does pose some problems for certain
 * reflection-related traversals. See reflect.c for more details.
 * 
 * INVARIANT: the first "implementsCount" entries in the itable
 * are those which this class declares it implements, in
 * left-to-right, order as written in the Java file (according to
 * the documentation for java.lang.Class.getInterfaces().) The remaining
 * entries are for super classes.
 *
 * The methodTableIndices field points to an array of indices that converts
 * the index of the method in the interface's methods array to the index of
 * the method in the class' methodTable array. It is not setup until the
 * class reaches the LINKED state.
 *
 * When a class is first loaded, the "interfaceTypeID" field contains the
 * typeID for the interface. When the class has its super clases loaded
 * (and moves to the SUPERCLASS_LOADED state), the interfaceTypeID is used
 * to lookup the cb for the "interfaceCb" field. When the class moves to
 * the LINKED state, the interfaceTypeID is overwritten with the
 * the methodTableIndices field.
 *
 * If a class does not declare any interfaces, but does inherit one or
 * more from its superclass, then it will simply use its parents's
 * CVMcbInterfaces structure. The class' INHERITED_INTERFACES flag will
 * be set in this case so we know not to free the CVMcbInterfaces when
 * the class is unloaded. Since this flag is not set for romized classes,
 * the CVMcbImplementsCount() macro has been modified to check the
 * if the CVMcbInterfaces belongs to the parent class and return 0 if it does.
 */

struct CVMInterfaceTable {
    CVMClassBlock* interfaceCb;        /* ClassBlock of the interface */
    union {
	CVMUint16*     methodTableIndicesX; /* see comment above for details */
	CVMClassTypeID interfaceTypeIDX;    /* class typeID's of interface */
    } intfInfoX;
};

struct CVMInterfaces {
    CVMUint16      interfaceCountX;  /* # entries in itable */
    CVMUint16      implementsCountX; /* # of entries the class implements */
    CVMInterfaceTable itable[1];  /* info for all of the class' interfaces. */
};

/*
 * MethodBlock - Java, native, and interface methods all require some sort
 * of per method data structure, but they also each have their own unique 
 * requirements. For this reason MethodBlock only contains the information
 * needed by all three and a separate descriptor record is used to contain
 * information specific to the type of method. This saves us about 16 bytes
 * per interface method and 12 bytes per native method.
 *
 * Another optimization is to move the ClassBlock* out of the MethodBlock
 * and only save one copy. Since the MethodBlocks are stored in an array,
 * we just put the ClassBlock* in the 4 bytes that precede the array.
 * The MethodBlock.methodIndex field contains the method's index into this
 * array so it can then locate its ClassBlock*. This is similar to what
 * we do with FieldBlocks.
 */

struct CVMMethodBlock {
    CVMMethodTypeID  nameAndTypeIDX;/* name and type */
    CVMUint16  methodTableIndexX;/* index of this method into cb.methodTable */
    CVMUint8   argsSizeX;     /* calculated based on signature and if static */
    CVMUint8   invokerIdxX;   /* index of method invoker to use */
    CVMUint8   accessFlagsX;  /* could be merged with invokerIdx */
    CVMUint8   methodIndexX;  /* index into CB.methods.mb[] of this method */
    CVMUint16  checkedExceptionsOffsetX; /* index into CB.checkedExceptions */
    union {   /* code specific method information */
	CVMJavaMethodDescriptor* jmd;        /* more info for java methods */
	CVMUint8*  nativeCode;       /* code for native methods */
	CVMUint32  methodSlotIndex;  /* for interface methods, the real 16-bit
				      * index of the method. This could be a
				      * CVMUint16, but it's easier for JCC if
				      * it's an CVMUint32.
				      */
	CVMMethodBlock* interfaceMb; /* for miranda methods, the interface mb
				      * the miranda methods was created for.
				      * Sometimes NULL. 
				      * See CVMclassPrepareInterfaces.
				      */
    } codeX;
};

#define CVMmbIsMiranda(mb)						\
    (CVMmbInvokerIdx(mb) == CVM_INVOKE_NONPUBLIC_MIRANDA_METHOD ||	\
     CVMmbInvokerIdx(mb) == CVM_INVOKE_MISSINGINTERFACE_MIRANDA_METHOD)
#define CVMmbMirandaIsNonPublic(mb) \
    (CVMmbInvokerIdx(mb) == CVM_INVOKE_NONPUBLIC_MIRANDA_METHOD)

#define CVMmbIsJava(mb)    (!CVMmbIs(mb, NATIVE) && !CVMmbIs(mb, ABSTRACT))

struct CVMMethodRange {
    CVMClassBlock*     cb; /* ClassBlock for all methods in this MethodRange */
    CVMMethodBlock     mb[256];
};

struct CVMMethodArray {
    CVMMethodRange ranges[1];
};

#define CVMmbMethodRange(mb)       \
    ((CVMMethodRange*) ((CVMClassBlock**)((mb) - CVMmbMethodIndex(mb)) - 1))

#ifdef CVM_DEBUG_ASSERTS
#define CVMmbJmd(mb)	      (*(CVMassert(CVMmbIsJava(mb)),	\
				 &mb->codeX.jmd))
#define CVMmbNativeCode(mb)   (*(CVMassert(CVMmbIs(mb, NATIVE)),	\
				 &mb->codeX.nativeCode))
#define CVMmbMethodSlotIndex(mb)  				\
    (*(CVMassert(CVMmbIs(mb, ABSTRACT) && 			\
		 CVMcbIs(CVMmbClassBlock(mb), INTERFACE)),	\
       &(mb)->codeX.methodSlotIndex))
#define CVMmbMirandaInterfaceMb(mb)   (*(CVMassert(CVMmbIsMiranda(mb)),	\
					 &mb->codeX.interfaceMb))
#else
#define CVMmbJmd(mb)                   ((mb)->codeX.jmd)
#define CVMmbNativeCode(mb)            ((mb)->codeX.nativeCode)
#define CVMmbMethodSlotIndex(mb)       ((mb)->codeX.methodSlotIndex)
#define CVMmbMirandaInterfaceMb(mb)    ((mb)->codeX.interfaceMb)
#endif

#define CVMmbAccessFlags(mb)           ((mb)->accessFlagsX)
#define CVMmbArgsSize(mb)              ((mb)->argsSizeX)
#define CVMmbCapacity(mb)	       (CVMjmdMaxCapacity(CVMmbJmd(mb)))
/* This macro returns either NULL or an CVMCheckedExceptions* which is
   guaranteed to have a non-zero numExceptions field. */
#define CVMmbCheckedExceptions(mb)     ((CVMCheckedExceptions*)	\
    ((CVMmbCheckedExceptionsOffset(mb) == 0) ? NULL :		\
     &CVMcbCheckedExceptions(CVMmbClassBlock(mb))		\
			     [CVMmbCheckedExceptionsOffset(mb)]))
#define CVMmbCheckedExceptionsOffset(mb)  \
                                       ((mb)->checkedExceptionsOffsetX) 
#define CVMmbClassBlock(mb)            (CVMmbMethodRange(mb)->cb)
#define CVMmbCodeLength(mb)	       (CVMjmdCodeLength(CVMmbJmd(mb)))
#define CVMmbExceptionTable(mb)        (CVMjmdExceptionTable(CVMmbJmd(mb)))
#define CVMmbExceptionTableLength(mb)  \
     (CVMjmdExceptionTableLength(CVMmbJmd(mb)))
#define CVMmbInvokerIdx(mb)            (mb->invokerIdxX)
#define CVMmbJavaCode(mb)              (CVMjmdCode(CVMmbJmd(mb)))
#define CVMmbLineNumberTable(mb)       (CVMjmdLineNumberTable(CVMmbJmd(mb)))
#define CVMmbLineNumberTableLength(mb) \
     (CVMjmdLineNumberTableLength(CVMmbJmd(mb)))
#define CVMmbLocalVariableTable(mb)    (CVMjmdLocalVariableTable(CVMmbJmd(mb)))
#define CVMmbLocalVariableTableLength(mb) \
     (CVMjmdLocalVariableTableLength(CVMmbJmd(mb)))
#define CVMmbMaxLocals(mb)    	       (CVMjmdMaxLocals(CVMmbJmd(mb)))
#define CVMmbMethodIndex(mb)           (mb->methodIndexX)
#define CVMmbMethodTableIndex(mb)      (mb->methodTableIndexX)
#define CVMmbNameAndTypeID(mb)         (mb->nameAndTypeIDX)


typedef CVMUint16 CVMCheckedException;

struct CVMCheckedExceptions {
    CVMUint16             numExceptions;
    CVMCheckedException   exceptions[1];
};

struct CVMJavaMethodDescriptor {
    /* An optimization is made to speed up method calls.
       The total size of the frame is saved in "capacity"
       rather than just saving max_stack. This
       saves us some calcuating when a frame is pushed.
       */
    CVMUint16      maxLocalsX;    /* including arguments */
    CVMUint16	   flagsX;	  /* see CVM_JMD_xxx     */
    /* %comment fresko001 */  
    CVMUint32      capacityX;     /* maxLocals + sizeof(Frame) + max_stack */
    CVMUint16      exceptionTableLengthX;
    CVMUint16      codeLengthX;
#if 0
    CVMUint32      inliningX;         /* possible inlining code */
#endif
    /*
     * Indirect cell to stack maps. This structure is build 
     * at runtime, but we want the CVMJavaMethodDescriptor to 
     * be read-only.
     */
    CVMStackMaps**  stackMapsX;

#ifdef CVM_DEBUG_CLASSINFO
    CVMUint16      lineNumberTableLengthX;
    CVMUint16      localVariableTableLengthX;
#endif
    /* The byte codes, exception handlers, line number table, and
       local variable table all immediately follow the
       JavaMethodDescriptor data structure. Their offsets need to
       be calculated.

       CVMUint8 code[codeLength];
       CVMExceptionHandler exceptionTable[exceptionTableLength];
       CVMLineNumberEntry lineNumberTable[lineNumberTableLength];
       CVMLocalVariableEntry localVariableTable[localVariableTableLength];
       */
};

/*
 * A stackmap entry for a Java method
 */
struct CVMStackMapEntry {
    CVMUint16 pc;
    CVMUint16 state[1];
};

/*
 * The stackmaps for the method
 */
struct CVMStackMaps {
    CVMUint32         noGcPoints;
    CVMStackMapEntry  smEntries[1];
};

#define CVMjmdFlags(jmd)                    ((jmd)->flagsX)
#define CVMjmdCapacity(jmd)                 ((jmd)->capacityX)
#define CVMjmdCodeLength(jmd)               ((jmd)->codeLengthX)
#define CVMjmdStackMaps(jmd)                ((jmd)->stackMapsX)
#define CVMjmdExceptionTableLength(jmd)     ((jmd)->exceptionTableLengthX)
#define CVMjmdLineNumberTableLength(jmd)    ((jmd)->lineNumberTableLengthX)
#define CVMjmdLocalVariableTableLength(jmd) ((jmd)->localVariableTableLengthX)
#define CVMjmdMaxLocals(jmd)                ((jmd)->maxLocalsX)

/*
 * In case somebody outside of the interpreter wants to know.
 */
#define CVMjmdMaxStack(jmd) \
    ((jmd)->capacityX - (jmd)->maxLocalsX - \
     CVM_JAVAFRAME_SIZE / sizeof(CVMStackVal32))

#define CVMjmdCode(jmd) \
    ((CVMUint8*)jmd + CVMjmdCodeOffset(jmd))
#define CVMjmdExceptionTable(jmd) \
    ((CVMExceptionHandler*) \
     ((CVMUint8*)jmd + CVMjmdExceptionTableOffset(jmd)))
#define CVMjmdLineNumberTable(jmd) \
    ((CVMLineNumberEntry*) \
     ((CVMUint8*)jmd + CVMjmdLineNumberTableOffset(jmd)))
#define CVMjmdLocalVariableTable(jmd) \
    ((CVMLocalVariableEntry*) \
     ((CVMUint8*)jmd + CVMjmdLocalVariableTableOffset(jmd)))

#define CVMjmdCodeOffset(jmd)			\
    (sizeof(CVMJavaMethodDescriptor))
#define CVMjmdExceptionTableOffset(jmd)		\
    (CVMjmdCodeOffset(jmd) + ((CVMjmdCodeLength(jmd) + 3) & ~3))
#define CVMjmdLineNumberTableOffset(jmd)			\
    (CVMjmdExceptionTableOffset(jmd) + 				\
     (CVMjmdExceptionTableLength(jmd) * sizeof(CVMExceptionHandler)))
#define CVMjmdLocalVariableTableOffset(jmd)			\
    (CVMjmdLineNumberTableOffset(jmd) + 			\
     (CVMjmdLineNumberTableLength(jmd) * sizeof(CVMLineNumberEntry)))

/*
 * The total size of a JavaMethodDescriptor and all its trailing data
 * structures
 */
#ifdef CVM_DEBUG_CLASSINFO
#define CVMjmdTotalSize(jmd) \
     (CVMjmdLocalVariableTableOffset(jmd) + \
      CVMjmdLocalVariableTableLength(jmd) * sizeof(CVMLocalVariableEntry))
#else
#define CVMjmdTotalSize(jmd) \
     (CVMjmdExceptionTableOffset(jmd) + \
      CVMjmdExceptionTableLength(jmd) * sizeof(CVMExceptionHandler))
#endif

/*
 * CVMJavaMethodDescriptor-related flags.
 * For tagging methods that the preloader believes might need to be
 * rewritten, and those that have been rewritten by the stackmap
 * disambiguator.
 * Also, an indication that the method needs
 * 4-byte alignment (because it contains a switch).
 */
#define CVM_JMD_MAY_NEED_REWRITE	1
#define CVM_JMD_NEEDS_ALIGNMENT		2
#define CVM_JMD_DID_REWRITE		4
#define CVM_JMD_STRICT			8

/*
 * When the stackmap disambiguator rewrites a Java method, it replaces
 * the entire CVMJavaMethodDescriptor with a newly allocated one. In this
 * case, it must make room for two extra pieces of information: the
 * CVMStackMaps* pointed to by CVMjmdStackMaps(), and the address of the original
 * method descriptor.
 *
 * In truth, these are only needed very occasionally, the first when the
 * space for the original CVMStackMaps* might be deallocated (as when rewriting
 * clinit code), and the second when rewriting the first Java method of a class
 * so that the start of the original allocation can be found!.
 *
 * This will be placed at the end of the new allocation, and CVMjmdStackMaps()
 * will address its first entry.
 */
typedef struct CVMJavaMethodExtension {
    CVMStackMaps*		    mapp;
    struct CVMJavaMethodDescriptor* originalJmd; 
} CVMJavaMethodExtension;


struct CVMExceptionHandler {
    CVMUint16          startpc;
    CVMUint16          endpc;
    CVMUint16          handlerpc;
    CVMUint16          catchtype;
};

#ifdef CVM_DEBUG_CLASSINFO

struct CVMLineNumberEntry {
    CVMUint16          startpc;
    CVMUint16          lineNumber;
};

struct CVMLocalVariableEntry {
    CVMUint16          startpc;
    CVMUint16          length;
    CVMUint16          index;
    CVMFieldTypeID     nameAndTypeID; /* name and type */
};

#endif


/*
 * FieldBlocks have been optimized to only occupy 8 bytes. They are all located
 * in an array with the 4 byte entry just before the array containing the
 * ClassBlock* for all of the fields in the array. That data structure
 * that contains the ClassBlock* and the array is the CVMFieldRange.
 *
 * A FieldBlock locates its ClassBlock* by looking backwards FieldBlock.fbIndex
 * entries to find the start of the array and then another 4 bytes to 
 * locate the ClassBlock*. Since cbInfo is only 8 bits, the
 * CVMFieldRange must be repeated every 256 enties.
 */

struct CVMFieldBlock {
    CVMFieldTypeID nameAndTypeIDX; /* name and type */
    CVMUint8       accessFlagsX;
    CVMUint8       fbIndexX;   /* fields index into FieldRange.fb */
    CVMUint16      offsetX;    /* offset in words into the object for instance
				* variables or offset in words into
				* ClassBlock.statics for static variables
				*/
};

struct CVMFieldRange {
    CVMClassBlock*     cb; /* ClassBlock for all fields in this FieldRange */
    CVMFieldBlock      fb[256];
};

struct CVMFieldArray {
    CVMFieldRange ranges[1];
};

#define CVMfbFieldRange(fb)       \
    ((CVMFieldRange*) ((CVMClassBlock**)((fb) - CVMfbIndex(fb)) - 1))

#define CVMfbAccessFlags(fb)      ((fb)->accessFlagsX)
#define CVMfbClassBlock(fb)       (CVMfbFieldRange(fb)->cb)
#define CVMfbIndex(fb)            ((fb)->fbIndexX)
#define CVMfbNameAndTypeID(fb)    ((fb)->nameAndTypeIDX)
#define CVMfbOffset(fb)           ((fb)->offsetX)
#define CVMfbStaticField(fb)      \
    (CVMcbStatics(CVMfbClassBlock(fb))[CVMfbOffset(fb)])

 
/*
 * This indicates whether this field occupies one word or two,
 * or if it is a reference type. For use by {getfield,putfield}_quick_w.
 */
#define CVMfbIsDoubleWord(fb) \
    CVMtypeidFieldIsDoubleword(CVMfbNameAndTypeID(fb))
#define CVMfbIsRef(fb) \
    CVMtypeidFieldIsRef(CVMfbNameAndTypeID(fb))

/*
 * CVMInnerClassesInfo for storing information about inner classes.
 */

#define CVMcbInnerClassesInfo(cb)      ((cb)->innerClassesInfoX)
#define CVMcbInnerClassesInfoCount(cb) (CVMcbInnerClassesInfo(cb) == NULL    \
					? 0				    \
					: CVMcbInnerClassesInfo(cb)->count)
#define CVMcbInnerClassInfo(cb, idx)   (&CVMcbInnerClassesInfo(cb)->info[idx])

struct CVMInnerClassInfo {
    CVMUint16 innerClassIndex;
    CVMUint16 outerClassIndex;
    CVMUint16 unused; /* innerNameIndex is not supported */
    CVMUint16 innerClassAccessFlags;
};

struct CVMInnerClassesInfo {
    CVMUint32 count;
    CVMInnerClassInfo info[1];
};

/*
 * Method invoker indices. Stored in CVMMethodBlock.invokerIdx field.
 * Used to determine what code will handle a method's invocation.
 *
 * WARNING: The exact order and values of these macros may be important.
 * [ see jcc/CVMConst.java and executejava.c]
 */
#define CVM_INVOKE_JAVA_METHOD	       	0
#define CVM_INVOKE_JAVA_SYNC_METHOD    	1
#define CVM_INVOKE_CVM_METHOD     	2
#define CVM_INVOKE_JNI_METHOD		3
#define CVM_INVOKE_JNI_SYNC_METHOD	4
#define CVM_INVOKE_ABSTRACT_METHOD     	5
#define CVM_INVOKE_NONPUBLIC_MIRANDA_METHOD     	7
#define CVM_INVOKE_MISSINGINTERFACE_MIRANDA_METHOD     	8
#ifdef CVM_CLASSLOADING
#define CVM_INVOKE_LAZY_JNI_METHOD     	9
#endif

/* 
 * Class, field, and method access and modifier flags. Some of these
 * values are different from the red book so they'll fit in one byte.
 * CVMreflectCheckAccess (reflect.c) and CVMverifyMemberAccess2 
 * (interpreter.c) require the following invariants:
 *   CVM_METHOD_ACC_PUBLIC == CVM_FIELD_ACC_PUBLIC
 *   CVM_METHOD_ACC_PROTECTED == CVM_FIELD_ACC_PROTECTED
 *   CVM_METHOD_ACC_PRIVATE == CVM_FIELD_ACC_PRIVATE
 * If these are not met, some of the code for reflection and object
 * serialization will have to be reworked.)
 */
#define CVM_CLASS_ACC_PUBLIC        0x01 /* visible to everyone */
#define CVM_CLASS_ACC_PRIMITIVE     0x02 /* primitive type class */
#define CVM_CLASS_ACC_FINALIZABLE   0x04 /* has non-trivial finalizer */
#define CVM_CLASS_ACC_REFERENCE     0x08 /* is a flavor of weak reference */
#define CVM_CLASS_ACC_FINAL         0x10 /* no further subclassing */
#define CVM_CLASS_ACC_SUPER         0x20 /* funky handling of invokespecial */
#define CVM_CLASS_ACC_INTERFACE     0x40 /* class is an interface */
#define CVM_CLASS_ACC_ABSTRACT      0x80 /* may not be instantiated */

#define CVM_FIELD_ACC_PUBLIC        0x01 /* visible to everyone */
#define CVM_FIELD_ACC_PRIVATE       0x02 /* visible only to defining class */
#define CVM_FIELD_ACC_PROTECTED     0x04 /* visible to subclasses */
#define CVM_FIELD_ACC_STATIC        0x08 /* instance variable is static */
#define CVM_FIELD_ACC_FINAL         0x10 /* no subclassing/overriding */
#define CVM_FIELD_ACC_VOLATILE      0x40 /* cannot cache in registers */
#define CVM_FIELD_ACC_TRANSIENT     0x80 /* not persistent */

#define CVM_METHOD_ACC_PUBLIC       CVM_FIELD_ACC_PUBLIC     /* invariant */
#define CVM_METHOD_ACC_PRIVATE      CVM_FIELD_ACC_PRIVATE    /* invariant */
#define CVM_METHOD_ACC_PROTECTED    CVM_FIELD_ACC_PROTECTED  /* invariant */
#define CVM_METHOD_ACC_STATIC       0x08 /* method is static */
#define CVM_METHOD_ACC_FINAL        0x10 /* no further overriding */
#define CVM_METHOD_ACC_SYNCHRONIZED 0x20 /* wrap method call in monitor lock */
#define CVM_METHOD_ACC_NATIVE       0x40 /* implemented in C */
#define CVM_METHOD_ACC_ABSTRACT     0x80 /* no definition provided */

/*
 * Test for specific access flags 
 */
#define CVMprivateThingIs(thing_, what_, flag_) \
    ((((thing_)->accessFlagsX) & CVM_##what_##_ACC_##flag_) != 0)

#define CVMfbIs(fb_, flag_) \
    CVMprivateThingIs(fb_, FIELD, flag_)

#define CVMmbIs(mb_, flag_) \
    CVMprivateThingIs(mb_, METHOD, flag_)

#define CVMcbIs(cb_, flag_) \
    CVMprivateThingIs(cb_, CLASS, flag_)


/*
 * Macros for initializing, and in some cases declaring, CVMClassBlocks,
 * CVMMethodBlocks, and CVMFieldBlocks.
 */

#ifdef CVM_DEBUG_CLASSINFO
#define CVM_INIT_SOURCEFILENAME(sourceFileName) \
    sourceFileName,
#define CVM_INIT_LINENUMERTABLELENGTH(lineNumberTableLength) \
    lineNumberTableLength,
#define CVM_INIT_LOCALVARIABLETABLELENGTH(localVariableTableLength) \
    localVariableTableLength,
#else
#define CVM_INIT_SOURCEFILENAME(sourceFileName) 
#define CVM_INIT_LINENUMERTABLELENGTH(lineNumberTableLength)
#define CVM_INIT_LOCALVARIABLETABLELENGTH(localVariableTableLength)
#endif

#ifdef CVM_CLASSLOADING
#define CVM_INIT_CLASSLOADING_FIELDS \
    0, 0,    /* classLoader and protectionDomain */
#else
#define CVM_INIT_CLASSLOADING_FIELDS 
#endif

#define CVM_INIT_CLASSBLOCK(gcMap, classname, superclass, constantpool,	\
                        interfaces, methods, fields, statics,		\
			constantpool_count, method_count,		\
			field_count, method_table_count, instance_size,	\
			access_flags, romIndex,	javaInstance, clinitEE, \
			checkedExceptions, sourceFileName,		\
			methodTablePtr, innerClassesInfo)	       	\
			{{gcMap}, classname,				\
			  {(CVMClassBlock*)superclass},			\
			  {(CVMConstantPool*)constantpool},		\
			  (CVMInterfaces*)interfaces,			\
			  (CVMMethodArray *)methods,			\
			  (CVMFieldArray *)fields, {statics},		\
			  constantpool_count, method_count,		\
			  field_count, method_table_count,		\
			  instance_size,				\
			  access_flags, 0, romIndex, 0,			\
			  (CVMClassICell*)javaInstance,			\
			  CVM_INIT_CLASSLOADING_FIELDS                  \
			  {clinitEE}, (CVMUint16*)checkedExceptions,	\
			  CVM_INIT_SOURCEFILENAME(sourceFileName)	\
			  (CVMMethodBlock**)methodTablePtr,		\
			  (CVMInnerClassesInfo*)innerClassesInfo}

#define CVM_INIT_METHODBLOCK(nameAndTypeID, methodTableIndex,		\
			     argsSize, invokerIdx, accessFlags,		\
			     methodIndex, checkedExceptionsOffset,	\
			     code)					\
    {									\
	nameAndTypeID, methodTableIndex,				\
	argsSize, invokerIdx,						\
	accessFlags, methodIndex,					\
	checkedExceptionsOffset, {(CVMJavaMethodDescriptor*)code}	\
    }

#define CVM_INIT_JAVA_METHOD_DESCRIPTOR(				\
                max_stack, maxLocals, flags,				\
                exceptionTableLength, codeLength,			\
                stackMaps, 						\
		lineNumberTableLength, localVariableTableLength)	\
    {									\
	maxLocals, flags,						\
        CVM_JAVAFRAME_SIZE / sizeof(CVMStackVal32) +			\
	    maxLocals + max_stack,					\
        exceptionTableLength, codeLength,				\
        stackMaps,							\
	CVM_INIT_LINENUMERTABLELENGTH(lineNumberTableLength)		\
	CVM_INIT_LOCALVARIABLETABLELENGTH(localVariableTableLength)	\
    }

#define CVM_INIT_FIELDBLOCK(nameAndTypeID, accessFlags, fbIndex, offset) \
    {									 \
	  nameAndTypeID, accessFlags, fbIndex, offset			 \
    }

/***********************
 * API's
 ***********************/

/*
 * Class lookup by name or typeID. 
 *   -The "init" flag indicates if static initializers should be run.
 *   -The "throwError" flag will cause NoClassDefFoundError to be thrown
 *    instead of ClassNotFoundException.
 *   -An exception os thrown if a NULL result is returned.
 */

extern CVMClassBlock*
CVMclassLookupByNameFromClass(CVMExecEnv* ee, const char* name,
			      CVMBool init, CVMClassBlock* fromClass);

extern CVMClassBlock*
CVMclassLookupByTypeFromClass(CVMExecEnv* ee, CVMClassTypeID typeID, 
			      CVMBool init, CVMClassBlock* fromClass);

extern CVMClassBlock*
CVMclassLookupByNameFromClassLoader(CVMExecEnv* ee, const char* name, 
				    CVMBool init,
				    CVMClassLoaderICell* loader, 
				    CVMObjectICell* pd,
				    CVMBool throwError);

extern CVMClassBlock*
CVMclassLookupByTypeFromClassLoader(CVMExecEnv* ee, CVMClassTypeID typeID, 
				    CVMBool init,
				    CVMClassLoaderICell* loader, 
				    CVMObjectICell* pd,
				    CVMBool throwError);

/*
 * Get an array class with elements of type elemCb
 */
extern CVMClassBlock*
CVMclassGetArrayOf(CVMExecEnv* ee, CVMClassBlock* elemCb);

/*
 * Look for public or declared method of type 'tid'
 */
extern CVMMethodBlock*
CVMclassGetMethodBlock(const CVMClassBlock* cb, const CVMMethodTypeID tid,
		       CVMBool isStatic);

#define CVMclassGetNonstaticMethodBlock(cb, tid)	\
	CVMclassGetMethodBlock(cb, tid, CVM_FALSE)
#define CVMclassGetStaticMethodBlock(cb, tid)		\
	CVMclassGetMethodBlock(cb, tid, CVM_TRUE)

/*
 * Look for declared method of type 'tid'
 */
CVMMethodBlock*
CVMclassGetDeclaredMethodBlockFromTID(const CVMClassBlock* cb,
				      CVMMethodTypeID tid);

/*
 * Look for declared method of given name and signature
 */
extern CVMMethodBlock*
CVMclassGetDeclaredMethodBlock(CVMExecEnv *ee,
			       const CVMClassBlock* cb,
			       const char *name,
			       const char *sig);
/*
 * Look for public or declared field of type 'tid'
 */
extern CVMFieldBlock*
CVMclassGetFieldBlock(const CVMClassBlock* cb, const CVMFieldTypeID tid, 
		      CVMBool isStatic);

#define CVMclassGetNonstaticFieldBlock(cb, tid)		\
	CVMclassGetFieldBlock(cb, tid, CVM_FALSE)
#define CVMclassGetStaticFieldBlock(cb, tid)		\
	CVMclassGetFieldBlock(cb, tid, CVM_TRUE)

/*
 * Load a class' hierarchy and resolve symbolic links to its superclass
 * and interfaces.
 */
#ifdef CVM_CLASSLOADING
extern CVMBool
CVMclassLoadSuperclasses(CVMExecEnv* ee, CVMClassBlock* cb);
#endif

/*
 * Load a class using the specified class loader.
 */
#ifdef CVM_CLASSLOADING
extern CVMClassBlock*
CVMclassLoadClass(CVMExecEnv* ee, CVMClassLoaderICell* loader, 
		  const char* classname, CVMClassTypeID classTypeID);
#endif

/*
 * Given an array of bytes, create an CVMClassBlock.
 */
#ifdef CVM_CLASSLOADING
extern CVMClassBlock*
CVMclassCreateInternalClass(CVMExecEnv* ee,
			    const CVMUint8* externalClass, 
			    CVMUint32 classSize, 
			    CVMClassLoaderICell* loader, 
			    const char* classname,
			    const char* dirNameOrZipFileName);
#endif

/*
 * Free a Class' memory. Called by the gc when it detects that there
 * are no references to the class.
 */
extern void
CVMclassFree(CVMExecEnv* ee, CVMClassBlock* cb);

/*
 * Free up a Class' JavaMethodDescriptors and associated stack maps.
 * For dynamic classes this is called by CVMclassFree, above,
 * but the preloader needs it to deallocate dynamic memory associated
 * with preloaded classes, such as the stackmaps.
 */
extern void
CVMclassFreeJavaMethods( CVMClassBlock* cb, CVMBool isPreloaded );

/*
 * Frees up the CVMFieldTypeID's in the LocalVariableTable.
 */
extern void
CVMclassFreeLocalVariableTableFieldIDs(CVMExecEnv* ee, CVMMethodBlock* mb);

/*
 * Do all the class unloading. Called after gc has completed its scan.
 */
#ifdef CVM_CLASSLOADING
extern void
CVMclassDoClassUnloadingPass1(CVMExecEnv* ee,
			      CVMRefLivenessQueryFunc isLive,
			      void* isLiveData,
			      CVMRefCallbackFunc transitiveScanner,
			      void* transitiveScannerData,
			      CVMGCOptions* gcOpts);

extern void
CVMclassDoClassUnloadingPass2(CVMExecEnv* ee);
#endif

/*
 * CVMclassVerify - Verify a class, including its bytecodes.
 */
#ifdef CVM_CLASSLOADING

extern CVMBool
CVMclassVerify(CVMExecEnv* ee, CVMClassBlock* cb);

enum { CVM_VERIFY_NONE = 0, CVM_VERIFY_REMOTE, CVM_VERIFY_ALL };

#ifdef CVM_TRUSTED_CLASSLOADERS
#define CVM_NEED_VERIFY(cond) (CVM_FALSE)
#else
#define CVM_NEED_VERIFY(cond) \
    ((CVMglobals.classVerificationLevel == CVM_VERIFY_ALL) || \
     ((CVMglobals.classVerificationLevel == CVM_VERIFY_REMOTE) && (cond)))
#endif /* CVM_TRUSTED_CLASSLOADERS */

#endif /* CVM_CLASSLOADING */

/*
 * Link an already loaded class. This mainly involves preparing the
 * fields, methodtable, and interface table. Throws and exception and 
 * returns CVM_FALSE is unsuccessful.
 */
#ifdef CVM_CLASSLOADING
extern CVMBool
CVMclassLink(CVMExecEnv* ee, CVMClassBlock* cb);
#endif

/*
 * CVMclassInit - run static initializers for entire class hierarchy.
 */
extern CVMBool
CVMclassInit(CVMExecEnv* ee, CVMClassBlock* cb);

/*
 * CVMclassInitNoCRecursion - setup interpeter state so the interpreter
 * can run the static intializers.
 */
extern int
CVMclassInitNoCRecursion(CVMExecEnv* ee, CVMClassBlock* cb,
			 CVMMethodBlock **p_mb);

/*** End functions in classinitialize.c ***/

/*
 * CVMdefineClass - Used to define a new class. Used by CVMjniDefineClass()
 * and JVM_DefineClass().
 */
extern CVMClassICell*
CVMdefineClass(CVMExecEnv* ee, const char *name, CVMClassLoaderICell* loader,
	       const CVMUint8* buf, CVMUint32 bufLen, CVMObjectICell* pd);

/*
 * CVMclassCreateArrayClass - creates a fake array class. Needed for any
 * array class type that wasn't reference by romized code and we make 
 * a reference to it at runtime.
 */
extern CVMClassBlock*
CVMclassCreateArrayClass(CVMExecEnv* ee, CVMClassTypeID arrayTypeId,
			 CVMClassLoaderICell* loader, CVMObjectICell* pd);

/*
 * A 'class callback' called on each class that is iterated over.
 */
typedef void (*CVMClassCallbackFunc)(CVMExecEnv* ee,
				     CVMClassBlock* cb,
				     void* data);

typedef void (*CVMStackCallbackFunc)( CVMObject**, void*);

/*
 * Iterate over all classes, both romized and dynamically loaded,
 * and call 'callback' on each class.
 */
extern void
CVMclassIterateAllClasses(CVMExecEnv* ee, 
			  CVMClassCallbackFunc callback,
			  void* data);

/*
 * Iterate over the loaded classes table, and call 'callback'
 * on each class.
 */
extern void
CVMclassIterateDynamicallyLoadedClasses(CVMExecEnv* ee, 
					CVMClassCallbackFunc callback,
					void* data);
/*
 * Free all classes in the class table. This is used during vm shutdown.
 */
extern void
CVMclassTableFreeAllClasses(CVMExecEnv* ee);

/*
 * Class table functions. The class table is a simple database of all
 * dynamically loaded classes. It is used by gc to scan dynamically loaded
 * classes and is also needed by jvmdi to get a list of all loaded classes.
 * In both cases CVMclassIterateDynamicallyLoadedClasses() is used to
 * iterate over the class table.
 */

extern CVMClassBlock**
CVMclassTableAllocateSlot(CVMExecEnv* ee);

void
CVMclassTableFreeSlot(CVMExecEnv* ee, CVMClassBlock** cbPtr);

extern void
CVMclassTableMarkUnscannedClasses(CVMExecEnv* ee,
				  CVMRefLivenessQueryFunc isLive,
				  void* isLiveData);

extern void
CVMclassTableIterate(CVMExecEnv* ee,
		     CVMClassCallbackFunc callback,
		     void* data);

#ifdef CVM_DEBUG
extern void
CVMclassTableDump(CVMExecEnv* ee);
#endif /* CVM_DEBUG */

/*
 * Loader cache functions. The loader cache is a database of <loader,cb>
 * pairs. The loader is the initiating class loader. It is used to speed
 * up class lookups and assist in implementing loader constraints.
 */

#ifdef CVM_CLASSLOADING
extern CVMBool
CVMloaderCacheCheckPackageAccess(CVMExecEnv *ee, CVMClassLoaderICell* loader, 
				 CVMClassBlock* cb, CVMObjectICell* pd);
#endif

#define CVMloaderCacheLookup(ee, classID, loader) \
    CVMloaderCacheLookupWithProtectionDomain(ee, classID, loader, NULL)

extern CVMClassBlock*
CVMloaderCacheLookupWithProtectionDomain(CVMExecEnv* ee,
					 CVMClassTypeID classID,
					 CVMClassLoaderICell* loader,
					 CVMObjectICell* pd);

extern CVMBool
CVMloaderCacheAdd(CVMExecEnv* ee, CVMClassBlock* cb,
		  CVMClassLoaderICell* loader);

#ifdef CVM_CLASSLOADING
extern void
CVMloaderCacheMarkUnscannedClassesAndLoaders(CVMExecEnv* ee,
					      CVMRefLivenessQueryFunc isLive,
					      void* isLiveData);
extern void
CVMloaderCachePurgeUnscannedClassesAndLoaders(CVMExecEnv* ee);
#endif /* CVM_CLASSLOADING */

extern void
CVMloaderCacheInit();

extern void
CVMloaderCacheDestroy(CVMExecEnv* ee);

#ifdef CVM_DEBUG
extern void
CVMloaderCacheDump(CVMExecEnv* ee);
#endif /* CVM_DEBUG */

/*
 * Loader constraint functions.
 */

#ifdef CVM_CLASSLOADING
CVMBool 
CVMloaderConstraintsCheckMethodSignatureLoaders(CVMExecEnv* ee,
						CVMMethodTypeID methodID, 
						CVMClassLoaderICell* loader1,
						CVMClassLoaderICell* loader2);
CVMBool 
CVMloaderConstraintsCheckFieldSignatureLoaders(CVMExecEnv* ee,
					       CVMFieldTypeID fieldID, 
					       CVMClassLoaderICell* loader1,
					       CVMClassLoaderICell* loader2);
extern void 
CVMloaderConstraintsMarkUnscannedClassesAndLoaders(
    CVMExecEnv* ee,
    CVMRefLivenessQueryFunc isLive,
    void* isLiveData);

extern void 
CVMloaderConstraintsPurgeUnscannedClassesAndLoaders(CVMExecEnv* ee);

#ifdef CVM_DEBUG
extern void
CVMloaderConstraintsDump(CVMExecEnv* ee);
#endif

#endif /* CVM_CLASSLOADING */

/*
 * Traverse the scannable state of one class, and call 'callback' on
 * each reference in it. The scannable state of the class is its
 * statics, and also the ClassLoader and javaInstance fields.  
 */
extern void
CVMclassScan(CVMExecEnv* ee, CVMClassBlock* cb,
	     CVMRefCallbackFunc callback, void* data);

/*
 * Traverse the reference-typed statics of each class.
 */
extern void
CVMclassScanStatics(CVMExecEnv* ee, CVMClassBlock* cb,
	     CVMRefCallbackFunc callback, void* data);

/* 
 * Initialization and destruction of the classes module
 */
extern void
CVMclassModuleInit(CVMExecEnv* ee);

extern void
CVMclassModuleDestroy(CVMExecEnv* ee);

extern CVMBool
CVMclassBootClassPathInit(JNIEnv *env);

extern void
CVMclassBootClassPathDestroy();

#endif /* _INCLUDED_CLASSES_H */
