/*
 * @(#)gc_common.h	1.45 00/10/25
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*
 * This file includes functions common to all GCs. It is the interface
 * between a particular GC implementation and the rest of the VM.
 */

#ifndef _INCLUDED_GC_COMMON_H
#define _INCLUDED_GC_COMMON_H

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/stacks.h"
#include "javavm/include/gc/gc_impl.h"
#include "javavm/include/basictypes.h"
#include "javavm/include/preloader.h"
#include "javavm/include/weakrefs.h"
#include "generated/offsets/java_lang_Class.h"

/*
 * The default size of the heap
 */
#define CVM_DEFAULT_HEAP_SIZE_IN_BYTES (4 * 1024 * 1024)

/*
 * A structure indicating the current GC mode. This is passed around
 * the root scan phase.
 */
struct CVMGCOptions {
    CVMBool fullGC;
    CVMBool allClassesAreRoots;
    CVMBool discoverWeakReferences;
};

/*
 * Extracts 'minBytes' and 'maxBytes' from options, then
 * initializes a heap of at least 'minBytes' bytes, and at most 'maxBytes'
 * bytes.
 */
extern CVMBool 
CVMgcInitHeap(CVMOptions *options);

/*
 * Get GC attribute named 'attrName'.
 * Return NULL if not found.
 */
extern char *
CVMgcGetGCAttributeVal(char* attrName);

/*
 * Allocate heap object of type 'cb'.
 */
extern CVMObject*
CVMgcAllocNewInstance(CVMExecEnv* ee, CVMClassBlock* cb);

/*
 * Allocate heap array of length 'len'. The type of the array is arrayCb.
 * The elements are basic type 'typeCode'.
 */
extern CVMArrayOfAnyType*
CVMgcAllocNewArray(CVMExecEnv* ee, CVMBasicType typeCode,
		   CVMClassBlock* arrayCb, CVMJavaInt len);

/*
 * Return the number of bytes free in the heap
 */
extern CVMJavaLong
CVMgcFreeMemory(CVMExecEnv* ee);

/*
 * Return the amount of total memory in the heap, in bytes
 */
extern CVMJavaLong
CVMgcTotalMemory(CVMExecEnv* ee);

/*
 * Constants related to object reference map formats
 */
#define CVM_GCMAP_NUM_FLAGS     2    /* the no. of flags bits in the map */
#define CVM_GCMAP_NUM_WEAKREF_FIELDS  2 /* the no. of special fields in
					  java.lang.ref.Reference */
#define CVM_GCMAP_FLAGS_MASK      ((1 << CVM_GCMAP_NUM_FLAGS) - 1)

#define CVM_GCMAP_NOREFS          0x0  /* No references in this object */

#define CVM_GCMAP_SHORTGCMAP_FLAG 0x0  /* 0 - object with inline GC map */
#define CVM_GCMAP_ALLREFS_FLAG    0x1  /* 1 - array of all references */
#define CVM_GCMAP_LONGGCMAP_FLAG  0x2  /* 2 - object with long GC map */

/*
 * Walk refs of heap object 'obj' with first header word
 * 'firstHeaderWord', and perform refAction on each discovered object
 * reference field. Also scan class of object if it has not been scanned
 * yet.
 *
 * Within the body of refAction, CVMObject** refPtr refers to the
 * address of a reference typed slot (an array element, or an object
 * field).  
 *
 * CVMscanClassIfNeeded(CVMExecEnv* ee, CVMClassBlock* cb,
 *                      CVMRefCallbackFunc callback, void* data);
 */
#define CVMscanClassIfNeeded(ee, cb, callback, data)			    \
    if (!CVMcbIsInROM(cb) && !CVMcbGcScanned(cb)) {			    \
	CVMclassScan(ee, cb, callback, data);	    	    		    \
	CVMcbSetGcScanned(cb);						    \
    }

#define CVMobjectWalkRefsWithSpecialHandling(ee, gcOpts, obj,		    \
					     firstHeaderWord_, refAction,   \
					     callback, data)		    \
{									    \
    CVMClassBlock* cb_ = (CVMClassBlock*)				    \
        (((CVMUint32)(firstHeaderWord_)) & CVM_OBJECT_CLASS_MASK);	    \
    CVMscanClassIfNeeded(ee, cb_, callback, data);			    \
    /*									    \
     * If we are scanning a java.lang.Class instance, scan its 		    \
     * corresponding CVMClassBlock.					    \
     */									    \
    if (cb_ == CVMsystemClass(java_lang_Class)) {                           \
	CVMClassBlock* classBlockPtr = 					    \
            *((CVMClassBlock**)(obj) + 					    \
	      CVMoffsetOfjava_lang_Class_classBlockPointer);		    \
        CVMscanClassIfNeeded(ee, classBlockPtr, callback, data);	    \
    }                                                                       \
    CVMobjectWalkRefsAux(ee, gcOpts, obj, firstHeaderWord_,		    \
			 CVM_TRUE, refAction);   	    		    \
}

#define CVMobjectWalkRefs(ee, gcOpts, obj, firstHeaderWord_, refAction)	    \
    CVMobjectWalkRefsAux(ee, gcOpts, obj, firstHeaderWord_, CVM_FALSE, refAction)

/*
 * Walk refs of heap object 'obj' with first header word
 * 'firstHeaderWord', and perform refAction on each discovered object
 * reference field. Ignore ROM objects.
 *
 * Within the body of refAction, CVMObject** refPtr refers to the
 * address of a reference typed slot (an array element, or an object
 * field).  
 */
#define CVMobjectWalkRefsAux(ee, gcOpts, obj, firstHeaderWord_,		  \
			     handleWeakReferences, refAction)		  \
{									  \
    CVMClassBlock* cb_ = (CVMClassBlock*)				  \
        (((CVMUint32)(firstHeaderWord_)) & CVM_OBJECT_CLASS_MASK);	  \
    CVMUint32      map = CVMcbGcMap(cb_).map;				  \
    CVMObject**    refPtr;						  \
									  \
    if (map != CVM_GCMAP_NOREFS) {					  \
	CVMUint32 flags = map & CVM_GCMAP_FLAGS_MASK;			  \
	refPtr = (CVMObject**)&(obj)->fields[0];		  	  \
									  \
        if (flags == CVM_GCMAP_SHORTGCMAP_FLAG) {			  \
            /* object with inline GC map: the most common case */  	  \
            								  \
            /* First skip the flags */				  	  \
            map >>= CVM_GCMAP_NUM_FLAGS;			  	  \
            								  \
            if (handleWeakReferences && 				  \
		(gcOpts)->discoverWeakReferences) { 			  \
                if (CVMcbIs(cb_, REFERENCE)) {				  \
                    /* Skip referent, next in case we just discovered	  \
		       a new active weak reference object. */		  \
                    if (CVMweakrefField(obj, next) == NULL) {		  \
			CVMweakrefDiscover(ee, obj);			  \
		        map >>= CVM_GCMAP_NUM_WEAKREF_FIELDS;	  	  \
                        refPtr += CVM_GCMAP_NUM_WEAKREF_FIELDS;		  \
                    }							  \
		}							  \
	    }								  \
            /* Now get through all the references in the object */	  \
            while (map != 0) {					  	  \
                if ((map & 0x1) != 0) {				  	  \
                    refAction;					  	  \
                }							  \
		map >>= 1;						  \
		refPtr++;						  \
            }							  	  \
        } else if (flags == CVM_GCMAP_ALLREFS_FLAG) {			  \
	    /* an array of all refs. Scan each and every one. */	  \
	    CVMArrayOfRef* arrRefs = (CVMArrayOfRef*)(obj);		  \
	    CVMJavaInt     arrLen  = CVMD_arrayGetLength(arrRefs);	  \
									  \
	    /*								  \
	     * Address of the first element of the array		  \
             * (skip the length field)					  \
	     */								  \
	    refPtr += 1;						  \
	    while (arrLen-- > 0) {					  \
		refAction;						  \
		refPtr++;						  \
	    }								  \
	} else {							  \
            /* object with "big" GC map */			  	  \
            CVMBigGCBitMap* bigmap = 					  \
                         (CVMBigGCBitMap*)(map & ~CVM_GCMAP_FLAGS_MASK);  \
            CVMUint32       mapLen = bigmap->maplen;		  	  \
            CVMUint32*      mapPtr = bigmap->map;			  \
            CVMUint32*      mapEnd = mapPtr + mapLen;		  	  \
            CVMObject**     otherRefPtr = refPtr;			  \
            								  \
            CVMassert(flags == CVM_GCMAP_LONGGCMAP_FLAG);		  \
            /* Make sure there is something to do */			  \
            CVMassert(mapPtr < mapEnd);					  \
	    map = *mapPtr;				  	  	  \
            if (handleWeakReferences &&					  \
		(gcOpts)->discoverWeakReferences) { 			  \
                if (CVMcbIs(cb_, REFERENCE)) {				  \
                    /* Skip referent, next in case we just discovered	  \
		       a new active weak reference object. */		  \
                    if (CVMweakrefField(obj, next) == NULL) {		  \
			CVMweakrefDiscover(ee, obj);			  \
		        map >>= CVM_GCMAP_NUM_WEAKREF_FIELDS;	  	  \
                        refPtr += CVM_GCMAP_NUM_WEAKREF_FIELDS;		  \
                    }							  \
		}							  \
	    }								  \
            while (mapPtr < mapEnd) {				  	  \
		while (map != 0) {					  \
		    if ((map & 0x1) != 0) {				  \
			refAction;					  \
		    }						  	  \
		    map >>= 1;					  	  \
		    refPtr++;					  	  \
		}							  \
		mapPtr++;						  \
                /* This may be a redundant read that may fall off the	  \
		   edge of the world in the extremely unlikely		  \
		   case. To prevent this case, we've left an extra	  \
		   safety word in the big map. (nedim120299) */ 	  \
                map = *mapPtr;						  \
		/* Advance the object pointer to the next group of 32	  \
		 * fields, in case map becomes 0 before we iterate	  \
		 * through all the fields in group *mapPtr		  \
                 */		  					  \
		otherRefPtr += 32;					  \
		refPtr = otherRefPtr;				  	  \
            }							  	  \
	}								  \
    }									  \
}

/*
 * Initiate a GC. Acquire all GC locks, stop all threads, and then
 * call back to the particular GC to do the work. When the particular
 * GC is done, resume.
 *
 * Called with heap locked and while GC-unsafe.
 * return CVM_TRUE on success, CVM_FALSE if GC could not be run.
 */
extern CVMBool
CVMgcStopTheWorldAndGC(CVMExecEnv* ee, CVMUint32 numBytes);

/*
 * Start a GC cycle. Will clear class marks if necessary.
 */
extern void
CVMgcStartGC(CVMExecEnv* ee);

/*
 * Clear the class GC_SCANNED marks for dynamically loaded classes.
 */
extern void
CVMgcClearClassMarks(CVMExecEnv* ee, CVMGCOptions* gcOpts);

/*
 * End a GC cycle.
 */
extern void
CVMgcEndGC(CVMExecEnv* ee);

/*
 * Like above, but called without heap locked and while GC-safe.
 */
extern void
CVMgcRunGC(CVMExecEnv* ee);

/*
 * The null CVMFrameGCScannerFunc that doesn't do anything
 */
extern void
CVMnullFrameScanner(CVMFrame* frame, CVMStackChunk* chunk,
		    CVMRefCallbackFunc callback, void* data);

/*
 * Process special objects with liveness info from a particular GC
 * implementation. This covers special scans like string intern table,
 * weak references and monitor structures.
 *
 * isLive - a predicate that returns true if an object is strongly referenced
 * transitiveScanner - a callback that marks an object and all its children
 */
extern void
CVMgcProcessSpecialWithLivenessInfo(CVMExecEnv* ee, CVMGCOptions* gcOpts,
				    CVMRefLivenessQueryFunc isLive,
				    void* isLiveData,
                                    CVMRefCallbackFunc transitiveScanner,
				    void* transitiveScannerData);

/*
 * Scan and optionally update special objects. This covers special
 * scans like string intern table, weak references and monitor
 * structures.  
 */
extern void
CVMgcScanSpecial(CVMExecEnv* ee, CVMGCOptions* gcOpts,
		 CVMRefCallbackFunc callback, void* data);

/*
 * Scan the root set of collection
 */
extern void
CVMgcScanRoots(CVMExecEnv* ee, CVMGCOptions* gcOpts,
	       CVMRefCallbackFunc callback, void* data);

/*
 * Destroy the heap
 */
extern CVMBool 
CVMgcDestroyHeap();

#ifdef CVM_DEBUG
/*
 * Heap dump support: Iterate over a contiguous-allocated range of objects.
 */
extern void
CVMgcScanObjectRange(CVMExecEnv* ee, CVMUint32* base, CVMUint32* top,
		     CVMObjectCallbackFunc callback, void* callbackData);

extern void
CVMgcDumpHeapSimple();

extern void
CVMgcDumpHeapVerbose();

extern void
CVMgcDumpHeapStats();

#endif /* CVM_DEBUG */

#endif /* _INCLUDED_GC_COMMON_H */
