/*
 * @(#)classes.c	1.81 00/11/01
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/typeid.h"
#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/typeid.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/preloader.h"
#include "javavm/include/globalroots.h"
#include "javavm/include/globals.h"
#include "javavm/include/utils.h"
#include "javavm/include/clib.h"
 

/*
 * Get an array class with elements of type elemCb.
 */
CVMClassBlock*
CVMclassGetArrayOf(CVMExecEnv* ee, CVMClassBlock* elemCb)
{
    CVMassert(CVMD_isgcSafe(ee));

    if (CVMcbIs(elemCb, PRIMITIVE)) {
	return (CVMClassBlock*)
	    CVMbasicTypeArrayClassblocks[CVMcbBasicTypeCode(elemCb)];
    }

    /*
     * Look it up.
     */
    {
	CVMClassTypeID arrayID = 
	    CVMtypeidIncrementArrayDepth(ee, CVMcbClassName(elemCb), 1);
	CVMClassBlock * arrayClass =
	    CVMclassLookupByTypeFromClass(ee, arrayID, CVM_FALSE, elemCb);
	 CVMtypeidDisposeClassID( ee, arrayID );
	 return arrayClass;
    }
}

static CVMMethodBlock*
CVMclassprivateReturnDeclaredMethodInClassMethods(const CVMClassBlock* cb,
						  CVMMethodTypeID tid)
{
    int i;
    for (i = 0; i < CVMcbMethodCount(cb); i++) {
        CVMMethodBlock* mb = CVMcbMethodSlot(cb, i);
	if (CVMtypeidIsSame(CVMmbNameAndTypeID(mb), tid)) {
            return mb;
	}
    }
    return NULL;
}

static CVMMethodBlock*
CVMclassprivateReturnDeclaredMethodInClassMethodTable(const CVMClassBlock* cb,
						      CVMMethodTypeID tid)
{
    int i;
    for (i = 0; i < CVMcbMethodTableCount(cb); i++) {
        CVMMethodBlock* mb = CVMcbMethodTableSlot(cb, i);
	if (CVMtypeidIsSame(CVMmbNameAndTypeID(mb), tid)) {
            return mb;
	}
    }
    return NULL;
}

/*
 * Look for public or declared method of type 'tid'
 */
CVMMethodBlock*
CVMclassGetMethodBlock(const CVMClassBlock* cb,
		       CVMMethodTypeID tid, CVMBool isStatic)
{
    CVMMethodBlock* mb;

    if (!isStatic) {
	/* Look in methodTable of the given class */
	mb = CVMclassprivateReturnDeclaredMethodInClassMethodTable(cb, tid);
	if (mb != NULL) {
	    goto found;
	}

	/* Look in declared methods of the given class */
	mb = CVMclassprivateReturnDeclaredMethodInClassMethods(cb, tid);
	if (mb != NULL) {
	    goto found;
	}

	/*
	 * Now look in all implemented interfaces for the method.
	 */
	{
	    int i;
	    for (i = 0; i < CVMcbInterfaceCount(cb); i++) {
		mb = CVMclassprivateReturnDeclaredMethodInClassMethods(
		        CVMcbInterfacecb(cb, i), tid);
		if (mb != NULL) {
		    goto found;
		}
	    }
	}
    } else {
	const CVMClassBlock *cb0 = cb;

	while (cb0 != NULL) {
	    /* Look in declared methods of the current class */
	    mb = CVMclassprivateReturnDeclaredMethodInClassMethods(cb0, tid);
	    if (mb != NULL) {
		if (CVMmbIs(mb, PRIVATE) && cb0 != cb) {
		    return NULL;
		} else {
		    goto found;
		}
	    }

	    /* <clinit> from current class only, not superclasses */
	    if (CVMtypeidIsStaticInitializer(tid)) {
		return NULL;
	    }

	    /* We're still here! Keep searching. */
	    cb0 = CVMcbSuperclass(cb0);
	}
    }

    return NULL;

found:

    return CVMmbIs(mb, STATIC) == isStatic ? mb : NULL;
}

/*
 * Look for declared method of type 'tid'
 */
CVMMethodBlock*
CVMclassGetDeclaredMethodBlockFromTID(const CVMClassBlock* cb,
				      CVMMethodTypeID tid)
{
    /* Look in methodTable of the given class */
    return CVMclassprivateReturnDeclaredMethodInClassMethods(cb, tid);
}

/*
 * Look for declared method of given name and signature
 */
CVMMethodBlock*
CVMclassGetDeclaredMethodBlock(CVMExecEnv *ee,
			       const CVMClassBlock* cb,
			       const char *name,
			       const char *sig)
{
    /*
     * We need to use the "new" function, because the "lookup" function
     * is only valid when we know that the type already exists.
     */
    CVMMethodTypeID tid = CVMtypeidNewMethodIDFromNameAndSig(ee, name, sig);

    if (tid != CVM_TYPEID_ERROR) {
	CVMMethodBlock* resultMb =
	    CVMclassGetDeclaredMethodBlockFromTID(cb, tid);
	CVMtypeidDisposeMethodID(ee, tid);
	return resultMb;
    } else {
	CVMclearLocalException(ee);
	return NULL;
    }
}

static CVMFieldBlock* 
CVMclassprivateReturnDeclaredFieldInClass(const CVMClassBlock* cb, 
					  CVMFieldTypeID tid)
{
    int i;
    for (i = 0; i < CVMcbFieldCount(cb); i++) {
        CVMFieldBlock* fb = CVMcbFieldSlot(cb, i);
	if (CVMtypeidIsSame(CVMfbNameAndTypeID(fb), tid)) {
            return fb;
	}
    }
    return 0;
}

/*
 * Look for declared field of type 'tid'
 */
CVMFieldBlock*
CVMclassGetFieldBlock(const CVMClassBlock* cb, CVMFieldTypeID tid,
		      CVMBool isStatic)
{
    const CVMClassBlock *cb0 = cb;
    CVMFieldBlock* fb;

    while (cb0 != NULL) {
	fb = CVMclassprivateReturnDeclaredFieldInClass(cb0, tid);
	if (fb != NULL) {
	    if (CVMfbIs(fb, PRIVATE) && cb0 != cb) {
		return NULL;
	    } else {
		return fb;
	    }
	}

	/* We're still here! Keep searching. */
	cb0 = CVMcbSuperclass(cb0);
    }

    /*
     * OK, now let's look at implemented interfaces if we are looking
     * for a static 
     */
    if (isStatic) {
	int i;
	for (i = 0; i < CVMcbInterfaceCount(cb); i++) {
	    fb = CVMclassprivateReturnDeclaredFieldInClass(
                    CVMcbInterfacecb(cb, i), tid);
	    if (fb != NULL) {
		return fb;
	    }
	}
    }

    return NULL;
}

/*
 * Iterate over all classes, both romized and dynamically loaded,
 * and call 'callback' on each class.
 */
void
CVMclassIterateAllClasses(CVMExecEnv* ee, 
			  CVMClassCallbackFunc callback,
			  void* data)
{
    /* Iterate over all the romized classes first. */
    CVMpreloaderIterateAllClasses( ee, callback, data);

    /* Iterate over all dynamically loaded classes. */
    CVMclassIterateDynamicallyLoadedClasses(ee, callback, data);
}

/*
 * Iterate over the loaded classes, and call 'callback'
 * on each class.
 */
void
CVMclassIterateDynamicallyLoadedClasses(CVMExecEnv* ee, 
					CVMClassCallbackFunc callback,
					void* data)
{
    CVMclassTableIterate(ee, callback, data);
}

/*
 * Traverse the reference-typed statics of one class, and call 'callback' on
 * each reference in it.
 */
void
CVMclassScanStatics(CVMExecEnv* ee, CVMClassBlock* cb, 
		    CVMRefCallbackFunc callback, void* data)
{
    CVMUint32     numRefStatics = CVMcbNumStaticRefs(cb);
    CVMtraceGcScan(("Scanning statics of class %C (0x%x), "
		    "with %d ref-type statics\n",
		    cb, cb, numRefStatics));
    if (numRefStatics > 0) {
	CVMJavaVal32* refStatics = CVMcbStatics(cb);
	CVMwalkContiguousRefs(refStatics, numRefStatics, callback, data);
    }
}

/*
 * Traverse the scannable state of one class, and call 'callback' on
 * each reference in it. The scannable state of the class is its
 * statics, and also the ClassLoader and javaInstance fields.  
 */
void
CVMclassScan(CVMExecEnv* ee, CVMClassBlock* cb,
	     CVMRefCallbackFunc callback, void* data)
{
    CVMUint32     numRefStatics = CVMcbNumStaticRefs(cb);
    if (CVMcbClassName(cb) != 0) {
	CVMtraceGcScan(("Scanning class %C (0x%x), "
			"with %d ref-type statics\n",
			cb, cb, numRefStatics));
    } else {
	CVMtraceGcScan(("Scanning class 0x%x, with %d ref-type statics\n",
			cb, numRefStatics));
    }
    if (numRefStatics > 0) {
	CVMJavaVal32* refStatics = CVMcbStatics(cb);
	CVMwalkContiguousRefs(refStatics, numRefStatics, callback, data);
    }
    /*
     * The instance can be NULL if the class is being loaded and
     * linked.  
     */
    if (CVMcbJavaInstance(cb) != NULL) {
	CVMwalkOneICell(CVMcbJavaInstance(cb), callback, data);
    }
}

/*
 * Do the initialization of the classes module
 */
void
CVMclassModuleInit(CVMExecEnv* ee)
{
    CVMMethodTypeID methodTypeID;
    const CVMClassBlock* cb;

    CVMloaderCacheInit();

    /* lookup mb for java.lang.Class.runStaticInitializers() */
    methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
        ee, "runStaticInitializers", "()V");
    cb = CVMsystemClass(java_lang_Class);
    CVMglobals.java_lang_Class_runStaticInitializers = 
	CVMclassGetNonstaticMethodBlock(cb, methodTypeID);
    CVMassert(CVMglobals.java_lang_Class_runStaticInitializers != NULL);

    /* lookup mb for java.lang.Class.newInstance() */
    methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
        ee, "newInstance", "()Ljava/lang/Object;");
    cb = CVMsystemClass(java_lang_Class);
    CVMglobals.java_lang_Class_newInstance = 
	CVMclassGetNonstaticMethodBlock(cb, methodTypeID);
    CVMassert(CVMglobals.java_lang_Class_newInstance != NULL);

#ifdef CVM_CLASSLOADING
    /* lookup mb for java.lang.ClassLoader.NativeLibrary.getFromClass() */
    methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
        ee, "getFromClass", "()Ljava/lang/Class;");
    cb = CVMsystemClass(java_lang_ClassLoader_NativeLibrary);
    CVMglobals.java_lang_ClassLoader_NativeLibrary_getFromClass = 
	CVMclassGetStaticMethodBlock(cb, methodTypeID);
    CVMassert(CVMglobals.java_lang_ClassLoader_NativeLibrary_getFromClass
	      != NULL);
    
    /* lookup mb for java.lang.ClassLoader.addClass() */
    methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
        ee, "addClass", "(Ljava/lang/Class;)V");
    cb = CVMsystemClass(java_lang_ClassLoader);
    CVMglobals.java_lang_ClassLoader_addClass = 
	CVMclassGetNonstaticMethodBlock(cb, methodTypeID);
    CVMassert(CVMglobals.java_lang_ClassLoader_addClass != NULL);
    
    /* lookup mb for java.lang.ClassLoader.loadClass() */
    methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
        ee, "loadClassInternal", "(Ljava/lang/String;)Ljava/lang/Class;");
    cb = CVMsystemClass(java_lang_ClassLoader);
    CVMglobals.java_lang_ClassLoader_loadClass = 
	CVMclassGetNonstaticMethodBlock(cb, methodTypeID);
    CVMassert(CVMglobals.java_lang_ClassLoader_loadClass != NULL);
    
    /* lookup mb for java.lang.ClassLoader.findNative() */
    methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
        ee, "findNative", "(Ljava/lang/ClassLoader;Ljava/lang/String;)J");
    cb = CVMsystemClass(java_lang_ClassLoader);
    CVMglobals.java_lang_ClassLoader_findNative = 
	CVMclassGetStaticMethodBlock(cb, methodTypeID);
    CVMassert(CVMglobals.java_lang_ClassLoader_findNative != NULL);
    
    /* lookup mb for java.lang.ClassLoader.checkPackageAccess() */
    methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
        ee, "checkPackageAccess", 
	"(Ljava/lang/Class;Ljava/security/ProtectionDomain;)V");
    cb = CVMsystemClass(java_lang_ClassLoader);
    CVMglobals.java_lang_ClassLoader_checkPackageAccess = 
	CVMclassGetNonstaticMethodBlock(cb, methodTypeID);
    CVMassert(CVMglobals.java_lang_ClassLoader_checkPackageAccess != NULL);
#endif
}

/*
 * Destroy the classes module
 * Start by freeing all dynamically loaded classes.
 * We don't need to free any heap objects, of course.
 */

void
CVMclassModuleDestroy(CVMExecEnv* ee)
{
    CVMclassTableFreeAllClasses(ee);
    CVMloaderCacheDestroy(ee);
}
