/*
 * @(#)constantpool.c	1.21 00/06/20
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifdef CVM_CLASSLOADING

#include "javavm/include/interpreter.h"
#include "javavm/include/classes.h"
#include "javavm/include/utils.h"
#include "javavm/include/common_exceptions.h"
#include "javavm/include/globals.h"

/*
 * Resolve a NameAndType constant pool entry into an fb or mb.
 */
static CVMFieldBlock*
CVMcpResolveFieldref(CVMExecEnv* ee,
		     CVMClassBlock* currentCb, 
		     CVMConstantPool* cp, 
		     CVMClassBlock* fieldCb,
		     CVMUint16 typeIDIdx);

static CVMMethodBlock*
CVMcpResolveMethodref(CVMExecEnv* ee,
		      CVMClassBlock* currentCb, 
		      CVMConstantPool* cp, 
		      CVMClassBlock* methodCb,
		      CVMUint16 typeIDIdx);

static CVMMethodBlock*
CVMcpResolveInterfaceMethodref(CVMExecEnv* ee,
			       CVMClassBlock* currentCb, 
			       CVMConstantPool* cp, 
			       CVMClassBlock* methodCb,
			       CVMUint16 typeIDIdx);

/*
 * CVMcpLookupClassAndVerifyAccess - Lookup the specified classID and
 * verify that the current class can access it.
 */
static CVMClassBlock*
CVMcpLookupClassAndVerifyAccess(CVMExecEnv* ee,
				CVMClassBlock* currentCb,
				CVMClassTypeID classID)
{
    CVMClassBlock* resolvedCb = 
	CVMclassLookupByTypeFromClass(ee, classID, CVM_FALSE, currentCb);
    if (resolvedCb == NULL) {
        /* Exception already thrown */
        return NULL;
    } else if (!CVMverifyClassAccess(ee, currentCb, resolvedCb, CVM_TRUE)){
	CVMthrowIllegalAccessError(ee, 
				   "try to access class %C from class %C",
				   resolvedCb, currentCb);
	return NULL;
    } else {
	return resolvedCb;
    }
}

/* 
 * CVMprivate_cpResolveEntryFromClass - resolve a constant pool entry.
 *   currentCb: the "current class" for security and class loading purposes.
 *   cp:        the constant pool
 *   cpIndex:   index of entry in cp to resolve
 * 
 * This function should only be called via the public CVMcpResolveEntry
 * and CVMcpResolveEntryFromClass macros.
 */

CVMBool
CVMprivate_cpResolveEntryFromClass(CVMExecEnv* ee,
				   CVMClassBlock* currentCb, 
				   CVMConstantPool* cp, 
				   CVMUint16 cpIndex)
{
    CVMConstantPoolEntryType cpType;
    CVMFieldBlock*  resolvedFb = NULL;
    CVMMethodBlock* resolvedMb = NULL;

    /* The cp index, classID, and cb for any class we may resolve, 
     * including the class referred to by a FieldRef or MethodRef.
     */
    CVMUint16       classCpIndex;
    CVMClassBlock*  resolvedCb;
    CVMClassTypeID  classID = CVM_TYPEID_ERROR;
    /*
     * We need to make sure the cp entry is not resolved and get its
     * contents while holding the lock. Otherwise it could get resolved
     * after we do the check, but before we grab the entry, in which
     * case the data we get is not what we expect.
     */
    CVMcpLock(ee);
    if (CVMcpIsResolved(cp, cpIndex)) {
	CVMcpUnlock(ee);
	return CVM_TRUE;
    }

    cpType = CVMcpEntryType(cp, cpIndex);
    switch(cpType) {
    case CVM_CONSTANT_ClassTypeID: {
	classID = CVMcpGetClassTypeID(cp, cpIndex);
	CVMcpUnlock(ee);

	classCpIndex = cpIndex;
	resolvedCb = 
	    CVMcpLookupClassAndVerifyAccess(ee, currentCb, classID);
	if (resolvedCb == NULL) {
	    /* Exception already thrown */
	    return CVM_FALSE;
	}
	
	break;
    }
    
    case CVM_CONSTANT_Fieldref:
    case CVM_CONSTANT_Methodref: 
    case CVM_CONSTANT_InterfaceMethodref: {
	CVMClassLoaderICell* currentLoader = 
	    (currentCb != NULL) ? CVMcbClassLoader(currentCb) : NULL;
	CVMUint16 typeIDIdx = CVMcpGetMemberRefTypeIDIdx(cp, cpIndex);
	classCpIndex  = CVMcpGetMemberRefClassIdx(cp, cpIndex);

	/* First, resolve reference to the class */
	if (!CVMcpIsResolved(cp, classCpIndex)) {
	    classID = CVMcpGetClassTypeID(cp, classCpIndex);
	    CVMcpUnlock(ee);
	
	    resolvedCb = 
		CVMcpLookupClassAndVerifyAccess(ee, currentCb, classID);
	    if (resolvedCb == NULL) {
		/* Exception already thrown */
		return CVM_FALSE;
	    }
	} else {
	    CVMcpUnlock(ee);
	    resolvedCb = CVMcpGetCb(cp, classCpIndex);
	}

	if (!CVMcbCheckRuntimeFlag(resolvedCb, LINKED)) {
	    if (!CVMclassLink(ee, resolvedCb)) {
		return CVM_FALSE;
	    }
	}
	
	switch(cpType) {
	case CVM_CONSTANT_Fieldref:
	    resolvedFb = CVMcpResolveFieldref(ee, currentCb, cp, 
					      resolvedCb, typeIDIdx);
	    if (resolvedFb == NULL) {
		return CVM_FALSE;
	    }
	    /* Check loader constraints */
	    if (!CVMloaderConstraintsCheckFieldSignatureLoaders(
                    ee, CVMcpGetFieldTypeID(cp, typeIDIdx),
		    currentLoader,
		    CVMcbClassLoader(CVMfbClassBlock(resolvedFb)))) {
		return CVM_FALSE;
	    }
	    break;

	case CVM_CONSTANT_Methodref:
	    resolvedMb = CVMcpResolveMethodref(ee, currentCb, cp, 
					       resolvedCb, typeIDIdx);
	    if (resolvedMb == NULL) {
		return CVM_FALSE;
	    }
	    /* Check loader constraints */
	    if (!CVMloaderConstraintsCheckMethodSignatureLoaders(
                    ee, CVMcpGetMethodTypeID(cp, typeIDIdx),
	  	    currentLoader, 
		    CVMcbClassLoader(CVMmbClassBlock(resolvedMb)))) {
		return CVM_FALSE;
	    }
	    break;

	case CVM_CONSTANT_InterfaceMethodref: 
	    resolvedMb = CVMcpResolveInterfaceMethodref(ee, currentCb, cp,
							resolvedCb, typeIDIdx);
	    if (resolvedMb == NULL) {
		return CVM_FALSE;
	    }
	    /* Check loader constraints */
	    if (!CVMloaderConstraintsCheckMethodSignatureLoaders(
                    ee, CVMcpGetMethodTypeID(cp, typeIDIdx),
		    currentLoader,
		    CVMcbClassLoader(CVMmbClassBlock(resolvedMb)))) {
		return CVM_FALSE;
	    }
	    break;
	}

	break;
    }

    default: 
	CVMassert(CVM_FALSE);
	CVMcpUnlock(ee);
	return CVM_TRUE;
    }
	
    /*
     * We must set the new entry value and type atomically, so this
     * must be done inside the lock. Otherwise someone might think
     * this entry is not resolved when it really is. Checking if the
     * entry is already resolved just avoids asserts by CVMcpSetXXX.
     *
     * We always have to set a cb entry no matter what type of entry this
     * function was called to resolve, since they all either directly 
     * (for ClassTypeID entries) or indirectly (for FieldRef, Methodref,
     * and InterfaceMethodref entries) refer to a ClassTypeID entry.
     */
    CVMcpLock(ee);
    if (!CVMcpIsResolved(cp, classCpIndex)) {
	CVMcpSetCb(cp, classCpIndex, resolvedCb);
    } else {
	classID = CVM_TYPEID_ERROR; /* so we don't try to dispose it later */
    }
    if (!CVMcpIsResolved(cp, cpIndex)) {
	/*
	 * NOTE: since a CVM_CONSTANT_ClassTypeID would have just gotten
	 * resolved above, here we only need to check whether or not we
	 * are resolving a FieldRef.
	 */
	if (cpType == CVM_CONSTANT_Fieldref) {
	    /* Constant pool entries for static fields may be changed
	     * to point directly to the static data by the quickening code.
	     * Comment cjp08/26/99 in quicken.c explains why. For now it
	     * must be set to the fb.
	     */
	    CVMcpSetFb(cp, cpIndex, resolvedFb);
	} else {
	    CVMassert(cpType == CVM_CONSTANT_Methodref ||
		      cpType == CVM_CONSTANT_InterfaceMethodref);
	    CVMcpSetMb(cp, cpIndex, resolvedMb);
	}
    }
    CVMcpUnlock(ee);

    /*
     * Free the constantpool reference to the classID, but only if we
     * resolved a class entry. Note that we deferred doing this until
     * after releasing the cpLock. This probably isn't totally necessary,
     * but it means we hold the lock less and avoid any possible rank error.
     */
    if (classID != CVM_TYPEID_ERROR) {
	CVMtypeidDisposeClassID(ee, classID);
    }
    
    return CVM_TRUE;
}

/*
 * Returns the CVMFieldBlock specified by the nameAndType cp index
 * and fieldCb pair.
 */
static CVMFieldBlock*
CVMcpResolveFieldref(CVMExecEnv* ee,
		     CVMClassBlock* currentCb, 
		     CVMConstantPool* cp, 
		     CVMClassBlock* fieldCb,
		     CVMUint16 typeIDIdx)
{
    CVMFieldTypeID fieldTypeID = CVMcpGetFieldTypeID(cp, typeIDIdx);

    /*
     * Search the field's class and all of its super classes
     * for a field with a matching typeID.
     */
    CVMClassBlock* cb = fieldCb;
    while (cb != NULL) {
	CVMInt32 fieldIdx;
	for (fieldIdx = CVMcbFieldCount(cb) - 1;  fieldIdx >= 0; fieldIdx--) {
	    CVMFieldBlock* fb = CVMcbFieldSlot(cb, fieldIdx);
	    if (CVMtypeidIsSame(fieldTypeID, CVMfbNameAndTypeID(fb))) {
		if (CVMverifyMemberAccess(ee, currentCb, cb, 
					  CVMfbAccessFlags(fb), CVM_TRUE)) {
		    return fb;
		} else {
		    CVMthrowIllegalAccessError(
                        ee, "try to access field %C.%F from class %C",
			cb, fb, currentCb);
		    return NULL;
		}
	    }
	}
	cb = CVMcbSuperclass(cb);
    }

    /*
     * Field not found. Throw an exception.
     */
    CVMthrowNoSuchFieldError(ee, "%C: field %!F not found", 
			     fieldCb, fieldTypeID);
    return NULL;
}

/*
 * Returns the CVMMethodBlock specified by the nameAndType cp index
 * and methodCb pair.
 */
static CVMMethodBlock*
CVMcpResolveMethodref(CVMExecEnv* ee,
		      CVMClassBlock* currentCb, 
		      CVMConstantPool* cp, 
		      CVMClassBlock* methodCb,
		      CVMUint16 typeIDIdx)
{
    CVMMethodTypeID methodTypeID = CVMcpGetMethodTypeID(cp, typeIDIdx);
    CVMClassBlock*  cb = methodCb;
    CVMMethodBlock* mb;

    /*
     * The method must not be a method of an interface class.
     */
    if (CVMcbIs(methodCb, INTERFACE)) {
	CVMthrowIncompatibleClassChangeError(ee, "%C", methodCb);
	return NULL;
    }

    /*
     * Search the method's class and all of its super classes
     * for a method with a matching typeID.
     */
    while (cb != NULL) {
	CVMInt32 methodIdx;
	for (methodIdx = CVMcbMethodCount(cb) - 1;
	     methodIdx >= 0; methodIdx--) {
	    mb = CVMcbMethodSlot(cb, methodIdx);
	    if (CVMtypeidIsSame(methodTypeID, CVMmbNameAndTypeID(mb))) {
		goto found;
	    }
	}
	cb = CVMcbSuperclass(cb);
    }

    /* search for miranda methods */
    {
	CVMUint16 n;	
	CVMUint16 objectMethods =
	    CVMcbMethodTableCount(CVMsystemClass(java_lang_Object));
	for (n = CVMcbMethodTableCount(methodCb) - 1; n >= objectMethods ;n--){
	    mb = CVMcbMethodTableSlot(methodCb, n);
	    if (CVMtypeidIsSame(methodTypeID, CVMmbNameAndTypeID(mb))) {
		/*
		 * If the mb is in the methodtable, but not declared,
		 * it must be a miranda method.
		 */
		CVMassert(CVMmbIsMiranda(mb));
		cb = CVMmbClassBlock(mb);
		goto found;
	    }
	}
    }
    
    /*
     * Method not found. Throw an exception.
     */
 fail:
    CVMthrowNoSuchMethodError(ee, "%C: method %!M not found", 
			      methodCb, methodTypeID);
    return NULL;

 found:
    /*
     * Check if erroneously "inheriting" a constructor.
     */
    if ((methodCb != cb) && CVMtypeidIsConstructor(CVMmbNameAndTypeID(mb))) {
	goto fail;
    }

    if (CVMverifyMemberAccess(ee, currentCb, cb, 
			      CVMmbAccessFlags(mb), CVM_TRUE)) {
	return mb;
    } else {
	CVMthrowIllegalAccessError(ee,
				   "try to access method %C.%M from class %C",
				   cb, mb, currentCb);
	return NULL;
    }
}

/*
 * Returns the CVMMethodBlock specified by the nameAndType cp index
 * and methodCb pair.
 */
static CVMMethodBlock*
CVMcpResolveInterfaceMethodref(CVMExecEnv* ee,
			       CVMClassBlock* currentCb, 
			       CVMConstantPool* cp, 
			       CVMClassBlock* methodCb,
			       CVMUint16 typeIDIdx)
{
    CVMUint16 intfIdx;
    CVMMethodTypeID methodTypeID = CVMcpGetMethodTypeID(cp, typeIDIdx);

    /*
     * The method must a method of an interface class.
     */
    if (!CVMcbIs(methodCb, INTERFACE)) {
	CVMthrowIncompatibleClassChangeError(ee, "%C", methodCb);
	return NULL;
    }

    /* 
     * The CVMInterfaceTable of an interface contains this interface
     * in the first slot, and all of the superinterfaces in the
     * remaining slots. Check the interface and all of its superinterfaces
     * for a matching method.
     */
    for (intfIdx = 0; intfIdx < CVMcbInterfaceCount(methodCb); intfIdx++) {
	CVMClassBlock* intfCb = CVMcbInterfacecb(methodCb, intfIdx);
	CVMInt32 methodIdx;
	for (methodIdx = CVMcbMethodCount(intfCb) - 1; 
	     methodIdx >= 0; methodIdx--) {
	    CVMMethodBlock* mb = CVMcbMethodSlot(intfCb, methodIdx);
	    if (CVMtypeidIsSame(methodTypeID, CVMmbNameAndTypeID(mb))) {
		return mb;
	    }
	}
    }

    /*
     * Method not found. Throw an exception.
     */
    CVMthrowNoSuchMethodError(ee, "%C: method %!M not found", 
			      methodCb, methodTypeID);
    return NULL;
}

#endif /* CVM_CLASSLOADING */
