/*
 * @(#)gc_impl.c	1.60 00/11/30
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*
 * This file includes the implementation of a generational collector.
 */

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/directmem.h"
#include "javavm/include/porting/time.h"

/*
 * This file is generated from the GC choice given at build time.
 */
#include "generated/javavm/include/gc_config.h"

#include "javavm/include/gc_common.h"
#include "javavm/include/gc/gc_impl.h"

#include "javavm/include/gc/generational/generational.h"

#include "javavm/include/gc/generational/gen_semispace.h"
#include "javavm/include/gc/generational/gen_markcompact.h"

#define CVM_GEN_NUM_GENERATIONS 2

/*
 * The actual generations
 */
static CVMGeneration* CVMgenGenerations[CVM_GEN_NUM_GENERATIONS];

/*
 * Time of the last major GC
 */
static CVMInt64 lastMajorGCTime;

/*
 * Define to GC on every n-th allocation try
 */
/*#define CVM_GEN_GC_ON_EVERY_ALLOC*/
#define CVM_GEN_NO_ALLOCS_UNTIL_GC 1

/*
 * The card table barrier
 */
CVMUint8* cardTable;
CVMUint8* cardTableVirtualBase;
CVMUint32 cardTableSize;
static CVMUint8* cardTableEnd;

static struct {
    CVMUint32 youngGenSize;
} genGCAttributes;

/*
 * The object header table to indicate the nearest object header to
 * a card. Same size as the card table.
 */
CVMInt8* objectHeaderTable;

/*
 * The card summary table
 */
CVMGenSummaryTableEntry* summaryTable;

/*
 * The beginning of the heap space, which the card table maps to.
 */
CVMUint32* heapBase;
static CVMUint32* heapBaseMemoryArea; /* For freeing purposes */

/*
 * Initialize GC global state
 */
void
CVMgcimplInitGlobalState(CVMGCGlobalState* globalState)
{
    CVMtraceMisc(("GC: Initializing global state for generational GC\n"));
}


/*
 * %comment: fresko010
 */
void
CVMgenClearBarrierTable()
{
    memset(cardTable, CARD_CLEAN_BYTE, cardTableSize);
    memset(objectHeaderTable, 0, cardTableSize);
}

/*
 * Update the object headers table for objects in range [startRange, endRange)
 */
void
CVMgenBarrierObjectHeadersUpdate(CVMGeneration* gen, CVMExecEnv* ee,
				 CVMGCOptions* gcOpts,
				 CVMUint32* startRange,
				 CVMUint32* endRange)
{
    CVMUint32* curr = startRange;
    CVMtraceGcCollect(("GC[SS,%s]: "
		       "object headers update [%x,%x)\n",
		       gcOpts->fullGC ? "full" : "partial",
		       startRange, endRange));
    while (curr < endRange) {
	CVMObject* currObj    = (CVMObject*)curr;
	CVMClassBlock* currCb = CVMobjectGetClass(currObj);
	CVMUint32  objSize    = CVMobjectSizeGivenClass(currObj, currCb);
	CVMObject* currObjEnd = (CVMObject*)(curr + objSize / 4);
	CVMUint8*  startCard  = CARD_TABLE_SLOT_ADDRESS_FOR(currObj);
	CVMUint8*  endCard    = CARD_TABLE_SLOT_ADDRESS_FOR(currObjEnd);
	/*
	 * I don't need to do anything if object starts on a card boundary
	 * since the object headers table is initialized to 0's. Otherwise
	 * I'd have to make a special case of it here, and check
	 * whether cardBoundary == curr.
	 */
	if (startCard != endCard) {
	    CVMUint8*  card;
	    CVMInt8*   hdr;
	    CVMUint32* cardBoundary;
	    CVMInt32   numCards;

	    /* Object crosses card boundaries. 
	       Find the first card for which the header table entry would
	       be >= 0. It's either startCard in case currObj is at
	       a card boundary, or startCard+1 if currObj is not at a
	       card boundary. */
	    cardBoundary = ROUND_UP_TO_CARD_SIZE(currObj);
	    card = CARD_TABLE_SLOT_ADDRESS_FOR(cardBoundary);
	    CVMassert((card == startCard) || (card == startCard + 1));
	    if (card == cardTableEnd) {
		CVMassert(card == startCard + 1);
		return; /* nothing to do -- we are at the edge */
	    }
	    hdr = HEADER_TABLE_SLOT_ADDRESS_FOR_CARD(card);
	    CVMassert(hdr >= objectHeaderTable);
	    CVMassert(hdr <  objectHeaderTable + cardTableSize);

	    /* Mark the number of words from this header */
	    CVMassert(cardBoundary >= curr);
	    CVMassert(cardBoundary - curr <= 127);
	    *hdr = cardBoundary - curr;
	    /* Now go through all the rest of the cards that the
	       object spans, and inform them that the start of the
	       object is here */
	    card += 1;
	    hdr  += 1;
	    numCards = -1;
	    if (endCard == cardTableEnd) {
		endCard--;
	    }
	    if (endCard - card < 127) {
		/* Fast common case, <= 127 iterations. */
		while (card <= endCard) {
		    CVMassert(hdr <  objectHeaderTable + cardTableSize);
		    *hdr = numCards; /* Plant back-pointers */
		    numCards--;
		    hdr++;
		    card++;
		}
		CVMassert(numCards >= -128);
	    } else {
		/* Slower rarer case for large objects */
		while (card <= endCard) {
		    CVMassert(hdr <  objectHeaderTable + cardTableSize);
		    *hdr = numCards; /* Plant back-pointers */
		    numCards--;
		    if (numCards == -127) {
			numCards = -1; /* Chain reaction! */
		    }
		    hdr++;
		    card++;
		}
	    }
	    CVMassert(card == endCard + 1);
	    CVMassert(HEADER_TABLE_SLOT_ADDRESS_FOR_CARD(card) == hdr);
	}
	curr += objSize / 4;
    }
    CVMassert(curr == endRange); /* This had better be exact */
}

/*
 * Scan objects in contiguous range, without doing any special handling.
 *
 * Scan from head of objStart. Scan only within [regionStart,regionEnd).
 * Call out to 'callback' only if the reference found is
 * not in the range [oldGenLower, oldGenHigher) 
 * (i.e. an old to young pointer).
 *
 * Returns the number of non-NULL old-to-young pointers. We can clear the
 * card if this number is 0.
 */
static CVMUint32
scanAndSummarizeObjectsOnCard(CVMExecEnv* ee, CVMGCOptions* gcOpts,
			      CVMUint32* objStart,
			      CVMGenSummaryTableEntry* summEntry,
			      CVMUint32* regionStart, CVMUint32* regionEnd,
			      CVMUint32* oldGenLower,
			      CVMRefCallbackFunc callback, void* callbackData)
{
    CVMUint32  numRefs = 0;
    CVMUint32* curr    = (CVMUint32*)objStart;
    CVMUint32* top     = regionEnd;
    CVMUint32* cardBoundary = CARD_BOUNDARY_FOR(regionStart);
    /* sanity check */
    CVMassert(SUMMARY_TABLE_SLOT_ADDRESS_FOR_CARD(
	          CARD_TABLE_SLOT_ADDRESS_FOR(regionStart)) == summEntry);
    CVMassert(summEntry >= summaryTable);
    CVMassert(summEntry <  &summaryTable[cardTableSize]);
    CVMtraceGcCollect(("GC[SS,%s]: "
		       "Scanning card range [%x,%x), objStart=0x%x\n",
		       gcOpts->fullGC ? "full" : "partial",
		       regionStart, regionEnd, objStart));
    /*
     * Clear summary table entry.
     * The summary table is 0xff terminated.
     */
    summEntry->intVersion = 0xffffffff;
    
    /*
     * And scan the objects on the card
     */
    while (curr < top) {
	CVMObject* currObj = (CVMObject*)curr;
	CVMClassBlock* currCb = CVMobjectGetClass(currObj);
	CVMUint32  objSize    = CVMobjectSizeGivenClass(currObj, currCb);
	CVMobjectWalkRefs(ee, gcOpts, currObj, currCb, {
	    CVMUint32* heapRef = *((CVMUint32**)refPtr);
	    /*
	     * Handle the range check here. If the pointer we found
	     * points to the old generation, 
	     * don't do the callback. We are only interested in pointers
	     * to the young generation.
	     *
	     * This one filters out NULL pointers as well.
	     * 
	     * %comment: fresko011
	     */
	    if (heapRef < oldGenLower) {
		if (heapRef != 0) {
		    if ((CVMUint32*)refPtr >= regionEnd) {
			/* We ran off the edge of the card */
			return numRefs;
		    } else if ((CVMUint32*)refPtr >= regionStart) {
			/* Call the callback if we are within the bounds */
			if (numRefs < 4) {
			    summEntry->offsets[numRefs] =
				(CVMUint32*)refPtr - cardBoundary;
			}
			numRefs++;
			(*callback)(refPtr, callbackData);
		    }
		}
	    }
	});
	/* iterate */
	curr += objSize / 4;
    }
    return numRefs;
}

static CVMUint32* 
findObjectStart(CVMUint8* card, CVMUint32* heapAddrForCard)
{
    CVMInt8* hdr = HEADER_TABLE_SLOT_ADDRESS_FOR_CARD(card);
    CVMInt8  n;
    CVMassert(hdr >= objectHeaderTable);
    CVMassert(hdr <  objectHeaderTable + cardTableSize);
    n = *hdr;
    CVMassert(HEAP_ADDRESS_FOR_CARD(card) == heapAddrForCard);
    if (n >= 0) {
	return heapAddrForCard - n;
    } else {
	CVMUint32* heapAddr = heapAddrForCard;
	do {
	    hdr = hdr + n; /* Go back! */
	    heapAddr = heapAddr + n * NUM_WORDS_PER_CARD;
	    CVMassert(hdr >= objectHeaderTable);
	    CVMassert(hdr <  objectHeaderTable + cardTableSize);
	    n = *hdr;
	} while (n < 0);
	CVMassert(heapAddr < heapAddrForCard);
	return heapAddr - n;
    }
}


/*#define CARD_STATS*/
#ifdef CARD_STATS
typedef struct cardStats {
    CVMUint32 cardsScanned;
    CVMUint32 cardsSummarized;
    CVMUint32 cardsClean;
    CVMUint32 cardsDirty;
} cardStats;

static cardStats cStats = {0, 0, 0, 0};
#define cardStatsOnly(x) x
#else
#define cardStatsOnly(x)
#endif

static void
callbackIfNeeded(CVMExecEnv* ee, CVMGCOptions* gcOpts,
		 CVMUint8* card, 
		 CVMUint32* lowerLimit, CVMUint32* higherLimit, 
		 CVMUint32* genLower, CVMUint32* genHigher,
		 CVMRefCallbackFunc callback,
		 void* callbackData)
{		
    CVMassert(card >= cardTable);
    CVMassert(card < cardTable + cardTableSize);
    cardStatsOnly(cStats.cardsScanned++);
    if (*card == CARD_SUMMARIZED_BYTE) {
	/*
	 * This card has not been dirtied since it was summarized. Scan the
	 * summary.
	 */
	CVMUint32* cardBoundary = CARD_BOUNDARY_FOR(lowerLimit);
	CVMGenSummaryTableEntry* summEntry =
	    SUMMARY_TABLE_SLOT_ADDRESS_FOR_CARD(card);
	CVMUint32 i;
	CVMUint8  offset;

	i = 0;
	/* If this card is summarized, it must have at least one entry */
	CVMassert(summEntry->offsets[0] != 0xff);
	while ((i < 4) && ((offset = summEntry->offsets[i]) != 0xff)) {
	    CVMObject** refPtr = (CVMObject**)(cardBoundary + offset);
	    /* We could not have summarized a NULL pointer */
	    CVMassert(*refPtr != NULL);
	    (*callback)(refPtr, callbackData);
	    i++;
	}
	CVMassert(i <= 4);
	cardStatsOnly(cStats.cardsSummarized++);
    } else if (*card == CARD_DIRTY_BYTE) {
	CVMUint32* objStart;
	CVMUint32  numOldYoungRefs;
	CVMGenSummaryTableEntry* summEntry;

	/* Make sure the heap pointer and the card we are scanning are in
           alignment */
	CVMassert(HEAP_ADDRESS_FOR_CARD(card) ==
		  CARD_BOUNDARY_FOR(lowerLimit));
	CVMassert(higherLimit - lowerLimit <= NUM_WORDS_PER_CARD);
	objStart = findObjectStart(card, CARD_BOUNDARY_FOR(lowerLimit));
	summEntry = SUMMARY_TABLE_SLOT_ADDRESS_FOR_CARD(card);
	numOldYoungRefs = scanAndSummarizeObjectsOnCard(ee, gcOpts, objStart,
			      summEntry, lowerLimit, higherLimit, genLower,
			      callback, callbackData);
	if (numOldYoungRefs == 0) {
	    /*
	     * No relevant pointers found. No need to scan until dirty again.  
	     */
	    *card = CARD_CLEAN_BYTE;
	} else if (numOldYoungRefs <= 4) {
	    *card = CARD_SUMMARIZED_BYTE;
	}
	cardStatsOnly(cStats.cardsDirty++);
    } else {
	CVMassert(*card == CARD_CLEAN_BYTE);
	cardStatsOnly(cStats.cardsClean++);
    }
		      
}

/*
 * Traverse all recorded pointers, and call 'callback' on each.
 */
void
CVMgenBarrierPointersTraverse(CVMGeneration* gen, CVMExecEnv* ee,
			      CVMGCOptions* gcOpts,
			      CVMRefCallbackFunc callback,
			      void* callbackData)
{
    CVMUint8*  lowerCardLimit;    /* Card to begin scanning              */
    CVMUint8*  higherCardLimit;   /* Card to end scanning                */
    CVMUint32* cardPtrWord;       /* Used for batch card scanning        */
    CVMUint32* heapPtr;           /* Track card boundaries in heap       */
    CVMInt32   remainder;         /* Batch scanning spillover            */
    CVMUint32* genLower;          /* Lower limit of live objects in gen  */
    CVMUint32* genHigher;         /* Higher limit of live objects in gen */

    /*
     * Scan cards between lowerCardLimit and higherCardLimit
     *
     * Iterate only over those pointers that belong to the old
     * generation. Stores may have been recorded for the young one
     * as well, but we can ignore those.
     */
    CVMtraceGcCollect(("GC[SS,%d,%s]: "
		       "Scanning barrier records for range [%x,%x)\n",
		       gen->generationNo,
		       gcOpts->fullGC ? "full" : "partial",
		       gen->allocBase, gen->allocPtr));

    /* We have to cover all pointers in [genLower,genHigher) */
    genLower        = gen->allocBase;
    genHigher       = gen->allocMark;

    /* ... and look at all the cards in [lowerCardLimit,higherCardLimit) */
    lowerCardLimit  = CARD_TABLE_SLOT_ADDRESS_FOR(genLower);
    higherCardLimit = CARD_TABLE_SLOT_ADDRESS_FOR(genHigher);

    /* This is the heap boundary corresponding to the next card */
    heapPtr         = NEXT_CARD_BOUNDARY_FOR(genLower);
    
    /*
     * Scan the first card in range, which may be partial.
     */

    /* Check whether the amount of data is really small or not */
    if (genHigher < heapPtr) {
	/* This is really rare. So we might as well check for the
	   condition that genLower == genHigher, i.e. an empty old
	   generation! */
	if (genLower == genHigher) {
	    return;
	}
	callbackIfNeeded(ee, gcOpts, lowerCardLimit,
			 genLower, genHigher,
			 genLower, genHigher,
			 callback, callbackData);
	return; /* Done, only one partial card! */
    } else {
	callbackIfNeeded(ee, gcOpts, lowerCardLimit,
			 genLower, heapPtr,
			 genLower, genHigher,
			 callback, callbackData);
	lowerCardLimit++;
    }
    
    /*
     * Have we already reached the end?
     */
    if (lowerCardLimit == higherCardLimit) {
	if (heapPtr == genHigher) {
	    return; /* nothing to do */
	}
	/*
	 * Make sure this is really the last card.
	 */
	CVMassert(genHigher - heapPtr < NUM_WORDS_PER_CARD);
	callbackIfNeeded(ee, gcOpts, lowerCardLimit,
			 heapPtr, genHigher,
			 genLower, genHigher,
			 callback, callbackData);
	return; /* Done, only one partial card! */
    }

    /*
     * How many individual card bytes are we going to look at until
     * we get to an integer boundary?
     */
    remainder =	CVMalignWordUp(lowerCardLimit) - (CVMUint32)lowerCardLimit;
    CVMassert(CARD_BOUNDARY_FOR(genLower) == genLower);
    /*
     * Get lowerCardLimit to a word boundary
     */
    while ((remainder-- > 0) && (lowerCardLimit < higherCardLimit)) {
	callbackIfNeeded(ee, gcOpts, lowerCardLimit,
			 heapPtr, heapPtr + NUM_WORDS_PER_CARD,
			 genLower, genHigher,
			 callback, callbackData);
	lowerCardLimit++;
	heapPtr += NUM_WORDS_PER_CARD;
    }
    /*
     * Nothing else to do
     */
    if (lowerCardLimit == higherCardLimit) {
	if (heapPtr == genHigher) {
	    return; /* nothing to do */
	}
	/* Make sure that we are "on the same page", literally! */
	CVMassert(HEAP_ADDRESS_FOR_CARD(higherCardLimit) == heapPtr);
	CVMassert(genHigher - heapPtr < NUM_WORDS_PER_CARD);
	callbackIfNeeded(ee, gcOpts, higherCardLimit,
			 heapPtr, genHigher,
			 genLower, genHigher,
			 callback, callbackData);
	return;
    }

    /*
     * lowerCardLimit had better be at a word boundary
     */
    CVMassert(CVMalignWordDown(lowerCardLimit) == (CVMUint32)lowerCardLimit);

    /*
     * Now adjust the higher card limit to a word boundary for batch
     * scanning.
     */
    remainder = (CVMUint32)higherCardLimit - CVMalignWordDown(higherCardLimit);
    higherCardLimit -= remainder;
    CVMassert(CVMalignWordDown(higherCardLimit) == (CVMUint32)higherCardLimit);

    /*
     * Now go through the card table in blocks of four for faster
     * zero checks.
     */
    for (cardPtrWord = (CVMUint32*)lowerCardLimit;
	 cardPtrWord < (CVMUint32*)higherCardLimit;
	 cardPtrWord++, heapPtr += NUM_WORDS_PER_CARD * 4) {
	if (*cardPtrWord != FOUR_CLEAN_CARDS) {
	    CVMUint32* hptr = heapPtr;
	    CVMUint8*  cptr = (CVMUint8*)cardPtrWord;
	    CVMUint8*  cptr_end = cptr + 4;
	    for (; cptr < cptr_end; cptr++, hptr += NUM_WORDS_PER_CARD) {
		callbackIfNeeded(ee, gcOpts, cptr,
				 hptr, hptr + NUM_WORDS_PER_CARD,
				 genLower, genHigher,
				 callback, callbackData);
	    }
	}
    }

    /*
     * And finally, the remaining few cards, if any, that "spilled" out
     * of the word boundary.
     */
    while (remainder-- > 0) {
	callbackIfNeeded(ee, gcOpts, higherCardLimit,
			 heapPtr, heapPtr + NUM_WORDS_PER_CARD,
			 genLower, genHigher,
			 callback, callbackData);
	higherCardLimit++;
	heapPtr += NUM_WORDS_PER_CARD;
    }
    /*
     * Now if there is a partial card left, handle it.
     */
    if (heapPtr < genHigher) {
	/* Make sure that we are "on the same page", literally! */
	CVMassert(HEAP_ADDRESS_FOR_CARD(higherCardLimit) == heapPtr);
	CVMassert(genHigher - heapPtr < NUM_WORDS_PER_CARD);
	callbackIfNeeded(ee, gcOpts, higherCardLimit,
			 heapPtr, genHigher,
			 genLower, genHigher,
			 callback, callbackData);
    }
}

struct CVMClassScanOptions {
    CVMRefCallbackFunc callback;
    void* callbackData;
    CVMBool allClassesAreRoots;
    CVMGeneration* generation; /* Scan only classes in this generation */
};

typedef struct CVMClassScanOptions CVMClassScanOptions;

/*
 * Callback function to scan a promoted class' statics
 */
static void
CVMgenScanClassPointersInGeneration(CVMExecEnv* ee, CVMClassBlock* cb,
				    void* opts)
{
    CVMObject* instance = *((CVMObject**)CVMcbJavaInstance(cb));
    CVMClassScanOptions* options = (CVMClassScanOptions*)opts;
    CVMGeneration* gen = options->generation;
    CVMassert(!options->allClassesAreRoots);
    /*
     * Scan the class state if the java.lang.Class instance is in this
     * generation.
     */
    if (gen->inGeneration(gen, instance)) {
	CVMscanClassIfNeeded(ee, cb, options->callback, options->callbackData);
    }
}

/*
 * Find all class instances in generation 'gen' and scan them
 * for pointers. 
 */
static void
CVMgenScanClassPointersInGen(CVMGeneration* gen,
			     CVMExecEnv* ee, CVMGCOptions* gcOpts,
			     CVMRefCallbackFunc callback,
			     void* callbackData)
{
    CVMClassScanOptions opts;

    opts.callback = callback;
    opts.callbackData = callbackData;
    opts.allClassesAreRoots = gcOpts->allClassesAreRoots;
    opts.generation = gen;
    CVMclassIterateDynamicallyLoadedClasses(
        ee, CVMgenScanClassPointersInGeneration, &opts);
}


/*
 * Initialize a heap of 'heapSize' bytes. heapSize is guaranteed to be
 * divisible by 4.  */
CVMBool
CVMgcimplInitHeap(CVMGCGlobalState* globalState,
		  CVMUint32 heapSize,
		  CVMUint32 maxHeapSize)
{
    CVMGeneration* youngGen;
    CVMGeneration* oldGen;
    CVMUint32  totBytes; /* Total memory space for all generations */
    CVMUint32  totYoungGenSize; /* Total for both semispaces */
    CVMUint32  roundedSemispaceSize;
    char* youngGenAttr;
    /*
     * Let the minimum total heap size be 1K.
     */
#define MIN_TOTAL_HEAP (1 << 10)

    /*
     * Make a sane minimum check on the total size of the heap.
     */
    if (heapSize < MIN_TOTAL_HEAP) {
	CVMdebugPrintf(("GC[generational]: "
                        "Stated heap size %d bytes is smaller than "
			"minimum %d bytes\n",
			heapSize, MIN_TOTAL_HEAP));
	CVMdebugPrintf(("\tTotal heap size now %d bytes\n", MIN_TOTAL_HEAP));
	heapSize = MIN_TOTAL_HEAP;
    }

    /*
     * Parse relevant attributes
     */
    youngGenAttr = CVMgcGetGCAttributeVal("youngGen");

    if (youngGenAttr == NULL) {
	genGCAttributes.youngGenSize = 1 << 20; /* The default size */
    } else {
	CVMInt32 size = CVMatoml(youngGenAttr);
	if (size < 0) {
	    /* ignore -- the empty string, or no value at all */
	    size = 1 << 20;
	}
	genGCAttributes.youngGenSize = size;
    }

    /*
     * At this point, we might find that the total size of the heap is
     * smaller than the stated size of the young generation. Make the
     * heap all young in that case. */

    if (heapSize < genGCAttributes.youngGenSize) {
	CVMdebugPrintf(("GC[generational]: "
                        "Total heap size %d bytes < "
			"young gen size of %d bytes\n",
			heapSize, genGCAttributes.youngGenSize));
	CVMdebugPrintf(("\tYoung generation size now %d bytes\n",
			heapSize));
	genGCAttributes.youngGenSize = heapSize;
    }
    
    /*
     * Now that the user specified sizes of the young generation and
     * total heap have stabilized, do one final sanity check -- a
     * comparison between young gen size and old gen size. If the
     * young gen size is less than 1M, it should not be smaller than,
     * say, around 1/8 the total heap size. If it's too small with
     * regards to the mark-compact generation, the mark-compact
     * algorithm will not have enough scratch space to work with,
     * causing eventual crashes.  
     */
    if ((genGCAttributes.youngGenSize < (1 << 20)) &&
	(genGCAttributes.youngGenSize < heapSize / 8)) {
	CVMdebugPrintf(("GC[generational]: "
			"young gen size of %d bytes < 1M\n",
			genGCAttributes.youngGenSize));
	CVMdebugPrintf(("\tYoung generation size now %d bytes\n",
			heapSize / 8));
	genGCAttributes.youngGenSize = heapSize / 8;
    }
    /*
     * %comment fresko001 
     */
    roundedSemispaceSize = 
	(CVMUint32)ROUND_UP_TO_CARD_SIZE(genGCAttributes.youngGenSize);
    totYoungGenSize = roundedSemispaceSize * 2;

    /* Add the size of one 'wasted' semispace. */
    totBytes = heapSize + roundedSemispaceSize; 

    /*
     * Round up to card size and add a card size for good measure.
     * We're going to need the slack, because we are going to round
     * the heap up to a card boundary.
     */
    totBytes = (CVMUint32)ROUND_UP_TO_CARD_SIZE(totBytes) + NUM_BYTES_PER_CARD;

    /*
     * A heap space allocator might be required here, where a porter
     * may have the option of allocating the heap with something other
     * than a malloc.
     */
    heapBaseMemoryArea = (CVMUint32*)calloc(1, totBytes);
    if (heapBaseMemoryArea == NULL) {
	/* The caller will signal heap initialization failure */
	return CVM_FALSE;
    }
    /*
     * Remember the beginning of this space. This is how we are going to
     * map a store into a card table entry.
     */
    heapBase = ROUND_UP_TO_CARD_SIZE(heapBaseMemoryArea);

    /*
     * Now totBytes is an integral number of cards. Subtract the slack,
     * and you get a safe region that is [heapBase, heapBase+totBytes)
     */
    totBytes -= NUM_BYTES_PER_CARD;

    /*
     * Allocate the extra data structures necessary
     */
    cardTableSize = totBytes / NUM_BYTES_PER_CARD;
    CVMassert(cardTableSize * NUM_BYTES_PER_CARD == totBytes);
    cardTable = (CVMUint8*)calloc(1, cardTableSize);
    if (cardTable == NULL) {
	/* The caller will signal heap initialization failure */
	free(heapBaseMemoryArea);
	return CVM_FALSE;
    }
    cardTableEnd = cardTable + cardTableSize;
    /*
     * The objectHeaderTable records locations of the nearest object headers
     * for a given card. Its size is the same as the card table.
     */
    objectHeaderTable = (CVMInt8*)calloc(1, cardTableSize);
    if (objectHeaderTable == NULL) {
	/* The caller will signal heap initialization failure */
	free(cardTable);
	free(heapBaseMemoryArea);
	return CVM_FALSE;
    }
    /*
     * The summaryTable records up to 4 locations of pointers
     * per card. If the card is not dirtied after it is summarized,
     * we can use the summary instead of scanning the card again.
     */
    summaryTable = 
	(CVMGenSummaryTableEntry*)calloc(cardTableSize,
					 sizeof(CVMGenSummaryTableEntry));
    if (summaryTable == NULL) {
	/* The caller will signal heap initialization failure */
	free(cardTable);
	free(heapBaseMemoryArea);
	free(objectHeaderTable);
	return CVM_FALSE;
    }
    /*
     * Computing this virtual origin for the card table makes marking
     * the dirty byte very cheap.
     */
    cardTableVirtualBase =
	&cardTable[-((CVMInt32)(((CVMUint32)heapBase)/NUM_BYTES_PER_CARD))];
    
    CVMassert((totYoungGenSize & (NUM_BYTES_PER_CARD - 1)) == 0);

    youngGen = CVMgenSemispaceAlloc(heapBase, totYoungGenSize); 

    if (youngGen == NULL) {
	free(cardTable);
	free(heapBaseMemoryArea);
	free(objectHeaderTable);
	return CVM_FALSE;
    }

    youngGen->generationNo = 0;
    CVMgenGenerations[0] = youngGen;

    /*
     * And allocate the desired amount of heap space
     */
    oldGen = CVMgenMarkCompactAlloc(heapBase + totYoungGenSize / 4,
				    totBytes - totYoungGenSize);

    /*
     * Make sure old space is on a card boundary!
     */
    CVMassert((((CVMUint32)(heapBase + totYoungGenSize / 4)) & 
	      (NUM_BYTES_PER_CARD-1)) == 0);
    if (oldGen == NULL) {
	free(cardTable);
	free(heapBaseMemoryArea);
	free(objectHeaderTable);
	CVMgenSemispaceFree((CVMGenSemispaceGeneration*)youngGen);
	return CVM_FALSE;
    }

    oldGen->generationNo = 1;
    CVMgenGenerations[1] = oldGen;

    youngGen->nextGen = oldGen;
    youngGen->prevGen = NULL;

    oldGen->nextGen = NULL;
    oldGen->prevGen = youngGen;

    /*
     * Initialize the barrier
     */
    CVMgenClearBarrierTable();
    /*
     * Initialize GC times. Let heap initialization be the first
     * major GC.
     */
    lastMajorGCTime = CVMtimeMillis();
    CVMdebugPrintf(("GC[generational]: Auxiliary data structures\n"));
    CVMdebugPrintf(("\theapBaseMemoryArea=[0x%x,0x%x)\n",
		    heapBaseMemoryArea,
		    heapBaseMemoryArea + (totBytes + NUM_BYTES_PER_CARD) / 4));
    CVMdebugPrintf(("\tcardTable=[0x%x,0x%x)\n",
		    cardTable, cardTable + cardTableSize));
    CVMdebugPrintf(("\tobjectHeaderTable=[0x%x,0x%x)\n",
		    objectHeaderTable, objectHeaderTable + cardTableSize));
    CVMdebugPrintf(("\tsummaryTable=[0x%x,0x%x)\n",
		    summaryTable, summaryTable + cardTableSize));
    CVMtraceMisc(("GC: Initialized heap for generational GC\n"));
    return CVM_TRUE;
}

/*
 * Return the number of bytes free in the heap. 
 */
CVMUint32
CVMgcimplFreeMemory(CVMExecEnv* ee)
{
    CVMGeneration* youngGen = CVMgenGenerations[0];
    CVMGeneration* oldGen = CVMgenGenerations[1];
    return youngGen->freeMemory(youngGen, ee) +
	oldGen->freeMemory(oldGen, ee);
}

/*
 * Return the amount of total memory in the heap, in bytes. Take into
 * account one semispace size only, since that is the useful amount.
 */
CVMUint32
CVMgcimplTotalMemory(CVMExecEnv* ee)
{
    CVMGeneration* youngGen = CVMgenGenerations[0];
    CVMGeneration* oldGen = CVMgenGenerations[1];
    return youngGen->totalMemory(youngGen, ee) +
	oldGen->totalMemory(oldGen, ee);
}

/*
 * Some generational service routines
 */

/*
 * Scan the root set of collection, as well as all pointers from other
 * generations
 */
void
CVMgenScanAllRoots(CVMGeneration* thisGen,
		   CVMExecEnv *ee, CVMGCOptions* gcOpts,
		   CVMRefCallbackFunc callback, void* data)
{
    CVMtraceGcCollect(("GC[SS,%d,%s]: "
		       "Scanning roots of generation\n",
		       thisGen->generationNo,
		       gcOpts->fullGC ? "full" : "partial"));
    /*
     * Clear the class marks if we are going to scan classes by
     * reachability.  
     */
    if (!gcOpts->allClassesAreRoots) {
	CVMgcClearClassMarks(ee, gcOpts);
    }
    /*
     * First scan pointers to this generation from others, depending
     * on whether we are scanning roots for the young or the old
     * generation.
     */
    if (thisGen->generationNo == 0) {
	/* Collecting young generation */

	/* First scan older to younger pointers */
	CVMGeneration* oldGen = CVMgenGenerations[1];
	/*
	 * Scan older to younger pointers up until 'allocMark'.
	 * These are pointers for which the object header table
	 * has already been updated in the last old gen GC.
	 */
	oldGen->scanOlderToYoungerPointers(oldGen, ee,
					   gcOpts, callback, data);
	/*
	 * Get any allocated pointers in the old generation for which
	 * object headers and card tables have not been updated yet.
	 *
	 * This is going to happen the first time CVMgenScanAllRoots()
	 * is called. The next time, scanOlderToYoungerPointers will
	 * pick all old->young pointers up.  
	 */
	if (oldGen->allocPtr > oldGen->allocMark) {
	    oldGen->scanPromotedPointers(oldGen, ee, gcOpts, callback, data);
	    CVMassert(oldGen->allocPtr == oldGen->allocMark);
	}
	if (!gcOpts->allClassesAreRoots) {
	    /* And scan classes in old generation that are pointing to
	       young */
	    CVMgenScanClassPointersInGen(oldGen, ee, gcOpts, callback, data);
	}
    } else {
	/* Collecting old generation */
	CVMGeneration* youngGen = CVMgenGenerations[0];
	/*
	 * For now, assume one or two generations
	 */
	CVMassert(thisGen->generationNo == 1);

	/* Fist scan younger to older pointers */
	youngGen->scanYoungerToOlderPointers(youngGen, ee,
					     gcOpts, callback, data);
	if (!gcOpts->allClassesAreRoots) {
	    /* And scan classes in young generation that are pointing to
	       old */
	    CVMgenScanClassPointersInGen(youngGen, ee, gcOpts, callback, data);
	}
    }
    /*
     * And now go through "system" pointers to this generation.
     */
    CVMgcScanRoots(ee, gcOpts, callback, data);
}

/*
 * This routine is called by the common GC code after all locks are
 * obtained, and threads are stopped at GC-safe points. It's the
 * routine that needs a snapshot of the world while all threads are
 * stopped (typically at least a root scan).
 *
 * GC enough to satisfy allocation of 'numBytes' bytes.
 */

void
CVMgcimplDoGC(CVMExecEnv* ee, CVMUint32 numBytes)
{
    CVMGCOptions gcOpts;
    CVMBool success;
    CVMGeneration* oldGen;
    CVMGeneration* youngGen;
/*#define SOLARIS_TIMING*/
#ifdef SOLARIS_TIMING
#include <sys/time.h>
    static CVMUint32 totGCTime = 0;
    hrtime_t time;
    CVMUint32 ms;

    time = gethrtime();
#endif /* SOLARIS_TIMING */
    gcOpts.fullGC = CVM_TRUE;
    gcOpts.allClassesAreRoots = CVM_FALSE;
    gcOpts.discoverWeakReferences = CVM_FALSE;

    /*
     * Do generation-specific initialization
     */
    youngGen = CVMgenGenerations[0];
    oldGen = CVMgenGenerations[1];
    youngGen->startGC(youngGen, ee, &gcOpts);
    oldGen->startGC(oldGen, ee, &gcOpts);

    /* Full collection */
    success = youngGen->collect(youngGen, ee, numBytes, &gcOpts);
    
    /*
     * Try an old-space GC if no success
     */
    if (!success) {
	/*
	 * Now clear the class marks again. Classes may have to be scanned
	 * multiple times to get pointers to all generations, since each
	 * generation filters out what it does not know how to handle.
	 */
	oldGen->collect(oldGen, ee, numBytes, &gcOpts);
	lastMajorGCTime = CVMtimeMillis();
    }

    youngGen->endGC(youngGen, ee, &gcOpts);
    oldGen->endGC(oldGen, ee, &gcOpts);

#ifdef SOLARIS_TIMING
    time = gethrtime() - time;
    ms = time / 1000000;
    totGCTime += ms;
    CVMconsolePrintf("GC took %d ms (%d ms total)\n", ms, totGCTime);
#endif
    cardStatsOnly({
	CVMconsolePrintf("CARDS: scanned=%d clean=%d (%d%%) "
			 "dirty=%d (%d%%) summarized=%d (%d%%)\n",
			 cStats.cardsScanned,
			 cStats.cardsClean,
			 cStats.cardsClean * 100 / cStats.cardsScanned,
			 cStats.cardsDirty,
			 cStats.cardsDirty * 100 / cStats.cardsScanned,
			 cStats.cardsSummarized,
			 cStats.cardsSummarized * 100 /
			 cStats.cardsScanned);
	memset(&cStats, 0, sizeof(cStats));
    });
}

/*
 * Allocate uninitialized heap object of size numBytes
 */
CVMObject*
CVMgcimplAllocObject(CVMExecEnv* ee, CVMUint32 numBytes)
{
    CVMObject* allocatedObj;
    CVMGeneration* youngGen = CVMgenGenerations[0];
    CVMGeneration* oldGen = CVMgenGenerations[1];
#define LARGE_OBJECTS_TREATMENT
/*#define LARGE_OBJECT_STATS*/
#ifdef LARGE_OBJECT_STATS
    static CVMUint32
	largeAllocations = 0,
	smallAllocations = 0,
	totalAllocations = 0;
#endif

#ifdef CVM_GEN_GC_ON_EVERY_ALLOC
    static CVMUint32 allocCount = 1;
    if ((allocCount % CVM_GEN_NO_ALLOCS_UNTIL_GC) == 0) {
	/* Do one for stress-test. If it was unsuccessful, try again in the
	   next allocation */
	if (CVMgcStopTheWorldAndGC(ee, numBytes)) {
	    allocCount = 1;
	}
    } else {
	allocCount++;
    }
#endif
#ifdef LARGE_OBJECTS_TREATMENT
#define LARGE_OBJECT_THRESHOLD 50000
    if (numBytes > LARGE_OBJECT_THRESHOLD) {
	CVMgenContiguousSpaceAllocate(oldGen, numBytes, allocatedObj);
    } else {
	CVMgenContiguousSpaceAllocate(youngGen, numBytes, allocatedObj);
    }
#ifdef LARGE_OBJECT_STATS
    if (numBytes > LARGE_OBJECT_THRESHOLD) {
	largeAllocations++;
    } else {
	smallAllocations++;
    }
    totalAllocations++;
    if (totalAllocations == 100000) {
	CVMconsolePrintf("ALLOCATIONS: "
			 "tot=%d, small=%d, large=%d, %% large=%d\n",
			 totalAllocations, smallAllocations, largeAllocations,
			 largeAllocations * 100 / totalAllocations);
	totalAllocations = largeAllocations = smallAllocations = 0;
    }
#endif
#else
    /*
     * No special treatment for large objects. Always try to allocate
     * from young space first.
     */
    CVMgenContiguousSpaceAllocate(youngGen, numBytes, allocatedObj);
#endif

    if (allocatedObj == NULL) {
	/* GC and re-try allocation */
	if (CVMgcStopTheWorldAndGC(ee, numBytes)) {
	    /*
	     * Re-try in the young generation, but only if GC has been
	     * able to run. If not, don't even bother.
	     */
	    CVMgenContiguousSpaceAllocate(youngGen, numBytes, allocatedObj);
	}
	if (allocatedObj == NULL) {
	    /* Try hard if we've already GC'ed! */
	    CVMgenContiguousSpaceAllocate(oldGen, numBytes, allocatedObj);
	}
    }
    return allocatedObj;
}

/*
 * Destroy GC global state
 */
void
CVMgcimplDestroyGlobalState(CVMGCGlobalState* globalState)
{
    CVMtraceMisc(("Destroying global state for generational GC\n"));
}

/*
 * Destroy heap
 */
CVMBool
CVMgcimplDestroyHeap(CVMGCGlobalState* globalState)
{
    CVMtraceMisc(("Destroying heap for generational GC\n"));

    free(heapBaseMemoryArea);
    free(cardTable);
    free(objectHeaderTable);
    free(summaryTable);

    heapBaseMemoryArea = NULL;
    cardTable = NULL;
    objectHeaderTable = NULL;
    summaryTable = NULL;

    CVMgenSemispaceFree((CVMGenSemispaceGeneration*)CVMgenGenerations[0]);
    CVMgenMarkCompactFree((CVMGenMarkCompactGeneration*)CVMgenGenerations[1]);
	
    CVMgenGenerations[0] = NULL;
    CVMgenGenerations[1] = NULL;

    return CVM_TRUE;
}

/*
 * JVM_MaxObjectInspectionAge() support
 */
CVMInt64
CVMgcimplTimeOfLastMajorGC()
{
    return lastMajorGCTime;
}

#ifdef CVM_DEBUG

/*
 * Heap iteration. Call (*callback)() on each object in the heap.
 */
void
CVMgcimplIterateHeap(CVMExecEnv* ee, 
		     CVMObjectCallbackFunc callback, void* callbackData)
{
    int i;
    /*
     * Iterate over objects in all generations
     */
    for (i = 0; i < CVM_GEN_NUM_GENERATIONS; i++) {
	CVMGeneration* gen  = CVMgenGenerations[i];
	CVMUint32*     base = gen->allocBase;
	CVMUint32*     top  = gen->allocPtr;
	CVMgcScanObjectRange(ee, base, top, callback, callbackData);
    }
}

#endif /* CVM_DEBUG */
