/*
 * -----------------------------------------------------------------
 * jvm.c : Implementations of jvm.h functions.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)jvm.c	1.207 00/11/10
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/* Implement jvm.h functions here, unless they more suitably belong
 * to somewhere else.
 */
#include "javavm/export/jvm.h"
#include "javavm/export/jni.h"
#include "native/common/jni_util.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/gc_common.h"
#include "javavm/include/porting/threads.h"
#include "javavm/include/porting/time.h"
#include "javavm/include/porting/int.h"
#include "javavm/include/porting/doubleword.h"
#include "javavm/include/clib.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/globalroots.h"
#include "javavm/include/localroots.h"
#include "javavm/include/preloader.h"
#include "javavm/include/common_exceptions.h"
#include "generated/offsets/java_lang_Class.h"
#include "generated/offsets/java_lang_Thread.h"
#include "generated/offsets/java_lang_Throwable.h"

#ifdef CVM_JVMDI
#include "javavm/include/jvmdi_impl.h"
#endif

#ifdef HPROF
#include "vmprofiler.h"
#endif

#ifdef CVM_REFLECT
# include "javavm/include/reflect.h"
#endif

#ifdef CVM_DYNAMIC_LINKING
# include "javavm/include/porting/linker.h"
#endif

#include "javavm/include/mem_registry.h"

JNIEXPORT jint JNICALL
JVM_GetInterfaceVersion(void)
{
    return JVM_INTERFACE_VERSION;
}

JNIEXPORT jint JNICALL
JVM_GetLastErrorString(char *buf, int len)
{
    return CVMioGetLastErrorString(buf, len);
}

/*
 * Find primitive classes
 */
JNIEXPORT jclass JNICALL
JVM_FindPrimitiveClass(JNIEnv *env, const char *utf)
{
    CVMExecEnv*    ee  = CVMjniEnv2ExecEnv(env);
    CVMClassTypeID tid;
    CVMClassBlock* cb;

    /*
     * We need to use the "new" function, because the "lookup" function
     * is only valid when we know that the type already exists.
     */
    tid = CVMtypeidNewClassID(ee, utf, strlen(utf));

    if (tid == CVM_TYPEID_ERROR) {
	/* CVMtypeidNewClassID() will always succeed if the type already
	 * exist, so it must not be primitive.
	 */
	CVMclearLocalException(ee);
	return NULL;
    }

    cb = CVMpreloaderLookupPrimitiveClassFromType(tid);
    CVMtypeidDisposeClassID(ee, tid);

    if (cb != NULL) {
	return (*env)->NewLocalRef(env, CVMcbJavaInstance(cb));
    } else {
	return NULL;
    }
}

/*
 * Link the class
 */
JNIEXPORT void JNICALL
JVM_ResolveClass(JNIEnv *env, jclass cls)
{
#ifdef CVM_CLASSLOADING
    CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(CVMjniEnv2ExecEnv(env),
						     cls);
    CVMclassLink(CVMjniEnv2ExecEnv(env), cb);
#else
    return; /* the class must be romized and is therefore linked */
#endif
}

JNIEXPORT jclass JNICALL
JVM_FindClassFromClassLoader(JNIEnv *env, const char *name, jboolean init,
			     jobject loader, jboolean throwError)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb =
	CVMclassLookupByNameFromClassLoader(ee, name, init, 
					    loader, NULL, throwError);

    if (cb != NULL) {
	return (*env)->NewLocalRef(env, CVMcbJavaInstance(cb));
    }
    return NULL;
}

JNIEXPORT jclass JNICALL
JVM_DefineClass(JNIEnv *env, const char *name, jobject loaderRef,
		const jbyte *buf, jsize bufLen, jobject pd)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    return CVMdefineClass(ee, name, loaderRef, (CVMUint8*)buf, bufLen, pd);
}

/*
 * Reflection support functions
 */

JNIEXPORT jstring JNICALL
JVM_GetClassName(JNIEnv* env, jclass cls)
{
    /* Try to avoid 1 malloc */
    char buf[128];
    char* className = buf;
    /* Consider caching the classname String with the class. */
    CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(CVMjniEnv2ExecEnv(env),
						     cls);
    jobject res;

    if (!CVMtypeidClassNameToCString(CVMcbClassName(cb), buf, 128)) {
	className = 
	    CVMtypeidClassNameToAllocatedCString(CVMcbClassName(cb));
    }
    TranslateFromVMClassName(className);
    res = (*env)->NewLocalRef(env, (*env)->NewStringUTF(env, className));
    if (className != buf)
	free(className);
    return res;
}


JNIEXPORT jobjectArray JNICALL
JVM_GetClassInterfaces(JNIEnv *env, jclass cls)
{
#ifdef CVM_REFLECT
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    jobject obj = CVMjniCreateLocalRef(ee);
    if (obj == NULL) { /* exception thrown */
	return NULL;
    }
    CVMreflectInterfaces(ee,
			 CVMgcSafeClassRef2ClassBlock(ee, cls),
			 (CVMArrayOfRefICell*) obj);
    if (CVMexceptionOccurred(ee)) {
	CVMjniDeleteLocalRef(env, obj);
	return NULL;
    }
    return obj;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

JNIEXPORT jobject JNICALL
JVM_GetClassLoader(JNIEnv *env, jclass cls)
{
#ifdef CVM_CLASSLOADING
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(ee, cls);
    return (*env)->NewLocalRef(env, CVMcbClassLoader(cb));
#else
    return NULL;
#endif
}

JNIEXPORT jboolean JNICALL
JVM_IsInterface(JNIEnv *env, jclass cls)
{
    CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(CVMjniEnv2ExecEnv(env),
						     cls);
    return CVMcbIs(cb, INTERFACE);
}

#ifdef JDK12
/*
 * Get the class's signers, if any. Make sure to return a new array,
 * not the original. Returning the original is a security hole as
 * the caller could modify it.
 */
JNIEXPORT jobjectArray JNICALL
JVM_GetClassSigners(JNIEnv *env, jclass cls)
{
    Hjava_lang_Class *cb = (Hjava_lang_Class *)DeRef(env, cls);
    HArrayOfObject *copy, *orig;
    HObject **copy_data, **orig_data;
    int i;
    size_t n;

    orig = (HArrayOfObject *)  unhand(cb)->signers;
    if (orig == NULL) 
	return NULL;

    n = obj_length(orig);

    copy = (HArrayOfObject *)ArrayAlloc(T_CLASS, n);
    if (copy == NULL) {
	ThrowOutOfMemoryError(CVMjniEnv2ExecEnv(env), 0);
	return NULL;
    }

    copy_data = (HObject **) unhand(copy)->body;
    orig_data = (HObject **) unhand(orig)->body;

    /* Use i <= n, NOT i < n, to get copy_data[n] 
       which contains the type of objects in the array */

    for (i=0; i <= n; i++) {
	copy_data[i] = orig_data[i];
    }

    KEEP_POINTER_ALIVE(copy_data);
    KEEP_POINTER_ALIVE(orig_data);

    return MkRefLocal(env, copy);
}

JNIEXPORT void JNICALL
JVM_SetClassSigners(JNIEnv *env, jclass cls, jobjectArray signers)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    unhand(cb)->signers = (HArrayOfObject *)DeRef(env, signers);
}
#endif

JNIEXPORT jobject JNICALL
JVM_GetProtectionDomain(JNIEnv *env, jclass cls)
{
#ifdef CVM_CLASSLOADING
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb;
    cb = CVMgcSafeClassRef2ClassBlock(ee, cls);
    return (*env)->NewLocalRef(env, CVMcbProtectionDomain(cb));
#else
    return NULL;
#endif
}

JNIEXPORT void JNICALL
JVM_SetProtectionDomain(JNIEnv *env, jclass cls, jobject pd)
{
#ifdef CVM_CLASSLOADING
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(ee, cls);

    /* Allocate a global root for the protection domain if necessary. */
    if (CVMcbProtectionDomain(cb) == NULL) {
	CVMcbProtectionDomain(cb) = CVMID_getGlobalRoot(ee);
	if (CVMcbProtectionDomain(cb) == NULL) {
	    CVMthrowOutOfMemoryError(ee, NULL);
	    return;
	}
    }

    CVMID_icellAssign(ee, CVMcbProtectionDomain(cb), pd);
#endif
}

JNIEXPORT jboolean JNICALL
JVM_IsArrayClass(JNIEnv *env, jclass cls)
{
    CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(CVMjniEnv2ExecEnv(env),
						     cls);
    return CVMisArrayClass(cb);
}

JNIEXPORT jboolean JNICALL
JVM_IsPrimitiveClass(JNIEnv *env, jclass cls)
{
    CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(CVMjniEnv2ExecEnv(env),
						     cls);
    return CVMcbIs(cb, PRIMITIVE);
}

JNIEXPORT jclass JNICALL
JVM_GetComponentType(JNIEnv *env, jclass cls)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(ee, cls);
    CVMClassBlock* elementCb;
    if (!CVMisArrayClass(cb)) {
	return NULL;
    }

    elementCb = CVMarrayElementCb(cb);
    CVMassert(elementCb != NULL);

    return (*env)->NewLocalRef(env, CVMcbJavaInstance(elementCb));
}

/* Remember that ACC_SUPER is added by the VM, and must be stripped. */
JNIEXPORT jint JNICALL
JVM_GetClassModifiers(JNIEnv *env, jclass cls)
{
    CVMClassBlock* cb = 
	CVMgcSafeClassRef2ClassBlock(CVMjniEnv2ExecEnv(env), cls);
    int icount = CVMcbInnerClassesInfoCount(cb);

    /* Check if this class happens to be a member class. */
    if (icount != 0) {
	int i;
	for (i = 0; i < icount; i++) {
	    CVMConstantPool* cp = CVMcbConstantPool(cb);
	    int classIdx = CVMcbInnerClassInfo(cb, i)->innerClassIndex;
	    CVMClassBlock* innerCb;
#ifdef CVM_CLASSLOADING
	    if (CVMcpIsResolved(cp, classIdx)) {
		innerCb = CVMcpGetCb(cp, classIdx);
	    } else {
		CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
		CVM_LOADERCACHE_LOCK(ee);
		innerCb =
		    CVMloaderCacheLookup(ee,
					 CVMcpGetClassTypeID(cp, classIdx),
					 CVMcbClassLoader(cb));
		CVM_LOADERCACHE_UNLOCK(ee);
	    }
#else
	    innerCb = CVMcpGetCb(cp, classIdx);
#endif
	    if (cb == innerCb) {
		/* Aha, this is really a member class. */
		jint access =
		    CVMcbInnerClassInfo(cb, i)->innerClassAccessFlags;
		/* Unlike CVMcbAccessFlags(), these access flags are 
		 * not modifed so they will fit in one bytes. Therefore,
		 * no translation is necessary.
		 */
		return (access & (~JVM_ACC_SUPER));
	    }
	}
    }

    /*
     * It not a member class. We need to translate the flags since they
     * were modified to make CVMcbAccessFlags() fit in one byte.
     */
    {
	jint access = 0;
	if (CVMcbIs(cb, PUBLIC))
	    access |= JVM_ACC_PUBLIC;
	if (CVMcbIs(cb, FINAL))
	    access |= JVM_ACC_FINAL;
	if (CVMcbIs(cb, INTERFACE))
	    access |= JVM_ACC_INTERFACE;
	if (CVMcbIs(cb, ABSTRACT))
	    access |= JVM_ACC_ABSTRACT;
	return access;
    }
}

/*
 * NOTE: The following routines are conditionally compiled. These can
 * not simply be removed from the build, since java.lang.Class refers
 * to them and we don't have JavaFilter incorporated yet. Therefore
 * they fail by calling CVMthrowUnsupportedOperationException when
 * reflection is not configured in.
 * JVM_GetClassFields
 * JVM_GetClassMethods
 * JVM_GetClassConstructors
 * JVM_GetClassField
 * JVM_GetClassMethod
 * JVM_GetClassConstructor
 * JVM_GetDeclaredClasses
 * JVM_GetDeclaringClass
 */

JNIEXPORT jobjectArray JNICALL
JVM_GetClassFields(JNIEnv *env, jclass cls, jint which)
{
#ifdef CVM_REFLECT
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    jobject obj = CVMjniCreateLocalRef(ee);
    if (obj == NULL) /* exception thrown */
	return NULL;
    CVMreflectFields(ee, CVMgcSafeClassRef2ClassBlock(ee, cls),
		     which, (CVMArrayOfRefICell*) obj);
    if (CVMexceptionOccurred(ee)) {
	CVMjniDeleteLocalRef(env, obj);
	return NULL;
    }
    return obj;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

JNIEXPORT jobjectArray JNICALL
JVM_GetClassMethods(JNIEnv *env, jclass cls, jint which)
{
#ifdef CVM_REFLECT
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    jobject obj = CVMjniCreateLocalRef(ee);
    if (obj == NULL) /* exception thrown */
	return NULL;
    CVMreflectMethods(ee, CVMgcSafeClassRef2ClassBlock(ee, cls),
		      which, (CVMArrayOfRefICell*) obj);
    if (CVMexceptionOccurred(ee)) {
	CVMjniDeleteLocalRef(env, obj);
	return NULL;
    }
    return obj;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

JNIEXPORT jobjectArray JNICALL
JVM_GetClassConstructors(JNIEnv *env, jclass cls, jint which)
{
#ifdef CVM_REFLECT
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    jobject obj = CVMjniCreateLocalRef(ee);
    if (obj == NULL) /* exception thrown */
	return NULL;
    CVMreflectConstructors(ee, CVMgcSafeClassRef2ClassBlock(ee, cls),
			   which, (CVMArrayOfRefICell*) obj);
    if (CVMexceptionOccurred(ee)) {
	CVMjniDeleteLocalRef(env, obj);
	return NULL;
    }
    return obj;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

JNIEXPORT jobject JNICALL
JVM_GetClassField(JNIEnv *env, jclass cls, jstring name, jint which)
{
#ifdef CVM_REFLECT
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    jobject obj = CVMjniCreateLocalRef(ee);
    const char *nameChars;
    if (obj == NULL) /* exception thrown */
	return NULL;

    nameChars = (*env)->GetStringUTFChars(env, name, NULL);
    if (nameChars == NULL) {
	(*env)->DeleteLocalRef(env, obj);
	return NULL;
    }

    CVMreflectField(ee, CVMgcSafeClassRef2ClassBlock(ee, cls),
		    nameChars, which, (CVMObjectICell*) obj);
    (*env)->ReleaseStringUTFChars(env, name, nameChars);
    if (CVMexceptionOccurred(ee)) {
	CVMjniDeleteLocalRef(env, obj);
	return NULL;
    }
    return obj;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

JNIEXPORT jobject JNICALL
JVM_GetClassMethod(JNIEnv *env, jclass cls, jstring name, jobjectArray types,
		   jint which)
{
#ifdef CVM_REFLECT
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    jobject obj = CVMjniCreateLocalRef(ee);
    const char *nameChars;
    if (obj == NULL) /* exception thrown */
	return NULL;

    nameChars = (*env)->GetStringUTFChars(env, name, NULL);
    if (nameChars == NULL) {
	(*env)->DeleteLocalRef(env, obj);
	return NULL;
    }

    CVMreflectMethod(ee, CVMgcSafeClassRef2ClassBlock(ee, cls),
		     nameChars, (CVMArrayOfRefICell*) types,
		     which, (CVMObjectICell*) obj);
    (*env)->ReleaseStringUTFChars(env, name, nameChars);
    if (CVMexceptionOccurred(ee)) {
	CVMjniDeleteLocalRef(env, obj);
	return NULL;
    }
    return obj;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

JNIEXPORT jobject JNICALL
JVM_GetClassConstructor(JNIEnv *env, jclass cls, jobjectArray types,
			jint which)
{
#ifdef CVM_REFLECT
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    jobject obj = CVMjniCreateLocalRef(ee);
    if (obj == NULL) /* exception thrown */
	return NULL;
    CVMreflectConstructor(ee, CVMgcSafeClassRef2ClassBlock(ee, cls),
			  (CVMArrayOfRefICell*) types,
			  which, (CVMObjectICell*) obj);
    if (CVMexceptionOccurred(ee)) {
	CVMjniDeleteLocalRef(env, obj);
	return NULL;
    }
    return obj;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

JNIEXPORT jobjectArray JNICALL
JVM_GetDeclaredClasses(JNIEnv *env, jclass ofClass)
{
#ifdef CVM_REFLECT
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(ee, ofClass);
    jobject obj = CVMjniCreateLocalRef(ee);
    if (obj == NULL)
	return NULL;
    CVMreflectInnerClasses(ee, cb, (CVMArrayOfRefICell*) obj);
    if (CVMexceptionOccurred(ee)) {
	CVMjniDeleteLocalRef(env, obj);
	return NULL;
    }
    return obj;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

JNIEXPORT jclass JNICALL
JVM_GetDeclaringClass(JNIEnv *env, jclass ofClass)
{
#ifdef CVM_REFLECT
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(ee, ofClass);
    CVMClassBlock* parentCb;

    parentCb = CVMreflectGetDeclaringClass(ee, cb);
    if (parentCb == NULL) {
	return NULL;
    }

    return (*env)->NewLocalRef(env, CVMcbJavaInstance(parentCb));
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

JNIEXPORT jclass JNICALL
JVM_GetCallerClass(JNIEnv *env, int n)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = CVMgetCallerClass(ee, n);
    if (cb == NULL) {
	return NULL;
    } else {
	return (*env)->NewLocalRef(env, CVMcbJavaInstance(cb));
    }
}

/* Note: JVM_NewInstance is not needed in CVM because Class.newInstance
   is implemented in Java. */

/*
 * Called from System.c
 */

JNIEXPORT jlong JNICALL
JVM_CurrentTimeMillis(JNIEnv *env, jclass ignored)
{
    return CVMtimeMillis();
}

static void
CVMcopyRefArrays(CVMExecEnv* ee,
		 CVMArrayOfRef* srcArr, jint src_pos,
		 CVMArrayOfRef* dstArr, jint dst_pos,
		 CVMClassBlock* dstElemCb,
		 jint length)
{
    CVMJavaInt i, j;
    if ((dstArr != srcArr) || (dst_pos < src_pos)) {
	for (i = src_pos, j = dst_pos; i < src_pos + length;
	     i++, j++) {
	    CVMObject* elem;
	    CVMD_arrayReadRef(srcArr, i, elem);
	    if (elem != NULL) {
		CVMClassBlock* rhsType = CVMobjectGetClass(elem);
		if ((rhsType != dstElemCb) &&
		    !CVMisAssignable(ee, rhsType, dstElemCb)) {
			if (!CVMlocalExceptionOccurred(ee)) {
			    CVMthrowArrayStoreException(ee, NULL);
			}
			return;
		    }
	    }
	    CVMD_arrayWriteRef(dstArr, j, elem);
	}
    } else {
	for (i = src_pos + length - 1, j = dst_pos + length - 1; 
	     i >= src_pos;
	     i--, j--) {
	    CVMObject* elem;
	    CVMD_arrayReadRef(srcArr, i, elem);
	    if (elem != NULL) {
		CVMClassBlock* rhsType = CVMobjectGetClass(elem);
		if ((rhsType != dstElemCb) &&
		    !CVMisAssignable(ee, rhsType, dstElemCb)) {
			if (!CVMlocalExceptionOccurred(ee)) {
			   CVMthrowArrayStoreException(ee, NULL);
			}
			return;
		    }
	    }
	    CVMD_arrayWriteRef(dstArr, j, elem);
	}
    }
}

JNIEXPORT void JNICALL
JVM_ArrayCopy(JNIEnv *env, jclass ignored,
	      jobject src, jint src_pos, 
	      jobject dst, jint dst_pos, jint length)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock *srcCb, *dstCb;
    CVMClassBlock *srcElemCb, *dstElemCb;
    CVMArrayOfAnyType *srcArr;
    CVMArrayOfAnyType *dstArr;
    size_t srclen;
    size_t dstlen;

    if ((src == 0) || (dst == 0)) {
	CVMthrowNullPointerException(ee, NULL);
	return;
    }
    CVMD_gcUnsafeExec(ee, {

	srcArr = (CVMArrayOfAnyType *)CVMID_icellDirect(ee, src);
	dstArr = (CVMArrayOfAnyType *)CVMID_icellDirect(ee, dst);

	srcCb = CVMobjectGetClass(srcArr);
	dstCb = CVMobjectGetClass(dstArr);

	/*
	 * First check whether we are dealing with source and destination
	 * arrays here. If not, bail out quickly.
	 */
	if (!CVMisArrayClass(srcCb) || !CVMisArrayClass(dstCb)) {
	    CVMthrowArrayStoreException(ee, NULL);
	    goto bail;
	}

	srcElemCb = CVMarrayElementCb(srcCb);
	dstElemCb = CVMarrayElementCb(dstCb);

	srclen = CVMD_arrayGetLength(srcArr);
	dstlen = CVMD_arrayGetLength(dstArr);

	/*
	 * If we are trying to copy between
	 * arrays that don't have the same type of primitive elements,
	 * bail out quickly.
	 *
	 * The more refined assignability checks for reference arrays
	 * will be performed in the CVM_T_CLASS case below.
	 */
	/*
	 * We could make this check slightly faster by comparing
	 * type-codes of the elements of the array, if we had a
	 * quick way of getting to that.
	 */
	if ((srcElemCb != dstElemCb) && (CVMcbIs(srcElemCb, PRIMITIVE) ||
					 CVMcbIs(dstElemCb, PRIMITIVE))) {
	    CVMthrowArrayStoreException(ee, NULL);
	    goto bail;
	} else if ((length < 0) || (src_pos < 0) || (dst_pos < 0) ||
		   (length + src_pos > srclen) ||
		   (length + dst_pos > dstlen)) {
	    CVMthrowArrayIndexOutOfBoundsException(ee, NULL);
	    goto bail;
	}
	
	switch(CVMarrayBaseType(srcCb)) {
	case CVM_T_BOOLEAN: 
	    CVMD_arrayCopyBoolean((CVMArrayOfBoolean*)srcArr, src_pos,
				  (CVMArrayOfBoolean*)dstArr, dst_pos,
				  length);
	    break;
	case CVM_T_BYTE: 
	    CVMD_arrayCopyByte((CVMArrayOfByte*)srcArr, src_pos,
			       (CVMArrayOfByte*)dstArr, dst_pos,
			       length);
	    break;
	    
	case CVM_T_SHORT:
	    CVMD_arrayCopyShort((CVMArrayOfShort*)srcArr, src_pos,
				(CVMArrayOfShort*)dstArr, dst_pos,
				length);
	    break;
	case CVM_T_CHAR:
	    CVMD_arrayCopyChar((CVMArrayOfChar*)srcArr, src_pos,
			       (CVMArrayOfChar*)dstArr, dst_pos,
			       length);
	    break;
	case CVM_T_INT:
	    CVMD_arrayCopyInt((CVMArrayOfInt*)srcArr, src_pos,
			      (CVMArrayOfInt*)dstArr, dst_pos,
			      length);
	    break;
	case CVM_T_LONG:
	    CVMD_arrayCopyLong((CVMArrayOfLong*)srcArr, src_pos,
			       (CVMArrayOfLong*)dstArr, dst_pos,
			       length);
	    break;
	case CVM_T_FLOAT: 
	    CVMD_arrayCopyFloat((CVMArrayOfFloat*)srcArr, src_pos,
				(CVMArrayOfFloat*)dstArr, dst_pos,
				length);
	    break;
	case CVM_T_DOUBLE: 
	    CVMD_arrayCopyDouble((CVMArrayOfDouble*)srcArr, src_pos,
				 (CVMArrayOfDouble*)dstArr, dst_pos,
				 length);
	    break;
	case CVM_T_CLASS: {
	    /*
	     * Do a quick check here for the easy case of the same
	     * array type, or copying into an array of objects. No
	     * need for element-wise checks in that case.  
	     */
	    if ((srcElemCb == dstElemCb) ||
		(dstElemCb == CVMsystemClass(java_lang_Object))) {
		CVMD_arrayCopyRef((CVMArrayOfRef*)srcArr, src_pos,
				  (CVMArrayOfRef*)dstArr, dst_pos,
				  length);
	    } else {
		/* Do it the hard way. */
		CVMcopyRefArrays(ee, 
				 (CVMArrayOfRef*)srcArr, src_pos, 
				 (CVMArrayOfRef*)dstArr, dst_pos, 
				 dstElemCb, length);
	    }
	    break;
	case CVM_T_VOID: 
	    CVMassert(CVM_FALSE);
	}
	}
    bail:;
    });
}

/* For compatibility only; no longer called by java.lang.Runtime */

#ifdef JDK12
JNIEXPORT void JNICALL
JVM_Exit(jint code)
{
    if (code == 0) {
        RunFinalizersOnExit();
    }
    Exit((int)code);
}
#endif

JNIEXPORT void JNICALL
JVM_Halt(jint code)		/* Called from java.lang.Shutdown */
{
    CVMexit(code);
}

JNIEXPORT void JNICALL
JVM_OnExit(void (*func)(void))
{
    CVMatExit(func);
}

JNIEXPORT void JNICALL
JVM_GC(void)
{
    CVMgcRunGC(CVMgetEE());
}

JNIEXPORT jlong JNICALL
JVM_MaxObjectInspectionAge()
{
    CVMInt64 diff = CVMlongSub(CVMtimeMillis(), CVMgcimplTimeOfLastMajorGC());
    return diff;
}

JNIEXPORT void JNICALL
JVM_TraceInstructions(jboolean on)
{
#ifdef CVM_DEBUG
    if (on) {
	CVMsetDebugFlags(CVM_DEBUGFLAG(TRACE_OPCODE));
    } else {
	CVMclearDebugFlags(CVM_DEBUGFLAG(TRACE_OPCODE));
    }
#endif
}

JNIEXPORT void JNICALL
JVM_TraceMethodCalls(jboolean on)
{
#ifdef CVM_DEBUG
    if (on) {
	CVMsetDebugFlags(CVM_DEBUGFLAG(TRACE_METHOD));
    } else {
	CVMclearDebugFlags(CVM_DEBUGFLAG(TRACE_METHOD));
    }
#endif
}

JNIEXPORT jlong JNICALL
JVM_TotalMemory(void)
{
    return CVMgcTotalMemory(CVMgetEE());
}

/* Called from Runtime.c */

JNIEXPORT jlong JNICALL
JVM_FreeMemory(void)
{ 
    return CVMgcFreeMemory(CVMgetEE());
}

/* Called from Throwable.c */

#undef min
#define min(x, y) (((int)(x) < (int)(y)) ? (x) : (y))

/* NOTE: The second argument to printStackTrace() is expected to be an object
 * with a void println(char[]) method, such as a PrintStream or a PrintWriter.
 */

JNIEXPORT void JNICALL
JVM_PrintStackTrace(JNIEnv *env, jobject throwable, jobject printable)
{
    CVMprintStackTrace(CVMjniEnv2ExecEnv(env), throwable, printable);
}

void 
CVMprintStackTrace(CVMExecEnv *ee, CVMThrowableICell* throwableICell,
		   CVMObjectICell* printableICell)

{
#ifndef CVM_DEBUG_STACKTRACES
    CVMconsolePrintf("\t<no backtrace available in non-debug build>\n");
#else
    CVMID_localrootBegin(ee) {
	CVMID_localrootDeclare(CVMArrayOfRefICell, backtraceICell);
	CVMID_localrootDeclare(CVMArrayOfIntICell, pcArrayICell);
	CVMID_localrootDeclare(CVMStringICell, stringICell);
	CVMID_localrootDeclare(CVMObjectICell, objICell);
	CVMID_localrootDeclare(CVMObjectICell, tempICell);
	CVMUint32 count;
	CVMUint32 pcArrayLen;

	/*
	 * Read in the backtrace array field of the throwable.
	 */
	CVMID_fieldReadRef(ee, throwableICell,
			   CVMoffsetOfjava_lang_Throwable_backtrace,
			   tempICell);
	CVMID_icellAssign(ee, backtraceICell, (CVMArrayOfRefICell*)tempICell);
	if (CVMID_icellIsNull(backtraceICell)) {
	    return;
	}
	
	/*
	 * The first element of the backtrace array is an array of pc values.
	 */
	CVMID_arrayReadRef(ee, backtraceICell, 0, tempICell);
	CVMID_icellAssign(ee, pcArrayICell, (CVMArrayOfIntICell*)tempICell);
	if (CVMID_icellIsNull(pcArrayICell)) {
	    return;
	}
	
	/*
	 * Print each frame in the backtrace.
	 */
	CVMID_arrayGetLength(ee, pcArrayICell, pcArrayLen);
	for (count = 0; count < pcArrayLen; count++) {
	    CVMBool         isString;
	    
	    /*
	     * Element 0 of the backtrace array was the array of pc values.
	     * The elements that follow are normally java.lang.Class objects
	     * that correpond to the class owning the pc. However, in some
	     * cases it is actually an already computed java.lang.String.
	     *
	     * If the object is NULL then that means the array was bigger
	     * than we needed and we are done now.
	     */
	    CVMID_arrayReadRef(ee, backtraceICell, count+1, objICell);
	    if (CVMID_icellIsNull(objICell)) {
		goto localRootEnd;
	    }
	    CVMD_gcUnsafeExec(ee, {
		isString = CVMgcUnsafeIsInstanceOf(
                    ee, CVMID_icellDirect(ee, objICell),
		    CVMsystemClass(java_lang_String));
	    });
	    if (isString) {
		CVMID_icellAssign(ee, stringICell, objICell);
	    } else {
		/* obj is a java.lang.Class. Compute the backtrace string. */
		CVMJavaInt     pc;
		char           buf[256];
		int            i = 0;
		CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(ee, objICell);
		
		/* get pc value */
		CVMID_arrayReadInt(ee, pcArrayICell, count, pc);
		
		strncpy(buf + i, "\tat ", 4);
		i += 4;
		
		/* make a string out of the pc value. */
		CVMpc2string(NULL, CVMpc2method(cb, (CVMUint8*)pc), 
			     (CVMUint8*)pc, buf + i, buf + sizeof(buf));
		CVMnewStringUTF(ee, stringICell, buf);
	    }

	    /* call the println method of the printable object */
	    if (!CVMID_icellIsNull(stringICell)) {
		CVMMethodBlock* mb; 
		CVMClassBlock*  cb;
		JNIEnv*         env = CVMexecEnv2JniEnv(ee);
#ifdef CVM_DEBUG
		/*
		 * The only way to get a NULL printableICell is if you are
		 * debugging and call this function manually. In that case
		 * we use System.out.
		 */
		if (printableICell == NULL) {
		    printableICell = (*env)->GetStaticObjectField(
                        env, 
			CVMcbJavaInstance(CVMsystemClass(java_lang_System)),
			CVMglobals.java_lang_System_out);
		}
#endif
		CVMD_gcUnsafeExec(ee, {
		    cb = CVMobjectGetClass(
                            CVMID_icellDirect(ee, printableICell));
		});
		mb = CVMclassGetNonstaticMethodBlock(
                        cb, CVMglobals.printlnTid);
		(*env)->CallVoidMethod(env, printableICell, mb, stringICell);
		/* maybe we should only check local exceptions,
		   or disable remote exceptions? */
		if ((*env)->ExceptionCheck(env)) {
		    break;
		}
	    }
	}

    localRootEnd:;
    } CVMID_localrootEnd();
#endif
}

/*
 * Fill in the specified Throwable object with the current backtrace,
 */

JNIEXPORT void JNICALL
JVM_FillInStackTrace(JNIEnv *env, jobject throwable)
{
    CVMfillInStackTrace(CVMjniEnv2ExecEnv(env), throwable);    
}

static void 
CVMgcUnsafeFillInStackTrace(CVMExecEnv *ee, CVMThrowableICell* throwableICell);

void 
CVMfillInStackTrace(CVMExecEnv *ee, CVMThrowableICell* throwableICell) {
    if (CVMD_isgcSafe(ee)) {
	CVMD_gcUnsafeExec(ee, {
	    CVMgcUnsafeFillInStackTrace(ee, throwableICell);
	});
    } else {
	CVMgcUnsafeFillInStackTrace(ee, throwableICell);
    }
}

static void 
CVMgcUnsafeFillInStackTrace(CVMExecEnv *ee, CVMThrowableICell* throwableICell)
{
#ifdef CVM_DEBUG_STACKTRACES
    /*
     * The pc's are saved as an int[] (to make it easy for the GC).  The
     * following constant indicates how may array elements are needed per pc
     * It will be 1 on 32-bit machines and 2 on 64-bit machines.
     */
    #define MULTIPLIER (sizeof(char*)/sizeof(CVMUint32))

    CVMFrame*        startFrame;
    CVMFrame*        frame;
    CVMMethodBlock*  mb;
    CVMArrayOfRef*   backtrace;
    CVMArrayOfInt*   pcArray;
    CVMObject*       tempObj;
    CVMInt32         backtraceSize;
    CVMInt32         count;
 
    startFrame = CVMeeGetCurrentFrame(ee);

    /* Go backwards, removing all frames that are just part of the <init> of
     * the throwable object.
     */
    mb = startFrame->mb;
    if (mb != NULL) { /* NULL indicates the initial dummy frame */
	if (mb == CVMglobals.java_lang_Throwable_fillInStackTrace) {
	    startFrame = CVMframePrev(startFrame);
	}
	mb = startFrame->mb;
		
	while ( CVMtypeidIsConstructor(CVMmbNameAndTypeID(mb)) &&
		CVMisSubclassOf(ee, CVMmbClassBlock(mb), 
				CVMsystemClass(java_lang_Throwable))) {
	    startFrame = CVMframePrev(startFrame);
	    mb = startFrame->mb;
	}
	if (CVMlocalExceptionOccurred(ee)) {
		return;
	}
    }

    /* how many useful frames are there on the stack? */
    CVMD_gcSafeExec(ee, {
	backtraceSize = 0;
	frame = startFrame;
	while (frame != NULL) {
	    if (frame->mb != NULL && 
		(CVMframeIsJNI(frame) || CVMframeIsJava(frame))) {
		backtraceSize++;
	    }
	    frame = CVMframePrev(frame);
	}
    });


    /*
     * The backtrace contains a pointer to an int[] and a list of
     * java.lang.Class objects.  The first item is the int[].  The size of
     * the array must be the length of the stack trace + 1.
     */
    CVMD_fieldReadRef(CVMID_icellDirect(ee, throwableICell),
		      CVMoffsetOfjava_lang_Throwable_backtrace,
		      tempObj);
    backtrace = (CVMArrayOfRef*)tempObj;
    if ((backtrace == NULL) || 
	(CVMD_arrayGetLength(backtrace) < backtraceSize + 1))
    {
	backtrace = (CVMArrayOfRef*)CVMgcAllocNewArray(
            ee, CVM_T_CLASS, 
	    (CVMClassBlock*)CVMbasicTypeArrayClassblocks[CVM_T_CLASS],
	    backtraceSize + 1);

	tempObj = (CVMObject*)backtrace;
	CVMD_fieldWriteRef(CVMID_icellDirect(ee, throwableICell), 
			   CVMoffsetOfjava_lang_Throwable_backtrace,
			   tempObj);
	if (backtrace == NULL) {
	    return;
	}
    }

    /*
     * The pcArray is a java array of ints and is stored as the first
     * element of the backtrace array.
     */
    CVMD_arrayReadRef(backtrace, 0, tempObj);
    pcArray = (CVMArrayOfInt*)tempObj;
    
    /* if pcArray is null or not long enough, increase it */
    if ((pcArray == NULL) || 
	(CVMD_arrayGetLength(pcArray) < backtraceSize * MULTIPLIER)) 
    {
	pcArray = (CVMArrayOfInt*)CVMgcAllocNewArray(
            ee, CVM_T_INT, 
	    (CVMClassBlock*)CVMbasicTypeArrayClassblocks[CVM_T_INT],
	    backtraceSize * MULTIPLIER);
	/* recache backtrace incase we become gcsafe during allocation. */
	CVMD_fieldReadRef(CVMID_icellDirect(ee, throwableICell),
			  CVMoffsetOfjava_lang_Throwable_backtrace,
			  tempObj);
	backtrace = (CVMArrayOfRef*)tempObj;

	tempObj = (CVMObject*)pcArray;
	CVMD_arrayWriteRef(backtrace, 0, tempObj);
	if (pcArray == NULL) {
	    return;
	}
    }


    /* Fill in the pcArray with the pc's */
    frame = startFrame;
    count = 0;
    while (frame != NULL) {
	/*
	 * We don't want to stay gcunsafe for too long, so we periodically
	 * offer a checkpoint. We need to restore backtrace and pcArray
	 * if a gc happened.
	 */
	CVMD_gcSafeCheckPoint(ee, {}, {
	    CVMD_fieldReadRef(CVMID_icellDirect(ee, throwableICell),
			      CVMoffsetOfjava_lang_Throwable_backtrace,
			      tempObj);	    
	    backtrace = (CVMArrayOfRef*)tempObj;
	    CVMD_arrayReadRef(backtrace, 0, tempObj);
	    pcArray = (CVMArrayOfInt*)tempObj;
	});

	if (frame->mb != NULL && 
	    (CVMframeIsJNI(frame) || CVMframeIsJava(frame))) {
	    CVMMethodBlock* mb = frame->mb;
	    /* Write the pc value into the pcArray. */
	    if (CVMmbIs(mb, NATIVE)) {
		CVMD_arrayWriteInt(pcArray, count, 
				   (CVMJavaInt)CVMmbNativeCode(mb));
	    } else {
		CVMD_arrayWriteInt(pcArray, count,
				   (CVMJavaInt)CVMframePc(frame));
            }

	    if (CVMtypeidIsStaticInitializer(CVMmbNameAndTypeID(mb))) {
		/* We must eagerly compute the line number information for
		 * <clinit> because it may be freed before the class is GC'ed.
		 */
#undef CVM_PC2STRING_BUF_SIZE
#define CVM_PC2STRING_BUF_SIZE 256
	        char  *buf = malloc(CVM_PC2STRING_BUF_SIZE); 
		int   i = 0;
		char *tmp = NULL; 

    		if (buf == NULL) {
		    buf = tmp = "<PC information unavailable due to low memory>"; 
    		}
		else {
		    strncpy(buf + i, "\tat ", 4);
		    i += 4;
		    /* make a string out of the pc value. */
		    CVMpc2string(frame, NULL, NULL, buf + i, 
		        buf + CVM_PC2STRING_BUF_SIZE);
		}

		CVMD_gcSafeExec(ee, {
		    CVMID_localrootBegin(ee) {
			CVMID_localrootDeclare(CVMStringICell, stringICell);
			CVMID_localrootDeclare(CVMArrayOfRefICell,
					       backtraceICell);
			CVMID_localrootDeclare(CVMObjectICell, tempICell);
			CVMnewStringUTF(ee, stringICell, buf);
			CVMID_fieldReadRef(
				  ee, throwableICell,
				  CVMoffsetOfjava_lang_Throwable_backtrace,
				  tempICell);
			CVMID_icellAssign(ee, backtraceICell,
					  (CVMArrayOfRefICell*)tempICell);
			CVMID_arrayWriteRef(ee, backtraceICell, count + 1, 
					   stringICell);
		    } CVMID_localrootEnd();
		});
		/* recache since we became gcsafe for a while. */
		CVMD_fieldReadRef(CVMID_icellDirect(ee, throwableICell),
				  CVMoffsetOfjava_lang_Throwable_backtrace,
				  tempObj);	    
		backtrace = (CVMArrayOfRef*)tempObj;
		CVMD_arrayReadRef(backtrace, 0, tempObj);
		pcArray = (CVMArrayOfInt*)tempObj;
		if (buf != tmp)
		    free(buf);
	    } else {
		/* Write the class object into the backtrace array. */
		CVMClassICell* classICell = 
		    CVMcbJavaInstance(CVMmbClassBlock(mb));
		CVMD_arrayWriteRef(backtrace, count + 1, 
				   CVMID_icellDirect(ee, classICell));
	    }
	    count++;
        }
	frame = CVMframePrev(frame);
    } 
    /* Zero out last entry if there is room. */
    if (count + 1 < CVMD_arrayGetLength(backtrace)) {
	CVMD_arrayWriteRef(backtrace, count + 1, 0);
    }
#endif
}


/* Called from object.c */
JNIEXPORT jint JNICALL
JVM_IHashCode(JNIEnv *env, jobject obj)
{
    if (obj == NULL) {
	return 0;
    }
    return CVMobjectGetHashSafe(CVMjniEnv2ExecEnv(env), obj);
}

JNIEXPORT void JNICALL
JVM_MonitorWait(JNIEnv *env, jobject obj, jlong millis)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    if (CVMlongLtz(millis)) {
	CVMthrowIllegalArgumentException(ee, "timeout value is negative");
        return;
    }

#ifdef HPROF
    jvmpi_monitor_wait(DeRef(env, obj), millis);
#endif

    /* NOTE: CVMgcSafeObjectWait() may throw certain exceptions: */

    /* Dio: enable changing the scheduler */
    ee->canChangeScheduler = 1;

    CVMgcSafeObjectWait(ee, obj, millis);
    /* Dio: disable changing the scheduler*/ 
    ee->canChangeScheduler = 0;

#ifdef HPROF
    jvmpi_monitor_waited(DeRef(env, obj));
#endif
}

JNIEXPORT void JNICALL
JVM_MonitorNotify(JNIEnv *env, jobject obj)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    /* NOTE: CVMgcSafeObjectNotify() may throw certain exceptions: */
    CVMgcSafeObjectNotify(ee, obj);
}

JNIEXPORT void JNICALL
JVM_MonitorNotifyAll(JNIEnv *env, jobject obj)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    /* NOTE: CVMgcSafeObjectNotifyAll() may throw certain exceptions: */
    CVMgcSafeObjectNotifyAll(ee, obj);
}

JNIEXPORT jobject JNICALL
JVM_Clone(JNIEnv *env, jobject obj)
{
    CVMClassBlock*  cb;
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    jobject result;

    CVMID_localrootBegin(ee) {
	CVMID_localrootDeclare(CVMObjectICell, cloneObj);
	CVMID_objectGetClass(ee, obj, cb);
	/*
	 * Now let's see whether we are cloning an array or an object
	 */
	if (CVMisArrayClass(cb)) {
	    CVMArrayOfAnyTypeICell* arrObj = (CVMArrayOfAnyTypeICell*)obj;
	    CVMBasicType typeCode = CVMarrayElemTypeCode(cb);
	    CVMJavaInt   arrLen;

	    CVMID_arrayGetLength(ee, arrObj, arrLen);
	    /* Make the copy first */
	    CVMID_allocNewArray(ee, typeCode, cb, arrLen, cloneObj);
	    if (CVMID_icellIsNull(cloneObj)) {
		/* Allocation failed */
		result = NULL;
		CVMthrowOutOfMemoryError(ee, "cloning array object");
		goto bailOut;
	    }
	    JVM_ArrayCopy(env, NULL, obj, 0, cloneObj, 0, arrLen);
	} else {
	    if (!CVMimplementsInterface(ee, cb,
				    CVMsystemClass(java_lang_Cloneable))) {
		result = NULL;
		if (!CVMlocalExceptionOccurred(ee)) {
		   CVMthrowCloneNotSupportedException(ee,
						   "cloning regular object");
		}
		goto bailOut;
	    }
	    CVMID_allocNewInstance(ee, cb, cloneObj);
	    if (CVMID_icellIsNull(cloneObj)) {
		/* Allocation failed */
		result = NULL;
		CVMthrowOutOfMemoryError(ee, "cloning regular object");
		goto bailOut;
	    }
	    CVMD_gcUnsafeExec(ee, {
		CVMObject* srcObj = CVMID_icellDirect(ee, obj);
		CVMObject* dstObj = CVMID_icellDirect(ee, cloneObj);
		CVMClassBlock* cbp;
		CVMwithAssertsOnly(
		    CVMUint32 coveredSize = sizeof(CVMObjectHeader);
		);
		
		/*
		 * Walk all instance fields in superclass order,
		 * and copy each with the right CVMD_ call.
		 */
		for (cbp = cb; cbp != NULL; cbp = CVMcbSuperclass(cbp)) {
		    CVMUint32 i;
		    /*
		     * Go through the declared fields
		     */
		    for (i = 0; i < CVMcbFieldCount(cbp); i++) {
			CVMFieldBlock* fb = CVMcbFieldSlot(cbp, i);
			if (!CVMfbIs(fb, STATIC)) {
			    CVMUint32 offset = CVMfbOffset(fb);
			    CVMClassTypeID type =
				CVMtypeidGetType(CVMfbNameAndTypeID(fb));
			    switch (type) {
				case CVM_TYPEID_INT:
				case CVM_TYPEID_SHORT:
				case CVM_TYPEID_CHAR:
				case CVM_TYPEID_BYTE:
			        case CVM_TYPEID_BOOLEAN: {
				    CVMJavaInt elem;
				    CVMD_fieldReadInt(srcObj, offset, elem);
				    CVMD_fieldWriteInt(dstObj, offset, elem);
				    CVMwithAssertsOnly(
				        coveredSize += sizeof(CVMJavaInt);
				    );
				    break;
				}
			        case CVM_TYPEID_LONG: {
				    CVMJavaLong l;
				    CVMD_fieldReadLong(srcObj, offset, l);
				    CVMD_fieldWriteLong(dstObj, offset, l);
				    CVMwithAssertsOnly(
				        coveredSize += sizeof(CVMJavaLong);
				    );
				    break;
				}
			        case CVM_TYPEID_DOUBLE: {
				    CVMJavaDouble d;
				    CVMD_fieldReadDouble(srcObj, offset, d);
				    CVMD_fieldWriteDouble(dstObj, offset, d);
				    CVMwithAssertsOnly(
				        coveredSize += sizeof(CVMJavaDouble);
				    );
				    break;
				}
			        case CVM_TYPEID_FLOAT: {
				    CVMJavaFloat f;
				    CVMD_fieldReadFloat(srcObj, offset, f);
				    CVMD_fieldWriteFloat(dstObj, offset, f);
				    CVMwithAssertsOnly(
				        coveredSize += sizeof(CVMJavaFloat);
				    );
				    break;
				}
			        default: /* Must be a ref */ {
				    CVMObject* o;
				    CVMD_fieldReadRef(srcObj, offset, o);
				    CVMD_fieldWriteRef(dstObj, offset, o);
				    CVMwithAssertsOnly(
				        coveredSize += sizeof(CVMObject*);
				    );
				    break;
				}
			    }
			}
		    }
		}
		/*
		 * Now make sure that we have copied just the right
		 * number of bytes from one object to the other
		 */
		CVMassert(coveredSize == CVMcbInstanceSize(cb));
	    });
	}
	result = (*env)->NewLocalRef(env, cloneObj);
    bailOut:;
    } CVMID_localrootEnd();
    return result;
}

/*
 * Thread class calls
 */

/* This is the (void *) argument passed to start_func, below, through
   CVMthreadCreate. This data structure is required to be allocated
   using malloc() by the caller; start_func is responsible for
   freeing it. */
typedef struct {
    /* This is required to be non-NULL and to point to a valid global
       root which contains a java.lang.Thread object. start_func is
       responsible for holding on to the Thread object (by storing it
       in the new thread's execution environment's currentThreadICell)
       and deallocating the global root. */
    CVMThreadICell* threadICell;
    /* This function pointer must be NULL for all Java threads (i.e.,
       those started by a call to JVM_StartThread). The JVMDI and
       JVMPI, however, need to be able to create "system threads",
       which are arbitrary function pointers which execute in the
       context of an CVMExecEnv, so they can make JNI calls. */
    void (*nativeFunc)(void *);
    /* This must be NULL for non-system threads. For system threads,
       if this argument is non-NULL, it must be allocated on the heap
       by the caller. start_func does not free this argument, only
       passes it down to the native function. */
    void* nativeFuncArg;
    /* RTSJ
     * This field is added to support memory area stack inheritance
     * for RealtimeThreads
     * - Pratik (13th July 2001)
     */
    PerThreadMemoryInfo *inheritedStack;
    CVMBool isDaemon;
} ThreadStartInfo;


/* ISHIN */


#define EDF_PRI  40
#define EDF_MIN  40
#define EDF_MAX 100

extern pthread_mutex_t threadQmtx;


//static int getEligibleRange (JNIEnv *env, jobject this)
static int getEligiblePriority0 (CVMExecEnv *ee)
{
   CVMExecEnv **firstPtr = &CVMglobals.threadList;
   CVMExecEnv *ptr = *firstPtr;
   int P_Array[256], i;


   /* struct sched_param p; */
   CVMInt32 priority;
   jlong deadline, td;
   int max=EDF_MAX, min=EDF_MIN;

   for (i=0; i<256; i++) P_Array[i] = 0;

   deadline = ee->startNextPeriodMicros + ee->deadline;

   printf("     -----------------------------------------------------------------\n");
   printf("      ThreadID  pthreadID  priority   nextPeriod     deadline\n");

   ptr = *firstPtr;
   while (ptr != 0) {
      CVMthreadGetPriority(&(ptr->threadInfo), &priority);
      if (ptr->deadline > 0 && ee->threadID != ptr->threadID ) {
	  printf("*          %d        %d        %3d     %9Ld   %9Ld(%Ld)\n",
		ptr->threadID, (int)ptr->threadInfo.pthreadCookie,
		priority,
		ptr->startNextPeriodMicros % 1000000000,
		(ptr->startNextPeriodMicros + ptr->deadline) % 1000000000,
		ptr->deadline % 1000000000);
	  td = ptr->startNextPeriodMicros + ptr->deadline;
	  if (td == deadline && priority >= min) {max = min = priority; break;}
	  else if (deadline < td && priority >= min) min = priority+1;
	  else if (deadline > td && priority <= max) max = priority-1;
	  P_Array[priority]++;
      }
      ptr = ptr->nextEE;
   }
   printf("     ---- max : %d, min : %d -----------------------------------------\n",
   	max, min);

   if (max > min) priority = min + 1;
   else if (min > EDF_MIN && max < EDF_MAX && max == min) priority = min;
   else { //priority = min + 1;
      int nEmpty=0, nMove=0, pEmpty, pMove, j;
      double ratio;

      i = min-1;
      while (i > EDF_MIN) {
         if (P_Array[i] == 0) break;
	 else {i--; nMove++;}
      }
      pMove = i+1;

      j=i;
      if (i>EDF_MIN) {
          while (i > EDF_MIN) {
             if (P_Array[i] == 1) break;
	     else {i--; nEmpty++;}
          }

          pEmpty = i+1;

	  nMove++;
          nEmpty = nEmpty + nMove;

          ratio = (double)nEmpty / (nMove+1);

	  /*
          printf (" nMove - %d, nEmpty - %d, ratio - %f\n", nMove, nEmpty, ratio);
          printf (" pMove - %d, pEmpty - %d, ratio - %f\n", pMove, pEmpty, ratio);
	  */

          for (i=0; i<nMove-1; i++) {
	      ptr = *firstPtr;
              while (ptr != 0) {
                  CVMthreadGetPriority(&(ptr->threadInfo), &priority);
                  if (priority == pMove) {
		      int inc;
		      struct sched_param p;
		      inc = ratio * (2*i+1) * 0.5;
//		      if (inc == 0) inc = 1;
                      p.sched_priority = pEmpty + inc;
		      printf (" * %d * Change priority from %d to %d\n",
	              (int)ptr->threadInfo.pthreadCookie, pMove,
		      p.sched_priority);
                      pthread_setschedparam(ptr->threadInfo.pthreadCookie,
		                            SCHED_FIFO, &p);
		      break;
		  }
                  ptr = ptr->nextEE;
              }
              pMove++;
          }
	  nMove--;
	  priority = pEmpty + ratio * (2*(nMove)+1) * 0.5;
      }
      else {
          int nEmpty=0, nMove=0, pEmpty, pMove, j;
          double ratio;
/*

          for (i=0; i<256; i++) P_Array[i] = 0;

          ptr = *firstPtr;
          while (ptr != 0) {
             CVMthreadGetPriority(&(ptr->threadInfo), &priority);
             if (ptr->deadline > 0 && ee->threadID != ptr->threadID ) {
	        printf("*          %d        %d        %3d     %9Ld   %9Ld(%Ld)\n",
		ptr->threadID, (int)ptr->threadInfo.pthreadCookie,
		priority,
		ptr->startNextPeriodMicros % 1000000000,
		(ptr->startNextPeriodMicros + ptr->deadline) % 1000000000,
		ptr->deadline % 1000000000);
	        P_Array[priority]++;
	     }
             ptr = ptr->nextEE;
          }

*/
  	  i = max+1;
          while (i < EDF_MAX) {
              if (P_Array[i] == 0) break;
	      else {i++; nMove++;}
          }
          pMove = i-1;

          j = i;
	  while (i < EDF_MAX) {
             if (P_Array[i] == 1) break;
	     else {i++; nEmpty++;}
          }
	  pEmpty = i-1;
          nEmpty = nEmpty + nMove;

          ratio = (double)nEmpty / (nMove+2);

          printf (" nMove - %d, nEmpty - %d, ratio - %f\n", nMove, nEmpty, ratio);
          printf (" pMove - %d, pEmpty - %d, ratio - %f\n", pMove, pEmpty, ratio);

          for (i=0; i<nMove; i++) {
	      ptr = *firstPtr;
              while (ptr != 0) {
                  CVMthreadGetPriority(&(ptr->threadInfo), &priority);
                  if (priority == pMove) {
		      int dec;
		      struct sched_param p;
		      dec = ratio * (2*i+1) * 0.5;
                      p.sched_priority = pEmpty - dec;
		      printf (" * %d * Change priority from %d to %d (dec:%d)\n",
	              (int)ptr->threadInfo.pthreadCookie, pMove,
		      p.sched_priority, dec);
                      pthread_setschedparam(ptr->threadInfo.pthreadCookie,
		                            SCHED_FIFO, &p);
		      break;
		  }
                  ptr = ptr->nextEE;
              }
              pMove--;
          }
	  priority = pEmpty - ratio * (2*nMove+1) * 0.5;
      }

      for (i=0; i<256; i++) P_Array[i] = 0;

      ptr = *firstPtr;
      while (ptr != 0) {
         int pri;
         CVMthreadGetPriority(&(ptr->threadInfo), &pri);
         if (ptr->deadline > 0 && ee->threadID != ptr->threadID ) {
	     printf("*          %d        %d        %3d     %9Ld   %9Ld(%Ld)\n",
		ptr->threadID, (int)ptr->threadInfo.pthreadCookie,
		pri,
		ptr->startNextPeriodMicros % 1000000000,
		(ptr->startNextPeriodMicros + ptr->deadline) % 1000000000,
		ptr->deadline % 1000000000);
	 }
	 ptr = ptr->nextEE;
      }
      printf("     -----------------------------------------------------------------\n");
   }

   return priority;
}

/* ISHIN */



static void start_func(void *arg)
{
    CVMExecEnv ee;
    JNIEnv *env;
    jclass clazz;
    jmethodID methodID;
    ThreadStartInfo *info = (ThreadStartInfo *) arg;
    CVMThreadICell *threadCell = info->threadICell;
    CVMThreadICell *eeThreadCell;
    void (*nativeFunc)(void *) = info->nativeFunc;
    CVMJavaLong eetopLongVal;

    eetopLongVal = CVMvoidPtr2Long(&ee);
#ifdef CVM_DEBUG
    /*printf("start_func started\n");*/
#endif
    printf("     [JVM--] A new thread is created. \n");

    if (!CVMinitExecEnv(&ee, !info->isDaemon)) {
	return; /* exception NOT thrown */
    }



    eeThreadCell = CVMcurrentThreadICell(&ee);

    ee.perThreadStack = info->inheritedStack;
    if(ee.perThreadStack == NULL)
    {
        MemoryAreaInfo *root;
        root = CVMJTregistryGetMemoryAreaInfoForObject(&ee, threadCell);
        CVMassert(root != NULL &&
                    (root->memType == MEMTYPE_IMMORTAL || root->memType == MEMTYPE_HEAP));

        ee.perThreadStack = CVMJTregistryCreatePTMI(&ee);
        ee.perThreadStack->info = root;
    }

    // Make top of perThreadStack as current memory area
    CVMID_icellAssign(&ee, CVMmemoryAreaICell(&ee), ee.perThreadStack->info->memoryArea);

//    CVMdebugPrintf(("start_func: ee thread icell is %I... passed thread icell is %I\n", eeThreadCell, threadCell));
    CVMID_icellAssign(&ee, eeThreadCell, threadCell);

    /* Store eetop */
    {
	CVMJavaLong eetop = CVMvoidPtr2Long(&ee);
	CVMID_fieldWriteLong(&ee, eeThreadCell,
			     CVMoffsetOfjava_lang_Thread_eetop,
			     eetop);
    }

    CVMID_freeGlobalRoot(&ee, threadCell);
    env = CVMexecEnv2JniEnv(&ee);

#ifdef CVM_JVMDI
    /* This will work for getting thread start/stop
       notification for Java threads, but when we support JVMDI's
       "RunDebugThread" or JVMPI's "CreateSystemThread" then it's
       likely this function won't be called from CVMthreadCreate. The
       specs aren't precise about whether these events are generated
       for "native" threads created using these routines. If they are
       expected to be (and the only way to find out is to look through
       the JDK source) then we'll have to be very careful when
       implementing RunDebugThread/CreateSystemThread to ensure the
       events get generated. */

    if (CVMglobals.jvmdiDebuggingEnabled) {
	CVMjvmdiNotifyDebuggerOfThreadStart(&ee, CVMcurrentThreadICell(&ee));
    }
#endif

    /* Core Extension Hookup */
//	ee.pid = getpid();

    {
      extern intFuncPtr JVMCeiFilters[] ;      
      extern int JVMCeiInitialized;
      jobject threadObject;
      threadObject = (*env)->NewLocalRef(env, CVMcurrentThreadICell(&ee));
#ifdef CVM_DEBUG
      /*printf("jvm: trying to call core extension hookups\n");*/
#endif
      if (JVMCeiInitialized)
        {
          if (JVMCeiFilters[JVMCEF_AFTER_CREATE_SYS_THREAD] != NULL)
            {
              struct arg_t{ JNIEnv *env; jobject tObject; PerThreadMemoryInfo *inheritedStack;} arg;
              arg.env = env; arg.tObject = threadObject; arg.inheritedStack = info->inheritedStack;

#ifdef CVM_DEBUG
              /*printf("calling rtthreadadapter\n");*/
#endif

              if(JVMCeiFilters[JVMCEF_AFTER_CREATE_SYS_THREAD]((void *) &arg) == -1)
              {
                  CVMconsolePrintf("Thread could not be started. Terminating thread.\n");
                  goto finish;
              }

#ifdef CVM_DEBUG
              /*printf("after calling rtthreadadapter\n");*/
#endif
            }
          /* Memory hookup - Pratik*/
          /*if(JVMCeiFilters[JVMCEF_INIT_HEAP_MEMORY] != NULL)
            {
              //struct arg_t{ JNIEnv *env; jobject tObject; } arg;
              //arg.env = env; arg.tObject = threadObject;
              printf("calling HeapMemoryAdapter\n");
              JVMCeiFilters[JVMCEF_INIT_HEAP_MEMORY]((void *) &env);
              printf("after calling HeapMemoryAdapter\n");

            }*/
        }
      else
        {
#ifdef CVM_DEBUG
          printf("It is not yet initialsed???\n");
#endif
        }
#ifdef CVM_DEBUG
      /*printf("jvm: core extension hookups over\n");*/
#endif
    }


#if 0
    {
      extern int JVMCeiInitialized;
      jfieldID fID;
      jclass threadClazz;
      int rs;
      jobject threadObject;

      if (JVMCeiInitialized)
	{
	  threadObject = (*env)->NewLocalRef(env, CVMcurrentThreadICell(&ee));

	/* verify if this thread is an RealtimeThread */

	  threadClazz = (*env)->FindClass(env, "javax/realtime/RealtimeThread");

	if (threadClazz == NULL)
	  {
	    printf("Thread Create Error getting class javax.realtime.RealtimeThread\n");
	    return ;
	  }
	
	
	if ( (*env)->IsInstanceOf(env, threadObject, threadClazz))
	  {    
	    extern intFuncPtr JVMCeiFilters[] ;      

	    fID = (*env)->GetFieldID(env, threadClazz, "resourceSetHandle","I");
	    rs = (*env)->GetIntField(env, threadObject, fID);

	    if (JVMCeiFilters[JVMCEF_AFTER_CREATE_SYS_THREAD] != NULL)
	      {
		struct attach_reserve_struct {
		  int pid; 
		  int rs;
		  timespec period;
		} attach_rs;

		long nsecs,msecs;
		attach_rs.pid = getpid();
		attach_rs.rs = rs;
		jmethodID mID = (*env)->GetMethodID(env, threadClazz, "getPeriodMillis","()L");
		msecs = (*env)->CallLongMethod(env, threadObject, mID);

		jmethodID mID = (*env)->GetMethodID(env, threadClazz, "getPeriodNanos","()L");
		nsecs = (*env)->CallLongMethod(env, threadObject, mID);

		attach_rs.period.tv_sec = msecs / 1000;
		attach_rs.period.tv_nsec = (msecs % 1000) + nsecs;

		if  (attach_rs.rs != 0)
		  JVMCeiFilters[JVMCEF_AFTER_CREATE_SYS_THREAD]((void*) &attach_rs);
	      }
	  }
      }
    }

#endif


    /************************ ISHIN **************************/
    {

    }
    {
	int prio;
	CVMthreadGetPriority(&(ee.threadInfo), &prio);
	ee.threadInfo.lastPriority = prio;
	//	printf ("ISHIN - pthread priority - %d\n", prio);
    }
    {

      struct sched_param p;
      CVMInt32 pri;
      jlong deadline;
      int min=40, max=100;// edf_min = 100 + 1; // EDF_MIN = 40, EDF_MAX = 50

      jclass clazz = CVMcbJavaInstance(CVMsystemClass(javax_realtime_RealtimeThread));
      jmethodID methodID = (*env)->GetStaticMethodID(env, clazz, "getIntScheduler", "()I");
      int scheduler = (*env)->CallStaticIntMethod (env, clazz, methodID);
      printf ("     ------------ Current Scheduler : ");
      switch (scheduler) {
	   case 1: printf (" Priority Scheduler ------------\n"); break;
	   case 2: printf (" EDF Scheduler ------------\n"); break;
      }

//      if (scheduler == 2 && ee.threadInfo.lastPriority == 40) { // EDF_PRI = 40
      if (scheduler == 2 && ee.deadline > 0) {

	CVMthreadGetPriority(&(ee.threadInfo), &pri);
	deadline = ee.startNextPeriodMicros + ee.deadline;

//	pthread_mutex_lock(&threadQmtx);
	CVMsysMutexLock(&ee, &CVMglobals.threadLock);
//
	printf("     [JVM] A new thread is created. \n");
	//printf("     [HACK] EDF thread is identified by its initial priority of 40.\n");
	printf("     =================================================================\n");
	printf ("      NEW Thread %d with deadline (%Ld) and priority (%d)\n",
		ee.threadID, deadline % 1000000000, pri);
	//      printf("     ThreadID pthreadID priority     deadline\n");

        pri = getEligiblePriority0 (&ee);

	printf ("      EDF thread %d priority is set to %d (%d, %d)\n",
		ee.threadID, pri, min, max);
	printf("     =================================================================\n");

	p.sched_priority = pri;
	pthread_setschedparam(ee.threadInfo.pthreadCookie, SCHED_FIFO, &p);
	CVMsysMutexUnlock(&ee, &CVMglobals.threadLock);
//	pthread_mutex_unlock(&threadQmtx);
      }

    }
    /************************ ISHIN **************************/
    


    clazz = CVMcbJavaInstance(CVMsystemClass(java_lang_Thread));

    /* Call Thread.startup() */
    methodID = (*env)->GetMethodID(env, clazz, "startup", "(JZ)V");
    CVMassert(methodID != NULL);

    if (nativeFunc != NULL) {
	ee.nativeRunInfo = (void *)info;
    } else {
	free(info);
    }

    (*env)->CallVoidMethod(env, eeThreadCell, methodID,
	eetopLongVal, nativeFunc != NULL);

    /* *** RTSJ *** */
    {
        jthrowable exception = NULL;
        exception=(*env)->ExceptionOccurred(env);
        if(exception != NULL)
        {
            // Exception was thrown. This means that the Thread and ThreadGroup
            // classes could not successfully handle the exception.
            // This is most probably an IllegalAssignmentError or MemoryAccessError. 
            // Handle it here by printing an error statement.
            // 
            // - Pratik & Kunal
            CVMClassBlock *cb;
            CVMID_objectGetClass(&ee, exception, cb);
            if(cb == CVMsystemClass(javax_realtime_IllegalAssignmentError) ||
               cb == CVMsystemClass(javax_realtime_MemoryAccessError))
            {
                CVMconsolePrintf("Thread %I terminated with exception %I\n", eeThreadCell, exception);
                (*env)->ExceptionClear(env);
            }
        
        }
    }
    /* *** JSTR *** */
    CVMassert(!(*env)->ExceptionCheck(env));

finish:
#ifdef CVM_JVMDI
    if (CVMglobals.jvmdiDebuggingEnabled) {
	CVMjvmdiNotifyDebuggerOfThreadEnd(&ee, CVMcurrentThreadICell(&ee));
    }
#endif

    {
      extern intFuncPtr JVMCeiFilters[] ;      
      extern int JVMCeiInitialized;
      jobject threadObject;
      threadObject = (*env)->NewLocalRef(env, CVMcurrentThreadICell(&ee));

      /*printf("jvm: trying to call core extension hookup before destroying thread\n");*/
      if (JVMCeiInitialized)
	  {
	    if (JVMCeiFilters[JVMCEF_BEFORE_DESTROY_SYS_THREAD] != NULL)
	    {
	      struct arg_t{ JNIEnv *env; jobject tObject; } arg;
	      arg.env = env; arg.tObject = threadObject;
	      /*printf("calling rtthreadadapter.destroySysThread\n");*/
	      JVMCeiFilters[JVMCEF_BEFORE_DESTROY_SYS_THREAD]((void *) &arg);
	      /*printf("after calling rtthreadadapter.destroySysThread\n");*/
	    }
	  }
	}

    CVMdestroyExecEnv(&ee);

    /*
     * Thread is effectively dead from the VM's point of view, so
     * no VM data structures can be accessed after this point.
     */
}

JNIEXPORT void JNICALL
JVM_RunNativeThread(JNIEnv *env, jobject this)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    ThreadStartInfo *info = (ThreadStartInfo *)ee->nativeRunInfo;
    void *arg = info->nativeFuncArg;
    void (*nativeFunc)(void *) = info->nativeFunc;

    ee->nativeRunInfo = NULL;
    free(info);

    (*nativeFunc)(arg);
}

JNIEXPORT void JNICALL
JVM_StartThread(JNIEnv *env, jobject this, jint priority)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    ThreadStartInfo *info = (ThreadStartInfo *)ee->nativeRunInfo;
    CVMThreadICell *threadICell = CVMID_getGlobalRoot(ee);
    int policy=0;

    /*********** ISHIN **********/
    //    printf ("ISHIN - JVM_StartThread - priority - %d\n", priority);
    //    ee->threadInfo.lastPriority = priority;
    /*********** ISHIN **********/


    if (threadICell == NULL) {
	CVMthrowOutOfMemoryError(ee, NULL);
	return;
    }

    if (info == NULL) {
	/* Normal thread */
	info = malloc(sizeof(ThreadStartInfo));
	info->nativeFunc = NULL;
	info->nativeFuncArg = NULL;
    }
    info->threadICell = threadICell;

    {
	CVMID_icellAssign(ee, threadICell, this);
	CVMID_fieldReadInt(ee, threadICell,
			   CVMoffsetOfjava_lang_Thread_daemon,
			   info->isDaemon);

	CVMassert(info->nativeFunc == NULL || info->isDaemon);

	CVMaddThread(ee, !info->isDaemon);

	{
	  extern intFuncPtr JVMCeiFilters[] ;      
	  extern int JVMCeiInitialized;

	  if (JVMCeiInitialized)
	    {
	      if (JVMCeiFilters[JVMCEF_BEFORE_CREATE_SYS_THREAD] != NULL)
		{
		  struct  arg_t {
		    JNIEnv *env; 
		    jobject tObject; 
		    CVMInt16 *ppolicy;
		    CVMInt16 *ppriority;
		  } arg;
		  
		  arg.env = env; 
		  arg.tObject = (*env)->NewLocalRef(env, threadICell);
		  arg.ppriority = (CVMInt16*) &priority;
		  arg.ppolicy = (CVMInt16*) &policy;
		  
#ifdef CVM_DEBUG
		  /*printf("calling rtthreadadapter\n");*/
#endif
		  
		  JVMCeiFilters[JVMCEF_BEFORE_CREATE_SYS_THREAD]((void *) &arg);
		  
#ifdef CVM_DEBUG
		  /*printf("after calling rtthreadadapter\n");*/
#endif

		  
		}
	    }
	}

    /* *** RTSJ *** */
    /* RTSJ Memory Area stack inheritance
     * - Pratik & Kunal (16th July 2001)
     */
    {
        PerThreadMemoryInfo *tempStack = NULL, *tempNode = NULL, *tos = NULL;
        PerThreadMemoryInfo *traverser = ee->perThreadStack;
        
        MemoryAreaInfo *threadArea = CVMJTregistryGetMemoryAreaInfoForObject(ee, this);
        if(threadArea->memType == MEMTYPE_IMMORTAL ||
                threadArea->memType == MEMTYPE_HEAP)
        {
            //no copying needed
            info->inheritedStack = NULL;
        }
        else
        {
            //first go to the memory area that the thread is allocated in
            while(traverser != NULL)
            {
                if(traverser->info == threadArea)
                    break;
                traverser = traverser->upInStack;
            }
            if(traverser == NULL)
            {
                printf("PANIC! Trying to start a thread allocated in a scope which is not on the parent's stack. Asserting\n");
                CVMassert(CVM_FALSE);
            }
                    
            while(traverser != NULL)
            {
                tempStack = CVMJTregistryCreatePTMI(ee);
                tempStack->info = traverser->info;
                tempStack->downInStack = tempNode;
                if(tempNode != NULL)
                {
                    tempNode->upInStack = tempStack;
                }
                else
                {
                    tos = tempStack;
                }
                tempNode = tempStack;
                traverser = traverser->upInStack;
            }
            info->inheritedStack = tos;
        }
    }
	if (!CVMthreadCreate(CVMglobals.nativeStackSize,
	    priority, start_func, info))
	{
	    CVMremoveThread(ee, !info->isDaemon);
	    CVMthrowOutOfMemoryError(ee, NULL);
	}
	/* this is temporary.. -Kalpesh */
//	ee->pid = getpid();
   }
}

void
JVM_StartSystemThread(JNIEnv *env, jobject this,
		      void(*nativeFunc)(void *), void* nativeFuncArg)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    ThreadStartInfo *info = malloc(sizeof(ThreadStartInfo));
    info->nativeFunc = nativeFunc;
    info->nativeFuncArg = nativeFuncArg;
    ee->nativeRunInfo = info;
  //  ee->pid = getpid();

    {
	/* Call Thread.start() */
	jclass clazz = CVMcbJavaInstance(CVMsystemClass(java_lang_Thread));
	jmethodID methodID = (*env)->GetMethodID(env, clazz, "start", "()V");
	CVMassert(methodID != NULL);

	(*env)->CallNonvirtualVoidMethod(env, this, clazz, methodID);
    }

    ee->nativeRunInfo = NULL;
}

#if defined(CVM_HAVE_DEPRECATED) || defined(CVM_JVMDI)

JNIEXPORT void JNICALL
JVM_SuspendThread(JNIEnv *env, jobject thread)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMThreadICell *thisThreadCell;
    CVMExecEnv *targetEE;
    CVMJavaLong eetopVal;
    CVMBool isSelf;
   
    thisThreadCell = CVMcurrentThreadICell(ee);
    CVMID_icellSameObject(ee, thisThreadCell, thread, isSelf);
    if (!isSelf) {
	CVMlocksForThreadSuspendAcquire(ee);

	CVMID_fieldReadLong(ee, thread,
			    CVMoffsetOfjava_lang_Thread_eetop,
			    eetopVal);
	targetEE = (CVMExecEnv *)CVMlong2VoidPtr(eetopVal);

	if (targetEE != NULL) {
	    CVMthreadSuspendConsistentRequest(ee, targetEE);
	    targetEE->threadState |= CVM_THREAD_SUSPENDED;
	    CVMthreadSuspend(&targetEE->threadInfo);
	    CVMthreadSuspendConsistentRelease(ee, targetEE);
	}
	CVMlocksForThreadSuspendRelease(ee);
    } else {
	/* %comment: rt034 */
	ee->threadState = CVM_THREAD_SUSPENDED;
	CVMthreadSuspend(&ee->threadInfo);
    }
}

JNIEXPORT void JNICALL
JVM_ResumeThread(JNIEnv *env, jobject thread)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMExecEnv *targetEE;
    CVMJavaLong eetopVal;

    CVMsysMutexLock(ee, &CVMglobals.threadLock);

    CVMID_fieldReadLong(ee, thread,
			CVMoffsetOfjava_lang_Thread_eetop,
			eetopVal);
    targetEE = (CVMExecEnv *)CVMlong2VoidPtr(eetopVal);

    if (targetEE != NULL) {
	if (targetEE->threadState & CVM_THREAD_SUSPENDED) {
	    CVMassert(targetEE != ee);
	    CVMthreadResume(&targetEE->threadInfo);
	    targetEE->threadState &= ~CVM_THREAD_SUSPENDED;
	}
    }
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
}

#endif /* CVM_HAVE_DEPRECATED || CVM_JVMDI */

/* We want to rely on priorities as set/got via the threadX functions.
 * However, it is legal to call setPriority() and getPriority() after
 * 'new' and before start() is called, in which case the thread's TID has
 * been allocated but its thread not yet created.  So we also maintain
 * tid->priority to bridge the gap.
 *
 * The priority should already have been check to fall within a
 * valid range by the software above.  All we do is pass it on to
 * the system level.  The JAVA level also set the priority field.
 */
JNIEXPORT void JNICALL
JVM_SetThreadPriority(JNIEnv *env, jobject thread, jint prio)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMExecEnv *targetEE;
    CVMJavaLong eetopVal;
   
    CVMsysMutexLock(ee, &CVMglobals.threadLock);

    CVMID_fieldReadLong(ee, thread,
			CVMoffsetOfjava_lang_Thread_eetop,
			eetopVal);
    targetEE = (CVMExecEnv *)CVMlong2VoidPtr(eetopVal);

    if (targetEE != NULL) {
	CVMthreadSetPriority(&targetEE->threadInfo, prio);
    }

    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
}

JNIEXPORT void JNICALL
JVM_Yield(JNIEnv *env, jclass threadClass)
{
    CVMthreadYield();
}

JNIEXPORT void JNICALL
JVM_Sleep(JNIEnv *env, jclass threadClass, jlong millis)
{
  CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);

#ifdef JDK12
    if (CVMlongLtz(millis)) {
	ThrowIllegalArgumentException(CVMjniEnv2ExecEnv(env),
	    "timeout value is negative");
	return;
    }
#endif

#ifdef HPROF
    jvmpi_monitor_wait(NULL, millis);
#endif
 
   /* Dio: enable changing Scheduler */
    ee->canChangeScheduler = 1;

    {
	if (CVMlongEqz(millis)) {
	    CVMthreadYield();
	} else {
	    CVMWaitStatus st;
	    CVMSysMonitor mon;
	    CVMsysMonitorInit(&mon, NULL, 0);
	    CVMsysMonitorEnter(ee, &mon);
	    st = CVMsysMonitorWait(ee, &mon, millis);
	    CVMsysMonitorExit(ee, &mon);
	    CVMsysMonitorDestroy(&mon);
	    if (st == CVM_WAIT_INTERRUPTED) {
		CVMthrowInterruptedException(ee, "operation interrupted");
	    }
	}
    }

#ifdef HPROF
    jvmpi_monitor_waited(NULL);
#endif

    /* Dio: disable changing the scheduler */
    ee->canChangeScheduler = 0;

    return;
}

JNIEXPORT jobject JNICALL
JVM_CurrentThread(JNIEnv *env, jclass threadClass)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    return (*env)->NewLocalRef(env, CVMcurrentThreadICell(ee));
}

#ifdef CVM_HAVE_DEPRECATED

/* Unlike JDK, this implementation counts native methods. */
JNIEXPORT jint JNICALL
JVM_CountStackFrames(JNIEnv *env, jobject thread)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);

    if (thread == NULL) {
	CVMthrowNullPointerException(ee, 0);
	return -1;
    } else {
	CVMExecEnv *targetEE;
	CVMJavaLong eetopVal;
	jint count = 0;
   
	CVMsysMutexLock(ee, &CVMglobals.threadLock);

	CVMID_fieldReadLong(ee, thread,
			    CVMoffsetOfjava_lang_Thread_eetop,
			    eetopVal);
	targetEE = (CVMExecEnv *)CVMlong2VoidPtr(eetopVal);

	if (targetEE != NULL) {
	    CVMFrame *frame = CVMeeGetCurrentFrame(targetEE);
	    while (frame != NULL) { 
		if (!CVMframeIsTransition(frame)) {
		    ++count;
		}
		frame = CVMframePrev(frame);
	    }
	} else {
	    /* Thread hasn't any state yet. */
	}
	CVMsysMutexUnlock(ee, &CVMglobals.threadLock);

	return count;
    }
}

#endif /* CVM_HAVE_DEPRECATED */

JNIEXPORT void JNICALL
JVM_Interrupt(JNIEnv *env, jobject thread)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMExecEnv *targetEE;
    CVMJavaLong eetopVal;
   
    CVMsysMutexLock(ee, &CVMglobals.threadLock);

    CVMID_fieldReadLong(ee, thread,
			CVMoffsetOfjava_lang_Thread_eetop,
			eetopVal);
    targetEE = (CVMExecEnv *)CVMlong2VoidPtr(eetopVal);

    /* %comment: rt035 */
    if (targetEE != NULL) {
	CVMthreadInterruptWait(CVMexecEnv2threadID(targetEE));
    }

    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
}

JNIEXPORT jboolean JNICALL
JVM_IsInterrupted(JNIEnv *env, jobject thread, jboolean clearInterrupted)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMExecEnv *targetEE;
    CVMThreadICell *thisThreadCell;
    CVMJavaLong eetopVal;
    jboolean result = JNI_FALSE;
    CVMBool isSelf;

    thisThreadCell = CVMcurrentThreadICell(ee);
    CVMID_icellSameObject(ee, thisThreadCell, thread, isSelf);

    if (isSelf) {
	/*
	 * The thread is the current thread, so we can avoid the
	 * locking because we know we aren't going away.
	 */
	result = CVMthreadIsInterrupted(CVMexecEnv2threadID(ee),
	    clearInterrupted);
    } else {
	/* a thread can only clear its own interrupt */
	CVMassert(!clearInterrupted);
	CVMsysMutexLock(ee, &CVMglobals.threadLock);
	CVMID_fieldReadLong(ee, thread,
			    CVMoffsetOfjava_lang_Thread_eetop,
			    eetopVal);
	if (!CVMlongEqz(eetopVal)) {
	    targetEE = (CVMExecEnv *)CVMlong2VoidPtr(eetopVal);
	    result = CVMthreadIsInterrupted(CVMexecEnv2threadID(targetEE),
		clearInterrupted);
	}
	CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
    }

    return result;
}

/*
 * SecurityManager support
 */

/* 
 * Implemented in startup code.
 *     bool_t InitializeSystemClassLoader(void)
 *
 * Implemented in interpreter.c as CVMisTrustedClassLoader().
 *     bool_t IsTrustedClassLoader(CVMClassLoaderICell *loader)
 */

#undef TRUSTED_CLASSLOADER
#ifdef CVM_CLASSLOADING
#define TRUSTED_CLASSLOADER(ee, loader)	\
	CVMisTrustedClassLoader((ee), (loader))
#else
#define TRUSTED_CLASSLOADER(ee, loader)	\
	(CVMassert((loader) == NULL), CVM_TRUE)
#endif

static CVMBool
is_trusted_frame(CVMExecEnv *ee, CVMFrame *frame)
{
    if (ee->privilege != NULL && (ee->privilege->frame == frame)) {
	CVMClassLoaderICell *loader;
	CVMClassBlock *cb = ee->privilege->clazz;
	if (cb != NULL && ((loader = CVMcbClassLoader(cb)) == NULL ||
	    TRUSTED_CLASSLOADER(ee, loader)))
	{
	    return CVM_TRUE;
	}
    }
    return CVM_FALSE;
}

static CVMClassBlock *
current_loaded_class(JNIEnv *env)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMFrame *frame = CVMeeGetCurrentFrame(ee);
    CVMClassBlock *cb;

    frame = CVMgetCallerFrameSpecial(frame, 0, CVM_FALSE);
    while (frame != NULL) {
	/* if a method in a class in a trusted loader is in a doPrivileged,
	   return NULL */

	if (is_trusted_frame(ee, frame)) {
	    return NULL;
	}

	CVMassert(frame->mb != NULL && !CVMframeIsTransition(frame));

	{
	    CVMClassLoaderICell *loader;
	    cb = CVMmbClassBlock(frame->mb);
	    if (cb != NULL && (loader = CVMcbClassLoader(cb)) != NULL &&
		!TRUSTED_CLASSLOADER(ee, loader))
	    {
	        return cb;
	    }
        }
	frame = CVMgetCallerFrameSpecial(frame, 1, CVM_FALSE);
    }
    return NULL;
}

JNIEXPORT jclass JNICALL
JVM_CurrentLoadedClass(JNIEnv *env)
{
    CVMClassBlock *cb = current_loaded_class(env);
    return (cb == NULL) ? NULL
	   : (*env)->NewLocalRef(env, CVMcbJavaInstance(cb));
}

JNIEXPORT jobject JNICALL
JVM_CurrentClassLoader(JNIEnv *env)
{
    CVMClassBlock *cb = current_loaded_class(env);
    return (cb == NULL) ? NULL
	   : (*env)->NewLocalRef(env, CVMcbClassLoader(cb));
}


JNIEXPORT jobjectArray JNICALL
JVM_GetClassContext(JNIEnv *env)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMFrame* frame;
    int n;
    jobjectArray result;
    CVMClassBlock* java_lang_class_arrayCb = 
	CVMclassGetArrayOf(ee, CVMsystemClass(java_lang_Class));

    /* count the classes on the execution stack */
    frame = CVMeeGetCurrentFrame(ee);
    /* skip current frame (SecurityManager.getClassContext) */
    frame = CVMframePrev(frame);
    n = 0;
    while (frame != NULL) { 
	n++;
	/* skip reflection frames */
	frame = CVMgetCallerFrameSpecial(frame, 1, CVM_TRUE);
    }

    /* allocate the Class array */
    CVMID_localrootBegin(ee) {
	CVMID_localrootDeclare(CVMObjectICell, classContextICell);
	CVMID_allocNewArray(ee, CVM_T_CLASS, java_lang_class_arrayCb,
			    n, classContextICell);
	/* The allocation nulls out the contents */

	if (!CVMID_icellIsNull(classContextICell)) {
	    result = (*env)->NewLocalRef(env, classContextICell);
	} else {
	    result = NULL;
	    CVMthrowOutOfMemoryError(ee, NULL);
	}
    } CVMID_localrootEnd();
    if (result == NULL) {
	return NULL;
    }

    /* fill in the Class array */
    {
	CVMArrayOfRefICell* classArray = (CVMArrayOfRefICell*)result;
	frame = CVMeeGetCurrentFrame(ee);
	/* skip current frame (SecurityManager.getClassContext) */
	frame = CVMframePrev(frame);
	n = 0;
	while (frame != NULL) {
	    CVMClassICell* clas =
		CVMcbJavaInstance(CVMmbClassBlock(frame->mb));
	    CVMID_arrayWriteRef(ee, classArray, n, clas);
	    n++;
	    /* skip reflection frames */
	    frame = CVMgetCallerFrameSpecial(frame, 1, CVM_TRUE);
	}
    }
    
    return result;
}

#ifdef CVM_HAVE_DEPRECATED

/* NOTE: The name string could probably be handled better... */

JNIEXPORT jint JNICALL
JVM_ClassDepth(JNIEnv *env, jstring name)
{
#ifdef JDK12
    struct javaframe *frame, frame_buf;
    ClassClass	*cb;
    char buf[STK_BUF_LEN], *p;
    jint depth = 0;

    javaString2CString((Hjava_lang_String *)DeRef(env, name), buf, sizeof(buf));
    for (p = buf ; *p ; p++) {
	if (*p == '.') {
	    *p = '/';
	}
    }

    for (frame = CVMjniEnv2ExecEnv(env)->current_frame ; frame != 0 ; ) {
	if ((frame->mb != 0) &&
	    !(frame->mb->fb.access & ACC_NATIVE)) {
	    cb = fieldclass(&frame->mb->fb);
	    if (cb && !strcmp(cbName(cb), buf)) {
		return depth;
	    }
	    depth++;
	}
	frame = FRAME_PREV(frame, &frame_buf);
    }
#else
    CVMconsolePrintf("JVM_ClassDepth() not implemented yet!\n");
    CVMassert(CVM_FALSE);
    return -1;
#endif
}

JNIEXPORT jint JNICALL
JVM_ClassLoaderDepth(JNIEnv *env)
{
#ifdef JDK12
    struct javaframe *frame, frame_buf;
    ClassClass	*cb;
    jint depth = 0;

    for (frame = CVMjniEnv2ExecEnv(env)->current_frame ; frame != 0 ; ) {
	/* if a method in a class in a trusted loader is in a doPrivileged,
	   return -1 */

	if (is_trusted_frame(env, frame))
	    return -1;

	if ((frame->mb != 0) && 
	    !(frame->mb->fb.access & ACC_NATIVE)) {
	    cb = fieldclass(&frame->mb->fb);
	    if (cb && cbLoader(cb) && !IsTrustedClassLoader(cbLoader(cb))) {
		return depth;
	    }
            depth++;
        }
	frame = FRAME_PREV(frame, &frame_buf);
    }
#else
    CVMconsolePrintf("JVM_ClassLoaderDepth() not implemented yet!\n");
    CVMassert(CVM_FALSE);
    return -1;
#endif
}

#endif /* CVM_HAVE_DEPRECATED */

/*
 * The following functions from ObjectInputStream.c
 * cannot be done with JNI or reflection
 *
 * (should we rethink how these might be implemented
 * using the already-existing mechanisms?)
 */

#define CVM_JAVA_IO_PACKAGE "java/io/"

/*
 * Allocate an initialized object of the specified class.
 *
 * See ObjectInputStream.java, inputObject, for more details. This
 * function seems a little strange because currClass, the actual class
 * of the object, does not necessarily have to be the same as
 * initClass, the class from which the default no-argument constructor
 * is called. The explanation is in the middle of initObject: if on
 * the receiving machine the class to be instantiated (call it B) is
 * not declared Serializable, the VM walks up the class hierarchy to
 * find class A, the most specialized superclass declaring
 * serializability. B is allocated, but A's no-argument constructor is
 * the constructor called; this has the effect of setting the rest of
 * B's fields to their default values.
 */
JNIEXPORT jobject JNICALL
JVM_AllocateNewObject(JNIEnv *env, jobject this, jclass currClass,
		      jclass initClass)
{
#ifdef CVM_SERIALIZATION
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb;
    CVMClassBlock* initCb;
    CVMMethodBlock* mb = NULL;
    int len;
    CVMMethodTypeID noArgConstructorTid;
    jobject result;

    cb = CVMgcSafeClassRef2ClassBlock(ee, currClass);
    initCb = CVMgcSafeClassRef2ClassBlock(ee, initClass);

    if (CVMcbIs(cb, INTERFACE) || CVMcbIs(cb, ABSTRACT) ||
	cb == CVMsystemClass(java_lang_Class)) {
	CVMthrowInstantiationException(ee, "%C", cb);
	return NULL;
    }

    noArgConstructorTid = CVMglobals.initTid;

    for (len = CVMcbMethodCount(initCb) - 1; len >= 0; len--) {
	mb = CVMcbMethodSlot(initCb, len);
	if (CVMtypeidIsSame(CVMmbNameAndTypeID(mb), noArgConstructorTid)) {
	    break;
	}
    }
    if (len < 0) { 
	CVMthrowNoSuchMethodError(ee, "<init>");
	return NULL;
    }

    if (cb == initCb && !CVMmbIs(mb, PUBLIC)) {
	/* Calling the constructor for class 'cb'. 
	 * Only allow calls to a public no-arg constructor.
	 * This path corresponds to creating an Externalizable object.
	 */
	CVMthrowIllegalAccessException(ee, "%M", mb);
	return NULL;
    }
    else if (!CVMverifyMemberAccess(ee, cb, initCb, CVMmbAccessFlags(mb),
				    CVM_FALSE)) {
	/* subclass 'cb' does not have access to no-arg constructor of
           'initCb'.*/
	CVMthrowIllegalAccessException(ee, "%M", mb);
	return NULL;
    }

    result = CVMjniCreateLocalRef(ee);
    if (result == NULL) {
	return NULL;  /* exception already thrown */
    }

    /* Allocate the object  */
    CVMID_allocNewInstance(ee, cb, result);
    if (CVMID_icellIsNull(result)) {
	(*env)->DeleteLocalRef(env, result);
	CVMthrowOutOfMemoryError(ee,
				 "allocating receiver-side serialized object");
        return NULL;
    }

    /* Invoke the no-arg constructor */
    (*env)->CallVoidMethod(env, result, mb);

    if ((*env)->ExceptionCheck(env)) {
	/* If the constructor threw an exception, we won't need the return
	   local ref. Delete, and return NULL. */
	(*env)->DeleteLocalRef(env, result);
	return NULL;
    }

    return result;
#else /* CVM_SERIALIZATION */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_SERIALIZATION */
}

/*
 * Allocate a new array of the specified length.
 *
 * currClass is the Java class corresponding to the desired array type.
 *
 * Does not initialize the array's cells.
 */
JNIEXPORT jobject JNICALL
JVM_AllocateNewArray(JNIEnv *env, jobject this, jclass currClass,
		     jint length)
{
#ifdef CVM_SERIALIZATION

    /* NOTE that this depends on reflection. This is okay because
       object serialization is meaningless without reflection support.
       In addition, our makefiles make sure to set CVM_REFLECT when
       CVM_SERIALIZATION is set. */

    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = CVMgcSafeClassRef2ClassBlock(ee, currClass);
    jobject result;

    if (!CVMisArrayClass(cb)) {
	CVMthrowInvalidClassException(ee, "%C", cb);
	return NULL;
    }
    result = CVMjniCreateLocalRef(ee);
    if (result == NULL) {
	return NULL;  /* exception already thrown */
    }
    CVMreflectNewArray(ee, CVMarrayElementCb(cb),
		       length, (CVMArrayOfAnyTypeICell*) result);
    if (CVMID_icellIsNull(result)) {
	(*env)->DeleteLocalRef(env, result);
	return NULL;  /* exception already thrown */
    }
    return result;
#else /* CVM_SERIALIZATION */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_SERIALIZATION */
}

/*
 * Return the first non-null class loader up the execution stack, or null
 * if only code from the null class loader is on the stack.
 */
JNIEXPORT jobject JNICALL
JVM_LatestUserDefinedLoader(JNIEnv *env)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMFrame *frame = CVMeeGetCurrentFrame(ee);

    while (frame != NULL) {
	if (frame->mb != NULL) {
	    CVMClassLoaderICell *loader =
	        CVMcbClassLoader(CVMmbClassBlock(frame->mb));
	    if (loader != NULL) {
		return (*env)->NewLocalRef(env, loader);
	    }
	}
	frame = CVMframePrev(frame);
    }

    return NULL;
}

/*
 * Code supporting reflection: used by Array.c
 */

JNIEXPORT jint JNICALL
JVM_GetArrayLength(JNIEnv *env, jobject arr)
{
#ifdef CVM_REFLECT
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* arrayCb;
    CVMObjectICell* arrObj = (CVMObjectICell*) arr;
    CVMJavaInt arrLen;
    
    if ((arrObj == NULL) || (CVMID_icellIsNull(arrObj))) {
	CVMthrowNullPointerException(ee, "Array.getLength()");
	return 0;
    }

    CVMID_objectGetClass(ee, arrObj, arrayCb);
    if (!CVMisArrayClass(arrayCb)) {
	CVMthrowIllegalArgumentException(ee, "object not of array type");
	return 0;
    }

    CVMID_arrayGetLength(ee, (CVMArrayOfAnyTypeICell*) arrObj, arrLen);

    return arrLen;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return 0;
#endif /* CVM_REFLECT */
}

JNIEXPORT jobject JNICALL
JVM_GetArrayElement(JNIEnv *env, jobject arr, jint index)
{
#ifdef CVM_REFLECT
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* arrayCb;
    CVMClassBlock* elementCb;
    CVMArrayOfAnyTypeICell* arrObj = (CVMArrayOfAnyTypeICell*) arr;
    CVMObjectICell* result;
    CVMJavaInt arrLen;

    if ((arrObj == NULL) || (CVMID_icellIsNull(arrObj))) {
	CVMthrowNullPointerException(ee, "Array.get()");
	return NULL;
    }

    CVMID_objectGetClass(ee, (CVMObjectICell*) arrObj, arrayCb);
    if (!CVMisArrayClass(arrayCb)) {
	CVMthrowIllegalArgumentException(ee, "object not of array type");
	return NULL;
    }
    
    elementCb = CVMarrayElementCb(arrayCb);
    CVMassert(elementCb != NULL);

    result = CVMjniCreateLocalRef(ee);
    if (result == NULL) {
	return NULL;
    }

    if (CVMcbIs(elementCb, PRIMITIVE)) {
	jvalue v = JVM_GetPrimitiveArrayElement(env, arr, index,
						CVMcbBasicTypeCode(elementCb));
	if (CVMexceptionOccurred(ee) | 
	    !CVMgcSafeJavaWrap(ee, v, CVMcbBasicTypeCode(elementCb), result)) {
	    (*env)->DeleteLocalRef(env, result);
	    return NULL;
	}
    } else {
	/* This clever idea is from ExactVM. No need to test index >= 0.
	   Instead, do an unsigned comparison. Negative indices will
	   appear as positive numbers >= 2^31, which is just out of the
	   range of legal Java array indices */
	CVMID_arrayGetLength(ee, arrObj, arrLen);
	if ((CVMUint32)index >= (CVMUint32)arrLen) {
	    CVMthrowArrayIndexOutOfBoundsException(ee,
						   "accessing array element");
	    (*env)->DeleteLocalRef(env, result);
	    return NULL;
	}

	CVMID_arrayReadRef(ee, arrObj, index, result);
	if (CVMID_icellIsNull(result)) {
	    (*env)->DeleteLocalRef(env, result);
	    result = NULL;
	}
    }
    return result;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

JNIEXPORT jvalue JNICALL
JVM_GetPrimitiveArrayElement(JNIEnv *env, jobject arr, jint index, jint wCode)
{
#ifdef CVM_REFLECT
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* arrayCb;
    CVMClassBlock* elementCb;
    CVMArrayOfAnyTypeICell* arrObj = (CVMArrayOfAnyTypeICell*) arr;
    CVMJavaInt arrLen;
    CVMBasicType fromType;
    CVMBasicType toType = (CVMBasicType) wCode;
    jvalue v;

    /* Initialize v so compiler doesn't complain */
    v.i = 0;

    if ((arrObj == NULL) || (CVMID_icellIsNull(arrObj))) {
	CVMthrowNullPointerException(ee, "Array.getPrimitiveElement()");
	return v;
    }

    CVMID_objectGetClass(ee, (CVMObjectICell*) arrObj, arrayCb);
    if (!CVMisArrayClass(arrayCb)) {
	CVMthrowIllegalArgumentException(ee, "object not of array type");
	return v;
    }
    
    elementCb = CVMarrayElementCb(arrayCb);
    CVMassert(elementCb != NULL);

    /* This clever idea is from ExactVM. No need to test index >= 0.
       Instead, do an unsigned comparison. Negative indices will
       appear as positive numbers >= 2^31, which is just out of the
       range of legal Java array indices */
    CVMID_arrayGetLength(ee, arrObj, arrLen);
    if ((CVMUint32)index >= (CVMUint32)arrLen) {
	CVMthrowArrayIndexOutOfBoundsException(ee, "accessing array element");
	return v;
    }

    if (!CVMcbIs(elementCb, PRIMITIVE)) {
	CVMthrowIllegalArgumentException(ee,
					 "array object does not contain "
					 "elements of primitive type");
	return v;
    }

    fromType = CVMcbBasicTypeCode(elementCb);
    switch (fromType) {
    case CVM_T_BOOLEAN:
	CVMID_arrayReadBoolean(ee, (CVMArrayOfBooleanICell*) arrObj,
			       index, v.z);
	break;
    case CVM_T_CHAR:
	CVMID_arrayReadChar(ee, (CVMArrayOfCharICell*) arrObj,
			    index, v.c);
	break;
    case CVM_T_FLOAT:
	CVMID_arrayReadFloat(ee, (CVMArrayOfFloatICell*) arrObj,
			     index, v.f);
	break;
    case CVM_T_DOUBLE:
	CVMID_arrayReadDouble(ee, (CVMArrayOfDoubleICell*) arrObj,
			      index, v.d);
	break;
    case CVM_T_BYTE:
	CVMID_arrayReadByte(ee, (CVMArrayOfByteICell*) arrObj,
			    index, v.b);
	break;
    case CVM_T_SHORT:
	CVMID_arrayReadShort(ee, (CVMArrayOfShortICell*) arrObj,
			     index, v.s);
	break;
    case CVM_T_INT:
	CVMID_arrayReadInt(ee, (CVMArrayOfIntICell*) arrObj,
			   index, v.i);
	break;
    case CVM_T_LONG:
	CVMID_arrayReadLong(ee, (CVMArrayOfLongICell*) arrObj,
			    index, v.j);
	break;
    default:
	CVMassert(CVM_FALSE);
	return v;
    }
    
    if (fromType != toType)
	REFLECT_WIDEN(v, fromType, toType, fail);
    
    return v;

fail:
    CVMthrowIllegalArgumentException(ee, "widening primitive value");
    return v;
#else /* CVM_REFLECT */
    jvalue v;
    v.i = 0;
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return v;
#endif /* CVM_REFLECT */
}

JNIEXPORT void JNICALL
JVM_SetArrayElement(JNIEnv *env, jobject arr, jint index, jobject val)
{
#ifdef CVM_REFLECT
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* arrayCb;
    CVMClassBlock* elementCb;
    CVMArrayOfAnyTypeICell* arrObj = (CVMArrayOfAnyTypeICell*) arr;
    CVMJavaInt arrLen;
    
    if ((arrObj == NULL) || (CVMID_icellIsNull(arrObj))) {
	CVMthrowNullPointerException(ee, "Array.set()");
	return;
    }

    CVMID_objectGetClass(ee, (CVMObjectICell*) arrObj, arrayCb);
    if (!CVMisArrayClass(arrayCb)) {
	CVMthrowIllegalArgumentException(ee, "object not of array type");
	return;
    }
    
    elementCb = CVMarrayElementCb(arrayCb);
    CVMassert(elementCb != NULL);

    if (CVMcbIs(elementCb, PRIMITIVE)) {
	jvalue v;
	CVMBasicType fromType;

	if (CVMgcSafeJavaUnwrap(ee, (CVMObjectICell*) val,
				&v, &fromType, NULL))
	    JVM_SetPrimitiveArrayElement(env, arr, index, v, fromType);
    } else {
	/* See whether the object is of the appropriate type */
	CVMBool isOfCorrectType;

	if (val != NULL) {
	    CVMD_gcUnsafeExec(ee, {
		CVMObject* directObj = CVMID_icellDirect(ee, val);
		isOfCorrectType = CVMgcUnsafeIsInstanceOf(ee, directObj,
							  elementCb);
	    });
	    
	    if (!isOfCorrectType) {
		CVMthrowIllegalArgumentException(ee, "argument object is not "
						 "an instance of the array's "
						 "element type");
		return;
	    }
	}

	/* This clever idea is from ExactVM. No need to test index >= 0.
	   Instead, do an unsigned comparison. Negative indices will
	   appear as positive numbers >= 2^31, which is just out of the
	   range of legal Java array indices */
	CVMID_arrayGetLength(ee, arrObj, arrLen);
	if ((CVMUint32)index >= (CVMUint32)arrLen) {
	    CVMthrowArrayIndexOutOfBoundsException(ee,
						   "accessing array element");
	    return;
	}
	if (val != NULL) {
	    CVMID_arrayWriteRef(ee, arrObj, index, val);
	} else {
	    CVMID_localrootBegin(ee) {
		CVMID_localrootDeclare(CVMObjectICell, nullICell);
		CVMID_icellSetNull(nullICell);
		CVMID_arrayWriteRef(ee, arrObj, index, nullICell);
	    } CVMID_localrootEnd();
	}
    }
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
#endif /* CVM_REFLECT */
}

JNIEXPORT void JNICALL
JVM_SetPrimitiveArrayElement(JNIEnv *env, jobject arr, jint index, jvalue v,
			     unsigned char vCode)
{
#ifdef CVM_REFLECT
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* arrayCb;
    CVMClassBlock* elementCb;
    CVMArrayOfAnyTypeICell* arrObj = (CVMArrayOfAnyTypeICell*) arr;
    CVMJavaInt arrLen;
    CVMBasicType toType;
    
    if ((arrObj == NULL) || (CVMID_icellIsNull(arrObj))) {
	CVMthrowNullPointerException(ee, "Array.setPrimitiveElement()");
	return;
    }

    CVMID_objectGetClass(ee, (CVMObjectICell*) arrObj, arrayCb);
    if (!CVMisArrayClass(arrayCb)) {
	CVMthrowIllegalArgumentException(ee, "object not of array type");
	return;
    }
    
    elementCb = CVMarrayElementCb(arrayCb);
    CVMassert(elementCb != NULL);
    if (!CVMcbIs(elementCb, PRIMITIVE)) {
	CVMthrowIllegalArgumentException(ee, "illegal array element type");
	return;
    }

    /* This clever idea is from ExactVM. No need to test index >= 0.
       Instead, do an unsigned comparison. Negative indices will
       appear as positive numbers >= 2^31, which is just out of the
       range of legal Java array indices */
    CVMID_arrayGetLength(ee, arrObj, arrLen);
    if ((CVMUint32)index >= (CVMUint32)arrLen) {
	CVMthrowArrayIndexOutOfBoundsException(ee,
					       "accessing array element");
	return;
    }

    toType = CVMcbBasicTypeCode(elementCb);

    if (vCode != toType)
	REFLECT_WIDEN(v, vCode, toType, fail);

    switch (toType) {
    case CVM_T_BOOLEAN:
	CVMID_arrayWriteBoolean(ee, (CVMArrayOfBooleanICell*) arrObj,
				index, v.b);
	break;
    case CVM_T_CHAR:
	CVMID_arrayWriteChar(ee, (CVMArrayOfCharICell*) arrObj,
			     index, v.c);
	break;
    case CVM_T_FLOAT:
	CVMID_arrayWriteFloat(ee, (CVMArrayOfFloatICell*) arrObj,
			      index, v.f);
	break;
    case CVM_T_DOUBLE:
	CVMID_arrayWriteDouble(ee, (CVMArrayOfDoubleICell*) arrObj,
			       index, v.d);
	break;
    case CVM_T_BYTE:
	CVMID_arrayWriteByte(ee, (CVMArrayOfByteICell*) arrObj,
			     index, v.b);
	break;
    case CVM_T_SHORT:
	CVMID_arrayWriteShort(ee, (CVMArrayOfShortICell*) arrObj,
			      index, v.s);
	break;
    case CVM_T_INT:
	CVMID_arrayWriteInt(ee, (CVMArrayOfIntICell*) arrObj,
			    index, v.i);
	break;
    case CVM_T_LONG:
	CVMID_arrayWriteLong(ee, (CVMArrayOfLongICell*) arrObj,
			     index, v.j);
	break;
    default:
	CVMassert(CVM_FALSE);
    }

    return;

fail:
    CVMthrowIllegalArgumentException(ee, "widening primitive value");
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
#endif /* CVM_REFLECT */
}

JNIEXPORT jobject JNICALL
JVM_NewArray(JNIEnv *env, jclass eltClass, jint length)
{
#ifdef CVM_REFLECT
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMJavaInt cbAddr;
    CVMClassBlock* cb;
    CVMObjectICell* classObj = (CVMObjectICell*) eltClass;
    jobject result;
    
    if ((classObj == NULL) || (CVMID_icellIsNull(classObj))) {
	CVMthrowNullPointerException(ee, "null class object");
	return NULL;
    }
    
    if (length < 0) {
	CVMthrowNegativeArraySizeException(ee, NULL);
	return NULL;
    }
    
    CVMID_fieldReadInt(ee, eltClass,
		       CVMoffsetOfjava_lang_Class_classBlockPointer,
		       cbAddr);
    cb = (CVMClassBlock*) cbAddr;
    CVMassert(cb != NULL);

    if (CVMisArrayClass(cb)) {
	/* Check for out-of-bounds array dimensions */
	int eltDims = CVMarrayDepth(cb);
	/* If eltDims is already 255, the we can't go on to create
	 * something of dims 256. */
	
	if (eltDims >= 255) {
	    CVMassert(eltDims > 255); /* this should never happen because
				       * arrays always have 255 or fewer
				       * dimensions */
	    CVMthrowIllegalArgumentException(ee, "too many dimensions");
	    return NULL;
	}
    }
    
    result = (*env)->NewLocalRef(env, CVMcurrentThreadICell(ee));
    if (result == NULL) {
	return NULL;
    }

    CVMreflectNewArray(ee, cb, length, (CVMArrayOfAnyTypeICell*)result);

    if (CVMexceptionOccurred(ee)) {
	(*env)->DeleteLocalRef(env, result);
	result = NULL;
    }

    return result;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

JNIEXPORT jobject JNICALL
JVM_NewMultiArray(JNIEnv *env, jclass eltClass, jintArray dim)
{
#ifdef CVM_REFLECT
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMJavaInt cbAddr;
    CVMClassBlock* cb;
    CVMClassTypeID arrayClassID;
    CVMBool	   computedArrayClassID = CVM_FALSE;
    CVMObjectICell* classObj = (CVMObjectICell*) eltClass;
    CVMArrayOfIntICell* arrayObj = (CVMArrayOfIntICell*) dim;
    CVMJavaInt arrLen;
    jobject res;
    int i;
    CVMJavaInt dimensionTmp;
    CVMInt32* dimensionsTmp;
    CVMInt32 effectiveDimension = -1; /* In case one of the entries is zero.
					 Because the java.lang.reflect.Array
					 package is so poorly specified,
					 I'm going to make this work just
					 like opc_multianewarray. */
    CVMInt32 elementDepth; /* For elements which are themselves arrays */
    
    if ((classObj == NULL) || (CVMID_icellIsNull(classObj))) {
	CVMthrowNullPointerException(ee, "null class object");
	return NULL;
    }
    
    if ((arrayObj == NULL) || (CVMID_icellIsNull(arrayObj))) {
	CVMthrowNullPointerException(ee, "null dimensions array");
	return NULL;
    }
    
    CVMID_arrayGetLength(ee, arrayObj, arrLen);
    if (arrLen == 0) {
	CVMthrowIllegalArgumentException(ee,
					 "zero-dimensional dimensions array");
	return NULL;
    }
    for (i = 0; i < arrLen; i++) {
	CVMID_arrayReadInt(ee, arrayObj, i, dimensionTmp);
	if (dimensionTmp < 0) {
	    CVMthrowNegativeArraySizeException(ee, NULL);
	    return NULL;
	} else if (dimensionTmp == 0) {
	    if (effectiveDimension == -1) {
		effectiveDimension = i + 1;
	    }
	}
    }

    /* %comment: rt036 */
    if (effectiveDimension == 1) {
	CVMthrowIllegalArgumentException(ee,
					 "zero-dimensional dimensions array");
	return NULL;
    }

    if (effectiveDimension == -1)
	effectiveDimension = arrLen;

    CVMID_fieldReadInt(ee, eltClass,
		       CVMoffsetOfjava_lang_Class_classBlockPointer,
		       cbAddr);
    cb = (CVMClassBlock*)cbAddr;
    CVMassert(cb != NULL);

    if (CVMisArrayClass(cb)) {
	/* Check total number of dimensions against 255
           limit.  See also JVM_NewArray above. */
	elementDepth = CVMarrayDepth(cb);
	if ((elementDepth + effectiveDimension) > 255) {
 	    CVMassert(elementDepth <= 255); /* this should never happen because
					     * arrays always have 255 or fewer
					     * dimensions */
	    CVMthrowIllegalArgumentException(ee, "too many dimensions");
	    return NULL;
	}
    }

    /* %comment: cjp021 */
    if (CVMcbIs(cb, PRIMITIVE)) {
	cb = CVMclassGetArrayOf(ee, cb);
	if (effectiveDimension > 1) {
	    arrayClassID = 
		CVMtypeidIncrementArrayDepth(ee, CVMcbClassName(cb),
					     effectiveDimension - 1);
	    computedArrayClassID = CVM_TRUE;
	} else {
	    arrayClassID = CVMcbClassName(cb);
	}
    } else {
	arrayClassID = CVMtypeidIncrementArrayDepth(ee, CVMcbClassName(cb),
						    effectiveDimension);
	computedArrayClassID = CVM_TRUE;
    }

    cb = CVMclassLookupByTypeFromClass(ee, arrayClassID, CVM_FALSE, cb);

    if (computedArrayClassID) {
	CVMtypeidDisposeClassID(ee, arrayClassID);
    }
    if (cb == NULL) {
	return NULL;
    }

    /* Okay, now we have the classblock. The only other thing we need
       to be able to use CVMmultiArrayAlloc (interpreter.[ch]) is a
       copy of the dimensions array. */
    dimensionsTmp = 
	(CVMInt32*) malloc(sizeof(CVMInt32) * effectiveDimension);
    for (i = 0; i < effectiveDimension; i++) {
	CVMID_arrayReadInt(ee, arrayObj, i, dimensionsTmp[i]);
    }
    
    res = CVMjniCreateLocalRef(ee);
    if (res != NULL) {
	CVMmultiArrayAlloc(ee, effectiveDimension, dimensionsTmp, cb,
			   (CVMObjectICell*) res);
    }
    free(dimensionsTmp);
    if (CVMexceptionOccurred(ee)) {
	(*env)->DeleteLocalRef(env, res);
	res = NULL;
    }
    return res;
#else /* CVM_REFLECT */
    CVMthrowUnsupportedOperationException(CVMjniEnv2ExecEnv(env), NULL);
    return NULL;
#endif /* CVM_REFLECT */
}

/* NOTE: JVM_GetField is implemented differently in CVM. See
   javavm/native/java/lang/reflect/Field.c. */

/* NOTE: JVM_GetPrimitiveField is implemented differently in CVM. See
   javavm/native/java/lang/reflect/Field.c. */

/* NOTE: JVM_SetField is implemented differently in CVM. See
   javavm/native/java/lang/reflect/Field.c. */

/* NOTE: JVM_SetPrimitiveField is implemented differently in CVM. See
   javavm/native/java/lang/reflect/Field.c. */

/* NOTE: JVM_InvokeMethod is implemented differently in CVM. See
   javavm/native/java/lang/reflect/Method.c. */

/* NOTE: JVM_NewInstanceFromConstructor is implemented differently in 
   CVM. See javavm/native/java/lang/reflect/Constructor.c. */

/*
 * Note that this function expects a full path name.  If the file
 * doesn't exist, then it is going to throw an exception.  In case
 * you are managing LD_LIBRARY_PATH yourself, like
 * ClassLoader.NativeLibrary does, then you had better deal with the
 * exception correctly.
 */
JNIEXPORT void * JNICALL 
JVM_LoadLibrary(const char *name)
{
#ifdef CVM_DYNAMIC_LINKING
    void *handle;

    /* I believe that in JDK 1.2 and above, the existence of the
       library in java.library.path is checked by Java. Therefore
       having an error detail message provided by in CVMdynlinkOpen is
       less useful. In addition, it's a property of dynamic linking on
       Solaris that such a detail message is available (via
       dlerror()); for example, VxWorks has no such facility. */
    handle = CVMdynlinkOpen(name);
    if (handle == NULL) {
	CVMthrowUnsatisfiedLinkError(CVMgetEE(), "%s", name);
    }
    return handle;
#else /* #defined CVM_DYNAMIC_LINKING */
    CVMthrowUnsupportedOperationException(CVMgetEE(), NULL);
    return NULL;
#endif  /* #defined CVM_DYNAMIC_LINKING */
}

JNIEXPORT void JNICALL 
JVM_UnloadLibrary(void * handle)
{
#ifdef CVM_DYNAMIC_LINKING
    CVMdynlinkClose(handle);
#else /* #defined CVM_DYNAMIC_LINKING */
    CVMthrowUnsupportedOperationException(CVMgetEE(), NULL);
#endif  /* #defined CVM_DYNAMIC_LINKING */
}

JNIEXPORT void * JNICALL 
JVM_FindLibraryEntry(void * handle, const char *name)
{
#ifdef CVM_DYNAMIC_LINKING
    return CVMdynlinkSym(handle, name);
#else /* #defined CVM_DYNAMIC_LINKING */
    CVMthrowUnsupportedOperationException(CVMgetEE(), NULL);
    return NULL;
#endif  /* #defined CVM_DYNAMIC_LINKING */
}

JNIEXPORT jboolean JNICALL
JVM_IsNaN(jdouble d)
{
    return CVMdoubleCompare(d, d, -1) != 0;
}

JNIEXPORT jboolean JNICALL
JVM_IsSupportedJNIVersion(jint version)
{
    return version == JNI_VERSION_1_1 || version == JNI_VERSION_1_2;
}

JNIEXPORT void * JNICALL
JVM_RawMonitorCreate(void)
{
    CVMReentrantMutex *rm = calloc(1, sizeof *rm);
    if (rm == NULL) {
	return NULL;
    }
    CVMreentrantMutexInit(rm, NULL, 0);
    return rm;
}

JNIEXPORT void JNICALL
JVM_RawMonitorDestroy(void *mon)
{
    CVMreentrantMutexDestroy(mon);
    free(mon);
}

JNIEXPORT jint JNICALL
JVM_RawMonitorEnter(void *mon)
{
    CVMreentrantMutexLock(CVMgetEE(), (CVMReentrantMutex *)mon);
    return 0;
}

JNIEXPORT void JNICALL
JVM_RawMonitorExit(void *mon)
{
    CVMreentrantMutexUnlock(CVMgetEE(), (CVMReentrantMutex *)mon);
}

JNIEXPORT void * JNICALL
JVM_GetThreadInterruptEvent(void)
{
#ifdef JDK12
    return sysThreadInterruptEvent();
#else
    CVMconsolePrintf("JVM_GetThreadInterruptEvent() not implemented yet!\n");
    CVMassert(CVM_FALSE);
    return NULL;
#endif
}

/* JDK1.2 has this in classresolver.c */

JNIEXPORT jclass JNICALL
JVM_FindLoadedClass(JNIEnv *env, jobject loader, jstring name)
{
    CVMClassBlock* cb;
    CVMClassTypeID classTypeID;
    char slash_name_buf[256];
    char *slash_name;
    int uni_len = (*env)->GetStringLength(env, name);
    int utf_len = (*env)->GetStringUTFLength(env, name);
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);

    if (utf_len >= sizeof(slash_name_buf)) {
        slash_name = malloc(utf_len + 1);
	if (slash_name == NULL) {
	    CVMthrowOutOfMemoryError(ee, NULL);
	    return NULL;
	}
    } else {
        slash_name = slash_name_buf;
    }
    
    (*env)->GetStringUTFRegion(env, name, 0, uni_len, slash_name);

    /* When the passed in name is an empty string */
    if (slash_name[0] == '\0') {
	return CVMjniNewLocalRef(env, NULL);
    }

    VerifyFixClassname(slash_name);

    CVMtraceClassLookup(("LU: JVM_FindLoadedClass called for \"%s\" using "
			  "0x%p class loader\n", slash_name, loader));

    /*
     * We need to use the "new" function, because the "lookup" function
     * is only valid when we know that the type already exists.
     */
    classTypeID = CVMtypeidNewClassID(ee, slash_name, strlen(slash_name));

    if (classTypeID != CVM_TYPEID_ERROR) {
	CVMBool success;
	/*
	 * Any time CVMloaderCacheLookup() is called and we want to make sure
	 * the class we look up is completely done loading, including having
	 * superclasses loaded, then we need to synchronize on the loader
	 * instance first. Otherwise we can get a class that is still in
	 * the process of loading.
	 *
	 * The loader's loadClass() method is the only client of this code,
	 * and is suppose to by synchronized, but we don't trust it to be.
	 */
	success = CVMgcSafeObjectLock(ee, loader);

	if (!success) {
	    CVMthrowOutOfMemoryError(ee, NULL);
	    cb = NULL;
	} else {
	    CVM_LOADERCACHE_LOCK(ee);
	    cb = CVMloaderCacheLookup(ee, classTypeID, loader);
	    CVM_LOADERCACHE_UNLOCK(ee);
	    
	    success = CVMgcSafeObjectUnlock(ee, loader);
	    CVMassert(success);
	}

	CVMtypeidDisposeClassID(ee, classTypeID);
    } else {
	cb = NULL;
    }

    if (slash_name != slash_name_buf) {
        free(slash_name);
    }

    return CVMjniNewLocalRef(env, (cb == NULL ? NULL : CVMcbJavaInstance(cb)));
}


/*
 * The following have been moved to packages.
 *
 * JVM_GetSystemPackage(JNIEnv *env, jstring str)
 * JVM_GetSystemPackages(JNIEnv *env)
 */


/* JDK1.2 has this in javai.c */

/*
 * Short cuts
 */
#define PUTPROP(props, key, val)			\
    if (!CVMputProp(env, putID, props, key, val)) {	\
        return NULL;					\
    }

#define PUTPROP_ForPlatformCString(props, key, val)			\
    if (!CVMputPropForPlatformCString(env, putID, props, key, val)) {	\
        return NULL;							\
    }

/*
 * Initialize VM specific properties.
 * sprops contains system properties that are platform-dependent.
 */
JNIEXPORT jobject JNICALL
JVM_InitProperties(JNIEnv *env, jobject props, java_props_t* sprops)
{
    /*
     * Use the setProperty method to enforce the use of strings as
     * key and value, and also to make sure that we have a properties
     * table at hand and not just a hashtable.
     */
    jmethodID putID = (*env)->GetMethodID(env,
					  (*env)->GetObjectClass(env, props),
					  "setProperty",
            "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/Object;");

    if (putID == NULL) return NULL;

    PUTPROP(props, "java.version", "J2ME " J2ME_PRODUCT_NAME " 1.0");

    PUTPROP(props, "java.vm.specification.version", "1.0");
    PUTPROP(props, "java.vm.specification.name",
	    "Java Virtual Machine Specification");
    PUTPROP(props, "java.vm.specification.vendor", "Sun Microsystems Inc.");

    PUTPROP(props, "java.vm.version", J2ME_VERSION_STRING);
    /* Modified by Pratik - 26th Nov 2001 */
//    PUTPROP(props, "java.vm.name", "CVM");
    PUTPROP(props, "java.vm.name", "TJVM");
    PUTPROP(props, "java.vm.info", "native threads");
    /* Modified by Pratik - 26th NOv 2001 */
//    PUTPROP(props, "java.vm.vendor", "Sun Microsystems Inc.");
    PUTPROP(props, "java.vm.vendor", "TimeSys Corporation");

#ifdef CVM_CLASSLOADING
    /*
     * Initial value of sun.boot.class.path
     */
    PUTPROP_ForPlatformCString(props, "sun.boot.class.path",
			       sprops->sysclasspath);
#endif
    
    /*
     * No compiler for now
     */
    PUTPROP_ForPlatformCString(props, "java.compiler", "NONE");
    PUTPROP_ForPlatformCString(props, "java.home", sprops->java_home);
    PUTPROP_ForPlatformCString(props, "sun.boot.library.path",
			       sprops->dll_dir);
    PUTPROP_ForPlatformCString(props, "java.library.path",
			       sprops->library_path);
    PUTPROP_ForPlatformCString(props, "java.ext.dirs", sprops->ext_dirs);
	
    return (*env)->NewLocalRef(env, props);
}

/*
 * java.lang.Compiler
 */

JNIEXPORT void JNICALL
JVM_InitializeCompiler (JNIEnv *env, jclass compCls)
{
}

JNIEXPORT jboolean JNICALL
JVM_IsSilentCompiler(JNIEnv *env, jclass compCls)
{
    return JNI_TRUE;
}

JNIEXPORT jboolean JNICALL
JVM_CompileClass(JNIEnv *env, jclass compCls, jclass cls)
{
    return JNI_FALSE;
}

JNIEXPORT jboolean JNICALL
JVM_CompileClasses(JNIEnv *env, jclass cls, jstring jname)
{
    return JNI_FALSE;
}

JNIEXPORT jobject JNICALL
JVM_CompilerCommand(JNIEnv *env, jclass compCls, jobject arg)
{
    return NULL;
}

JNIEXPORT void JNICALL
JVM_EnableCompiler(JNIEnv *env, jclass compCls)
{
}

JNIEXPORT void JNICALL
JVM_DisableCompiler(JNIEnv *env, jclass compCls)
{
}

#ifdef JDK13

JNIEXPORT void * JNICALL
JVM_RegisterSignal(jint sig, void *handler)
{
    CVMconsolePrintf("JVM_RegisterSignal() not implemented yet!\n");
    CVMassert(CVM_FALSE);
    return (void *)NULL;
}

JNIEXPORT void JNICALL
JVM_RaiseSignal(jint sig)
{
    CVMconsolePrintf("JVM_RaiseSignal() not implemented yet!\n");
    CVMassert(CVM_FALSE);
}

JNIEXPORT jint JNICALL
JVM_FindSignal(const char *name)
{
    CVMconsolePrintf("JVM_FindSignal() not implemented yet!\n");
    CVMassert(CVM_FALSE);
    return -1;
}

#endif
