/*
 * @(#)quicken.c	1.24 00/11/09
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifdef CVM_CLASSLOADING

#include "javavm/include/interpreter.h"
#include "javavm/include/classes.h"
#include "javavm/include/utils.h"
#include "javavm/include/preloader.h"
#include "javavm/include/common_exceptions.h"
#include "javavm/include/porting/int.h"
#include "generated/javavm/include/opcodes.h"
#include "javavm/include/globals.h"

#ifdef CVM_JVMDI
#include "javavm/include/jvmdi_impl.h"
#endif

/*
 * GET_INDEX - Macro used for getting an unaligned unsigned short from
 * the byte codes.
 */
#undef  GET_INDEX
#define GET_INDEX(ptr) (CVMgetUint16(ptr))

/*
 * CVMopcodeWasQuickened - check to see if the opcode was already quickened.
 * Note: If breakpoints are enabled and we have to deal with
 * two threads trying to quicken at the same time.
 *
 * WARNING:  Must be executed within the code lock.
 *
 *   p_opcode: Pointer to the opcode we are quickening. The current opcode
 *             at *pc is also returned here. Once this is updated, the code
 *             lock is not released, so the opcode remains valid.
 *   pc:       Pointer to the instruction we are quickening
 */
static CVMBool
CVMopcodeWasQuickened(CVMExecEnv* ee, CVMUint8* p_opcode, CVMUint8* pc)
{
#ifdef CVM_JVMDI
    if (*pc == opc_breakpoint) {
	/* If the original opcode is an opc_breakpoint, then we need call
	 * the debugger to get the real opcode and make sure that it
	 * hasn't been quickened already.
	 */
	CVMUint8 breakpointOpcode =
	    CVMjvmdiGetBreakpointOpcode(ee, pc, CVM_FALSE);
	if (*p_opcode != breakpointOpcode ) {
	    /* The opcode we are quickening is not the same as the 
	     * breakpoint opcode so it must have been quickened by
	     * another thread.
	     */
	    *p_opcode = breakpointOpcode;  /* alway return updated opcode */
	    return CVM_TRUE;
	}
    } else if (*pc != *p_opcode) {
	/* It's not a breakpoint opcode and it's not the unquickened
	 * opcode, so it must have been quickened by another thread.
	 */
	*p_opcode = *pc;  /* alway return updated opcode */
	return CVM_TRUE;
    }
    *p_opcode = *pc;  /* alway return updated opcode */
    return CVM_FALSE;
#else
    /* IMPORTANT NOTE: This is not allowed to block, since it is executed
       under the global CODE_LOCK microlock. */
    CVMBool wasQuickened = *pc != *p_opcode;
    *p_opcode = *pc;  /* alway return updated opcode */
    return wasQuickened;
#endif
}

/* 
 * CVMisSpecialSuperCall - decides if the specified mb should be turned
 * into an invokesuper_quick instead of a invokenonvirtual_quick?
 *
 * The four conditions that have to be satisfied:
 *    The ACC_SUPER flag is set in the current class
 *    The method isn't private
 *    The method isn't an <init> method
 *    The method's class is a superclass (and not equal) to the current class.
 */ 
static CVMBool
CVMisSpecialSuperCall(CVMClassBlock* currClass, CVMMethodBlock* mb) {
    if (CVMcbIs(currClass, SUPER)
	  && (!CVMmbIs(mb, PRIVATE))
	  && (!CVMtypeidIsConstructor(CVMmbNameAndTypeID(mb)))
	  && (currClass != CVMsystemClass(java_lang_Object))) {
	CVMClassBlock* methodClass = CVMmbClassBlock(mb);
	CVMClassBlock* cb = CVMcbSuperclass(currClass);
	while (cb != NULL) {
	    if (cb == methodClass) {
		return CVM_TRUE;
	    }
	    cb = CVMcbSuperclass(cb);
	}
    }
    return CVM_FALSE;
}

/*
 * CVMquickenOpcode - quickens all opcodes.
 *
 * quickening:
 *            A pointer to a three-byte location. On entry, the first byte
 *            contains the instruction to be quickened. On exit, quickening[]
 *            contains the new instruction (and/or the operands) to be written
 *            into the instruction stream.
 * pc:        A pointer to the current instruction.
 * cp:        A pointer to the constant pool.
 * p_cb:      A place to store the CVMClassBlock* in case static initializers
 *            still need to be run. The interpreter needs this. Otherwise
 *            it would need to enter the code lock and repeat what
 *            CVMopcodeWasQuickened() has already done in order to get
 *            the cb from the constant pool.
 * clobbersCpIndex: true if quickening the opcode will cause the constant
 *            pool index in the instruction to be overwritten.
 *
 * Possible return values are:
 *
 *    CVM_QUICKEN_SUCCESS_OPCODE_ONLY:
 *        quickening succeeded, only opcode changed.
 *    CVM_QUICKEN_SUCCESS_OPCODE_AND_OPERANDS
 *        quickening succeeded, opcode and operands changed.
 *    CVM_QUICKEN_NEED_TO_RUN_STATIC_INITIALIZERS
 *        need to run static initializers, *p_cb updated. The interpreter
 *        needs to run static initializers first and then call
 *        CVMquickenOpcode() again.
 *    CVM_QUICKEN_ALREADY_QUICKENED
 *        someone else already raced us to it
 *            OR 
 *        in the JVMDI case, the instruction stream is already
 *        updated in CVMquickenOpcode().
 *    CVM_QUICKEN_ERROR
 *        an error occurred, exception pending
 *
 * On exit  quickening[0] contains the new opcode.
 *          quickening[1,2] contain the new operands if they have changed.
 *
 */
CVMQuickenReturnCode
CVMquickenOpcode(CVMExecEnv* ee, CVMUint8* quickening, CVMUint8* pc, 
		 CVMConstantPool* cp, CVMClassBlock** p_cb,
		 CVMBool clobbersCpIndex)
{
    CVMUint8        opcode = quickening[0];
    CVMUint16       cpIndex;
    CVMBool         isStaticOpcode = CVM_FALSE; /* default to false */
    CVMBool         isPutOpcode = CVM_FALSE;    /* default to false */
    CVMClassBlock*  cb = NULL;
    CVMMethodBlock* mb = NULL; /* only setup for some invoke opcodes */
    CVMFieldBlock*  fb = NULL; /* only setup for field opcodes */

    /*
     * Get the constant pool index. Note: If quickening the
     * instruction overwrites the index. If clobbersCpIndex is true
     * then we need to take some extra precautions to make sure we
     * got a valid cpIndex.
     */
    if (opcode == opc_ldc) {
	cpIndex = pc[1];
    } else {
	cpIndex = GET_INDEX(pc + 1);
    }
    if (clobbersCpIndex) {
	/* We can't trust the cpIndex that we fetch from the instruction 
	 * stream unless we also verify that the opcode hasn't been quickened.
	 * If the opcode was already quickened then there is nothing for 
	 * us to do.
	 */
	CVMBool wasQuickened;
	/* Grabbing the code lock ensures that no one is in the middle
	 * of quickening when we check to see if the opcode has been
	 * quickened yet. We want to avoid the race condition where the
	 * index has been overwritten, but the opcode has not been
	 * written yet. By calling CVMopcodeWasQuickened
	 * from within the code lock, we guarantee that if the opcode
	 * wasn't quickened then the cpIndex is good. If the opcode
	 * has been quickened, then we don't care about cpIndex.
	 */
	CVM_CODE_LOCK(ee);
	wasQuickened = CVMopcodeWasQuickened(ee, quickening, pc);
        CVM_CODE_UNLOCK(ee);
	if (wasQuickened) { 
	    /* opcode has already been quickened */
	    return CVM_QUICKEN_ALREADY_QUICKENED;
	}
    }

    if (!CVMcpResolveEntry(ee, cp, cpIndex)) {
	return CVM_QUICKEN_ERROR;
    }

    /*
     * Do any required error checking an handling.
     */
    switch (opcode) {
    case opc_new: {
	cb = CVMcpGetCb(cp, cpIndex);

#ifndef CVM_TRUSTED_CLASSLOADERS	
	/*
	 * Make sure we are not trying to instantiate an iterface or an
	 * abstract class.
	 */
	if (CVMcbIs(cb, INTERFACE) || CVMcbIs(cb, ABSTRACT)) {
	    CVMthrowInstantiationError(ee, "%C", cb);
	    return CVM_QUICKEN_ERROR;
	}
#endif
	/*
	 * If static intializers still need to be run for this class, then
	 * leave it up to the interpreter to do this for us first. 
	 */
	if (!CVMcbCheckRuntimeFlag(cb, INITIALIZED)
	    && CVMcbClinitEE(cb) != ee) {
	    *p_cb = cb; /* return cb to the interpreter. */
	    return CVM_QUICKEN_NEED_TO_RUN_STATIC_INITIALIZERS;
	}
	break;
    }
    case opc_putstatic:
	isStaticOpcode = CVM_TRUE;
	/* fall through */
    case opc_putfield:
	isPutOpcode = CVM_TRUE;
	goto check_field;
    case opc_getstatic:
	isStaticOpcode = CVM_TRUE;
	/* fall through */
    case opc_getfield:
    check_field: {
	    CVMBool isStaticField = CVM_FALSE;
	    CVMBool isFinalField = CVM_FALSE;
	    
	    /* 
	     * cjp08/26/99
	     * For non-static fields, resolved Fieldrefs will always be
	     * FieldBlock type entry. For static fields, resolved Fieldrefs
	     * will be either a FieldBlock entry or a StaticXXX entry,
	     * depending on a few factors.
	     *
	     * If we are doing lossless quickening, then the entry will
	     * always be a FieldBlock.
	     *
	     * If we are doing lossy quickening, then the entry for a static
	     * field will be a FieldBlock if the entry was just resolved by
	     * the above CVMcpResolveEntry() call (either by this thread or
	     * another one we are in a harmless race condition with). In this
	     * case we will further resolve the entry to a StaticXXX entry.
	     * This two stage resolution is needed because the first time we
	     * quicken a getstatic or putstatic, we need the CVMFieldBlock
	     * so we can run static initializers if necessary. After that
	     * the entry can just be a StaticXXX, which contains a direct
	     * pointer to the static data. Note that we don't actually
	     * change the entry to a StaticXXX until a bit later when we
	     * are sure that there are no errors and static initializers
	     * have been run.
	     */
#ifdef CVM_NO_LOSSY_OPCODES
	    /* We always have a FieldBlock in the lossless case */
	    fb = CVMcpGetFb(cp, cpIndex);
#else
	    /* Grab the lock because the FieldBlock entry may be changed
	     * to a StaticXXX entry on us if we don't.
	     */
	    CVMcpLock(ee);
	    if (CVMcpTypeIs(cp, cpIndex, FieldBlock)) {
		fb = CVMcpGetFb(cp, cpIndex);
		CVMcpUnlock(ee);
	    } else {
		CVMcpUnlock(ee);
		CVMassert(CVMcpTypeIs(cp, cpIndex, StaticRef) ||
			  CVMcpTypeIs(cp, cpIndex, Static32) ||
			  CVMcpTypeIs(cp, cpIndex, Static64));
		fb = NULL;
		isStaticField = CVM_TRUE;
		isFinalField  = CVMcpCheckStaticFieldIsFinal(cp, cpIndex);
	    }
#endif /* CVM_NO_LOSSY_OPCODES */
	    
	    /* If we have an fb then we still need to setup isStaticField
	     * and isFinalField.
	     */
	    if (fb != NULL) {
		cb = CVMfbClassBlock(fb);
		isStaticField = CVMfbIs(fb, STATIC);
		/* 
		 * We only consider the field to be final if the
		 * field's class is not the same as the current class.  
		 *
		 * This covers static fields as well as instance fields.
		 */
		if (CVMfbIs(fb, FINAL) &&
		    (cb != CVMmbClassBlock(CVMeeGetCurrentFrame(ee)->mb))) {
		    isFinalField = CVM_TRUE;
		} else {
		    isFinalField = CVM_FALSE;
		}
	    }

#ifndef CVM_TRUSTED_CLASSLOADERS
	    /*
	     * Make sure that both the opcode and the fb agree on whether or
	     * not the field is static.
	     */
	    if (isStaticField != isStaticOpcode) {
		if (fb != NULL) {
		    CVMthrowIncompatibleClassChangeError(
                        ee,
			"%C: field %F %s to be static",
			cb, fb, 
			isStaticOpcode ? "used" : "did not use");
		} else {
		    CVMthrowIncompatibleClassChangeError(
                        ee, "field %s to be static",
			isStaticOpcode ? "used" : "did not use");
		}
		return CVM_QUICKEN_ERROR;
	    }

	    /*
	     * Make sure we aren't trying to store into a final field.
	     * The VM spec says that you can only store into a final field
	     * when initializing it. The JDK seems to interpret this as meaning
	     * that a class has write access to its final fields, so we do
	     * the same here (see how isFinalField gets setup above).
	     */
	    if (isPutOpcode && isFinalField) 
	    { 
		if (fb != NULL) {
		    CVMthrowIllegalAccessError(ee, "%C: field %F is final",
					       cb, fb);
		} else {
		    CVMthrowIllegalAccessError(ee, "field is final");
		}
		return CVM_QUICKEN_ERROR;
	    } 
#endif /* CVM_TRUSTED_CLASSLOADERS */

#ifndef CVM_NO_LOSSY_OPCODES
	    /* If the field is static and hasn't been fully resolved to
	     * a StaticXXX field yet, then fully resolve it, but only
	     * if static initalizers have been run or are being run
	     * by this thread.
	     */
	    if (isStaticField && fb != NULL) {
		/*
		 * If static intializers still need to be run for this class,
		 * then leave it up to the interpreter to do this for us 
		 * first.
		 */
		if (!CVMcbCheckRuntimeFlag(cb, INITIALIZED) 
		    && CVMcbClinitEE(cb) != ee) {
		    *p_cb = cb; /* return cb to the interpreter. */
		    return CVM_QUICKEN_NEED_TO_RUN_STATIC_INITIALIZERS;
		}
		/*
		 * cjp08/26/99
		 * Now that we have passed all checks and have made sure
		 * static initializers have been run, we can
		 * safely resolve the FieldBlock entry to a StaticXXX
		 * entry. This may have already been done if there was a
		 * race condition, in which case we'll do nothing.
		 */
		CVMcpLock(ee);
		if (CVMcpTypeIs(cp, cpIndex, FieldBlock)) {
		    CVMFieldTypeID tid = CVMfbNameAndTypeID(fb);
		    if (CVMtypeidFieldIsRef(tid)) {
			CVMcpSetStaticRef(cp, cpIndex, &CVMfbStaticField(fb));
		    } else if (CVMtypeidFieldIsDoubleword(tid)) {
			CVMcpSetStatic64(cp, cpIndex, &CVMfbStaticField(fb));
		    } else {
			CVMcpSetStatic32(cp, cpIndex, &CVMfbStaticField(fb));
		    }
		    if (isFinalField) {
			CVMcpSetStaticFieldIsFinal(cp, cpIndex);
		    }
		}
		CVMcpUnlock(ee);
	    }
#endif /* CVM_NO_LOSSY_OPCODES */
	    break;
	}
    case opc_invokestatic:
	isStaticOpcode = CVM_TRUE;
	goto check_method;
    case opc_invokevirtual:
    case opc_invokespecial:
    check_method: {
	    mb = CVMcpGetMb(cp, cpIndex);
	    cb = CVMmbClassBlock(mb);
#ifndef CVM_TRUSTED_CLASSLOADERS
	    /*
	     * Make sure that both the opcode and the mb agree on whether or
	     * not the method is static.
	     */
	    if (CVMmbIs(mb, STATIC) != isStaticOpcode) {
		CVMthrowIncompatibleClassChangeError(
                    ee, "%C: method %M %s to be static",
		    cb, mb, isStaticOpcode ? "used" : "did not use");
		return CVM_QUICKEN_ERROR;
	    } 
#endif
	    /*
	     * If static intializers still need to be run for this class, then
	     * leave it up to the interpreter to do this for us first. 
	     */
	    if ((opcode == opc_invokestatic)
		&& !CVMcbCheckRuntimeFlag(cb, INITIALIZED) 
		&& CVMcbClinitEE(cb) != ee) {
		*p_cb = cb; /* return cb to the interpreter. */
		return CVM_QUICKEN_NEED_TO_RUN_STATIC_INITIALIZERS;
	    }
	    break;
	}
    case opc_invokeinterface:
    case opc_anewarray:
    case opc_multianewarray:
    case opc_checkcast:
    case opc_instanceof:
    case opc_ldc:
    case opc_ldc_w:
    case opc_ldc2_w:
	/* nothing to do for these opcodes */
	break;
    default:
	CVMassert(CVM_FALSE);
    }

    /*
     * Get the new instruction bytes for the quickened opcode.
     */
    {
	CVMUint8  newOpcode = opc_nop; /* opcode we will quicken to. */
	CVMUint8  operand1 = 0;  /* 1st instruction operand if needed. */
	CVMUint8  operand2 = 0;  /* 2nd instruction operand if needed. */

	CVMBool   changesOperands = CVM_FALSE; /* true if quickening changes 
						* the operands */

	/*
	 * Assign newOpcode to the opcode we are going to quicken to.
	 * If the 1st or 2nd operand bytes are also modified, set operand1
	 * and operand2 to the proper values and set the changesOperands flag.
	 */
	switch (opcode) {
	case opc_new:
	    newOpcode = opc_new_quick;
	    break;
	case opc_anewarray:
	    newOpcode = opc_anewarray_quick;
	    break;
	case opc_multianewarray:
	    newOpcode = opc_multianewarray_quick;
	    break;
	case opc_checkcast:
	    newOpcode = opc_checkcast_quick;
	    break;
	case opc_instanceof:
	    newOpcode = opc_instanceof_quick;
	    break;
	case opc_ldc:
	    if (CVMcpTypeIs(cp, cpIndex, StringICell)) {
		newOpcode = opc_aldc_ind_quick;
	    } else {
		newOpcode = opc_ldc_quick;
	    }
	    break;
	case opc_ldc_w:
	    if (CVMcpTypeIs(cp, cpIndex, StringICell)) {
		newOpcode = opc_aldc_ind_w_quick;
	    } else {
		newOpcode = opc_ldc_w_quick;
	    }
	    break;
	case opc_ldc2_w:
	    newOpcode = opc_ldc2_w_quick;
	    break;
	case opc_invokeinterface:
	    newOpcode = opc_invokeinterface_quick;
	    break;
	case opc_invokestatic:
	    newOpcode = opc_invokestatic_quick;
	    break;
	case opc_invokevirtual:
	    if (CVMmbIs(mb, PRIVATE)) {
		newOpcode = opc_invokenonvirtual_quick;
	    }
#ifndef CVM_NO_LOSSY_OPCODES
	    else if (CVMmbMethodTableIndex(mb) <= 0xFF) {
		if (CVMmbClassBlock(mb) == CVMsystemClass(java_lang_Object)) {
		    newOpcode = opc_invokevirtualobject_quick;
		} else switch(CVMtypeidGetReturnType(CVMmbNameAndTypeID(mb))) {
		case CVM_TYPEID_VOID:
		    newOpcode = opc_vinvokevirtual_quick;
		    break;
		case CVM_TYPEID_OBJ:
		    newOpcode = opc_ainvokevirtual_quick;
		    break;
		case CVM_TYPEID_LONG:
		case CVM_TYPEID_DOUBLE:
		    newOpcode = opc_dinvokevirtual_quick;
		    break;
		default:
		    newOpcode = opc_invokevirtual_quick;
		    break;
		}
		operand1 = CVMmbMethodTableIndex(mb);
		operand2 = CVMmbArgsSize(mb);
		changesOperands = CVM_TRUE;
	    }
#endif
	    else {
		newOpcode = opc_invokevirtual_quick_w;
	    }
	    break;
	case opc_invokespecial: {
	    CVMMethodBlock* new_mb = mb;
	    CVMClassBlock* currClass = 
		CVMmbClassBlock(CVMeeGetCurrentFrame(ee)->mb);
	    if (CVMisSpecialSuperCall(currClass, mb)) {
		new_mb = CVMcbMethodTableSlot(CVMcbSuperclass(currClass),
					      CVMmbMethodTableIndex(mb));
	    }
	    if (mb == new_mb) {
		newOpcode = opc_invokenonvirtual_quick;
	    } else {
		newOpcode = opc_invokesuper_quick;
		operand1 = CVMmbMethodTableIndex(mb) >> 8;
		operand2 = CVMmbMethodTableIndex(mb) && 0xFF;
		changesOperands = CVM_TRUE;
	    }
	    break;
	}

	case opc_putstatic:
	case opc_putfield:
	case opc_getstatic:
	case opc_getfield: {
	    if (isStaticOpcode) {	    
#ifdef CVM_NO_LOSSY_OPCODES
		if (CVMfbIsRef(fb)) {
		    newOpcode = (isPutOpcode ? opc_aputstatic_checkinit_quick
				             : opc_agetstatic_checkinit_quick);
		} else if (CVMfbIsDoubleWord(fb)) {
		    newOpcode = (isPutOpcode ? opc_putstatic2_checkinit_quick
				             : opc_getstatic2_checkinit_quick);
		} else {
		    newOpcode = (isPutOpcode ? opc_putstatic_checkinit_quick
				             : opc_getstatic_checkinit_quick);
		}
#else
		if (CVMcpTypeIs(cp, cpIndex, StaticRef)) {
		    newOpcode = (isPutOpcode ? opc_aputstatic_quick
				             : opc_agetstatic_quick);
		} else if (CVMcpTypeIs(cp, cpIndex, Static64)) {
		    newOpcode = (isPutOpcode ? opc_putstatic2_quick
				             : opc_getstatic2_quick);
		} else {
		    CVMassert(CVMcpTypeIs(cp, cpIndex, Static32));
		    newOpcode = (isPutOpcode ? opc_putstatic_quick
				             : opc_getstatic_quick);
		}
#endif /* CVM_NO_LOSSY_OPCODES */
	    } else {
#ifndef CVM_NO_LOSSY_OPCODES
		CVMUint16 offset = CVMfbOffset(fb);
		if (offset <= 0xFF) {
		    if (CVMfbIsDoubleWord(fb)) {
			newOpcode = (isPutOpcode ? opc_putfield2_quick
				                 : opc_getfield2_quick);
		    } else if (CVMfbIsRef(fb)) { 
			newOpcode = (isPutOpcode ? opc_aputfield_quick
				                 : opc_agetfield_quick);
		    } else {
			newOpcode = (isPutOpcode ? opc_putfield_quick
				                 : opc_getfield_quick);
		    }
		    operand1 = offset;
		    operand2 = 0; /* not really needed */
		    changesOperands = CVM_TRUE;
		} else
#endif /* CVM_NO_LOSSY_OPCODES */
		{
		    newOpcode = (isPutOpcode ? opc_putfield_quick_w 
				             : opc_getfield_quick_w);
		}
	    }
	}
	break;
	
	default:
	    CVMassert(CVM_FALSE);
	    break;
	}

#if 0	
	CVMdebugPrintf(("\t%squickening %s to %s\n", 
			(changesOperands ? "+" : "-"),
			CVMopnames[opcode], CVMopnames[newOpcode]));
#endif
	
	/*
	 * Write out the new instruction. 
	 */
	quickening[0] = newOpcode;	/* newOpcode is known by now */ 
#ifdef CVM_JVMDI
	/* 
	 * If the initial opcode was opc_breakpoint, then notify
	 * the debugger of the new opcode for the breakpoint. Otherwise
	 * just write out the new opcode.
	 */
	if (*pc == opc_breakpoint) {
	    if (changesOperands) {
		CVM_CODE_LOCK(ee);
		CVM_WRITE_CODE_BYTE(pc, 1, operand1);
		CVM_WRITE_CODE_BYTE(pc, 2, operand2);
		/*
		 * The JDK doesn't do this part inside the lock, but you
		 * have to. If another thread interrupts us at this point
		 * and calls get_breakpoint_opcode, it will get the non-quick
		 * opcode. This is ok if the code lock is already grabbed,
		 * since the thread will end up blocking when it tries to
		 * grab the code lock so it can quicken the instruction.
		 */
		CVMjvmdiSetBreakpointOpcode(ee, pc, newOpcode);
		CVM_CODE_UNLOCK(ee);
	    } else {
		CVMjvmdiSetBreakpointOpcode(ee, pc, newOpcode);
	    }
	} else {
	    /* If we have to modify the instruction operands, then
	     * the instruction must be quickened within the code
	     * lock. Also, we must always modify the operands
	     * before the opcode so the interpreter will never
	     * fetch a quick opcode before the operands are
	     * written. */
	    if (changesOperands) {
		CVM_CODE_LOCK(ee);
		CVM_WRITE_CODE_BYTE(pc, 1, operand1);
		CVM_WRITE_CODE_BYTE(pc, 2, operand2);
		CVM_WRITE_CODE_BYTE(pc, 0, newOpcode);
		CVM_CODE_UNLOCK(ee);
	    } else {
		CVM_WRITE_CODE_BYTE(pc, 0, newOpcode);
	    }
	}
	return CVM_QUICKEN_ALREADY_QUICKENED;
#else
	if (changesOperands) {
	    quickening[1] = operand1;
	    quickening[2] = operand2;
	    return CVM_QUICKEN_SUCCESS_OPCODE_AND_OPERANDS;
	} else {
	    return CVM_QUICKEN_SUCCESS_OPCODE_ONLY;
	}
#endif

    }
}

#endif /* CVM_CLASSLOADING */
