/*
 * @(#)CConstants.java	1.2 00/05/19
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

package runtime;
import Const;
import components.*;
import vm.*;
import util.*;
import java.util.Enumeration;
import DataFormatException;

/*
 * Class to help CWriter print out the
 * in-core data structures for a module
 * for C-based systems.
 *
 * This class manipulates and writes constants,
 * especially Strings.
 */

class CConstants {

    private CCodeWriter out;
    private CWriter    writer;
    private boolean formatError = false;

    private String stringTableName;
    private String stringTableSub;

    public CStringTable stringTable = new CStringTable();
    public AsciiTable       aTable      = new AsciiTable();

    private boolean         pureTypeTables;

    CWriter.Alias          typeTableAlias;
    CWriter.Alias          constantpoolAlias;

    CConstants( CCodeWriter o, CWriter x, boolean pureTypeTables ){
	out = o;
	writer   = x;
	this.pureTypeTables = pureTypeTables;
	typeTableAlias = writer.new Alias("ROMPureTypeTable", 
					  "ROMImpureTypeTable");
	constantpoolAlias = writer.constantpoolAlias;
	stringTableName = writer.stringTableAlias.name();
	stringTableSub = "&"+stringTableName;
    }

    final void writeUTFPointer( int offset ) {
	out.print( stringTableSub );
	if (writer.stringTableAlias.hasMaxSize()) {
	    int section = offset/writer.stringTableAlias.getMaxEntries();
	    out.print(section);
	    offset = offset%writer.stringTableAlias.getMaxEntries();
	}
	out.write('[');
	out.printHexInt( offset );
	out.write(']');
    }
	
    final void writeUTFPointer( UnicodeConstant a ){
	out.print( stringTableSub );
	if ( a.stringTableOffset < 0 ){
	    // this must never happen.
	    System.err.println(Localizer.getString("cconstants.reference_to_dead_string", a));
	}
	int offset;
	if (writer.stringTableAlias.hasMaxSize()) {
	    int section = a.stringTableOffset/writer.stringTableAlias.getMaxEntries();
	    out.print(section);
	    offset = a.stringTableOffset%writer.stringTableAlias.getMaxEntries();
	} else	
	    offset = a.stringTableOffset;
	
	out.write('[');
	out.printHexInt( offset );
	out.write(']');
    }

    private final static int maxAscizOpLength = 128; // arbitrary.
    
  
    static private byte[] stringCommentBuffer = new byte[20];

    static public void stringComment(String string, CCodeWriter out) { 
        stringComment(string, -1, out);
    } 

    synchronized
    static public void stringComment(String string, int offset, CCodeWriter out) { 
        byte buffer[] = stringCommentBuffer;
	int bufferLength = buffer.length;
        int length = string.length();
	if (offset < 0) { 
	    out.print("\t/* \"");
	} else { 
	    out.print("\t/* (");
	    out.printHexInt(offset);
	    out.print(") \"");
	}
	if (length <= bufferLength) { 
	    string.getBytes(0, length, buffer, 0);
	    sanitizeOutput(buffer, 0, length, out);
	} else { 
	    string.getBytes(0, bufferLength - 3, buffer, 0);
	    sanitizeOutput(buffer, 0, bufferLength - 3, out);
	    out.print("...");
	}
	out.print("\" */");
    }


    /*
     * escape nasty characters in string destined for a C string.
     * This hardly ever happens.
     */
    static private void 
    sanitizeOutput( byte data[], int begin, int l, CCodeWriter out){
	int i;
	int c;
	int run;
	int max = begin+l;
	for ( i = begin, run=begin; i < max; i++ ){
	    c = data[i];
	    if ( ( c >= (byte)' ' ) && ( c <= (byte)'~' )
		 && ( c != (byte)'"') && (c  != (byte)'\\' )  
		 && ( c != (byte)'/')) {
		    // the only dangers between space and ~ are " and \  
	            // We must also be careful about writing */ by accident.
		    continue;
	    } else {
		if ( run < i ){
		    // there is a non-zero "run" of normal characters.
		    // Write them out.
		    out.write( data, run, i-run );
		}
		run = i+1;
		// now deal with the funky one.
		switch ( c ){
		case '\\':
		    out.write('\\');
		    out.write('\\');
		    break;
		case '\n':
		    out.write('\\');
		    out.write('n');
		    break;
		case '\r':
		    out.write('\\');
		    out.write('r');
		    break;
		case '\t':
		    out.write('\\');
		    out.write('t');
		    break;
		case '"':
		    out.write('\\');
		    out.write('"');
		    break;
		case '/':
		    // we only have a problem if the previous char was an
		    // asterisk.  It looks like we're ending the comment.
		    if (i > begin && data[i-1] == (byte)'*')
		        out.write('\\');
		    out.write('/');
		    break;
		default:
		    out.write('\\');
		    out.write('?');
		    break;
		  /*		  
		    c &= 0xff; // trim to normal byte bounds.
		    out.print( ( c < 010 ) ? "\\00" : ( c < 0100 ) ? "\\0" : "\\" );
		    out.print( Integer.toOctalString( c ) );
		    break;
		  */
		}
	    }
	}
	if ( run < max ){
	    // there is a non-zero "run" of normal characters at end.
	    // Write them out.
	    out.write( data, run, max-run );
	}
    }

    final void recordUTFStrings( ConstantObject c[] ){
	ConstantObject obj;
	int n = c.length;
	for( int i = 1; i < n ; i+=obj.nSlots ){
	    obj = c[i];
	    if ( (obj instanceof UnicodeConstant)  ){
		UnicodeConstant u = (UnicodeConstant) obj;
		if ( u.references > 0 ){
		    aTable.addString( u );
		} else {
		    u.stringTableOffset = -1;
		}
	    }
	}
    }

    final void writeUTFStrings(CCodeWriter stringWriter){
	aTable.computeAsciiTable();
	CWriter.Alias stringTableAlias = writer.stringTableAlias;
	boolean hasMaxSize = stringTableAlias.hasMaxSize();
	writer.declare(stringWriter, "const char " + 
		       stringTableAlias.name() + (hasMaxSize ? "0" : "") + 
		       "[]");
	Enumeration strings = aTable.getTable().elements();
	byte[] strng = new byte[maxAscizOpLength];
	int off = 0;
	while ( strings.hasMoreElements() ){
	    UnicodeConstant u = (UnicodeConstant)strings.nextElement();
	    String s = u.UTFstring;
	    int sLength = s.length();
	    // Don't write strings which are suffixes of other strings.
	    if (u.isSuffix) {
		u.stringTableOffset += off;
	    } else {
		if ((hasMaxSize) && 
		    ((stringTableAlias.count + sLength) >= stringTableAlias.getMaxEntries())) {
		    int padding = stringTableAlias.getMaxEntries() - stringTableAlias.count;
		    off += padding;
		    while (--padding >= 0)
			stringWriter.print(" PAD,");
		    stringTableAlias.count = 0;
		    stringTableAlias.tables++;
		    stringWriter.println("\n};\n\n");
		    writer.declare(stringWriter, "const char " + 
				   stringTableAlias.name() + 
				   stringTableAlias.tables + "[]");
		}
		u.stringTableOffset += off;
		stringComment(s, u.stringTableOffset, stringWriter);
		for (int i = 0; i < sLength; i++) { 
		    if (i % 64 == 0) stringWriter.print("\n\t");
		    byte chars[] = CCodeWriter.charString[s.charAt(i)];
		    stringWriter.write(chars, 0, chars.length);
		    stringWriter.print(", ");
		}
		stringTableAlias.count += s.length()+1;
		stringWriter.println("0,");
	    }
	}
	stringWriter.println("};\n\n");
    }

    void writeConstants(ClassClass[] classes) 
                               throws DataFormatException { 
	writeConstants(classes, null);
    }

    void writeConstants(ClassClass[] classes, ConstantPool cp) 
                               throws DataFormatException { 
	internAllStrings(classes);
	stringTable.writeStrings(out, writer);
	if (cp == null) {
	    writeAllTypeTables(classes);
	    writeAllConstantPools(classes);
	} else {
	    writeSharedTypeTable(cp);
	    writeSharedConstantPool(classes, cp);
	}
    }

    private void internAllStrings(ClassClass[] classes) {
        int n = classes.length;
        for ( int i = 0; i < n; i++ ) {
	    ConstantObject cp[] = classes[i].ci.constants;
	    int num = (cp == null ? 0 : cp.length);
	    for(int j = 1; j < num ; ) {
	        ConstantObject obj = cp[j];
		if ( obj instanceof StringConstant ){
		    stringTable.intern( (StringConstant) obj );
		}
		j += obj.nSlots;
	    }
	}
    }

    private void writeAllTypeTables(ClassClass classes[]) 
                   throws DataFormatException { 
	CCodeWriter out = this.out;
        int n = classes.length;
	for (int purity = pureTypeTables ? 0 : 1; purity <= 1; purity++) { 
	    boolean isPure = (purity == 0);
	    int offset = 0;
	    for ( int i = 0; i < n; i++ ) {
		ClassClass c = classes[i];
		ClassInfo ci = c.ci;
		ConstantObject cp[] = ci.constants;
		if (cp == null || cp.length == 0)
		    continue;
		if (c.impureConstants == isPure) 
		    continue;
		if (offset == 0) { 
		    /* This is the first item in the table, create table */
		    String tableName = typeTableAlias.name(isPure);
		    out.println("static " + (isPure ? "const" : "") + 
				" char " + tableName + " [] = {");
		}
		typeTableAlias.define(ci, offset, isPure);
		out.comment("Class " + ci.className + " (" + offset + ")");
		out.print("\t  0,");
		int size = cp.length;
		int mod = 1;
		for (int j = 1; j < size ; j++ ){
		    if (mod == 0) out.print("\t");
		    int v = (cp[j-1] instanceof DoubleValueConstant) ? 0 :
		            (cp[j].tag+(cp[j].isResolved()?0x80:0));
		    out.write( out.byteString[v], 0, 3 );
		    out.write(',');
		    if (++mod == 16) { 
		        out.write('\n'); mod = 0;
		    }
		}
		if (mod != 0) out.write('\n');
		offset += size;
	    }
	    // If offset == 0, then there is no array.
	    if (offset > 0) { 
	        out.println("};\n\n");
	    }
	}
    }

    private void writeSharedTypeTable(ConstantPool scp)
                   throws DataFormatException {
        CCodeWriter out = this.out;
        ConstantObject cp[] = scp.getConstants();
        if (cp == null || cp.length == 0)
            return;
        if (scp.impureConstants == false)
            return;
        /* This is the first item in the table, create table */
        String tableName = typeTableAlias.name(!scp.impureConstants);
        out.println("static char " + tableName + " [] = {");
        typeTableAlias.define(scp, 0, !scp.impureConstants);
        out.print("\t  0,");
        int size = cp.length;
        int mod = 1;
        for (int j = 1; j < size ; j++ ){
            if (mod == 0) out.print("\t");
            int v = (cp[j-1] instanceof DoubleValueConstant) ? 0 :
                    (cp[j].tag+(cp[j].isResolved()?0x80:0));
            out.write( out.byteString[v], 0, 3 );
            out.write(',');
            if (++mod == 16) {
                out.write('\n'); mod = 0;
            }
        }    
        out.println("};\n\n");
    }    

    private void writeAllConstantPools(ClassClass classes[]) 
                   throws DataFormatException { 
	CCodeWriter out = this.out;
        int n = classes.length;
	boolean hasImpureConstants = false;
	for (int purity = 0; purity <= 1; purity++) { 
	    boolean isPure = (purity == 0);
	    if (!isPure && !hasImpureConstants) {
	        // don't bother going through the loop the second time, if
	        // there was no impure constants the first time through
	        continue;
	    }
	    String tableName = constantpoolAlias.name(isPure);
	    int offset = 0;
	    out.println("static " + 
			(isPure ? "const void* const " : "const void* ") + 
			tableName + " [] = {");
	    for ( int i = 0; i < n; i++ ) {
		ClassClass c = classes[i];
		ClassInfo ci = c.ci;
		ConstantObject cp[] = ci.constants;
		if (cp == null || cp.length == 0)
		    continue;
		if (c.impureConstants == isPure) { 
		    if (isPure) 
		        hasImpureConstants = true;
		    continue;
		}
		out.comment("Class " + ci.className + " (" + offset + ")");
		constantpoolAlias.define(ci, offset, isPure);
		out.write('\t');
		if (isPure && !pureTypeTables) { 
		    out.write('0');
		} else {
		    typeTableAlias.write(ci, out);
		} 
		out.println(",");
		writeConstantArray(out, cp);
		offset += cp.length;
	    }
	    out.println(" };\n\n");
	    /***
	    if (isPure) { 
	        out.println("is" + tableName +"(const void **const cp) {");
		out.println("    return " + tableName + " <= cp && cp < &" + 
			    tableName + "[" + offset + "];");
		out.println("}\n\n");
	    }
	    ***/
	}
    }

    private void writeSharedConstantPool(ClassClass classes[], ConstantPool scp)
                   throws DataFormatException {
        CCodeWriter out = this.out;

	boolean isPure = !scp.impureConstants;
	String tableName = constantpoolAlias.name(isPure);
	out.println("static " +
                        (isPure ? "const void* const " : "const void* ") +
                        tableName + " [] = {");

	// Write the constant object from shared constant pool
	constantpoolAlias.define(scp, 0, isPure);
	out.write('\t');
	if (isPure) {
	    out.write('0');
	} else {
	    typeTableAlias.write(scp, out);
	}
        out.println(",");
	
	ConstantObject cp[] = scp.getConstants();
        int offset = cp.length;
	writeConstantArray(out, cp);

	if (!isPure) {
	    offset = 0;
	    tableName = constantpoolAlias.name(true);
	    out.println(" };\n\n");
	    out.println("static "+ "const void* const "+ tableName + " [] = {");
	}

	// If ClassClass is an array class, write the constant object from
	// private constant pool.
        int n = classes.length;
        for ( int i = 0; i < n; i++ ) {
            ClassClass c = classes[i];
 	    if (!c.isArrayClass()) 
	        continue;
            ClassInfo ci = c.ci;
            cp = ci.constants;
            if (cp == null || cp.length == 0)
                continue;
            out.comment("Class " + ci.className + " (" + offset + ")");
            constantpoolAlias.define(ci, offset, true);
            out.write('\t');
            out.write('0');
            out.println(",");
            writeConstantArray(out, cp);
            offset += cp.length;
        }
        out.println(" };\n\n");
    }    

    private void writeConstantArray(CCodeWriter out, ConstantObject[] cp) {

	int size = cp.length;
	for (int j = 1; j < size ; j++ ) {
            out.print('\t');
	    ConstantObject obj = cp[j];
	
            switch ( obj.tag ){
                case Const.CONSTANT_LONG:
                case Const.CONSTANT_DOUBLE:
                    DoubleValueConstant dvc = (DoubleValueConstant) obj;

		    long bignum = ((long)(dvc.highVal) << 32) +
                              (((long)(dvc.lowVal)) << 32 >>> 32);
                    out.print(obj.tag == Const.CONSTANT_DOUBLE
                              ? "DOUBLE((void *)" : "LONG((void *)");
                    out.printHexInt(dvc.highVal);
                    out.print(",(void *)");
                    out.printHexInt(dvc.lowVal);
                    out.print(")\t/* ");
                    if (obj.tag == Const.CONSTANT_DOUBLE) {
                        out.print(Double.longBitsToDouble(bignum));
                    } else {
                        out.print(bignum + "= 0x" +
                               Long.toHexString(bignum));
                    }
                    out.print(" */");
		    j++;
                    break;

                case Const.CONSTANT_INTEGER:
                case Const.CONSTANT_FLOAT:
                    out.print("(void *)");
                    out.printHexInt( ((SingleValueConstant)obj).value );
                    break;
                case Const.CONSTANT_STRING: {
                    StringConstant sobj = (StringConstant)obj;
                    writer.stringAlias.write(sobj, out);
                    out.print(".handle");
                    stringComment(sobj.str.string, out);
                    break;
                }

                case Const.CONSTANT_FIELD:
                    FieldConstant fc = (FieldConstant) obj;
                    if ( obj.isResolved() ){
                        writer.printFieldname(fc);
                    } else {
                        int ID = (fc.clas.index<<16) + (fc.sig.index);
                        out.print("(void *)");
                        out.printHexInt(ID);
                    }
                    out.print("\t/* ");
                    out.print(fc.clas.name.string);
                    out.write('.');
                    out.print(fc.sig.name.string);
                    out.print(" */");
                    break;
 
                case Const.CONSTANT_METHOD:
                case Const.CONSTANT_INTERFACEMETHOD:
                    MethodConstant mc = (MethodConstant) obj;
                    if ( obj.isResolved() ){
                        writer.printMethodname(mc);
                    } else {
                        int ID = (mc.clas.index<<16) + (mc.sig.index);
                        out.print("(void *)");
                        out.printHexInt(ID);
                    }
                    out.print("\t/* ");
                    out.print(mc.clas.name.string);
                    out.write('.');
                    out.print(mc.sig.name.string);
                    out.print(" */");
                    break;

               case Const.CONSTANT_CLASS:
                    ClassConstant cc = (ClassConstant)obj;
                    if ( obj.isResolved() ){
                        ClassInfo cci =  cc.find();
                        writer.classAlias.write(cci, out);
                    } else {
                        out.print("(void *)" + cc.name.index );
                    }
                    out.print("\t/* ");
                    out.print(cc.name);
                    out.print(" */");
                    break;
 
               case Const.CONSTANT_NAMEANDTYPE:
                    NameAndTypeConstant ntc = (NameAndTypeConstant)obj;
                    out.print("(void *)");
                    if ( obj.isResolved() ) {
                        UnicodeConstant name = ntc.name;
                        UnicodeConstant type = ntc.type;
                        int ID = writer.altSigHash.getID(name, type);
                        out.printHexInt(ID);
                        out.print("\t/* ");
                        out.print(name.string+ " : " +  type.string);
                        out.print(" */");
                    } else {
                        out.printHexInt((ntc.name.index<<16)
                                      + (ntc.type.index) );
                    }
                    break;
                case Const.CONSTANT_UTF8:
                // its a UnicodeConstant
                    writeUTFPointer( (UnicodeConstant) obj );
                    stringComment(((UnicodeConstant)obj).string, out);
                    break;
                default:
                    out.write('\n');
                    out.write('X');// FIXME.
                    formatError = true;
                    out.write('\n');
            }
            out.write(',');
            out.write('\n');
	}
	writer.nconstants += size;
    }  

    private final void countUTFReferences( ClassMemberInfo a[] ){
	int n = a.length;
	for ( int i = 0; i < n; i++ ){
	    a[i].type.incReference();
	    a[i].name.incReference();
	}
    }

    private final void countUTFReferences( ClassInfo c ){
	c.thisClass.name.incReference(); // name of the class itself
	if ( c.superClass != null )
	    c.superClass.name.incReference(); // superfluous
	if ( c.fields != null ) countUTFReferences( c.fields );
	if ( c.methods != null ) countUTFReferences( c.methods );
	    
    }

    final void countUTFReferences( ConstantPool cp, ClassInfo classes[] ){
	/*
	 * We know which UTF strings are actually needed
	 * for running the program. Or, at least, we can compute
	 * by looking at all the data structures.
	 */
	cp.clearAllReferences();
	for( int i = 0; i < classes.length; i++ ){
	    countUTFReferences( classes[i] );
	}
    }

    public void writeStrings(){
    }

    public boolean checkError(){ return formatError; }

}
