/*
 * @(#)generational.h	1.20 00/05/19
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*
 * This file is specific and only visible to generational GC
 * implementation.  
 */

#ifndef _INCLUDED_GENERATIONAL_H
#define _INCLUDED_GENERATIONAL_H

/*
 * These are the definitions of generations and spaces. Each
 * generation, when collected, gives you a pointer to the start of a
 * contiguous allocation area.  
 */
typedef struct CVMGenSpace {
    CVMUint32* allocPtr;   /* Current contiguous allocation pointer */
    CVMUint32* allocBase;  /* The bottom of the allocation area */
    CVMUint32* allocTop;   /* The top of the allocation area */
} CVMGenSpace;

typedef struct CVMGeneration {
    CVMUint32* allocPtr;   /* Current contiguous allocation pointer */
    CVMUint32* allocBase;  /* The bottom of the allocation area */
    CVMUint32* allocTop;   /* The top of the allocation area */
    CVMUint32* allocMark;  /* An allocation pointer mark (for promotion) */
    CVMUint32  generationNo; /* The number of this generation */
    struct CVMGeneration* nextGen; /* The next generation after this */
    struct CVMGeneration* prevGen; /* The previous generation before this */

    /* Collect this generation. TRUE if successful, FALSE otherwise */
    CVMBool    (*collect)(struct CVMGeneration* gen,
			  CVMExecEnv* ee,
			  CVMUint32 numBytes,       /* Goal */
			  CVMGCOptions* gcOpts);  

    /* Iterate over pointers to a younger generation. Frequently used
       during young-space collections, and so should be fast (using
       write barrier information, for example) */
    void       (*scanOlderToYoungerPointers)(struct CVMGeneration* gen,
					     CVMExecEnv* ee,
					     CVMGCOptions* gcOpts,
					     CVMRefCallbackFunc callback,
					     void* callbackData);

    /* Iterate over pointers to an older generation. Used only
       during a full collection, and so does not really need to be fast. */
    void       (*scanYoungerToOlderPointers)(struct CVMGeneration* gen,
					     CVMExecEnv* ee,
					     CVMGCOptions* gcOpts,
					     CVMRefCallbackFunc callback,
					     void* callbackData);

    /* Promote 'objectToPromote' into the current generation. Returns
       the new address of the object, or NULL if the promotion failed. */
    CVMObject* (*promoteInto)(struct CVMGeneration* gen,
			      CVMObject* objectToPromote,
			      CVMUint32  objectSize);

    /* Iterate over promoted pointers in a generation. A young space
       collection accummulates promoted objects in an older
       generation. These objects need to be scanned for pointers into
       the young generation */
    void       (*scanPromotedPointers)(struct CVMGeneration* gen,
				       CVMExecEnv* ee,
				       CVMGCOptions* gcOpts,
				       CVMRefCallbackFunc callback,
				       void* callbackData);

    /* Indicate whether a pointer is from this generation */
    CVMBool    (*inGeneration)(struct CVMGeneration* gen,
			       CVMObject* ref);

    /*
     * Get extra, unused space for use by another generation.
     */
    CVMGenSpace* (*getExtraSpace)(struct CVMGeneration* gen);
				  
    /* Indicate to the generation that a GC is about to start, or has
       ended, so it can initialize and destroy its data structures,
       especially those used for inter-generational
       "communication". Information about the kind of GC is encoded in
       the CVMGCOptions passed in. */
    void       (*startGC)(struct CVMGeneration* gen,
			  CVMExecEnv* ee, CVMGCOptions* gcOpts);
    void       (*endGC)(struct CVMGeneration* gen,
			CVMExecEnv* ee, CVMGCOptions* gcOpts);

    /* Get total amount of usable space in this generation */
    CVMUint32  (*totalMemory)(struct CVMGeneration* gen, CVMExecEnv* ee);

    /* Get free amount of usable space in this generation */
    CVMUint32  (*freeMemory)(struct CVMGeneration* gen, CVMExecEnv* ee);
} CVMGeneration;

/*
 * Some useful generational GC service routines. Particular generations
 * use these either as convenience, or for "communicating" with other
 * generations, say for inter-generational pointer scanning, or object
 * promotion.
 */

/*
 * Scan the root set of collection, as well as all pointers from other
 * generations
 */
void
CVMgenScanAllRoots(CVMGeneration* thisGen,
		   CVMExecEnv *ee, CVMGCOptions* gcOpts,
		   CVMRefCallbackFunc callback, void* data);

/*
 * Update the object headers table for objects in range [startRange, endRange)
 */
void
CVMgenBarrierObjectHeadersUpdate(CVMGeneration* gen, CVMExecEnv* ee,
				 CVMGCOptions* gcOpts,
				 CVMUint32* startRange,
				 CVMUint32* endRange);

/*
 * A contiguous space allocator.
 * Parameters: CVMGeneration* gen, CVMUint32 numBytes
 */
#define CVMgenContiguousSpaceAllocate(gen, numBytes,    \
				      ret)              \
{							\
    CVMUint32* allocPtr = gen->allocPtr;		\
    CVMUint32* allocTop = gen->allocTop;		\
    CVMUint32* allocNext = allocPtr + numBytes / 4;	\
    if (allocNext <= allocTop) {			\
	gen->allocPtr = allocNext;			\
	ret = (CVMObject*)allocPtr;			\
    } else {						\
	ret = NULL;					\
    }							\
}

/*
 * Constants.
 *
 * An object is to be promoted if it survives CVM_GEN_PROMOTION_THRESHOLD
 * no of GC's.
 */
#define CVM_GEN_PROMOTION_THRESHOLD 2

/*
 * Write barrier
 */

/*
 * Card table definitions
 */
extern CVMUint8* cardTable;
extern CVMUint8* cardTableVirtualBase;
extern CVMUint32 cardTableSize;
/*
 * The object header table to indicate the nearest object header to
 * a card. Same size as the card table.
 */
extern CVMInt8* objectHeaderTable;

/*
 * 0xff terminated list of object offsets
 */
typedef union CVMGenSummaryTableEntry {
    CVMUint8  offsets[4];
    CVMUint32 intVersion; /* For zeroing quickly */
} CVMGenSummaryTableEntry;

extern CVMGenSummaryTableEntry* summaryTable;

/*
 * The beginning of the heap space, which the card table maps to.
 */
extern CVMUint32* heapBase;

#define NUM_BYTES_PER_CARD 512
#define NUM_WORDS_PER_CARD (NUM_BYTES_PER_CARD / 4)

/* Find the card corresponding to heap location 'ptr' */
#define CARD_TABLE_SLOT_ADDRESS_FOR(ptr) \
    (&cardTableVirtualBase[(CVMUint32)(ptr) / NUM_BYTES_PER_CARD])

/* Round up to card size multiples. */
#define ROUND_UP_TO_CARD_SIZE(obj) \
    ((CVMUint32*)((((CVMUint32)(obj)) + (NUM_BYTES_PER_CARD-1)) & ~(NUM_BYTES_PER_CARD-1)))

/* Find the card boundary for object 'obj' */
#define CARD_BOUNDARY_FOR(obj) \
    ((CVMUint32*)(((CVMUint32)(obj)) & ~(NUM_BYTES_PER_CARD-1)))

/* Find the next card boundary for object 'obj' */
#define NEXT_CARD_BOUNDARY_FOR(obj) \
    (CARD_BOUNDARY_FOR(obj) + NUM_WORDS_PER_CARD)

/* Find the object header table entry corresponding to a card */
#define HEADER_TABLE_SLOT_ADDRESS_FOR_CARD(crd) \
    ((crd) - cardTable + (CVMInt8*)objectHeaderTable)

/* Find the summary table entry corresponding to a card */
#define SUMMARY_TABLE_SLOT_ADDRESS_FOR_CARD(crd) \
    (&summaryTable[(CVMUint32)((crd) - cardTable)])

/* Find the heap address corresponding to a card */
#define HEAP_ADDRESS_FOR_CARD(crd) \
    (((crd) - cardTable) * NUM_WORDS_PER_CARD + heapBase)

/* A card is dirty */
#define CARD_DIRTY_BYTE 0

/* A card is clean */
#define CARD_CLEAN_BYTE 1

/* A card is summarized */
#define CARD_SUMMARIZED_BYTE 2

/* For faster checks, an integer that makes up four clean cards */
#define FOUR_CLEAN_CARDS \
    (((CVMUint32)CARD_CLEAN_BYTE << 24) |    \
     ((CVMUint32)CARD_CLEAN_BYTE << 16) |    \
     ((CVMUint32)CARD_CLEAN_BYTE <<  8) |    \
     ((CVMUint32)CARD_CLEAN_BYTE <<  0))


extern void
CVMgenClearBarrierTable();

/*
 * Traverse recorded pointers
 */
extern void
CVMgenBarrierPointersTraverse(CVMGeneration* gen, CVMExecEnv* ee,
			      CVMGCOptions* gcOpts,
			      CVMRefCallbackFunc callback,
			      void* callbackData);

#endif /* _INCLUDED_GENERATIONAL_H */

