/*
 * -----------------------------------------------------------------
 *  stacks.h :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)stacks.h	1.80 00/11/16
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifndef _INCLUDED_STACKS_H
#define _INCLUDED_STACKS_H

/****************************************************************************
 * CVMStack
 *
 * CVMStack is a C type for the four types of stacks we need:
 * interpreter stack, local roots stack, global roots stack, and jni
 * global refs stack.
 *
 * The stack is broken up into separate chunks. Initially there is only
 * one chunk, but as the stack grows more chunks may be added. If the
 * stack shrinks, chunks are not aggressively freed. Because of this,
 * currentStackChunk->next might not be NULL, but may in fact point
 * to a free chunk that can be used if that stack needs to grow again.
 ****************************************************************************/

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/gc_common.h"
#include "javavm/include/stackwalk.h"

/*
 * Generic 32-bit wide "Java slot" definition. This type occurs
 * in Java locals, object fields, constant pools, and
 * operand stacks (as a CVMStackVal32).
 */
typedef union CVMSlotVal32 {
    CVMJavaVal32    j;     /* For "Java" values */
    CVMUint8*       a;     /* a return created by jsr or jsr_w */
} CVMSlotVal32;

/*
 * Generic 32-bit wide stack slot definition.
 */
union CVMStackVal32 {
    CVMJavaVal32    j;     /* For "Java" values */
    CVMSlotVal32    s;     /* any value from a "slot" or locals[] */
    CVMObjectICell  ref;   /* JNI refs, local roots, global roots */
    CVMStackVal32*  next;  /* An element of a free list */
};

struct CVMStackChunk {
    CVMStackChunk*  prev;          /* Previous chunk of this stack */
    CVMStackChunk*  next;          /* Next chunk of this stack */
    CVMStackVal32*  end_data;      /* address of end of this chunk */
    CVMStackVal32   data[1];       /* actual data  */
};


struct CVMStack {
    CVMStackChunk*   firstStackChunk; 
    CVMStackChunk*   currentStackChunk;
    CVMUint32        minStackChunkSize;/* minimum size for new chunk */
    CVMUint32        maxStackSize;     /* max space used in all chunks */
    CVMUint32        stackSize;        /* Total space used in all chunks */
    CVMFrame*        volatile currentFrame;     /* Current stack frame */
    /* Stack chunk pointers. Both may not be needed and can
     * always be calculated. stackChunkEnd saves us an indirection on
     * method calls. stackChunkStart might not be of much value.
    */
    CVMStackVal32*   stackChunkStart;  /* &currentStackChunk->data[0] */
    CVMStackVal32*   stackChunkEnd;    /* currentStackChunk->end_data */
};

/****************************************************************************
 * CVMFrame
 *
 * CVMFrame is the common base class for all frames used to keep track of
 * local roots, java method calls, jni method calls, and frames created
 * by JNI PushLocalFrame. The last frame is always pointed to by the
 * owning stacks CVMStack.currentFrame field. It is not possible for
 * a frame to automatically know it's owning stack, but it can be deduced
 * from the frame type.
 ****************************************************************************/

/* Added by Kalpesh */
#define DEFERRED        0 /* Thread's ATC permission when thread is in Native or Simple Java Method */
#define SYNCH_DEFERRED  1 /* when thread is in synchronized method or block */
#define	NONDEFERRED	2 /* when thread is in method which throws(checked exceptions) 
			      * AsynchronouslyInterruptedException 			      */

typedef union ATCPermission{ 

    CVMStackVal32 stuff;//This should never be used
    struct data{
	char	  permission;   // Current ATC Permission 
        char      old_permission; // ATC Permission before entering synchronized blocks
        CVMUint16 depth;        //Depth of frame from "zero" frame//	
	//char      filler1;
	//char	    filler2;
    }val;
  
}ATCPermission;


struct CVMFrame{
    CVMFrame*             prevX;      /* previous frame: access using macros*/
    CVMFrameGCScannerFunc *scanner;   /* gc scanner for this frame */

    /*Added by Kalpesh*/
    ATCPermission ATCflags;

    /*Counter to encode synch nesting depth in a Java Method */  
    int synch_counter;

    /* topOfStack is used by CVMLocalRootsFrame to allocate the next
     * local root (localRoots field), by CVMJavaFrame to push and pop
     * operands (opstack field), and by CVMJNIFrame to allocate new
     * local refs (localRefs field).
     */
    CVMStackVal32*   volatile topOfStack; /* Pointer to top of stack.
					     This is the first free slot. */
    CVMMethodBlock*  mb;                  /* the method we are executing */

};

struct CVMFreelistFrame {
    CVMFrame          frame;

    CVMUint32         inUse;      /* Number of items in use under topOfStack */
    CVMStackVal32*    freeList;   /* A pointer to the free list */
    CVMStackVal32     vals[1];    /* Frame contents */
};

#define CVMfreelistFrameInit(frame_)	\
{					\
    ((CVMFrame*)(frame_))->mb = 0;     	\
    (frame_)->inUse = 0;		\
    (frame_)->freeList = 0;		\
}

/*
 * These are the capacities to ensure for a base frame and a
 * freelisting frame
 */
#define CVMbaseFrameCapacity \
    (sizeof(CVMFrame) / sizeof(CVMStackVal32))

#define CVMfreelistFrameCapacity  \
    (sizeof(CVMFreelistFrame) / sizeof(CVMStackVal32) - 1)

/****************************************************************************
 * Stack and Frame API's
 ****************************************************************************/

/* Many of the macros take arguments that could easily have been retrieved
 * from other arguments. This is done for performance because most of the
 * arguments are already be cached in registers. Also, some of the macros
 * don't set all fields in the stack and frame that they could set. This
 * is also done for performance to allow the interpreter to delay flushing
 * the value from a local to the stack or field only when a gc is needed.
 */


/*
 * Expand stack 's' to contain capacity words, and return a pointer to the
 * new area. If unsuccessful, returns NULL and throws an exception if the
 * throwException argument is true.
 *
 * If justChecking == TRUE, we are simply checking whether we have
 * this capacity or not. 
 */
extern 
CVMStackVal32* CVMexpandStack(CVMExecEnv* ee, CVMStack* s, 
			      CVMUint32 capacity,
			      CVMBool throwException,
			      CVMBool justChecking);

/*
 * Macros for dealing with frames that need special handling when they
 * are popped. We need a cheap way of finding out if we are popping the
 * first frame in a stack chunk, or if we are popping a frame that spans
 * multiple stack chunks. Both of these situations require special handling
 * by popFrame so information in the CVMStack can be updated. We use
 * the low bit in frame->prev to indicate that frame requires special
 * handling by popFrame. This is a cheap place for the flag because we
 * need to load frame->prev anyway when popping frame.
 */

/* Clear low bit in CVMFrame* */
#define CVMframeUnmasked(frame_) \
    ((CVMFrame*)((CVMUint32)(frame_) & ~1))

/* Set low bit in CVMFrame* */
#define CVMframeMasked(frame_, mask_) \
    ((CVMFrame*)((CVMUint32)(frame_) | (mask_)))

/* Get prevous frame without bit being set. Used to iterate over frames */
#define CVMframePrev(frame_) \
    CVMframeUnmasked((frame_)->prevX)

/* Clear the "special handling" bit */
#define CVMclearFrameNeedsSpecialHandlingBit(frame_) \
    (frame_) = CVMframeUnmasked((frame_))

/* Set the "special handling" bit */
#define CVMsetFrameNeedsSpecialHandlingBit(frame_) \
    (frame_) = CVMframeMasked((frame_), 1)

/* Check if the frame needs special handling in popFrame */
#define CVMframeNeedsSpecialHandling(frame_) \
    (((CVMUint32)(frame_) & 1) != 0)

/* 
 *   CVMBool 
 *   CVMcheckCapacity(CVMStack* s2_, CVMStackVal32* tos2,
 *                    CVMUint32 capacity2)
 *
 *   Checks if stack s2_ with current top of stack tos2_, can ensure
 *   a capacity of capacity2_. Return true if it can.
 */
#define CVMcheckCapacity(s2_, tos2_, capacity2_)		\
    (((tos2_) + (capacity2_)) <= (s2_)->stackChunkEnd)

/* 
 * Checks if the stack can ensure the specified capacity. It will attempt to
 * grow the stack if necessary. Returns CVM_TRUE if it can. Otherwise it
 * throws an exception and returns CVM_FALSE.
 */
extern CVMBool
CVMensureCapacity(CVMExecEnv* ee, CVMStack* stack, int capacity);

/*
 *  void
 *  CVMpushFrame(CVMExecEnv* ee,
 *      CVMStack* s_, CVMFrame* cur_frame_, CVMStackVal32* tos_,
 *      CVMUint32 capacity_, CVMUint32 frameOffset_,
 *      CVMFrameGCScannerFunc frameScanner_, failAction_, expandAction);
 *
 *  Push frame to stack s_, with current frame cur_frame_, current top
 *  of stack tos_, desired capacity_ in number of words, and a frame
 *  offset of frameOffset_ in number of words. cur_frame_ gets back
 *  a pointer to the new frame. The frame offset determines the
 *  'distance' from the previous frame's top of stack to the allocated
 *  CVMFrame. It is used to make room for java local variables.
 *
 *  On exit, cur_frame_ is set to the new frame. If the push failed (due
 *  to the failure of an expansion request) cur_frame_ is set to 0.
 *
 *  Must be called while gcUnsafe, but will become safe if the stack
 *  needs to be expanded.
 * 
 *  tos_ is not updated because java frames need to set the location of
 *  the new frame's local variables to the old tos minus the arg size.
 *
 *  frameScanner_ is a CVMFrameGCScannerFunc in order to scan roots in
 *  this frame at GC time.
 *
 *  failAction_ contains statements to executed if the frame cannot
 *  be allocated. It may contain a goto. CVMpushFrame will also throw 
 *  an exception and set cur_frame_ to NULL if unsuccessful.
 *
 *  expandAction_ contains statement to be executed only if the stack
 *  was successfully expanded. It is used by the interpreter to copy
 *  method arguments. It may contain a goto.
 */
#ifdef CVM_DEBUG
#define CVMassertPushFrameHasProperGCSafety(ee)				      \
    /* unfortunately ee->threadID is only available if CVM_DEBUG is defined */\
    CVMassert(CVMD_isgcUnsafe(ee) || ee->threadID == 0);
#else
#define CVMassertPushFrameHasProperGCSafety(ee)
#endif

#define CVMpushFrame(ee_, s_, cur_frame_, tos_, capacity_,		\
                     frameOffset_, frameScanner_,			\
		     failAction_, expandAction_)			\
{									\
    CVMFrame*       cf_    = (CVMFrame*)(cur_frame_);			\
									\
    /* The only time we should be gcSafe is during thread creation. */  \
    CVMassertPushFrameHasProperGCSafety(ee_);              		\
									\
    if (CVMcheckCapacity((s_), (tos_), (capacity_))) {			\
	(cur_frame_)       = (CVMFrame*)&tos_[(frameOffset_)];		\
        (cur_frame_)->synch_counter = 0;	/*Added by Kalpesh*/	\
        (cur_frame_)->ATCflags.val.permission = 0;			\
        (cur_frame_)->ATCflags.val.old_permission = 0;			\
        (cur_frame_)->prevX = cf_;					\
	if(cf_!=0){							\
		(cur_frame_)->ATCflags.val.depth = 			\
			((cur_frame_)->prevX)->ATCflags.val.depth + 1;	\
	}								\
	else{								\
		(cur_frame_)->ATCflags.val.depth = 0;			\
	}								\
	(ee_)->currentStackDepth = (cur_frame_)->ATCflags.val.depth;	\
        (cur_frame_)->scanner = frameScanner_;				\
        (s_)->currentFrame    = (cur_frame_);				\
    } else {								\
        CVMStackVal32* space_ =						\
	    CVMexpandStack(ee_, s_, (capacity_), CVM_TRUE, CVM_FALSE);	\
        if (space_ == 0) {						\
            (cur_frame_) = 0;	/* do this before failAction_ */	\
            failAction_;						\
        } else {							\
	    CVMsetFrameNeedsSpecialHandlingBit(cf_);			\
	    (cur_frame_)       = (CVMFrame*)&space_[(frameOffset_)];	\
            (cur_frame_)->prevX = cf_;					\
            (cur_frame_)->scanner = frameScanner_;			\
	    (cur_frame_)->synch_counter = 0;	/*Added by Kalpesh*/	\
	  (cur_frame_)->ATCflags.val.permission = 0;			\
	    (cur_frame_)->ATCflags.val.old_permission = 0;		\
 	    if(cf_!=0){							\
		(cur_frame_)->ATCflags.val.depth = 			\
			((cur_frame_)->prevX)->ATCflags.val.depth + 1;	\
	    }								\
	    else{							\
	 	(cur_frame_)->ATCflags.val.depth = 0;			\
	    }								\
	    (ee_)->currentStackDepth = (cur_frame_)->ATCflags.val.depth;\
            (s_)->currentFrame    = (cur_frame_);			\
            expandAction_;     /* do this after setting cur_frame_ */	\
        }								\
    }									\
}

/*
 *  void
 *  CVMpopFrame(CVMStack* s_, CVMFrame* cur_frame_);
 *
 *  Pop topmost frame from stack s_ with current frame
 *  cur_frame. cur_frame_ gets pointer to resulting new topmost frame.  
 *
 *  If the prev bit of the popped frame is tagged, this means one or both of
 *  two things: The frame spanned multiple stack chunks and/or 
 *  the frame is the first one in the current chunk. To handle these cases, 
 *  first find the chunk that the popped frame points to. If the new top 
 *  frame is not on the same chunk, update again.
 */
#define CVMpopFrame(s_, cur_frame_)					      \
{									      \
    CVMframeGetPreviousUpdateChunk((s_)->currentStackChunk, (cur_frame_),     \
				   {					      \
				       (s_)->currentStackChunk = chunk_;      \
				       (s_)->stackChunkStart = chunk_->data;  \
				       (s_)->stackChunkEnd = chunk_->end_data;\
				   });					      \
    (s_)->currentFrame = (cur_frame_);					      \
}

/*
 * Common to any code that wants to get to the previous frame,
 * and track stack chunks in the process.
 *
 * cur_chunk_     -- The last chunk of the current frame
 * cur_frame_     -- The current frame
 * updateAction_  -- The action to perform if we had to travel chunks
 *
 */
#define CVMframeGetPreviousUpdateChunk(cur_chunk_,			\
				       cur_frame_,			\
				       updateAction_)			\
{									\
    CVMFrame* prev_ = (cur_frame_)->prevX;				\
    if (CVMframeNeedsSpecialHandling(prev_)) {				\
	/*								\
	 * Either curFrame spanned multiple stack			\
	 * chunks or it is the first frame in				\
	 * chunk. In any case, we must update the			\
	 * new current chunk to refer to last chunk of			\
	 * previous frame.						\
	 */								\
        CVMStackChunk* chunk_   = (cur_chunk_);				\
	/*								\
	 * First bring the chunk to the chunk of			\
	 * the current frame						\
	 */								\
        while(!CVMaddressInStackChunk((cur_frame_), chunk_)) {		\
            chunk_ = chunk_->prev;					\
        }								\
	prev_ = CVMframeUnmasked(prev_);				\
	/*								\
	 * Now 'chunk' points to the starting chunk			\
	 * of curFrame.  If this frame was the				\
	 * first in the current chunk, the previous			\
	 * frame's last chunk is the previous				\
	 * chunk.							\
	 */								\
        if ((prev_ != NULL) && 						\
	    !CVMaddressInStackChunk(prev_->topOfStack, chunk_)) {	\
	    chunk_ = chunk_->prev;					\
	}								\
        updateAction_;							\
    }									\
    cur_frame_ = prev_;							\
}

/*
 * push/pop GC root frame pushes empty frames that will contain GC roots
 * like JNI locals and globals, CVM local and global roots.
 *
 * If unsuccessful, throws and exception and sets curr_frame3_ to NULL.
 *
 * Note that this can only be called when GC-safe!
 */
#define CVMpushGCRootFrame(ee3_, s3_, curr_frame3_, capacity3_, frameScanner_)\
{									      \
    CVMD_gcUnsafeExec((ee3_), {						      \
	CVMpushFrame((ee3_), (s3_), (curr_frame3_),			      \
		     ((curr_frame3_)->topOfStack),			      \
		     (capacity3_), 0, frameScanner_, {}, {});		      \
	if ((curr_frame3_) != 0) {					      \
	    (curr_frame3_)->topOfStack =				      \
		(CVMStackVal32*)(curr_frame3_) + (capacity3_);		      \
	}								      \
    });									      \
    CVMassert((curr_frame3_) != 0); 					      \
}

#define CVMpopGCRootFrame(s3_, curr_frame3_)	\
{						\
    CVMpopFrame((s3_), (curr_frame3_));		\
}

/* 
 * Walk the references in a given stack. Use the 'scanner' embedded in
 * each frame to do the scan. 
 */
#define CVMstackDoRoots(stack, frameAction) 			\
{								\
    CVMStackWalkContext swc;					\
    CVMFrame*           frame;					\
    CVMStackChunk*      chunk;					\
    								\
    CVMstackwalkInit(stack, &swc);				\
    frame = swc.frame;						\
    chunk = swc.chunk;						\
    do {							\
        frameAction;						\
	CVMstackwalkPrev(&swc);					\
	frame = swc.frame;					\
	chunk = swc.chunk;					\
    } while (frame != 0);					\
}

#define CVMstackScanRoots(ee, stack, callback, data, gcOpts)	\
    CVMstackDoRoots(stack, {					\
	frame->scanner(ee, frame, chunk, 			\
			callback, data, gcOpts);		\
    });

/*
 * Scan each root in a freelisting GC root frame.
 */
#define CVMstackScanGCRootFrameRoots(stack, callback, data)	  	  \
    CVMstackDoRoots(stack, {		  	  	  		  \
	CVMwalkRefsInGCRootFrame(frame, ((CVMFreelistFrame*)frame)->vals, \
				 chunk, callback, data, CVM_TRUE);	  \
    });

/* 
 * Walk the references in a given GC root frame. Call 'callback_' on
 * each reference discovered. The GC root frame may or may not have
 * free lists, as indicated by 'hasFreeLists_'.
 */
#define CVMwalkRefsInGCRootFrame(frame_,       				 \
                                 rootData_,				 \
				 chunk_, 				 \
				 callback_, 				 \
				 data_,					 \
				 hasFreeLists_)				 \
{									 \
    CVMStackVal32* tos_ = (frame_)->topOfStack;				 \
									 \
    /*									 \
     * First scan all roots across multiple chunks until we get to the	 \
     * chunk the frame is on.						 \
     */									 \
    while (!CVMaddressInStackChunk(frame_, chunk_)) {			 \
        CVMwalkRefsInGCRootFrameSegment(tos_, (chunk_)->data,		 \
					callback_, data_, hasFreeLists_);\
	(chunk_) = (chunk_)->prev;					 \
	tos_ = (chunk_)->end_data;					 \
    }									 \
    /*									 \
     * Now we are in the first chunk of the frame. Scan from the start	 \
     * of the root area to the top of stack (or the end of chunk, 	 \
     * depending on whether the previous loop executed or not).		 \
     */									 \
    CVMwalkRefsInGCRootFrameSegment(tos_, rootData_,			 \
				    callback_, data_, hasFreeLists_);	 \
    /* %comment rt001 */						 \
}

/*
 * This is a helper macro for CVMwalkRefsInGCRootFrame(). It walks a
 * contiguous segment of roots. It calls 'callback' only on the non-NULL
 * roots.
 *
 * It is also used as the implementation of CVMwalkContiguousRefs()
 * (see below)
 */
#define CVMwalkRefsInGCRootFrameSegment(higherAddr_, 			     \
					lowerAddr_,			     \
					callback_,			     \
					data_,				     \
					hasFreeLists_)			     \
{									     \
    CVMStackVal32* refPtr_;						     \
    for (refPtr_ = (higherAddr_) - 1; refPtr_ >= (lowerAddr_); refPtr_--) {  \
        CVMUint32 referent_ = *((CVMUint32*)refPtr_);			     \
        if (referent_ != 0) {						     \
	    if (hasFreeLists_) {					     \
		/*							     \
		 * Not a free list element. Walk it.    		     \
		 */							     \
		if ((referent_ & 0x1) == 0) {	 	     		     \
		    (*callback_)((CVMObject**)refPtr_, data_);		     \
		}							     \
	    } else {							     \
		(*callback_)((CVMObject**)refPtr_, data_);     		     \
	    }								     \
        }								     \
    }								 	     \
}

/*
 * Walk numRefs_ contiguous references starting at addr refsAddr_. Call
 * 'callback_' on each non-NULL reference found.
 */
#define CVMwalkContiguousRefs(refsAddr_, numRefs_, callback_, data_) 	      \
    CVMwalkRefsInGCRootFrameSegment((CVMStackVal32*)((refsAddr_)+(numRefs_)), \
				    (CVMStackVal32*)(refsAddr_),	      \
				    (callback_), (data_),		      \
				    CVM_FALSE)

#define CVMwalkOneICell(refICellPtr, callback, data)		\
{								\
    CVMObject* referent = *((CVMObject**)refICellPtr);		\
    if (referent != 0) {					\
	(*callback)((CVMObject**)refICellPtr, data);		\
    }								\
}



/*
 * CVMStackVal32*
 * CVMframeAllocFromTop(CVMExecEnv* ee_, CVMStack* s_, CVMFrame* cur_frame_,
 *                      CVMStackVal32* result_);
 *
 * Allocates a slot from the top of the specified CVMFrame.
 *
 * If unsuccessful, sets result_ to NULL. NO EXCEPTION IS THROWN.
 */
#define CVMframeAllocFromTop(ee_, s_, cur_frame_, result_)		\
{									\
    CVMFrame*      cf_    = (CVMFrame*)(cur_frame_);			\
    CVMStackVal32* tos_   = cf_->topOfStack;				\
    CVMStackVal32* space_;						\
									\
    /* We might end up expanding the stack. Assert GC-safety */	        \
    CVMassert(CVMD_isgcSafe(ee_));					\
									\
    if (CVMcheckCapacity((s_), (tos_), 1)) {				\
        space_ = (tos_);						\
        CVMID_icellSetNull(&space_->ref);				\
        (result_)         = space_;					\
        cf_->topOfStack   = space_ + 1;					\
    } else {								\
        CVMD_gcUnsafeExec(ee_, {					\
            space_ = CVMexpandStack(ee_, (s_), 1,			\
				    CVM_FALSE, CVM_FALSE);		\
            if (space_ == 0) {						\
                (result_) = 0;						\
            } else {							\
                CVMID_icellSetNull(&space_->ref);			\
	        (result_)         = space_;				\
	        cf_->topOfStack   = space_ + 1;				\
                if (tos_ != space_) {					\
                    CVMsetFrameNeedsSpecialHandlingBit(cf_->prevX);	\
               }							\
            }								\
        });								\
    }									\
}

/*
 * CVMStackVal32*
 * CVMframeAlloc(CVMExecEnv* ee_, CVMStack* s_, CVMFreelistFrame* cur_frame_,
 *               CVMStackVal32* result_);
 *
 * Allocates a slot from the CVMFreelistFrame. If a slot is on the free
 * list then it is allocated there. Otherewise it is allocated from
 * the top of the frame. Slots on the free list have their low bit set
 * so it needs to be cleared before returning the slot.
 *
 * If unsuccessful, sets result_ to NULL. NO EXCEPTION IS THROWN.
 */
#define CVMframeAlloc(ee_, s_, cur_frame_, result_)			   \
{									   \
    CVMStackVal32*  slot_  = (cur_frame_)->freeList;			   \
									   \
    /* We might end up expanding the stack. Assert GC-safety */		   \
    CVMassert(CVMD_isgcSafe(ee_));					   \
    if (slot_ != 0) {							   \
	(cur_frame_)->freeList = 					   \
            (CVMStackVal32*)((CVMUint32)((*slot_).next) & ~1); 		   \
        CVMID_icellSetNull(&slot_->ref);				   \
	(result_) = slot_;						   \
        (cur_frame_)->inUse++;						   \
    } else {								   \
	CVMframeAllocFromTop(ee_, (s_), (cur_frame_), (result_));  	   \
        if (result_ != NULL) {						   \
           (cur_frame_)->inUse++;					   \
        }								   \
    }									   \
}

/*
 * void
 * CVMframeFree(CVMFreelistFrame* cur_frame_, CVMStackVal32* slot_);
 *
 * Puts slot_ on the free list of cur_frame_. The low bit of the new
 * free list item is set to indicate that it is a freed item. This is done
 * so when gc scans the frame, it can tell which slots are in use and
 * which slots are free.
 */
#define CVMframeFree(cur_frame_, slot_)					   \
{									   \
    CVMStackVal32*  freeList_ = (cur_frame_)->freeList;			   \
									   \
    CVMassert(slot_ != NULL);						   \
    if (((CVMUint32)(slot_)->next & 1) != 0) { /* check if already free */ \
	CVMdebugPrintf(("CVMframeFree: %p is already free\n", slot_));     \
        CVMassert(CVM_FALSE);						   \
    } else {								   \
	(cur_frame_)->freeList = slot_;					   \
	(slot_)->next = (CVMStackVal32*)((CVMUint32)freeList_ | 1);	   \
	(cur_frame_)->inUse--;						   \
    }									   \
}

/*
 * Is the address in the given stack chunk?
 */
#define CVMaddressInStackChunk(address_, chunk_)		\
    (((CVMStackVal32*)(address_) < (chunk_)->end_data) &&	\
     ((CVMStackVal32*)(address_) >= &(chunk_)->data[0]))

/*
 * Initialize a stack before first use, and destroy after use.
 * If CVMinitStack is unsuccessful, it will return CVM_FALSE. No
 * exception is thrown.
 */
extern CVMBool
CVMinitStack(CVMExecEnv *ee, CVMStack* s,
	     CVMUint32 initialStackSize, CVMUint32 maxStackSize,
	     CVMUint32 minStackChunkSize, CVMUint32 initialFrameCapacity,
	     CVMFrameGCScannerFunc frameScanner);

extern void
CVMdestroyStack(CVMStack* stack);

/*
 * Initialize a GC root stack with a singleton frame
 * If CVMinitGCRootStack is unsuccessful, it will return CVM_FALSE. An
 * exception is NOT thrown.
 */
extern CVMBool
CVMinitGCRootStack(CVMExecEnv *ee, CVMStack* s, CVMUint32 initialStackSize, 
		   CVMUint32 maxStackSize, CVMUint32 minStackChunkSize,
		   CVMFrameGCScannerFunc frameScanner);

extern void
CVMdestroyGCRootStack(CVMStack* stack);

/*
 * Dump stack or frame contents for debugging
 */
#ifdef CVM_DEBUG_DUMPSTACK
extern void
CVMdumpStack(CVMStack* s, CVMBool verbose, CVMBool includeData,
	     CVMInt32 frameLimit);

extern CVMStackChunk*
CVMdumpFrame(CVMFrame* frame, CVMStackChunk* startChunk,
	     CVMBool verbose, CVMBool includeData);
#endif

/*
 * Check C stack size availability in the recursive function and
 * throw an exception.
 */
extern CVMBool
CVMCstackCheckSize(CVMExecEnv *ee, CVMUint32 redzone, char *func_name, CVMBool exceptionThrow);

/*
 * A callback function for each frame discovered in CVMscanStack()
 */
typedef void (*CVMFrameCallbackFunc)(CVMFrame *thisFrame);

/* Scans the in-use part of the stack and calls frameCallback for each
 * discovered frame. This should be a function since it's big and
 * slow anyway. 
 */
extern void 
CVMscanStack(CVMStack* stack, CVMFrameCallbackFunc frameCallback);

/* Scans the in-use part of the stack for garbage collection. This
 * should be a function since it's big and slow anyway.  
 */
extern void 
CVMscanStackForGC(CVMStack* stack);

#endif /* _INCLUDED_STACKS_H */
