/*
 * @(#)CVMWriter.java	1.104 00/11/09
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */
package runtime;
import Const;
import CVMConst;
import components.*;
import vm.*;
import jcc.Util;
import util.*;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;
import java.io.OutputStream;


/*
 * The CoreImageWriter for the Embedded VM
 */

public class CVMWriter implements CoreImageWriter, Const, CVMConst {

    /* INSTANCE DATA */
    protected  Exception failureMode = null; // only interesting on falure


    protected String    	  outputFileName;
    protected CCodeWriter 	  classOut;
    protected CCodeWriter 	  auxOut;
    protected OutputStream        xxx;
    protected OutputStream        zzz;

    protected String 		  headerFileName = null;
    protected CCodeWriter 	  headerOut;
    protected OutputStream         yyy;

    /*
     * The following only used if we're segmenting the output,
     * which happens if this.init gets a maxSegSize > 0 
     */
    protected boolean		  segmentedOutput;
    protected int		  maxClasses;
    protected int		  curClasses = 0;
    protected int		  nClassfileOut;
    protected BufferedPrintStream listOut;

    protected CVMStringTable	  stringTable;
    protected CVMClass	  	  classes[];

    private   VMClassFactory     classMaker = new CVMClassFactory();

    boolean	   formatError = false;
    boolean	   verbose     = false;
    boolean	   classDebug  = false;
    boolean        lossless    = false;

    private static final String		staticInitThreadStoreName="StaticInitThreadStorage";
    private static final String		stringArrayName = "CVM_ROMStrings";
    private static final String		masterStringArrayName = "CVM_ROMStringsMaster";
    private static final String		staticStoreName = "CVM_staticData";
    private static final String		masterStaticStoreName = "CVM_StaticDataMaster";

    private ClassnameFilterList         nativeTypes;
    private ClassnameFilter		invisibleClassList = new ClassnameFilter();

    private boolean			classLoading = true; // by default.

    // Place all byte codes into read-write memory.  Useful for setting
    // breakpoints.
    protected boolean			noPureCode = false;

    private boolean			doShared;
    private String			sharedConstantPoolName;
    private int				sharedConstantPoolSize;

    private CVMMethodType[]		methodTypes;
    private int[]			nMethodTypes;
    private int				totalMethodTypes;

    //
    // to keep track of which variants of classblock, method arrays, field arrays,
    // and method descriptors have been declared
    // see declareClassblock, declareMethodArray, and declareFieldArray
    // respectively
    //
    private java.util.BitSet		classblockSizes  = new java.util.BitSet();
    private java.util.BitSet		methodArraySizes = new java.util.BitSet();
    private java.util.BitSet		fieldArraySizes  = new java.util.BitSet();

    /*
     * Per-class data that is built up as we
     * print out the various data structures.
     * Rather than trying to pass around both name and size of each,
     * we collect the information here.
     */
    private String			classBlockName;
    private String			constantPoolName;
    private int				constantPoolSize;
    private String			methodArrayName;
    private int				methodTableSize;
    private String			checkedExceptionName;
    private String			interfaceTableName;
    private String			fieldTableName;
    private int				fieldTableSize;
    /*
     * Name of the array of stackmaps indirect cells, as referenced
     * by the JavaMethodDescriptors, and the next index into said array.
     */
    private static final String		stackmapIndirect = "CVM_stackmapIndirect";
    private int				stackmapIndex    = 0;


    /* for statistics only */
    int  ncodebytes;
    int  ncatchframes;
    int  nmethods;
    int  nfields;
    int  nconstants;
    int	 njavastrings;
    int  ninnerclasses;
    int	 nWritableMethodBlocks;
    int	 nClassesWithWritableMethodBlocks;

    public CVMWriter( ){ 
	stringTable = new CVMStringTable();
    }

    public void init( boolean classDebug, boolean lossless, ClassnameFilterList nativeTypes, boolean verbose, int maxSegmentSize ){
	this.verbose = verbose;
	this.classDebug = classDebug;
	this.lossless = lossless;
	this.nativeTypes = nativeTypes;
	if ( maxSegmentSize > 0 ){
	    this.segmentedOutput = true;
	    this.maxClasses = maxSegmentSize;
	} else {
	    this.segmentedOutput = false;
	}
    }

    // If attribute starts with attributeName, then parse the rest of the
    // string as a list of classes.
    private boolean parseClassListAttribute(String attributeName, 
					    String attribute, 
					    ClassnameFilter filter) { 
        if (!attribute.startsWith(attributeName))
	    return false;
	String val = attribute.substring(attributeName.length() );
	java.util.StringTokenizer tkn = 
	    new java.util.StringTokenizer(val, " ,", false );
	while ( tkn.hasMoreTokens() ){
	    String classname = tkn.nextToken().replace('.', '/');
	    filter.includeName(classname);
	}
	return true;
    }

    private boolean
    isClassInvisible( String classname ) {
        return invisibleClassList.accept(null, classname);
    }

    private boolean
    isClassCNI( String classname) { 
	// compare against interned string
	return nativeTypes.isType( classname, "CNI" );
    }

    private boolean
    isClassJNI( String classname) { 
	// compare against interned string
	return nativeTypes.isType( classname, "JNI" );
    }

    public boolean setAttribute( String attribute ){
	if ( parseClassListAttribute("invisible=", attribute,
				     invisibleClassList)){
	    return true;
	} else if ( attribute.equals("noClassLoading" ) ){
	    classLoading = false;
	    return true;
	} else if ( attribute.equals("noPureCode") ) {
	    noPureCode = true;
	    return true;
	} else
	    return false; // no attribute by this name.
    }


    /*
     * If this is a simple, unsegmented output file, then
     * the file name given is the name of the C file, from which
     * we compute the .h file, crudely. We open them here.
     *
     * But if this is a segmented output job, then the name given
     * serves as a prefix, and we open a whole bunch of files and
     * set up a whole bunch of state:
     *		filename+"List"  - file containing names of all .c files generated
     *				   opened as listOut;
     *		filename+".h"    - header file, as usual
     *		filename+"Aux.c" - file containing all type, string, and other miscellanious
     *				   generated stuff.
     *				   opened as auxOut;
     *		filename+"Class"+n+".c" - files containing directly class-related data,
     *				   whole classes, and a max of this.maxClasses of them
     *				   apiece.
     */
    public boolean open( String filename ){
	if ( classOut != null ) close();
	outputFileName = filename;
	if ( ! segmentedOutput ){
	    /* old-fashioned, single output case */
	    if ( filename == null ){
		xxx = System.out;
		headerFileName = "jcc.output.h";
	    } else {
		try {
		    xxx = new java.io.FileOutputStream( filename );
		} catch ( java.io.IOException e ){
		    failureMode = e;
		    return false;
		}
		headerFileName = filename+".h";
	    }
	    try {
		classOut = auxOut = new CCodeWriter( xxx );
		yyy = new java.io.FileOutputStream( headerFileName );
		headerOut = new CCodeWriter( yyy );
	    } catch ( java.io.IOException e ){
		failureMode = e;
		return false;
	    }
	    writeHeaderPrologue( headerOut );
	    writePrologue( classOut );
	    return true;
	} else {
	    /* segmented output.
	     * open the aux file, the first classOut file, the list file and
	     * the header file.
	     */
	    curClasses = 0;
	    nClassfileOut = 0;
	    headerFileName = outputFileName+".h";
	    try {
		yyy = new java.io.FileOutputStream( headerFileName );
		headerOut = new CCodeWriter( yyy );
		zzz = new java.io.FileOutputStream( filename+"Aux.c");
		auxOut = new CCodeWriter( zzz );
		listOut = new BufferedPrintStream( new java.io.FileOutputStream( outputFileName+"List" ));
		listOut.println( filename+"Aux.c");
		openNextClassFile();
	    } catch ( java.io.IOException e ){
		failureMode = e;
		return false;
	    }
	    writeHeaderPrologue( headerOut );
	    writePrologue( auxOut );
	    return true;
	}
    }

    private void
    openNextClassFile() throws java.io.IOException{
	if ( classOut != null ){
	    classOut.close();
	}
	String newName = outputFileName+(nClassfileOut++)+".c";
	xxx = new java.io.FileOutputStream( newName );
	classOut = new CCodeWriter( xxx );
	listOut.println( newName );
	writePrologue( classOut );
    }

    public void close(){
	if ( classOut != null ){
	    classOut.close();
	}
	if ( auxOut  != null && auxOut != classOut ){
	    auxOut.close();
	}
	auxOut = null;
	classOut = null;
	if ( listOut != null ){
	    listOut.close();
	    listOut = null;
	}
	if ( headerOut != null ){
	    headerOut.close();
	    headerOut = null;
	}
	outputFileName = null;
	return;
    }

    public void printError( java.io.PrintStream o ){
	if ( failureMode != null ){
	    failureMode.printStackTrace( o );
	} 
    }


    protected void declare( String declaration,  BufferedPrintStream curOut ){
	headerOut.print( "extern " );
	headerOut.print( declaration );
	headerOut.write(';');
	headerOut.write('\n');
	curOut.print( declaration );
    }

    public static String getUTF( UnicodeConstant u ){
	if (u.UTFstring == null ){
	   u.UTFstring = Util.unicodeToUTF( u.string );
	}
	return u.UTFstring;
    }

    /*
     * For Java methods,
     * write out the structure which will be initialized with a
     * JavaMethodDescriptor and its associated structures: code array,
     * line numbers, exceptions, local variables
     */
    private void writeCode(CVMMethodInfo meth) {
	MethodInfo 		  mi 	  = meth.method;
	LineNumberTableEntry[]	  lntab	  = mi.getLineNumberTable();
	LocalVariableTableEntry[] locvartab = mi.getLocalVariableTable();

	int	codeLengthRoundup = (mi.code.length+3)&~3;
	int 	nTryCatch 	  = (mi.exceptionTable==null) ? 0 
	       				  : mi.exceptionTable.length;
	int	nLineNumbers = (classDebug && lntab != null) ? lntab.length : 0;
	int	nLocals = (classDebug && locvartab != null) ? locvartab.length : 0;

	String	codeBlockName	  = meth.getNativeName();
	String	descriptorTypeTag = codeBlockName+"MDType";

	CCodeWriter declareOut;

	boolean impureCode = noPureCode || !meth.isCodePure();

	classOut.println("\n/* Code block for "+meth.method.qualifiedName()+" */");


	/*
	 * Declare the structure which will be initialized with a
	 * JavaMethodDescriptor and its associated structures: code array,
	 * line numbers, exceptions, local variables
	 * Unlike several of the other structures, we do not bother remembering
	 * and reusing prior definitions -- there are so many dimensions it doesn't
	 * see worthwhile.
	 * In fact, the only reason this is not an anonimous structure is the
	 * case where it has to be writable and initialized, for JVMDI.
	 */
	if ( ! impureCode ){
	    // this might as well be static since there will be
	    // no external references to it
	    classOut.print("STATIC ");
	    declareOut = classOut;
	} else {
	    declareOut = headerOut;
	}

	declareOut.print("const struct ");
	declareOut.print( descriptorTypeTag );
	declareOut.println(" {\n    union {");
	declareOut.println("        CVMJavaMethodDescriptor jmd;");
	declareOut.println("        CVMUint" +
		    ((meth.alignment() == 1) ? "8" : "32") +
		    " align;\n    } u;");
	declareOut.println("    CVMUint8 data["+codeLengthRoundup+"];");
	if ( nTryCatch != 0 ){
	    declareOut.println("    CVMExceptionHandler exceptionTable["+
			nTryCatch+"];");
	}
	if ( nLineNumbers != 0 ){
	    declareOut.println("    CVMLineNumberEntry lineNumberTable["+
			nLineNumbers+"];");
	}
	if ( nLocals != 0 ){
	    declareOut.println("    CVMLocalVariableEntry localVariableTable["+
			nLocals+"];");
	}
	declareOut.print("} ");

	/*
	 * Now we have the shape of the data we need to supply the data itself
	 * (themselves?)
	 *
	 * EITHER the MD has to be writable. in thise case the declaration above
	 * was to the header file. Append to it the names of the master and
	 * the working copy, emit the master, reserve space for the copy,
	 * and don't forget the initialization instructions.
	 *
	 * OR the MD is read-only, const, so this is the thing itself
	 */

	if ( impureCode ){
	    declareOut.print( codeBlockName );
	    declareOut.print("Master;\n struct ");
	    declareOut.print( descriptorTypeTag );
	    declareOut.write(' ');
	    declareOut.println( codeBlockName );
	    declareOut.println(";");

	    classOut.print( "const struct " );
	    classOut.print( descriptorTypeTag );
	    classOut.write(' ');
	    classOut.print( codeBlockName );
	    classOut.print( "Master" );

	    CVMInitInfo.addInfo( "&"+codeBlockName+"Master", "&"+codeBlockName, "sizeof( struct "+descriptorTypeTag+")" );
	} else {
	    classOut.print( codeBlockName );
	}

	classOut.print( " = {\n    { CVM_INIT_JAVA_METHOD_DESCRIPTOR(" );
	/*
	 * Make sure locals are big enough for any return value.
	 */
	classOut.print(mi.stack+", "+Math.max(2,mi.locals));
	classOut.print( ", ");
	if ( meth.codeHasJsr() ){
	    classOut.print("CVM_JMD_MAY_NEED_REWRITE|");
	    if (meth.alignment() != 1) {
		classOut.print("CVM_JMD_NEEDS_ALIGNMENT|");
	    }
	}
	if (meth.isStrictFP()) {
	    classOut.print("CVM_JMD_STRICT|");
	}
	classOut.print("0, " );
	classOut.print(nTryCatch+", "+mi.code.length+", &");
	classOut.print(stackmapIndirect+"["+(stackmapIndex++) );
	classOut.print("], "+nLineNumbers+", "+nLocals);
	classOut.print(") },\n    {");

	// output the code array
	int j = 99;
	byte[] methodCode = mi.code;
	int l = methodCode.length;
	for ( int i = 0; i <l; i++ ){
	    if ( j >= 12 ){
		classOut.print("\n\t");
		j = 1;
	    } else {
		j+= 1;
	    }
	    classOut.printHexInt( methodCode[i]& 0xff );
	    classOut.print(",");
	}
	classOut.print("\n    }");
	ncodebytes += mi.code.length;

	// output the CVMExceptionHandler[] (catch frames)
	if ( nTryCatch != 0 ){
	    classOut.println(",\n    {");
	    for ( int i = 0; i < nTryCatch; i++ ){
		ExceptionEntry e = mi.exceptionTable[i];
		classOut.println("\t{ "+e.startPC+", "+e.endPC+", "+e.handlerPC+", "
			    +(e.catchType == null ? 0 : e.catchType.index)
			    +"},");
	    }
	    classOut.print("    }");
	    ncatchframes += nTryCatch;
	}

	// output the CVMLineNumberEntry[]
	if ( nLineNumbers != 0 ){
	    classOut.println(",\n    {");
	    for ( int i = 0; i < nLineNumbers; i++ ){
		LineNumberTableEntry e = lntab[i];
		classOut.println("\t{ "+e.startPC+", "+e.lineNumber+"},");
	    }
	    classOut.print("    }");
	}

	// output the CVMLocalVariableEntry[]
	if ( nLocals != 0 ){
	    classOut.println(",\n    {");
	    for ( int i = 0; i < nLocals; i++ ){
		LocalVariableTableEntry e = locvartab[i];
		int typeid = CVMDataType.parseSignature(getUTF(e.sig));
		int nameid = CVMMemberNameEntry.lookupEnter(getUTF(e.name));
		String nameAndType = "0x"+Integer.toHexString((nameid<<16)+typeid);
		classOut.println("\t{ "+e.pc0+", "+e.length+", "+e.slot+
			    ", "+ nameAndType+"},");
	    }
	    classOut.print("    }");
	}

	classOut.println("\n};");


    }


    private void declareMethodArray( CVMClass c, boolean writeable ){
	String thisTableName = c.getNativeName()+"_methods";
	String tableAddress = "&"+thisTableName;
	int nmethods = c.methods.length;
	String methodArrayFlavor = "struct CVMMethodArray"+nmethods;
	if ( !methodArraySizes.get( nmethods ) ){
	    // need to declare a structure containing a classblock pointer
	    // plus "nmethods" CVMMethodBlock structures
	    int i;
	    headerOut.print(methodArrayFlavor+" {");
	    if (nmethods > 256)
		headerOut.println();
	    int numMethodArrays = (nmethods-1) / 256 + 1;
	    for (i = 1; i <= numMethodArrays; i++) {
		int nmethods2; // # methods in this method array
		if (i == numMethodArrays) {
		    nmethods2 = (nmethods-1) % 256 + 1;
		} else {
		    nmethods2 = 256;
		}
		headerOut.println(" CVMClassBlock const *cb" + i + ";" +
				  " CVMMethodBlock mb" + i + 
				  "[" + nmethods2 + "];");
		if (nmethods > 256) {
		    headerOut.println();
		}
	    }
	    headerOut.println(" };");
	    methodArraySizes.set( nmethods );
	}
	if (writeable ){
	    /*
	     * Declare actual array, record copying information.
	     * What we're about to initialize is the read-only master.
	     */
	    String masterName = thisTableName+"Master";

	    declare(methodArrayFlavor+" "+thisTableName, classOut);
	    classOut.println(";");
	    CVMInitInfo.addInfo( "&"+masterName, tableAddress, "sizeof( "+methodArrayFlavor+")" );
	    declare( "const "+methodArrayFlavor+" "+masterName, classOut );
	} else {
	    declare( "const "+methodArrayFlavor+" "+thisTableName, classOut);
	}
	methodArrayName = tableAddress;
	methodTableSize = nmethods;
    }

    void writeMethods(CVMClass c) {
	CVMMethodInfo m[] = c.methods;
	int nCheckedExceptions = 0;
	int currCheckedException = 0;
	int checkedExceptionVector[];
	boolean writableMethods = false;
	if ( (m == null) || (m.length == 0 )){
	    methodArrayName = "0";
	    checkedExceptionName = "0";
	    methodTableSize = 0;
	    return;
	}
	int nmethod =  m.length;
	String thisTableName = c.getNativeName()+"_methods";
	String thisExceptionName = null;
	for ( int i = 0; i < nmethod; i++ ){
	    CVMMethodInfo meth = m[i];
	    MethodInfo    mi   = meth.method;
	    if ( mi.exceptionsThrown != null ){
		int n = mi.exceptionsThrown.length;
		if ( n != 0 )
		    nCheckedExceptions += n+1; // "1" is for the count word.
	    }
	    if ( mi.code != null ){
		writeCode( meth );
	    }
	    if ( meth.codeHasJsr() ){
		writableMethods = true;
	    }
	}
	//
	// if any methods declare that they throw any exceptions,
	// we need to construct the exception vector.
	if ( nCheckedExceptions == 0 ){
	    checkedExceptionName = "0";
	    checkedExceptionVector = null;
	} else {
	    thisExceptionName = c.getNativeName()+"_throws";
	    checkedExceptionName = thisExceptionName;
	    // Ensure that item 0 in this array is always unused.
	    // See classes.h.
	    nCheckedExceptions++;
	    checkedExceptionVector = new int[ nCheckedExceptions ];
	    checkedExceptionVector[0] = 0;
	    currCheckedException = 1;
	}
	//
	// if any method has a JSR, then its method block, and thus
	// all of them, must be writable, so we can replace its code, bundled
	// with its JavaMethodDescriptor.
	declareMethodArray( c, writableMethods );
	classOut.print( " = {\n");

	// Include JNI header file if there are native methods
	for ( int i = 0; i < nmethod; i++ ){
	    CVMMethodInfo meth = m[i];
	    int access = meth.CVMflags();
	    if ((access&CVM_METHOD_ACC_NATIVE)!=0) {
		String subdir = isClassCNI(c.ci.className) ? "cni/" : "jni/";
		headerOut.println("#include \"generated/" + subdir +
		    Util.convertToClassName(c.ci.className)+".h\"");
		break;
	    }
	}

	for ( int i = 0; i < nmethod; i++ ){
	    CVMMethodInfo meth = m[i];
	    int mbIndex = i % 256;
	    int typeid = meth.sigType.entryNo;
	    int nameid = CVMMemberNameEntry.lookupEnter(getUTF(meth.method.name));
	    String nameAndType = "0x"+Integer.toHexString((nameid<<16)+typeid);
	    int access = meth.CVMflags();

	    boolean synchro = (access&CVM_METHOD_ACC_SYNCHRONIZED)!=0;
	    boolean isNative  = (access&CVM_METHOD_ACC_NATIVE)!=0;
	    boolean isAbstract= (access&CVM_METHOD_ACC_ABSTRACT) != 0;

	    int invoker = isAbstract ? CVM_INVOKE_ABSTRACT_METHOD
			  : isNative ? (
				isClassCNI(c.ci.className) ?
				    CVM_INVOKE_CVM_METHOD : (
				synchro ? CVM_INVOKE_JNI_SYNC_METHOD
					: CVM_INVOKE_JNI_METHOD ))
			  : (
				synchro ? CVM_INVOKE_JAVA_SYNC_METHOD
					: CVM_INVOKE_JAVA_METHOD );

	    //
	    // note the offset in the exception array,
	    // and add to that array as necessary.
	    int exceptions;
	    if ( (meth.method.exceptionsThrown != null) && (meth.method.exceptionsThrown.length!=0) ){
		ClassConstant thrown[] = meth.method.exceptionsThrown;
		int net = thrown.length;
	        exceptions = currCheckedException;
		checkedExceptionVector[currCheckedException++] = net;
		for ( int j = 0; j < net; j++ ){
		    checkedExceptionVector[currCheckedException++] = thrown[j].index;
		}
	    } else {
		// Note we have changed this from ((unsigned short)
		// -1) to 0 to be consistent with the documentation in
		// classes.h. A checkedExceptionsOffset of 0 indicates
		// no checked exceptions for the method. Also see the
		// allocation of the checkedExceptionVector above.
		exceptions = 0;
	    }
	    int argssize   = meth.method.argsSize;
	    int tableIndex = meth.methodOffset();
	    /* method offsets are always mod 256 and we have to emit a cb
	     * entry before each block of 256 methods. */
	    if (mbIndex == 0) {
		if (i != 0) {
		    classOut.print( "\t},\n");
		}
		classOut.print( "\t&"+classBlockName+",\n\t{\n" );
	    }
	    classOut.print("	CVM_INIT_METHODBLOCK("+nameAndType+","+tableIndex+","+argssize+","+invoker+","+access+","+mbIndex+","+exceptions+",");
	    if ( isAbstract ){
		classOut.print("(CVMJavaMethodDescriptor*)"+i+"),\n");
	    } else if ( isNative ){
		if (isClassCNI(c.ci.className)) {
		    String jniName = meth.method.getNativeName(true);
		    String eniName = "CNI" + jniName.substring(5);
		    classOut.print("(CVMJavaMethodDescriptor*)&"+eniName+"),\n");
		} else {
		    classOut.print("(CVMJavaMethodDescriptor*)&"+
			meth.method.getNativeName(true)+"),\n");
		}
	    } else {
		classOut.print("&("+meth.getNativeName()+".u.jmd)),\n" );
	    }
		
	}
	classOut.print("}};\n");
        this.nmethods += nmethod;
	if ( writableMethods ){
	    nClassesWithWritableMethodBlocks += 1;
	    nWritableMethodBlocks += nmethod;
	}
	//
	// now, if there are any exception-catch clauses
	// that need to be represented, print out the table.
	if ( nCheckedExceptions != 0 ){
	    classOut.println("STATIC const CVMUint16 "+thisExceptionName+"[] = {");
	    int j = 0;
	    for ( int i = 0; i < nCheckedExceptions; i++ ){
		if ( j >= 8 ){
		    classOut.print("\n    ");
		    j = 0;
		} else {
		    j += 1;
		}
		classOut.print( checkedExceptionVector[i]+", ");
	    }
	    classOut.println("\n};");
	}
    }

    /*
     * Write out the InnerClassInfo structure. This structure will contain
     * information about all referenced inner classes.
     */
    private void
    writeInnerClasses(CVMClass c) { 
	//Write information about each inner class contained in this class.
	int innerClassCount = c.ci.innerClassAttr.getInnerClassCount();
	InnerClassAttribute ica = c.ci.innerClassAttr;
	classOut.println("    " + innerClassCount + ", {");
	for (int j=0; j < innerClassCount; j++) {
	    // write out the indexes and access flags of this inner class.
	    classOut.print("        {");
	    classOut.print(ica.getInnerInfoIndex(j));
	    classOut.write(',');
	    classOut.print(ica.getOuterInfoIndex(j));
	    classOut.write(',');
	    /* innerNameIndex not supported because utf8 cp entries
	     * are not kept around.
	    classOut.print(ica.getInnerNameIndex(j));
	    */
	    classOut.print('0');
	    classOut.write(',');
	    classOut.print(ica.getAccessFlags(j));
	    classOut.print("},\n");
	}
	classOut.println("    }");
	ninnerclasses += innerClassCount;
    }

    void mungeAllIDsAndWriteExternals(ClassClass classes[]) { 
    }

    private static int CVMDataAccessFlags( int access ){
	int a = 0;
	if ( (access&ACC_PUBLIC) != 0 ) a |= CVM_FIELD_ACC_PUBLIC;
	if ( (access&ACC_PRIVATE) != 0 ) a |= CVM_FIELD_ACC_PRIVATE;
	if ( (access&ACC_PROTECTED) != 0 ) a |= CVM_FIELD_ACC_PROTECTED;
	if ( (access&ACC_STATIC) != 0 ) a |= CVM_FIELD_ACC_STATIC;
	if ( (access&ACC_FINAL) != 0 ) a |= CVM_FIELD_ACC_FINAL;
	if ( (access&ACC_VOLATILE) != 0 ) a |= CVM_FIELD_ACC_VOLATILE;
	if ( (access&ACC_TRANSIENT) != 0 ) a |= CVM_FIELD_ACC_TRANSIENT;
	return a;
    }

    private void declareFieldArray( CVMClass c ){
	// Note: Should deal with > 255 fields.
	// Today is not that day. Just do the simple case.
	FieldInfo m[] = c.ci.fields;
	int nfields = m.length;
	String thisTableName = c.getNativeName()+"_fields";
	String fieldArrayFlavor = "struct CVMFieldArray"+nfields;
	if ( !fieldArraySizes.get( nfields ) ){
	    // need to declare a structure containing a classblock pointer
	    // plus "nfields" CVMFieldBlock structures
	    int i;
	    headerOut.print(fieldArrayFlavor+" {");
	    if (nfields > 256)
		headerOut.println();
	    int numFieldArrays = (nfields-1) / 256 + 1;
	    for (i = 1; i <= numFieldArrays; i++) {
		int nfields2; // # fields in this field array
		if (i == numFieldArrays) {
		    nfields2 = (nfields-1) % 256 + 1;
		} else {
		    nfields2 = 256;
		}
		headerOut.print(" CVMClassBlock const *cb" + i + ";" + 
				" CVMFieldBlock fb" + i + 
				"[" + nfields2 + "];");
		if (nfields > 256) {
		    headerOut.println();
		}
	    }
	    headerOut.println(" };");
	    fieldArraySizes.set( nfields );
	}
	declare( "const "+fieldArrayFlavor+" "+thisTableName, classOut);
	fieldTableName = "&"+thisTableName;
	fieldTableSize = nfields;
    }

    void writeFields(CVMClass c) {
	// Note: Should deal with > 255 fields.
	// Today is not that day. Just do the simple case.
	FieldInfo m[] = c.ci.fields;
	if ( (m == null) || (m.length== 0)){
	    fieldTableName = "0";
	    fieldTableSize = 0;
	    return;
	}
	int nfields = m.length;
	declareFieldArray( c );
	classOut.print( " = {\n");
	for ( int i = 0; i < nfields; i++ ){
	    int fbIndex = i % 256;
	    FieldInfo f = m[i];
	    int typeid = CVMDataType.parseSignature(getUTF(f.type));
	    int nameid = CVMMemberNameEntry.lookupEnter(getUTF(f.name));
	    String nameAndType = "0x"+Integer.toHexString((nameid<<16)+typeid);
	    //System.out.println( nameAndType+" => "+Integer.toHexString(typeid) );
	    int access = CVMDataAccessFlags( f.access );
	    int offset = f.instanceOffset;
	    /* field offsets are always mod 256 and we have to emit a cb
	     * entry before each block of 256 fields. */
	    if (fbIndex == 0) {
		if (i != 0) {
		    classOut.print( "\t},\n");
		}
		classOut.print( "\t&"+classBlockName+",\n\t{\n" );
	    }
	    // static members use simple byte offset
	    // instance members include size of header info
	    if ( f.isStaticMember() ){
		classOut.print("	CVM_INIT_FIELDBLOCK("+nameAndType+","+access+","+fbIndex+", STATIC_OFFSET("+offset+")),\n");
	    } else {
		classOut.print("	CVM_INIT_FIELDBLOCK("+nameAndType+","+access+","+fbIndex+", FIELD_OFFSET("+offset+")),\n");
	    }
	}
	classOut.println("}};");
	this.nfields += nfields;

    }

    private static String methodBlockAddress( MethodInfo mi ){
	CVMMethodInfo em = (CVMMethodInfo)(mi.vmMethodInfo);
	CVMClass      ec = (CVMClass)(mi.parent.vmClass);
	int n = mi.index;
	return "&"+ec.getNativeName()+"_methods.mb"+(n/256+1)+"["+n%256+"]"; // had better not be an interface.
    }

    private static String fieldBlockAddress( FieldInfo f ){
	// pretty disgusting.
	// if this is frequent enough, find a better way!
	ClassInfo c = f.parent;
	FieldInfo cf[] = c.fields;
	int n = f.index;
	//for ( n = 0; cf[n] != f; n++ ) // TRY THIS.
	 //   ;
	// assume we found it.
	String parentName = ((CVMClass)(c.vmClass)).getNativeName();
	return "&"+parentName+"_fields.fb"+(n/256+1)+"["+n%256+"]";
    }

    private static String staticFieldAddress( FieldInfo f ){
	CVMClass c = (CVMClass)(f.parent.vmClass);
	return "&"+staticStoreName+"[STATIC_OFFSET("+f.instanceOffset+")]";
    }

    private static final int unresolvedCVMConstantMap[] = {
	CVM_CONSTANT_Invalid,
	CVM_CONSTANT_Utf8,
	CVM_CONSTANT_Invalid,
	CVM_CONSTANT_Integer|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_Float|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_Long|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_Double|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_ClassTypeID,
	CVM_CONSTANT_StringObj|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_Fieldref,    CVM_CONSTANT_Methodref,
	CVM_CONSTANT_InterfaceMethodref, CVM_CONSTANT_FieldTypeID
    };

    private static final int resolvedCVMConstantMap[] = {
	CVM_CONSTANT_Invalid,
	CVM_CONSTANT_Utf8,	CVM_CONSTANT_Invalid,
	CVM_CONSTANT_Integer|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_Float|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_Long|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_Double|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_ClassBlock|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_StringObj|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_FieldBlock|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_MethodBlock|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_MethodBlock|CVM_CONSTANT_POOL_ENTRY_RESOLVED,
	CVM_CONSTANT_FieldTypeID,
    };

    private static int CVMconstantTag( ConstantObject c, boolean doResolved ){
	int t;
	if ( doResolved && c.isResolved() ){
	    t = resolvedCVMConstantMap[ c.tag ];
	    switch (t){
	    case CVM_CONSTANT_FieldBlock|CVM_CONSTANT_POOL_ENTRY_RESOLVED:{
		/* may be a static of some form. investigate further */
		FieldInfo f = ((FieldConstant)c).find();
		if ( (f.access&ACC_STATIC) != 0 ){
		    if ( (f.access&ACC_REFERENCE) != 0 ){
			t = CVM_CONSTANT_StaticRef|CVM_CONSTANT_POOL_ENTRY_RESOLVED;
		    } else if ( (f.access&ACC_DOUBLEWORD) != 0 ){
			t = CVM_CONSTANT_Static64bit|CVM_CONSTANT_POOL_ENTRY_RESOLVED;
		    } else {
			t = CVM_CONSTANT_Static32bit|CVM_CONSTANT_POOL_ENTRY_RESOLVED;
		    }
		}
		}
		/* else FieldBlock is correct */
		break;

	    case CVM_CONSTANT_FieldTypeID:
		/* I believe this should never happen, as these 
		 * are less than resolved, or deleted entirely.
		 */
		String sig = ((NameAndTypeConstant)c).type.string;
		if ( sig.charAt(0) == SIGC_METHOD ){
		    t = CVM_CONSTANT_MethodTypeID;
		}
		/* else FieldTypeID is correct */
		break;
	    }
	} else {
	    t = unresolvedCVMConstantMap[ c.tag ];
	    switch (t){
	    case CVM_CONSTANT_FieldTypeID:
		String sig = ((NameAndTypeConstant)c).type.string;
		if ( sig.charAt(0) == SIGC_METHOD ){
		    t = CVM_CONSTANT_MethodTypeID;
		}
		/* else FieldTypeID is correct */
		break;
	    }
	}
	return t;
    }

    // return address of this type table.
    private String writeTypeTable( ConstantObject constants[], String name ){
	int clen = constants.length;
	String ttname = name+"_tt";
	auxOut.print("CVMUint8 "+ttname+"[] = { 0,");
	int j = 99;
	for ( int i = 1; i < clen; i+=constants[i].nSlots ){
	    if ( j >= 10 ){
		auxOut.print("\n    ");
		j = 1;
	    } else {
		j += 1;
	    }
	    auxOut.print( CVMconstantTag(constants[i], true)+"," );
	    if ( constants[i].nSlots == 2 )
		auxOut.print(" 0,"); // 2nd word of 2-word constant what should it be?
	}
	auxOut.println("\n};");
	return "&"+ttname;
    }

    // return address of this array Info.
    private String writeArrayInfo( ArrayClassInfo info, String name ){
	String arrayInfoName = name+"_ai";
	classOut.println("STATIC const CVMArrayInfo "+arrayInfoName+" = {");
	classOut.println("\t"+info.depth+", "+info.baseType+",");
	classOut.println("\t(CVMClassBlock*)&"+
		    cbName((CVMClass)info.baseClass.find().vmClass)+",");
	classOut.println("\t(CVMClassBlock*)&"+
		    cbName((CVMClass)info.subarrayClass.find().vmClass));
	classOut.println("};");
	return "&"+arrayInfoName;
    }

 
    // return address of this constant pool.
    private String writeConstants( ConstantObject constants[], String name ){
	boolean doTypeTable = false;
	int clen = constants.length;
	if ( (clen == 0) || (clen == 1 ) ){
	    return "0";
	}
	for ( int i = 1; i < clen; i+=constants[i].nSlots ){
	    if ( ! constants[i].isResolved() ){
		doTypeTable = true;
		break;
	    }
	}
	if ( doTypeTable && !classLoading ){
	    // in future, do something more useful here.
	    System.err.println(Localizer.getString("cwriter.no_class_loading"));
	    // write the type table out anyway, in case
	    // there is someone to look at it at runtime!
	}
	String typeTableName;
	if ( doTypeTable ){
	    typeTableName = writeTypeTable( constants, name );
	}else{
	    typeTableName = "0";
	}
	String constantPoolName = name+"_cp";
	classOut.print( doTypeTable ? "STATIC " : "STATIC const ");
	classOut.println("CVMInt32 "+constantPoolName+"["+(clen)+"] = {");
	classOut.println("	(CVMInt32)"+typeTableName+",");
	for ( int i = 1; i < clen; i+=constants[i].nSlots ){
	    classOut.print("    ");
	    writeConstant( constants[i], classOut );
	    classOut.println(",");
	}
	classOut.println("};");
	nconstants += clen;
	return "&"+constantPoolName;
    }

    private void declareClassblock( CVMClass c, int innerClassCount ){
	String myName = c.getNativeName();
	String classblockFlavor = "struct CVMClassBlock"+innerClassCount;
	if ( !classblockSizes.get( innerClassCount ) ){
	    // need to declare a classblock-like struct with innerClassCount
	    // elements in its CVMInnerClassInfo array
	    if ( innerClassCount < 1 ){
		headerOut.println(classblockFlavor+" { CVMClassBlock classclass; };");
	    } else {
		headerOut.println(classblockFlavor+" { CVMClassBlock classclass; CVMUint32 innerClassesCount; CVMInnerClassInfo innerClasses[ "+(innerClassCount)+"]; };");
	    }
	    classblockSizes.set( innerClassCount );
	    }
	declare( "const "+classblockFlavor+" "+myName+"_Classblock", classOut);
    }

    // helper for classBitMap
    private static int GCbits( FieldInfo slottable[], int first, int last )
    {
	int mapval = 0;
	for ( int i = first; i < last; i++ ){
	    switch ( slottable[i].type.string.charAt(0) ){
	    case 'L':
	    case '[':
		// this is a reference.
		mapval |= 1<<(i-first);
	    }
	}
	return mapval;
    }

    //
    // compute the GC bit map for a class.
    // if it is big, then write it out as well. In any case,
    // return a string to use to initialize the gcMap field in the CVMClassBlock
    //
    private String classBitMap( CVMClass c ){
	if ( c.isArrayClass() ){
	    // either an array of primitive or an array of objects.
	    ArrayClassInfo ac = (ArrayClassInfo)(c.ci);
	    if ( (ac.depth==1) && (ac.baseType!=T_CLASS) )
		return "0"; // contains no references
	    else
		return "1"; // contains nothing but references.
	}
	ClassInfo ci = c.ci;
	FieldInfo ft[] = ci.slottable;
	if ( (ft==null) || (ft.length==0) ) return "0";
	int nslots;
	if ( (nslots=ft.length) <= 30 ){
	    // easy case. Small object. Map can be represented in a word.
	    return "0x"+Integer.toHexString(GCbits(ft,0,nslots)<<2);
	} else {
	    // hard case.
	    String myName = c.getNativeName()+"_gcBits";
	    int    nWords = (nslots+31)/32;
	    classOut.print( "STATIC const struct { CVMUint16 maplen; CVMUint32 map["+nWords+"]; } ");
	    classOut.print( myName );
	    classOut.print( " = { ");
	    classOut.print( nWords );
	    classOut.print( ",\n    {" );
	    for ( int i = 0; i < nslots; i+= 32 ){
		classOut.print( "0x"+Integer.toHexString(GCbits( ft, i, Math.min( i+32, nslots))) );
		classOut.print(", ");
	    }
	    classOut.println("} };");
	    return "((CVMUint32)&"+myName+") + 0x2";
	}
    }

    private void writeClassinfo( int classno, CVMClass c ){

	String	   superName = ( c.ci.superClassInfo == null )? "0" : ("&"+cbName( (CVMClass)(c.ci.superClassInfo.vmClass) ) );
	
	String     myStaticName = (c.nStaticWords==0)?"0":("(CVMJavaVal32*)(&"+staticStoreName+")");
	String     methodTableName;
	int innerClassCount = (c.ci.innerClassAttr == null ?
			       0 : c.ci.innerClassAttr.getInnerClassCount());
	MethodInfo vtbl[] = c.ci.methodtable;
	int	   methodTableCount;
	int	   romindex = c.hasStaticInitializer ? ++maxROMindex : 0;
	String	   gcBits;
	/*
	 * Note: If this is a primitive class, the
	 * typecode is stored in the methodTableCount field.
	 * Since these things cannot be instantiated, they need
	 * no virtual methods.
	 */
	if ( c.isPrimitiveClass() ){
	    methodTableCount = c.typecode();
	    methodTableName = "0";
	} else {
	    // vtbl == null may occur for array classes! 
	    methodTableCount = (vtbl==null)?0:vtbl.length;
	    if (methodTableCount > 0) {
		methodTableName = "&"+c.getNativeName()+"_mt";
	    } else if (c.isArrayClass()) {
		methodTableName = "&java_lang_Object_mt";
	    } else { /* must be an interface */
		methodTableName = "0";
	    }
	}
	// classBitMap may do output, so it needs
	// to be called before declareClassblock!
	gcBits = classBitMap( c );

	declareClassblock( c, innerClassCount );

	int classid = c.classid();
	//System.out.println( c.ci.className+" => "+Integer.toHexString(classid) );

	classOut.println(" = {\n    CVM_INIT_CLASSBLOCK( "+gcBits+", " +
		    classid+",");
	classOut.println("    "+superName+",");
	classOut.println("    "+constantPoolName+", ");
	classOut.println("    "+interfaceTableName+",");
	classOut.println("    "+methodArrayName+",");
	classOut.println("    "+fieldTableName+", "+myStaticName+",");
	classOut.println("    "+constantPoolSize+", "+methodTableSize+", "+
		    fieldTableSize+", "+methodTableCount+
		    ", OBJ_SIZE("+c.instanceSize()+"), "+
		    c.CVMflags()+", "+romindex+",");
	classOut.println("    (CVMClassICell*)(&CVM_classICell["+classno+"]), "+
		    "&CVM_clinitEE["+classno+"],");
	classOut.println("    "+checkedExceptionName+", ");
	if (classDebug && c.ci.sourceFileAttr != null) {
	    classOut.println("    \""+c.ci.sourceFileAttr.sourceName+"\", ");
	} else {
	    classOut.println("    0, ");
	}

	/* write out the methodTablePtr. */
	classOut.println("    " + methodTableName + ",");


	/* write out the InnerClassesInfo */
	if (c.ci.innerClassAttr != null) {
	    classOut.println("    &" + c.getNativeName() +
			     "_Classblock.innerClassesCount),");
	    writeInnerClasses(c);
	} else {
	    classOut.println("    0),");
	}

	classOut.println("};");
    }

    void writeMethodTable(CVMClass c) {
	MethodInfo vtbl[] = c.ci.methodtable;
	int        vtblSize;

	/*
	 * Note: If this is a primitive class, the
	 * typecode is stored in the methodTableCount field.
	 * Since these things cannot be instantiated, they need
	 * no virtual methods.
	 */
	if ( c.isPrimitiveClass() ){
	    return;
	} else {
	    // vtbl == null may occur for array classes! 
	    vtblSize = (vtbl==null)?0:vtbl.length;
	}

	if (vtblSize > 0) {
	    /* write out the methodtable. */
	    String methodTableName = c.getNativeName()+"_mt"+"["+vtblSize+"]";
	    if (c.ci.superClass == null) { // is it java/lang/Object
		// java/lang/Object needs an extern methodTable declaration.
		declare( "const CVMMethodBlock* "+methodTableName, classOut);
	    } else {
		classOut.print( "STATIC const CVMMethodBlock* "+methodTableName);
	    }
	    classOut.println(" = {");
	    for ( int i = 0; i < vtblSize; i++ ){
		classOut.println( "    (CVMMethodBlock*)"+
				  methodBlockAddress( vtbl[i] )+",");
	    }
	    classOut.println("};\n");
	}
    }
	
    /* (These two belong directly in the CCodeWriter) */

    private void
    writeIntegerValue( int v, CCodeWriter out ){
	 // little things make gcc happy.
	 if (v==0x80000000)
	    out.print( "(CVMInt32)0x80000000" );
	else
	     out.print( v );
    }

    private void
    writeLongValue( String tag, int highval, int lowval, CCodeWriter out ){
	out.print( tag );
	out.write( '(' );
	writeIntegerValue( highval, out );
	out.print( ", " );
	writeIntegerValue( lowval, out );
	out.write( ')' );
    }

    private void
    writeConstant( ConstantObject value, CCodeWriter out ){
	switch ( value.tag ){
	case Const.CONSTANT_INTEGER:
	case Const.CONSTANT_FLOAT:
	     writeIntegerValue(((SingleValueConstant)value).value, out);
	     break;
	case Const.CONSTANT_UTF8:
	    out.print("(CVMInt32)");
	    out.write('"');
	    out.print( Util.unicodeToUTF(((UnicodeConstant)value).string ));
	    out.write('"');
	    // also, not sharing here at all.
	    break;
	case Const.CONSTANT_STRING:
	    out.print("(CVMInt32)&"+stringArrayName+"[");
	    out.print( ((StringConstant)value).unicodeIndex );
	    out.write(']');
	    break;
	case Const.CONSTANT_LONG:{
	    DoubleValueConstant dval = (DoubleValueConstant) value;
	    writeLongValue( "LONG",  dval.highVal, dval.lowVal, out );
	    }
	    break;
	case Const.CONSTANT_DOUBLE:{
	    DoubleValueConstant dval = (DoubleValueConstant) value;
	    writeLongValue( "DOUBLE",  dval.highVal, dval.lowVal, out );
	    }
	    break;
	case Const.CONSTANT_CLASS:
	    if ( value.isResolved() ){
		out.print("(CVMInt32)&"+cbName( (CVMClass)(((ClassConstant)value).find().vmClass)));
	    } else {
		int classid = CVMDataType.lookupClassname(
		    getUTF(((ClassConstant)value).name) );
		out.print( "(CVMInt32)");
		out.printHexInt( classid );
	    }
	    break;
	case Const.CONSTANT_METHOD:
	case Const.CONSTANT_INTERFACEMETHOD:
	    if ( value.isResolved() ){
		out.print("(CVMInt32)");
		out.print(methodBlockAddress( ((MethodConstant)value).find()));
	    } else {
		FMIrefConstant ref = (FMIrefConstant)value;
		out.print("SHORTPAIR("+ ref.clas.index+", "+ref.sig.index+")");
	    }
	    break;
	case Const.CONSTANT_FIELD:
	    if ( value.isResolved() ){
		FieldInfo ftarget = ((FieldConstant)value).find();
		ClassInfo ctarget = ftarget.parent;
		out.print("(CVMInt32)");
		if ( ftarget.isStaticMember() && !lossless && !ctarget.vmClass.hasStaticInitializer){
		    out.print(staticFieldAddress( ftarget));
		}else{
		    out.print(fieldBlockAddress( ftarget));
		}
	    } else {
		FMIrefConstant ref = (FMIrefConstant)value;
		out.print("SHORTPAIR("+ ref.clas.index+", "+ref.sig.index+")");
	    }
	    break;
	case CONSTANT_NAMEANDTYPE:
	    NameAndTypeConstant ntcon = (NameAndTypeConstant)value;
	    int nameid = CVMMemberNameEntry.lookupEnter(getUTF(ntcon.name));
	    String sigString = getUTF(ntcon.type);
	    int typeid = (sigString.charAt(0)==SIGC_METHOD)
	       ? (CVMMethodType.parseSignature(sigString).entryNo)
	       : (CVMDataType.parseSignature(sigString));
	    out.printHexInt( (nameid<<16)+typeid );
	    break;
	}
    }

    /*
     * Merge the static store for all classes into one area.
     * Sort static cells, refs first, and assign offsets.
     * Write the resulting master data glob and arrange for
     * it to be copied to writable at startup.
     */
    void writeStaticStore( ClassClass cvec[] ){
	int nclass = cvec.length;
	int nStaticWords = 1; // 1 header word assumed
	int nRef    = 0;
	/*
	 * Count all statics.
	 * Count refs, and count number of words.
	 */
	for ( int cno = 0; cno < nclass; cno++ ){
	    CVMClass c = (CVMClass)(cvec[cno]);
	    c.orderStatics();
	    nRef += c.nStaticRef;
	    nStaticWords += c.nStaticWords;
	}
	/*
	 * Assign offsets and at the same time
	 * record any initial values.
	 * write it later.
	 */
	int refOff = 1;
	int scalarOff = refOff+nRef;

	ConstantObject staticInitialValue[] = new ConstantObject[nStaticWords];
	short          staticInitialSize[]  = new short[nStaticWords];

	for ( int cno = 0; cno < nclass; cno++ ){
	    CVMClass c = (CVMClass)(cvec[cno]);
	    FieldInfo f[] = c.statics;
	    if ( (f == null) || (f.length == 0 ) ) continue; // this class has none
	    int     nFields = f.length;
	    for ( int i = 0; i < nFields; i++ ){
		FieldInfo fld = f[i];
		short     nslots = (short)(fld.nSlots);
		char toptype =  fld.type.string.charAt(0);
		if ( (toptype == 'L') || (toptype=='[')){
		    fld.instanceOffset = refOff;
		    staticInitialValue[refOff] = fld.value;
		    staticInitialSize[refOff] = 1;
		    refOff += 1;
		} else {
		    fld.instanceOffset = scalarOff;
		    staticInitialValue[scalarOff] = fld.value;
		    staticInitialSize[scalarOff] = nslots;
		    scalarOff += nslots;
		}
	    }
	}

	/*
	 * Allocate writable space for the static store.
	 * Start writing the static storage header.
	 * Remember that the first word is the number of refs.
	 * And many of those refs get initialized to zero.
	 * A static final String constant gets initialized with a pointer.
	 */
	auxOut.println("CVMInt32 "+staticStoreName+"["+nStaticWords+"];");
	auxOut.print("STATIC const CVMInt32 "+masterStaticStoreName+"[] = {\n    ");
	auxOut.print( nRef ); // the initial count word.
	auxOut.print(",");
	//System.out.println(nRef + " reference statics, total words "+nStaticWords );

	short nslots = 1;
	int nConstPerLine = 1; // the initial count word.
	for ( int i = 1; i < nStaticWords; i += nslots ){
	    ConstantObject v = staticInitialValue[i];
	    nslots = staticInitialSize[i];
	    if ( v == null ){
		auxOut.print( "0");
		if (nslots == 2 )
		    auxOut.print( ", 0");
	    } else {
		writeConstant( v, auxOut );
	    }
	    auxOut.print(", ");
	    nConstPerLine += nslots;
	    if ( nConstPerLine >= 8 ){
		auxOut.print("\n    ");
		nConstPerLine = 0;
	    }
	}
	auxOut.println("\n};");
	CVMInitInfo.addInfo( "&"+masterStaticStoreName, "&"+staticStoreName, nStaticWords+"*sizeof(CVMInt32)" );
    }

    int maxROMindex = 0;	// 
    private void
    writeRomMap(String name, String sizename) { 
        // The indices are 0 .. maxROMindex.  All classes that are already
        // initialized have index 0.  
        // So we set up an array in which the element 0 indicates "already
        // initialized", and the remainder are zero.

	auxOut.println("CVMUint8 "+name+"[" + (1+maxROMindex) + "];");
	auxOut.println("const int "+sizename+" = "+(1+maxROMindex)+";");

    }



    /*
     * typedef struct {
     *     char *hash;			 * the string for this entry * 
     *     unsigned long info;		 * high bit set if malloc'ed
     * 				 * low 31 bits are the low 31 bits of hash * 
     * #define MALLOCED_FLAG (1u << 31)
     * #define HASHBIT_MASK (MALLOCED_FLAG - 1)
     * } StrIDhashSlot;
     *
     * typedef struct StrIDhash {
     *     int size;			* number of entries *
     *     hash_fn hash;		* hash function for this table *
     *     struct StrIDhash *next;	* next bucket *
     *     short used;			* number of full entries *
     *     short baseid;		* ID for item in slot[0] *
     *     void **params;		* param table, if needed *
     *     StrIDhashSlot slot[1];	* expanded to be number of slots *
     * } StrIDhash;
     */

    private String idTypeDecl[] = {
       "extern void default_hash();",
       "typedef struct {",
       "    const char *hash;",
       "    unsigned long info;",
       "} StrIDhashSlot;",
       "#define STR_ID_HASH(n)  struct { \\",
       "    int size; \\",
       "    void (*hash)(); \\",
       "    void *next; \\",
       "    short used; \\",
       "    short baseid; \\",
       "    HString **params; \\",
       "    StrIDhashSlot slot[ n ]; \\",
       "} "
    };

    /*
     * These are macros to encapsulate some of the messiness
     * of data definitions. They are, perhaps, compiler-specific.
     */
    private static String stdHeader[] = {
	"#define IN_ROMJAVA 1",
	"#include \"javavm/include/defs.h\"",
	"#include \"javavm/include/objects.h\"",
	"#include \"javavm/include/classes.h\"",
	"#include \"javavm/include/interpreter.h\"",
	"#include \"javavm/include/jni_impl.h\"",
	"#include \"javavm/include/typeid.h\"",
	"#include \"javavm/include/typeid_impl.h\"",
	"#include \"javavm/include/string_impl.h\"",
	"#include \"javavm/include/preloader_impl.h\"",
	"#include \"javavm/include/porting/endianness.h\"",

	"#if (CVM_ENDIANNESS == CVM_BIG_ENDIAN)",
	"#define LONG(high, low) high, low",
        "#else",
	"#define LONG(high, low) low, high",
	"#endif /* endianness */",
	"#if (CVM_DOUBLE_ENDIANNESS == CVM_BIG_ENDIAN)",
	"#define DOUBLE(high, low) high, low",
        "#else",
	"#define DOUBLE(high, low) low, high",
	"#endif /* endianness for double's */",
	"#ifdef _MSC_VER",
	"#define STATIC /* static */",
	"#else",
	"#define STATIC static",
	"#endif",

	"#define SCALARTYPE_NO(n) (n)",
	"#define SHORTPAIR(a, b)  (((a)<<16)+(b))", /* may depend on target! */
	//"#define MSIG_NO(n) ((struct methodTypeTableEntry*)&CVMMethodSigTable.typeSig[n])",
	"#define NAME_NO(n) (n)",
	"#define SIG_DETAIL_NO(n) (&msig_details[n])",
	"",
	"extern CVMExecEnv* CVM_clinitEE[];",
	"extern CVMInt32 "+staticStoreName+"[];",
	"extern const CVMMethodBlock* java_lang_Object_mt[];",
	"",

    };

    /* this is target-dependent and belongs in a header file... */
    private static String targetHeader[] = {
    };
	
    private void writeHeaderPrologue( BufferedPrintStream o ){

	for ( int i = 0; i < targetHeader.length; i++ ){
	    o.println( targetHeader[i] );
	}
	for ( int i = 0; i < stdHeader.length; i++ ){
	    o.println( stdHeader[i] );
	}
    }

    private void writePrologue( BufferedPrintStream o ){
	if ( headerFileName != null ){
	    o.println("#include \""+headerFileName+"\"");
	}
    }

    private static String cbName( CVMClass c ){
	return c.getNativeName()+"_Classblock.classclass";
    }

    private void writeClass( int classno, CVMClass c ){

	classOut.println("\n/* ******** Class "+c.getNativeName()+" ******** */");

	classBlockName = cbName( c );

	writeMethods(c);
	// writeInterfaces(c);
	interfaceTableName = (c.inf.iv.length==0)?"0":("(CVMInterfaces*)&"+c.inf.name);

	writeFields(c);

	if (c.isArrayClass()) {
	    ArrayClassInfo info = (ArrayClassInfo)c.ci;
	    constantPoolName = writeArrayInfo(info, c.getNativeName() );
	    constantPoolSize = 0;
	} else {
	    if ( doShared ) {
		constantPoolName = sharedConstantPoolName;
		constantPoolSize = sharedConstantPoolSize;
	    } else {
		ConstantObject cpool[] = c.ci.constants;
		constantPoolName = writeConstants(cpool, c.getNativeName() );
		constantPoolSize = cpool.length;
	    }
	}

	writeMethodTable(c);
	writeClassinfo(classno, c);

    }

    private void processStrings( ConstantObject cp[] ){
	if ( cp == null ) return;
	int n = cp.length;
	for ( int i = 1; i < n; ){
	    ConstantObject obj = cp[i];
	    if ( obj instanceof StringConstant ){
		stringTable.intern( (StringConstant) obj );
	    }
	    i += obj.nSlots;
	}
    }

    private void classProcessStrings( ClassInfo c, boolean doShared ){
	FieldInfo f[] = c.fields;
	int fieldCount = ( f== null ) ? 0 : f.length;
	if ( ! doShared ){
	    processStrings( c.constants );
	}
	//
	// make sure that strings appearing ONLY as static final
	// initializer values get processed, too.
	//
	for ( int i = 0; i < fieldCount; i++ ){
	    FieldInfo fi = f[i];
	    if ( fi.isStaticMember() && fi.value instanceof StringConstant ){
		stringTable.intern( (StringConstant) (fi.value) );
	    }
	}
    }

    private void instantiateAllStrings(){
	njavastrings = stringTable.writeStrings( auxOut, masterStringArrayName);
	auxOut.println("struct java_lang_String "+stringArrayName+"["+njavastrings+"];");
	auxOut.println("const int CVM_nROMStrings = "+njavastrings+";");
	CVMInitInfo.addInfo( "&"+masterStringArrayName, "&"+stringArrayName, njavastrings+"*sizeof(struct java_lang_String)" );

	stringTable.writeStringInternTable( auxOut, "CVMInternTable", stringArrayName ); 
    }

    /*
     * Our representation of method types has four variations:
     *	0: inline form, inline details;
     *  1: inline form, outboard details;
     *  2: outboard form, inline details (rare);
     *  3: outboard form, outboard details, too.
     * For our own convenience, we are going to segregate them
     * by type. This affects the indexing scheme, so we must
     * assign that, too, after the sort.
     */
    private static int
    variety( CVMMethodType mt ){
	int v = 0;
	if ( mt.form.nSyllables> CVMSigForm.SYLLABLESPERDATUM) v+= 2;
	if ( mt.nDetails > 2 ) v += 1;
	return v;
    }

    private void
    enumerateMethodTypes(){
	methodTypes = new CVMMethodType[4];
	nMethodTypes = new int[4];
	Enumeration e = CVMMethodType.methodTypes.elements();
	while ( e.hasMoreElements() ){
	    CVMMethodType mt = (CVMMethodType)(e.nextElement() );
	    int v = variety( mt );
	    mt.listNext = methodTypes[v];
	    methodTypes[v] = mt;
	    nMethodTypes[v] += 1; }
	int curTypeNo = 0;
	for ( int i = 0; i < 4; i++ ){
	    for( CVMMethodType mt = methodTypes[i]; mt!= null; mt = mt.listNext ){
		mt.entryNo = curTypeNo++;
	    }
	}
	totalMethodTypes = curTypeNo;
    }

    /*
     * The CVM type structures are built up as a side effect of
     * turning signatures and names into 16-bit cookies while writing out
     * other structures. Thus they must be written AFTER all the entries are
     * computed. Fortunately there will be no forward-references to them,
     * as all references are by index values, not addresses.
     */
    private void writeCVMNameAndTypeTables(){
	writeCVMpkgHeader();
	writeCVMDataTypeTable();
	writeCVMpkg();
	writeCVMMethodTypeTable();
	writeCVMMemberNameTable();
    }

    /*
     * The value-type table is one of the larger type structures, as it also
     * holds class names. It has variant records. It contains an internal linked
     * list for lookup-by-name. It has reference counts, but we will set these
     * to the magic maximum value so that this can be READONLY.
     */
    private void writeCVMDataTypeTable(){
	Vector allDataTypes = CVMDataType.dataTypes; 
	int    nData = allDataTypes.size();
	/*
	 * May need to do an iteration through to determine the layout, in case
	 * we have some trouble initializing union types and have to do more
	 * specific declarations!
	 */
	auxOut.println("STATIC struct scalarTableSegment * dummyNextCell;");
	CVMInitInfo.addInfo( "NULL", "&dummyNextCell", "sizeof(dummyNextCell)" );
	auxOut.println("const struct { SCALAR_TABLE_SEGMENT_HEADER struct scalarTableEntry data["+nData+"]; }\nCVMFieldTypeTable = {" );
	auxOut.println("\t"+(CVMTypeCode.CVMtypeLastScalar+1)+", 0, -1,"+nData+", &dummyNextCell, {" );
	for ( int i = 0; i < nData; i++ ){
	    CVMDataType o = (CVMDataType)allDataTypes.elementAt( i );
	    /* Yet more debugging output ....
		if ( o != null && i != o.entryNo ){
		    System.out.println("Datatype entry "+o+" misplaced: is at vector["+i+"] and expected at ["+o.entryNo+"]");
		}
	    */
	    if ( o instanceof CVMClassDataType ){
		CVMClassDataType ctype = (CVMClassDataType) o;
		String next;
		next = (ctype.next == null ) ? "TYPEID_NOENTRY" : ("SCALARTYPE_NO("+ctype.next.typeNo+")");
		// int nameHash = CVMDataType.computeHash(ctype.classInPackage);
		auxOut.println("	INIT_CLASSTYPE( "+next+", PKG_NO("+ctype.pkg.entryNo+"), \""+ctype.classInPackage+"\"),");
	    } else {
		CVMArrayDataType atype = (CVMArrayDataType) o;
		String next;
		next = (atype.next == null ) ? "TYPEID_NOENTRY" : ("SCALARTYPE_NO("+atype.next.typeNo+")");
		auxOut.println("	INIT_ARRAYTYPE( "+next+", "+atype.depth+", "+atype.baseType+"),");
	    }
	}
	auxOut.println("}};");

    }

    /*
     * The package name data structures and hash tables have to go after the
     * general value-type table, as it contains little hash tables 
     * of pointers to the latter.
     * These are used for lookup by class name.
     */
    private void writeCVMpkgHeader(){
	int    nPkg = CVMpkg.pkgVector.size();
	auxOut.println("STATIC struct pkg CVM_ROMpackages["+(nPkg)+"];" );
	auxOut.println("#define PKG_NO(n) &CVM_ROMpackages[n]");
    }

    private void writeCVMpkg(){
	Vector allPackages = CVMpkg.pkgVector; 
	int    nPkg = allPackages.size();

	auxOut.println("STATIC const struct pkg CVM_ROMpackagesMaster["+(nPkg)+"] = {" );
	for ( int i= 0; i < nPkg; i++ ){
	    CVMpkg p = (CVMpkg) (allPackages.elementAt(i));
	    writeAnCVMpkg( p );
	    /* DEBUG
		if ( p.entryNo != i ){
		    System.out.println("Package entry "+p.pkgName+" misplaced: is at vector["+i+"] and expected at ["+p.entryNo+"]");
		}
	    */
	}
	auxOut.println("};");

	CVMInitInfo.addInfo( "CVM_ROMpackagesMaster", "CVM_ROMpackages", nPkg+"*sizeof(struct pkg)" );

	// that was the packages themselves.
	// this is the hash table that is the basis of looking them up.
	// this needs to be writable, since we add at the front of the
	// linked list.
	auxOut.print("STATIC const struct pkg *CVM_PkgHashtableMaster["+CVMpkg.NPACKAGEHASH+"] = {");
	for ( int i = 0; i< CVMpkg.NPACKAGEHASH; i++ ){
	    CVMpkg p = CVMpkg.packages[i];
	    if ( i%5 == 0 ){
		auxOut.print("\n\t");
	    }
	    auxOut.print( (p==null)?"0, ":("PKG_NO("+(p.entryNo)+"), "));
	}
	auxOut.println("\n};");
	auxOut.println("struct pkg *CVM_pkgHashtable["+CVMpkg.NPACKAGEHASH+"];");
	CVMInitInfo.addInfo( "CVM_PkgHashtableMaster", "CVM_pkgHashtable", CVMpkg.NPACKAGEHASH+"*sizeof(struct pkg*)" );

	auxOut.println("struct pkg *CVMnullPackage = PKG_NO(0);");
    }

    private void writeAnCVMpkg( CVMpkg p ){
	String next = (p.next==null)?"0" : ("PKG_NO("+p.next.entryNo+")");
	//int    nameHash = CVMDataType.computeHash(p.pkgName);
	auxOut.print("    { \""+p.pkgName+"\", "+next+", INROM, MAX_COUNT, {");
	for( int i = 0; i < CVMpkg.NCLASSHASH; i++ ){
	    String hashp = (p.typeData[i]==null)? "TYPEID_NOENTRY" : ("SCALARTYPE_NO("+p.typeData[i].typeNo+")");
	    if ( i%3==0){
		auxOut.print("\n\t");
	    }
	    auxOut.print(hashp);
	    auxOut.print(", ");
	}
	auxOut.println("\n    }},");
    }

    /*
     * The CVMMethodTypes are complex, multi-part beasts.
     * The first part is the Form or Terse Signature,
     * Then we need the details array
     * And finally the table of method types. There are 4 kinds of them!:
     *	0: inline form, inline details;
     *  1: inline form, outboard details;
     *  2: outboard form, inline details (rare);
     *  3: outboard form, outboard details, too.
     * They are all the same size!
     */
    public void writeCVMMethodTypeTable(){
	// The first part is the Form or Terse Signature.
	writeCVMForms();

	// then the details array.
	writeCVMMethodDetails();

	// write the indirect dummy.
	auxOut.println("STATIC struct methodTypeTableSegment *methodTypeDummy;");
	CVMInitInfo.addInfo( "NULL", "&methodTypeDummy", "sizeof(methodTypeDummy)");

	// write the specific declaration we need for this
	// set of types.
	auxOut.println("const struct { METHOD_TABLE_SEGMENT_HEADER");
	for ( int i = 0; i < 4; i++ ){
	    if ( nMethodTypes[i] == 0 ) continue;
	    auxOut.print("    struct { METHOD_TYPE_HEADER ");
	    if ( (i&2) == 0 ){
		// inline form
		auxOut.print("unsigned int formdata; ");
	    } else {
		// outline form
		auxOut.print("struct sigForm * formp; ");
	    }
	    if ( (i&1) == 0 ){
		// inline details
		auxOut.print("unsigned short data[2]; ");
	    } else {
		// outline details
		auxOut.print("const short* datap; ");
	    }
	    auxOut.print(" }typesig");
	    auxOut.println( i+"["+nMethodTypes[i]+"];");
	}
	auxOut.println("} CVMMethodTypeTable = {");

	// now the header info
	auxOut.println("    0, 0, -1, "+totalMethodTypes+", &methodTypeDummy,");

	// and the actual data.
	for ( int i = 0; i < 4; i++ ){
	    if ( nMethodTypes[i] == 0 ) continue;
	    auxOut.println("    {");
	    for ( CVMMethodType m = methodTypes[i]; m != null; m = m.listNext ){
		String next = (m.next==null) ? "TYPEID_NOENTRY" : String.valueOf(m.next.entryNo);
		auxOut.print("	{"+next+", MAX_COUNT, ROMSTATE "+(m.form.nSyllables-2));
		CVMSigForm f = m.form;
		if ( (i&2) == 0 ){
		    // inline form
		    auxOut.print(", 0x"+Integer.toHexString( f.data[0] ));
		} else {
		    // outline form
		    auxOut.print(", (struct sigForm*)(&CVM_ROMforms.form"+f.listNo+"["+f.entryNo+"])");
		}
		if ( (i&1) == 0 ){
		    // inline details
		    auxOut.print(", {");
		    for ( int j=0; j<2; j++){
			auxOut.print( (j<m.nDetails)?m.details[j]:0 );
			auxOut.write(',');
		    }
		    auxOut.print("}");
		} else {
		    // outline details
		    auxOut.print(", SIG_DETAIL_NO("+(m.detailOffset)+")");
		}
		auxOut.println("},");
	    }
	    auxOut.println("    },");
	}
	auxOut.println("};");

	/*
	 * A signature search begins with a hash.
	 * The hash table must be writeable.
	 */
	auxOut.println("CVMUint16	CVMMethodTypeHash[NMETHODTYPEHASH];");
	auxOut.print("STATIC CVMUint16 CVMMethodTypeHashMaster[NMETHODTYPEHASH]={");
	for ( int i = 0 ; i < CVMMethodType.NHASH; i++ ){
	    CVMMethodType mt = CVMMethodType.hashTable[i];
	    if ( i%4 == 0 ){
		auxOut.print("\n	");
	    }
	    if ( mt == null ){
		auxOut.print("TYPEID_NOENTRY, ");
	    } else {
		auxOut.print( mt.entryNo );
		auxOut.print(", ");
	    }
	}
	auxOut.println("};");
	CVMInitInfo.addInfo( "CVMMethodTypeHashMaster", "CVMMethodTypeHash", CVMMethodType.NHASH+"*sizeof(CVMMethodTypeHash[0])" );
    }

    /*
     * Method signature varieties 1 & 3 have details written in an
     * array-of-short.
     */
    private void
    writeCVMMethodDetails( ){
	if (nMethodTypes[1]+nMethodTypes[3] == 0 ) return; // nothing to do.
	// We could search for shared sub-sequences of detail sequences.
	// On the other hand, these so seldom occur at all
	// that it would make very little difference.
	int curOffset = 0;
	auxOut.print("STATIC const short msig_details[] = {");
	for ( int i = 1; i<4; i+= 2 ){ // i.e. for i in {1,3}
	    for ( CVMMethodType m = methodTypes[i]; m!=null; m = m.listNext ){
		m.detailOffset = curOffset;
		int ndetail = m.nDetails;
		int details[] = m.details;
		for ( int j = 0; j < ndetail; j++ ){
		    if ( curOffset%6==0){
			auxOut.print("\n	");
		    }
		    auxOut.print(details[j]);
		    auxOut.write(',');
		    auxOut.write(' ');
		    curOffset+= 1;
		}
	    }
	}
	auxOut.println("\n};");
    }


    private void writeCVMForms(){
	/*
	 * Declare a data structure to contain all the "big" forms.
	 * Since they are of varying lengths, this will require a
	 * custom declaration.
	 */
	auxOut.println("STATIC const struct {");
	for ( int i = 1; i < CVMSigForm.listLength.length; i++ ){
	    if ( CVMSigForm.listLength[i] == 0 ) continue; // never mind...
	    auxOut.println("	DECLARE_SIG_FORM_N( /**/, "+(i+1)+") form"+i+"["+CVMSigForm.listLength[i]+"];");
	}
	auxOut.println("} CVM_ROMforms = {");
	/*
	 * Now supply the data. They're all on a single linked list,
	 * which we construct (last-to-first) as we go.
	 */
	String link = "0";
	for ( int i = 1; i < CVMSigForm.listLength.length; i++ ){
	    int n = CVMSigForm.listLength[i];
	    if ( n == 0 ) continue; // never mind...
	    String segname = "&CVM_ROMforms.form"+i;
	    auxOut.println("    {");
	    int fno = 0;
	    for ( CVMSigForm f = CVMSigForm.formList[i]; f!= null; f = f.listNext ){
		auxOut.print("	");
		auxOut.print("{ (struct sigForm*)"+link+", MAX_COUNT, ROMSTATE ");
		auxOut.print( (f.nSyllables-2)+", { ");
		int nWordsLess1 = f.data.length-1;
		for ( int k=0; k<=nWordsLess1; k++ ){
		    auxOut.printHexInt( f.data[k] );
		    if ( k<nWordsLess1)
			auxOut.print(", ");
		}
		auxOut.println("} },");
		//f.listNo  = i;
		f.entryNo = fno;
		link = segname+"["+(fno++)+"]";
	    }
	    auxOut.println("    },");
	}
	auxOut.println("};");

	/*
	 * Finally, the root of the form list.
	 */
	auxOut.println("\nstruct sigForm *	CVMformTable;");
	auxOut.println("STATIC const struct sigForm * CVMformTableMaster = (struct sigForm *)"+link+";");
	CVMInitInfo.addInfo( "&CVMformTableMaster", "&CVMformTable", "sizeof(CVMformTableMaster)" );
    }

    private void
    writeCVMMemberNameTable(){
	int ndata = CVMMemberNameEntry.tableSize();
	auxOut.println("STATIC struct memberNameTableSegment * dummyNextMembernameCell;");
	CVMInitInfo.addInfo( "NULL", "&dummyNextMembernameCell", "sizeof(dummyNextMembernameCell)" );
	auxOut.println("const struct { MEMBER_NAME_TABLE_SEGMENT_HEADER struct memberName data["+ndata+"]; }\nCVMMemberNames = {" );
	auxOut.println("    0, 0, -1, "+ndata+", &dummyNextMembernameCell, {" );
	for ( int i = 0; i < ndata; i++ ){
	    CVMMemberNameEntry mne = (CVMMemberNameEntry)(CVMMemberNameEntry.table.elementAt( i ));
	    String next    = (mne.next==null) ? "TYPEID_NOENTRY" : ("NAME_NO("+mne.next.entryNo+")");
	    auxOut.println("\t{"+next+", MAX_COUNT, ROMSTATE \""+mne.name+"\"},");
	}
	auxOut.println("}};");


	auxOut.println("CVMUint16 CVMMemberNameHash[NMEMBERNAMEHASH];");
	auxOut.print("STATIC const CVMUint16 CVMMemberNameHashMaster[NMEMBERNAMEHASH]={");
	for ( int i = 0 ; i < CVMMemberNameEntry.NMEMBERNAMEHASH; i++ ){
	    CVMMemberNameEntry mne = CVMMemberNameEntry.hash[i];
	    if ( i%4 == 0 ){
		auxOut.print("\n	");
	    }
	    if ( mne == null ){
		auxOut.print("TYPEID_NOENTRY, ");
	    } else {
		auxOut.print( mne.entryNo );
		auxOut.print(", ");
	    }
	}
	auxOut.println("};");
	CVMInitInfo.addInfo( "CVMMemberNameHashMaster", "CVMMemberNameHash", CVMMemberNameEntry.NMEMBERNAMEHASH+"*sizeof(CVMMemberNameHash[0])" );
    }

    public void writeClassList( ){
	boolean  foundFirstNonPrimitive = false;
	boolean  foundFirstVector = false;
	boolean  foundLastVector   = false;
	int	 firstNonPrimitive = 0;
	int	 firstVector = 0;
	int 	 lastVector = 0;
	int n = (classes == null) ? 0 : classes.length;

	auxOut.println("const int CVM_nROMClasses = "+n+";");
	auxOut.println("struct java_lang_Class CVM_ROMClasses["+n+"];");
	auxOut.println("STATIC const struct java_lang_Class CVM_ROMClassesMaster[] = {" );
	for ( int i = 0; i < n; i++ ){
	    int hashCode = 0;
	    CVMClass c = classes[i];
	    int classid = c.classid();
	    auxOut.println("    CVM_CLASS_INIT(&"+cbName(c)+","+hashCode+"),");
	    if ( (! foundFirstNonPrimitive ) & ! c.isPrimitiveClass() ){
		firstNonPrimitive = i;
		foundFirstNonPrimitive = true;
	    }
            int classid_depth = classid & CVMTypeCode.CVMtypeArrayMask;
	    if ((classid_depth != 0) && (classid_depth != CVMTypeCode.CVMtypeBigArray)) {
		if ( !foundFirstVector ){
		    firstVector = i;
		    foundFirstVector = true;
		}
		if ( ((classid>>CVMTypeCode.CVMtypeArrayShift) > 1)
		    && !foundLastVector ){
			lastVector = i-1;
			foundLastVector = true;
		}
	    }
	}
        if (!foundFirstVector) {
            // If, by now, we've not found the first vector, then there is no
            //     preloaded array type.  Point the first vector to the end of
            //     the array:
            firstVector = n;
            foundFirstVector = true;
            lastVector = n;
            foundLastVector = true;
        } else if (!foundLastVector) {
            // If we get here, then we must have found the first vector but
            //     have not found the last vector.  This means that there are
            //     array types of depth 1 in the list but no array types of
            //     depth greater than 1.  Point the last vector to the last
            //     entry in the list:
            lastVector = n - 1;
            foundLastVector = true;
        }
	auxOut.print("};\nconst int CVM_firstROMVectorClass = ");
	auxOut.print(firstVector);
	auxOut.print(";\nconst int CVM_lastROMVectorClass = ");
	auxOut.print(lastVector);
	auxOut.print(";\nconst int CVM_firstROMNonPrimitiveClass = ");
	auxOut.print(firstNonPrimitive);
	auxOut.println(";");
	CVMInitInfo.addInfo( "CVM_ROMClassesMaster", "CVM_ROMClasses", n+"*sizeof(struct java_lang_Class)" );
	 
	//
	// array of stack map indirect cells
	//
	declare("CVMStackMaps*	"+stackmapIndirect+"["+stackmapIndex+"]", auxOut );
	auxOut.println(";");
	CVMInitInfo.addInfo( "NULL", stackmapIndirect, stackmapIndex+"*sizeof(CVMStackMaps*)" );

	declare("const CVMClassICell CVM_classICell[]", auxOut );
	auxOut.println(" = {");
	for ( int i = 0; i < n; i++ ){
	    auxOut.println("    { (CVMObject*)(&CVM_ROMClasses["+i+"])}," );
	}
	auxOut.println("};");
	auxOut.println("CVMExecEnv* CVM_clinitEE["+n+"];");
    }
    
    public boolean writeClasses( ConstantPool consts, 
				 ConstantPool sharedconsts ){

	// getClassVector sorts the classes with
	// super before sub, for reasonable processing without recursion.
	ClassClass arrayOfClasses[] = ClassClass.getClassVector( classMaker );
	ClassClass.setTypes();
	doShared = sharedconsts!=null;
	int nClasses = arrayOfClasses.length;

	classes = new CVMClass[nClasses];

	/*
	 * sort the array by classid. This makes it easier do lookups
	 * at runtime.
	 * The mechanics of this whole writer assume a single
	 * ordering for these classes, as they are often
	 * referred to by their class number.
	 * This is it.
	 * To avoid finding the primitive classes at all, they are
	 * placed at the very end.
	 */
	System.arraycopy( arrayOfClasses, 0, classes, 0, nClasses );
        Arrays.sort(classes, (java.util.Comparator)classMaker);

	enumerateMethodTypes();// after all classes processed once by classMaker
				  // ... but before we write the classes.
	
	// write out some constant pool stuff here,
	// if we're doing one big shared one...
	// gutted out for now...
	if ( doShared ) processStrings( sharedconsts.getConstants() );

	if (verbose)
	    System.out.println(Localizer.getString("cwriter.writing_classes"));
	try {
	    //mungeAllIDsAndWriteExternals(classes);
	    for ( int i = 0; i < nClasses; i++ ){
		classProcessStrings( classes[i].ci, doShared );
	    }
	    instantiateAllStrings();
	    CVMInterfaceMethodTable.writeInterfaceTables(classes, auxOut, headerOut);
	    writeStaticStore( classes );
	    for ( int i = 0; i < nClasses; i++ ){
		writeClass( i, (CVMClass)classes[i] );
		if ( segmentedOutput && ( curClasses++ >= maxClasses) ){
		    curClasses = 0;
		    openNextClassFile();
		}
	    }
	    writeClassList();

	    writeCVMNameAndTypeTables();

	    writeRomMap( "CVMROMClassesRuntimeFlags", 
			 "CVM_nROMClassesRuntimeFlags");
	    CVMInitInfo.write( auxOut, "CVM_preloaderInitTriple",
			       "CVM_preloaderInitMap",
			       "CVM_nPreloaderMapEntries");
	} catch (Exception e) { 
	    classOut.flush();
	    headerOut.flush();
	    System.out.println(e);
	    e.printStackTrace(System.out);
	    formatError = true;
	}
	return (!formatError) && (! classOut.checkError()) && (! headerOut.checkError());
    }

    public boolean writeClasses( ConstantPool consts ) {
	return writeClasses(consts, null);
    }


    public void printSpaceStats( java.io.PrintStream o ){
	//ClassClass classes[] = ClassClass.getClassVector( classMaker );
	o.println("	"+Localizer.getString("cwriter.total_classes", Integer.toString(classes.length)));
	o.println("		"+Localizer.getString("cwriter.method_blocks", Integer.toString(nmethods)));
	if ( nWritableMethodBlocks != 0 ){
	    o.println("		... of which "+nWritableMethodBlocks+" blocks are writable" );
	    o.println("		... for "+nClassesWithWritableMethodBlocks+" classes");
	}
	o.println("		"+Localizer.getString("cwriter.bytes_java_code", Integer.toString(ncodebytes)));
	o.println("		"+Localizer.getString("cwriter.catch_frames", Integer.toString(ncatchframes)));
	o.println("		"+Localizer.getString("cwriter.field_blocks", Integer.toString(nfields)));
	o.println("		"+Localizer.getString("cwriter.inner_classes", Integer.toString(ninnerclasses)));
	o.println("		"+Localizer.getString("cwriter.constant_pool_entries", Integer.toString(nconstants)));
	o.println("		"+Localizer.getString("cwriter.java_strings",Integer.toString(njavastrings)));
    }

}
