/*
 * -----------------------------------------------------------------
 *  cstates.c :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)cstates.c	1.20 00/11/01
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/cstates.h"
#include "javavm/include/globals.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/utils.h"
#include "javavm/include/porting/doubleword.h"

void
CVMcsInit(CVMCState *cs, const char *nm, int rank)
{
    cs->request = CVM_FALSE;
#ifdef CVM_DEBUG
    cs->count = 0;
    cs->reached = CVM_FALSE;
#endif
    cs->name = nm;
    CVMsysMutexInit(&cs->mutex, nm, rank);
    CVMcondvarInit(&cs->consistentCV, &cs->mutex.rmutex.mutex);
    CVMcondvarInit(&cs->resumeCV, &cs->mutex.rmutex.mutex);
}

void
CVMcsDestroy(CVMCState *cs)
{
    CVMcondvarDestroy(&cs->consistentCV);
    CVMcondvarDestroy(&cs->resumeCV);
    CVMsysMutexDestroy(&cs->mutex);
}

/*
 * Bring all other threads to a consistent state.
 */
void
CVMcsReachConsistentState(CVMExecEnv *ee, CVMCStateID csID)
{
    CVMCState *cs = CVM_CSTATE(csID);

    /* Check that current thread is consistent */
    CVMassert(CVMcsIsConsistent(CVM_TCSTATE(ee, csID)));
    /* and that a request hasn't already been made */
    CVMassert(!CVMcsCheckRequest(cs));

    CVMsysMutexLock(ee, &cs->mutex);

#ifdef CVM_DEBUG
    ++cs->count;
    CVMassert(cs->requester == NULL);
    cs->requester = ee;

    CVMassert(!cs->reached);
    CVMtraceCS(("<%d> CS<%d>: ee=0x%p, reaching consistent state %s\n",
	ee->threadID, cs->count, ee, cs->name));
#endif

    cs->inconsistentThreadCount = 0;
    cs->request = CVM_TRUE;

    /*
     * NOTE: on a multi-processor system, we might need to use a
     * memory barrier to make sure changes are visible to all threads.
     */

    /*
     * After the request is set, threads will only transition from
     * unsafe to safe, so it is OK to check the per-thread state
     * without a lock and before we suspend it.
     */

    /* Roll other threads forward if necessary */
 
    CVMsysMutexLock(ee, &CVMglobals.threadLock);
    {
        CVM_WALK_ALL_THREADS(ee, targetEE, {
	    if (targetEE != ee) {
		CVMTCState *tcs = CVM_TCSTATE(targetEE, csID);
		CVMUint32 suspended =
		    targetEE->threadState & CVM_THREAD_SUSPENDED;

		/*
		 * Check the state while the thread is running.
		 * State variable needs to be volatile and consistent
		 * with regard to cs->request (MP systems).
		 *
		 * A suspended thread is consistent, even though
		 * its state may be in flux if it was at a
		 * transition.
		 */
		tcs->wasConsistent = CVMcsIsConsistent(tcs);
		if (suspended || tcs->wasConsistent) {
		    /* thread is OK */
		    CVMtraceCS(("<%d> CS<%d>: %s thread <%d>\n",
			ee->threadID, cs->count,
			suspended ? "suspended" : "consistent",
			targetEE->threadID));
		} else {
		    ++cs->inconsistentThreadCount;
		    CVMtraceCS(("<%d> CS<%d>: inconsistent thread <%d>\n",
			ee->threadID, cs->count, targetEE->threadID));
		}
	    }
        });
    }
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);

    {
	CVMBool interrupted = CVM_FALSE;
	while (cs->inconsistentThreadCount > 0) {
	    if (!CVMsysMutexWait(ee, &cs->mutex, &cs->consistentCV,
		    CVMlongConstZero()))
	    {
		interrupted = CVM_TRUE;
	    }
	}
	if (interrupted) {
	    CVMtraceCS(("<%d> CS<%d>: interrupted!\n",
		ee->threadID, cs->count));
	    CVMthreadInterruptWait(CVMexecEnv2threadID(ee));
	}
    }

    CVMtraceCS(("<%d> CS<%d>: done waiting for count to reach 0\n",
	ee->threadID, cs->count));

#ifdef CVM_DEBUG
    cs->reached = CVM_TRUE;
#endif
    CVMsysMutexUnlock(ee, &cs->mutex);

    CVMtraceCS(("<%d> CS<%d>: mutex unlocked\n", ee->threadID, cs->count));

}

/*
 * Allow other threads to become inconsistent again.
 */
void
CVMcsResumeConsistentState(CVMExecEnv *ee, CVMCStateID csID)
{
    CVMCState *cs = CVM_CSTATE(csID);
    CVMtraceCS(("<%d> CS<%d>: ee=0x%p, resuming consistent state %s\n",
	ee->threadID, cs->count, ee, cs->name));

#ifdef CVM_DEBUG
    cs->reached = CVM_FALSE;
    CVMassert(cs->requester == ee);
    cs->requester = NULL;
#endif
    CVMsysMutexLock(ee, &cs->mutex);
    cs->request = CVM_FALSE;
    CVMcondvarNotifyAll(&cs->resumeCV);
    CVMsysMutexUnlock(ee, &cs->mutex);
}

/*
 * Rendezvous with the requestor thread.  We block only if "block" is
 * set to true.
 */
void
CVMcsRendezvous(CVMExecEnv *ee, CVMCState *cs, CVMTCState *tcs, CVMBool block)
{
    CVMBool isConsistent = tcs->isConsistent;
#ifdef CVM_DEBUG
    CVMUint32 count;

    /* The requester should not be making transitions during the request */
    CVMassert(cs->requester != ee);
#endif

    /* must block if inconsistent */
    CVMassert(block != isConsistent);

    /* Force safe-->unsafe transition to appear safe */
    tcs->isConsistent = CVM_TRUE;

    CVMsysMutexLock(ee, &cs->mutex);

#ifdef CVM_DEBUG
    count = cs->count;
    CVMtraceCS(("<%d> CS<%d>: ee=0x%p, performing rendezvous for "
	"consistent state %s\n",
	ee->threadID, count, ee, cs->name));
#endif

    CVMtraceCS(("<%d> CS<%d>: ee=0x%p, isConsistent %d (%d)"
	" block? %d request? %d\n",
	ee->threadID, count, ee, isConsistent, tcs->wasConsistent,
	block, CVMcsCheckRequest(cs)));
    if (!tcs->wasConsistent) {
	CVMassert(CVMcsCheckRequest(cs));
	CVMassert(cs->inconsistentThreadCount > 0);
	tcs->wasConsistent = CVM_TRUE;
	--cs->inconsistentThreadCount;
	CVMtraceCS(("<%d> CS<%d>: decrementing count to %d \n",
	    ee->threadID, count, cs->inconsistentThreadCount));
	if (cs->inconsistentThreadCount == 0) {
	    CVMtraceCS(("<%d> CS<%d>: performing notify\n",
		ee->threadID, count));
	    /* notify condition variable thread */
	    CVMcondvarNotify(&cs->consistentCV);
	}
    }

#ifdef CVM_DEBUG
    /* still on current cycle? */
    CVMassert(count == cs->count);
#endif

	/*Added by Kunal - for NoHeapRealtimeThread*/
	if(CVMisNoHeapRealtimeThread(ee))
    {
		/*isConsistent = CVM_TRUE;
    	tcs->isConsistent = CVM_TRUE; */
		block = CVM_FALSE;
	}

    if (block) {
	CVMBool interrupted = CVM_FALSE;
	while (CVMcsCheckRequest(cs)) {
	    CVMtraceCS(("<%d> CS<%d>: blocking\n",
		ee->threadID, cs->count));
	    if (!CVMsysMutexWait(ee, &cs->mutex, &cs->resumeCV,
		    CVMlongConstZero()))
	    {
		interrupted = CVM_TRUE;
	    }
	}
	if (interrupted) {
	    CVMtraceCS(("<%d> CS<%d>: interrupted!\n",
		ee->threadID, cs->count));
	    CVMthreadInterruptWait(CVMexecEnv2threadID(ee));
	}
	CVMtraceCS(("<%d> CS<%d>: done blocking\n", ee->threadID, count));
    } else {
	CVMtraceCS(("<%d> CS<%d>: not blocking\n", ee->threadID, count));
    }
    tcs->isConsistent = isConsistent;
    CVMtraceCS(("<%d> CS<%d>: isConsistent reset to %d\n",
	ee->threadID, count, isConsistent));

	/*Added by Kunal - for NoHeapRealtimeThread*/
    if(CVMisNoHeapRealtimeThread(ee))
    {
		isConsistent = CVM_TRUE;
    	/*tcs->isConsistent = CVM_TRUE;
		block = CVM_FALSE;*/
    }
#ifdef CVM_DEBUG
    /* Shouldn't be inconsistent if consistent state was reached */
	/*printf("cstates.c::Rendezvous() isConsistent %d cs->reached %d, block %d,
			tcs->isConsistent %d, thr-type %d\n",isConsistent, 
			cs->reached,block,tcs->isConsistent,CVMisNoHeapRealtimeThread(ee));*/
    CVMassert(isConsistent || !cs->reached);
#endif

    CVMtraceCS(("<%d> CS<%d>: resuming execution with isConsistent %d\n",
	ee->threadID, cs->count, tcs->isConsistent));

    CVMsysMutexUnlock(ee, &cs->mutex);

#ifdef CVM_DEBUG
    /* Shouldn't be inconsistent if consistent state was reached */
    CVMassert(isConsistent || !cs->reached);
#endif
}

const char * const CVMcstateNames[CVM_NUM_CONSISTENT_STATES] = {
    "gc-safe"	/* CVM_GC_SAFE */
};
