/*
 * -----------------------------------------------------------------
 * gc_common.c : The abstracted garbage collection api/code. This
 * plugs into any GC implementation (e.g. mark-sweep, generational, 
 * etc.).
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)gc_common.c	1.90 00/11/11
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/directmem.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/stackwalk.h"
#include "javavm/include/utils.h"
#include "javavm/include/string.h"
#include "javavm/include/sync.h"

#include "javavm/include/localroots.h"
#include "javavm/include/globalroots.h"
#include "javavm/include/common_exceptions.h"

#include "javavm/include/preloader.h"

#include "javavm/include/gc_common.h"
#include "javavm/include/gc/gc_impl.h"

#include "javavm/include/globals.h"
#include "javavm/include/basictypes.h"

#include "javavm/export/jvmcei.h"

/*
 * Parse the heap size from the -Xms command line option.
 * 
 */
static CVMBool
CVMgcParseOptions(CVMOptions *options, CVMUint32* sizeOut)
{
    if (sizeOut == NULL) {
	return CVM_FALSE;
    }
    
    if (options->heapSizeStr != NULL) {
	CVMInt32 tmpVal;
	
	tmpVal = CVMatoml(options->heapSizeStr);
	if (tmpVal < 0) {
	    return CVM_FALSE;
	}
	*sizeOut = (CVMUint32) tmpVal;
    } else {
	*sizeOut = CVM_DEFAULT_HEAP_SIZE_IN_BYTES;
    }
    return CVM_TRUE;
}


/*
 * This method parses the contents of CVMglobals.gcOptions which were
 * previously extracted from the VM command line option
 * "-Xgc:<attr1>,<attr2>,...".  CVMglobals.gcOptions will not contain
 * the "-Xgc:" part.  
 *
 * Separates and stores the GC attributes, so that they can be examined
 * by shared and particular GC implementations.
 *
 * Returns CVM_TRUE if successful, CVM_FALSE otherwise.
 */
static CVMBool
CVMgcParseAttributes(CVMOptions* options)
{
    char* gcAttrPtrBegin;
    char* gcAttrPtr;
    char* attrStr;
    int i, numAttr;

    /*
     * Count the number of entries in the GC attributes string, and allocate
     * a large enough array for it.
     *
     * Duplicate the string, so that we can corrupt it later on with
     * embedded '\0's.
     */
    if (options->gcAttributesStr == NULL) {
	/*
	 * Nothing to do, parsing succeeds.
	 */
	return CVM_TRUE;
    }
    gcAttrPtrBegin = strdup(options->gcAttributesStr);
    if (gcAttrPtrBegin == NULL) {
	return CVM_FALSE;
    }
    numAttr = 0;
    gcAttrPtr = gcAttrPtrBegin;
    while (gcAttrPtr != NULL) {
	numAttr++;
	gcAttrPtr = strchr(gcAttrPtr, ',');
	if (gcAttrPtr != NULL) {
	    gcAttrPtr += 1;
	}
    }
    /* There are 'numAttr' GC options */
    CVMglobals.numGCOptions = numAttr;
    CVMglobals.gcOptions = calloc(numAttr, sizeof(const char *));
    if (CVMglobals.gcOptions == NULL) {
	/* Out of memory */
	free(gcAttrPtrBegin);
	return CVM_FALSE;
    }
    /* Now that we've counted, rewind, and get the options */
    gcAttrPtr = gcAttrPtrBegin;
    i = 0;
    while (gcAttrPtr != NULL) {
	attrStr = gcAttrPtr;
	gcAttrPtr = strchr(gcAttrPtr, ',');
	if (gcAttrPtr != NULL) {
	    *gcAttrPtr = '\0';
	    gcAttrPtr += 1;
	}
	CVMglobals.gcOptions[i++] = attrStr;
    }
    CVMassert(i == numAttr);
    return CVM_TRUE;
}

/*
 * Initialize a heap of at least 'minBytes' bytes, and at most 'maxBytes'
 * bytes.
 */
CVMBool 
CVMgcInitHeap(CVMOptions *options)
{
    CVMUint32 heapSize, maxHeapSize;

    if (!CVMgcParseOptions(options, &heapSize)) {
	CVMconsolePrintf("Cannot start VM "
			 "(error parsing heap size command "
			 "line option)\n");
	return CVM_FALSE;
    }
    /*
     * For now, we assume a non-expandable heap, and set the maxHeapSize
     * to be the same as heapSize. This may change in the future.
     */
    maxHeapSize = heapSize;

    if (!CVMgcParseAttributes(options)) {
	CVMconsolePrintf("Cannot start VM "
			 "(error parsing GC attributes "
			 "command-line option)\n");
	 return CVM_FALSE;
    }

    /*
     * This is the place to parse shared attributes, if any. Currently
     * there aren't any that I can think of, so the particular GC 
     * implementation is the only client of GC attributes now.
     *
     * CVMgcParseSharedGCAttributes();
     */

    /*    CVMconsolePrintf("CVMgcInitHeap: heap size = %d, ", heapSize); */

    /* Let's round these up to nice double-word multiples, before
       passing them on */
    heapSize = CVMalignDoubleWordUp(heapSize);
    maxHeapSize = CVMalignDoubleWordUp(maxHeapSize);
    return CVMgcimplInitHeap(&CVMglobals.gc, heapSize, maxHeapSize);
}

/*
 * Get GC attribute named 'attrName'.
 * If 'attrName' is of the form 'attrName=val', return "val". 
 * If 'attrName' is a flag instead, just return attrName.
 * Return NULL if not found.
 */
char *
CVMgcGetGCAttributeVal(char* attrName)
{
    int i;
    int len = strlen(attrName);
    for (i = 0; i < CVMglobals.numGCOptions; i++) {
	char* option = CVMglobals.gcOptions[i];
	if (!strncmp(option, attrName, len)) {
	    /* Name prefix matches. Return the value */
	    option += len;
	    if (*option == '\0') {
		return attrName;
	    } else if (*option == '=') {
		return option + 1;
	    }
	}
    }
    return NULL;
}

/*
 * The heap lock guards all heap operations. It is private to
 * gc_common.c
 */

static void
CVMgcPrivateLockHeapSafe(CVMExecEnv* ee)
{
    CVMSysMutex* heapLock = &CVMglobals.heapLock;
    CVMassert(CVMD_isgcSafe(ee));
    CVMsysMutexLock(ee, heapLock);
}

static void
CVMgcPrivateLockHeapUnsafe(CVMExecEnv* ee)
{
    CVMSysMutex* heapLock = &CVMglobals.heapLock;
    /*
     * Try locking
     */
    if (!CVMsysMutexTryLock(ee, heapLock)) {
	/*
	 * Must use blocking version. Become GC-safe first.
	 */
	CVMD_gcSafeExec(ee, {
	    CVMgcPrivateLockHeapSafe(ee);
	});
    }
}

static void
CVMgcPrivateUnlockHeap(CVMExecEnv* ee)
{
    CVMSysMutex* heapLock = &CVMglobals.heapLock;
    CVMsysMutexUnlock(ee, heapLock);
}

/*
 * WARNING: GC-unsafe allocator. Use CVMID_allocNewInstance() in GC-safe
 * code.
 */
CVMObject*
CVMgcAllocNewInstance(CVMExecEnv* ee, CVMClassBlock* cb)
{
  CVMObject* newInstance;
  
  /** Real time extension hooks 
   *  -Pratik
   */
  CVMBool isClass = CVM_FALSE;
//  CVMObjectICell oldMemoryArea;
  CVMBool oldCheckConstraints = CVM_FALSE;
  DECLARE_VARS_FOR_MA_SLIDING(ee);
  
  // This field used to indicate if memory was allocated on heap or other
  // memory areas. I need to lock/unlock heap only if allocation is being 
  // done from heap.
  CVMBool isHeap = CVM_TRUE;
  {      
    extern intFuncPtr JVMCeiFilters[] ;
    extern int JVMCeiInitialized;
   
    if(JVMCeiInitialized)
    {
        if( /*CVMisAssignable(ee, cb, CVMsystemClass(java_lang_Thread)) ||
            CVMisAssignable(ee, cb, CVMsystemClass(java_lang_ThreadGroup)) ||*/
            CVMisAssignable(ee, cb, CVMsystemClass(java_lang_ClassLoader))// ||
            /*CVMisAssignable(ee, cb, CVMsystemClass(java_lang_Error)) */) 
             
        {
            extern CVMObjectICell* jtImmortalMemory;
 //           CVMID_icellAssignDirect(ee, &oldMemoryArea, CVMmemoryAreaICell(ee));
 //           CVMID_icellAssignDirect(ee, CVMmemoryAreaICell(ee), jtImmortalMemory);
            oldCheckConstraints = CVMcheckForConstraints(ee);
            CVMdisableCheckForConstraints(ee);
            SLIDE_UP_TO_MA(ee, jtImmortalMemory);

            isClass = CVM_TRUE;
        }
    }
    
    // Budgeted allocation
    // The following huge if condition results in true if
    // 1. JVMCei is initialised
    // 2. ee->currentThreadICell is not NULL
    // 3. ee->currentThreadICell is a RealtimeThread
    // 4. cb is NOT StackOverflowError (special condition - If I am 
    //      throwing a StackOverflowError, I shhouldn't be pushing 
    //      stuff on the stack and call a Java method bcos that itself 
    //      will result in another StackOverflowError and we will recurse 
    //      and recurse and crash...:()
    //      (Changed the last condition to Error so that it covers 
    //      IllegalAssignmentError and MemoryAccessError
    //      - 12th July 2001)
    // - Pratik.
    if(JVMCeiInitialized &&
        CVMcheckForConstraints(ee) &&
        !CVMID_icellIsNull(CVMcurrentThreadICell(ee)) &&
        CVMgcUnsafeIsInstanceOf(ee, CVMID_icellDirect(ee, CVMcurrentThreadICell(ee)), 
                                    CVMsystemClass(javax_realtime_RealtimeThread)) &&
        !CVMisAssignable(ee, cb, CVMsystemClass(java_lang_Error))) 
    {
      // I am a realtime thread. Check if allocation is allowed
      // The following 20 lines are copied from the call to Finalizer.register()
      // below.
      CVMBool allowed = CVM_FALSE;
	  CVMD_gcSafeExec(ee, {
		JNIEnv* env = CVMexecEnv2JniEnv(ee);
      	//jobject clazz =
	  	//	CVMcbJavaInstance(CVMsystemClass(javax_realtime_RealtimeThread));
	  	CVMMethodBlock* mb =
			CVMglobals.javax_realtime_RealtimeThread_allowAllocation;
      	/*
       	 * We need to push a local frame here so we can make
       	 * JNI calls at all points that instances may be allocated.
         */
      	if ((*env)->PushLocalFrame(env, 4) == JNI_OK) 
		{
			allowed = 
                (*env)->CallBooleanMethod(env, CVMcurrentThreadICell(ee), mb, 
                                          CVMmemoryAreaICell(ee), CVMcbInstanceSize(cb));
			(*env)->PopLocalFrame(env, NULL);
      	} 
		else
		{
		  newInstance = NULL;
	    }
		});
        //Allocation is not allowed
        if(!allowed)
        {
            CVMdebugPrintf(("CVMgcAllocNewInstance(): Tried to allocate beyond my limits.\n"));
            return NULL;
        }

    }
    //check for exceptions here.
    if(CVMlocalExceptionOccurred(ee))
    {
        //an exception happened. I am going to return NULL.
        CVMconsolePrintf("An exception (%O) occured while trying to allocate an object. Returning NULl which will result in an OutOfMemoryError", CVMgetLocalExceptionObj(ee));
        CVMclearLocalException(ee);
        return NULL;
    }
    
    /* The following condition should result in true when 
     * CEI is initialised,
     * the IS_HEAP_Memory filter is installed,
     * and the memory area for the current ExecEnv is NOT heap.
     */
    if(JVMCeiInitialized &&                                   /* check if filters are installed */
       JVMCeiFilters[JVMCEF_IS_HEAP_MEMORY] != NULL &&        /* check if filter is null */
       JVMCeiFilters[JVMCEF_IS_HEAP_MEMORY]((void*) ee) == 0) /* check if it is not heap */    
      {
		//Memory area isn't heap. I have to allocate myself
	        
		struct arg_t {
		  CVMExecEnv* env; 
		  CVMJavaInt size;
		  CVMObject* newInstance; 
		} arg;
	
		arg.env = ee;
		arg.size = CVMcbInstanceSize(cb);
		JVMCeiFilters[JVMCEF_ALLOC_NEW_INSTANCE]((void*) &arg);
		newInstance = arg.newInstance;
		isHeap = CVM_FALSE;
      }
    else
      {
		//I have to allocate from heap. Allocate from gcimpl
		CVMgcPrivateLockHeapUnsafe(ee);
		newInstance = CVMgcimplAllocObject(ee,CVMcbInstanceSize(cb));
			 
		isHeap = CVM_TRUE;
      }
  }

  /*
   * Caller responsible for handling failed allocation
   */
  if (newInstance == 0) 
  {
	if(isHeap)
	    CVMgcPrivateUnlockHeap(ee);
    return 0;
  }
  
  /* Zero out object */
  
  /* %comment fresko003 */  
  memset((void*)newInstance, 0, CVMcbInstanceSize(cb));

  /* Set header of new object */
  newInstance->hdr.clas = cb; /* sync type 0 */

  CVMobjMonitorSet(newInstance, 0, CVM_LOCKSTATE_UNLOCKED);
  /*
   * Make sure that our notion of what the default header word is
   * matches that created by the above CVMobjMonitorSet
   */
  CVMassert(CVMobjectVariousWord(newInstance) ==
	    CVM_OBJECT_DEFAULT_VARIOUS_WORD);
  
  CVMtraceGcAlloc(("GC_COMMON: Allocated <%C> (size=%d) => 0x%p\n",
		   cb, CVMcbInstanceSize(cb), newInstance));
  
  if(isHeap)
	  CVMgcPrivateUnlockHeap(ee);
  
  if (CVMcbIs(cb, FINALIZABLE)) {
    CVMObjectICell* theCell = CVMfinalizerRegisterICell(ee);
    /*
     * We should call java.lang.ref.Finalizer.register().
     * We are going to go GC-safe here, so we'd better stash
     * the new object in a safe place in the interim. The caller
     * has not stored it in a GC-scannable place yet.
     *
     * NOTE: The following assertion is necessary to prevent a
     * case of a finalizable allocation in Finalizer.register().
     * If there is such a case, then we are going to re-enter this
     * function, and we will want to overwrite theCell which is
     * big trouble.
     */
    CVMassert(CVMID_icellIsNull(theCell));
    CVMID_icellSetDirect(ee, theCell, newInstance); /* safe! */
    
	/* The following if condition was added to separate out 
	 * finalizable objects that get queued in java.lang.ref.Finalizer
	 * and realtime objects that should get queued in javax.realtime.Finalizer
	 * - Pratik (8th Feb 2001)
	 */
	if(isHeap)
	{
	  CVMD_gcSafeExec(ee, {
		JNIEnv* env = CVMexecEnv2JniEnv(ee);
		jobject clazz =
		  CVMcbJavaInstance(CVMsystemClass(java_lang_ref_Finalizer));
		CVMMethodBlock* mb =
		  CVMglobals.java_lang_ref_Finalizer_register;
		/*
		 * We need to push a local frame here so we can make
		 * JNI calls at all points that instances may be allocated.
		 */
//        CVMdebugPrintf(("gc_common.c: Invoking register for object %I\n", theCell));
		if ((*env)->PushLocalFrame(env, 4) == JNI_OK) {
		  (*env)->CallStaticVoidMethod(env, clazz, mb, theCell);
		  (*env)->PopLocalFrame(env, NULL);
		} else {
		  newInstance = NULL;
	    }
    });
	CVMtraceWeakrefs(("WR: Registered a finalizable %C\n", cb));
	} //end if(isHeap)
	else 
	{
  //      CVMdebugPrintf(("gc_common.c: Invoking register for object %O allocated in memory area %O\n", CVMID_icellDirect(ee, theCell), CVMID_icellDirect(ee, CVMmemoryAreaICell(ee))));
		CVMJTregistryAddFinalizableObject(ee,newInstance);
	} // end else of if(isHeap)
	/*
     * Now re-load, since the object's address may have changed.
     */
    if (newInstance != NULL) {
      newInstance = CVMID_icellDirect(ee, theCell);
    }

    CVMID_icellSetNull(theCell); /* done! */
  }

  if(isClass)
  {
//    CVMID_icellAssignDirect(ee, CVMmemoryAreaICell(ee), &oldMemoryArea);
    SLIDE_BACK_DOWN(ee);
    CVMsetCheckForConstraints(ee, oldCheckConstraints);
  }
  return newInstance;
}

/*
 * Allocate heap array of length 'len'. The array type is 'arrayCb'.
 */
CVMArrayOfAnyType*
CVMgcAllocNewArray(CVMExecEnv* ee, CVMBasicType typeCode,
		   CVMClassBlock* arrayCb, CVMJavaInt arrayLen)
{
    CVMArrayOfAnyType* newArray;
    CVMJavaInt         arrayObjectSize;
	CVMBool isHeap = CVM_TRUE;

    CVMassert(arrayCb != 0);
    CVMassert(CVMbasicTypeSizes[typeCode] != 0);

    if (arrayLen >= (1 << 28)) {
	/* Don't bother. arrayObjectSize would overflow for this if
	   the array type was a long or a double. The chances of such
	   an allocation being satisfied are slim to none anyway. */
	return 0;
    }
    /*
     * The size of the heap object to accommodate this array
     */
    arrayObjectSize = CVMalignWordUp(CVMoffsetof(CVMArrayOfAnyType, elems) +
	CVMbasicTypeSizes[typeCode] * arrayLen);


    /** Real time extension hooks 
     *  -Pratik
     */
    {      
      extern intFuncPtr JVMCeiFilters[] ;
      extern int JVMCeiInitialized;
      
          
      // Budgeted allocation
      if(JVMCeiInitialized &&
          CVMcheckForConstraints(ee) &&
          !CVMID_icellIsNull(CVMcurrentThreadICell(ee)) &&
          CVMgcUnsafeIsInstanceOf(ee, CVMID_icellDirect(ee, CVMcurrentThreadICell(ee)), 
                                    CVMsystemClass(javax_realtime_RealtimeThread)))
      {
          // I am a realtime thread. Check if allocation is allowed
          // The following 20 lines are copied from the call to Finalizer.register()
          // below.
          CVMBool allowed = CVM_FALSE;
//#ifdef CVM_DEBUG
#if 0
          CVMconsolePrintf("Checking if I can allocate an array of %d bytes in %O. Class = %C\n", arrayObjectSize, 
                                CVMID_icellDirect(ee, CVMmemoryAreaICell(ee)), arrayCb );
#endif
	      CVMD_gcSafeExec(ee, {
		    JNIEnv* env = CVMexecEnv2JniEnv(ee);
      	  //jobject clazz =
	  		//CVMcbJavaInstance(CVMsystemClass(javax_realtime_RealtimeThread));
	  	  CVMMethodBlock* mb =
			CVMglobals.javax_realtime_RealtimeThread_allowAllocation;
      	  /*
       	   * We need to push a local frame here so we can make
       	   * JNI calls at all points that instances may be allocated.
           */
      	  if ((*env)->PushLocalFrame(env, 4) == JNI_OK) 
		  {
			//(*env)->CallStaticVoidMethod(env, clazz, mb, CVMmemoryAreaICell(ee), theCell);
			allowed = 
                (*env)->CallBooleanMethod(env, CVMcurrentThreadICell(ee), mb, 
                                          CVMmemoryAreaICell(ee), arrayObjectSize);
			(*env)->PopLocalFrame(env, NULL);
      	  } 
		});
        
        // check for exceptions here.
        if(CVMlocalExceptionOccurred(ee))
        {
            //an exception happened. I am going to return NULL.
            CVMconsolePrintf("An exception (%O) occured while trying to allocate an object. Returning NULl which will result in an OutOfMemoryError", CVMgetLocalExceptionObj(ee));
            CVMclearLocalException(ee);
            return NULL;
        }
        
        //Allocation is not allowed
        if(!allowed)
        {
//#ifdef CVM_DEBUG
#if 0
            CVMconsolePrintf("CVMgcAllocNewInstance(): Tried to allocate an array beyond my limits.\n");
#endif
            return NULL;
        }
      }

      /* The following condition should result in true when 
       * CEI is initialised,
       * the IS_HEAP_Memory filter is installed,
       * and the memory area for the current ExecEnv is NOT heap.
       */
      if(JVMCeiInitialized &&                                   /* check if filters are installed */
		  JVMCeiFilters[JVMCEF_IS_HEAP_MEMORY] != NULL &&        /* check if filter is null */
		  JVMCeiFilters[JVMCEF_IS_HEAP_MEMORY]((void*) ee) == 0) /* check if it is not heap */
	  {
		//Memory area isn't heap. I have to call MemoryArea.newArray()
		struct arg_t {
		  CVMExecEnv* env; 
	   	  CVMJavaInt size;
	      CVMArrayOfAnyType* newArray; 
	    } arg;
	  
	    arg.env = ee;
	    //		  arg.typeCode = typeCode;
	    //arg.arrayCb = arrayCb;
	    //arg.arrayLen = arrayLen;
	    arg.size = arrayObjectSize;
	    JVMCeiFilters[JVMCEF_ALLOC_NEW_ARRAY]((void*) &arg);
	    newArray = arg.newArray;
		isHeap = CVM_FALSE;
	  }
      else
	  {
	    CVMgcPrivateLockHeapUnsafe(ee);
	    newArray = (CVMArrayOfAnyType*)CVMgcimplAllocObject(ee, arrayObjectSize);	  
		isHeap = CVM_TRUE;
	  }
    }

    /*
     * Caller responsible for handling failed allocation
     */
    if (newArray == 0) 
	{
	  if(isHeap)
		CVMgcPrivateUnlockHeap(ee);
	  return 0;
    }
    /* Zero out object */
    memset((void*)newArray, 0, arrayObjectSize);

    /* Set header of new object */
    newArray->hdr.clas = arrayCb; /* sync type 0 */
    CVMobjMonitorSet(newArray, 0, CVM_LOCKSTATE_UNLOCKED);


    /*
     * Make sure that our notion of what the default header word is
     * matches that created by the above CVMobjMonitorSet
     */
    CVMassert(CVMobjectVariousWord(newArray) ==
	      CVM_OBJECT_DEFAULT_VARIOUS_WORD);
    newArray->length   = arrayLen;

    CVMtraceGcAlloc(("GC_COMMON: Allocated <%C> (size=%d, len=%d) => 0x%p\n",
		     arrayCb, arrayObjectSize, arrayLen, newArray));

	if(isHeap)
  	  CVMgcPrivateUnlockHeap(ee);

    return newArray;
}

void
CVMnullFrameScanner(CVMFrame* frame, CVMStackChunk* chunk,
		    CVMRefCallbackFunc callback, void* data)
{
    CVMassert(!CVMframeIsTransition(frame) &&
	      !CVMframeIsJava(frame) &&
	      !CVMframeIsJNI(frame));
    return;
}

/*
 * The following is private to the root scanner, and is used to cart
 * arguments to CVMclassScanCallback
 */
struct CVMClassScanOptions {
    CVMRefCallbackFunc callback;
    void* data;
    CVMBool allClassesAreRoots;
};

typedef struct CVMClassScanOptions CVMClassScanOptions;

/*
 * This wrapper to CVMclassScan() is passed on to 
 * CVMclassIterateDynamicallyLoadedClasses().
 */
static void
CVMclassScanCallback(CVMExecEnv* ee, CVMClassBlock* cb, void* opts)
{
    CVMClassScanOptions* options = (CVMClassScanOptions*)opts;
    /*
     * We'd better make sure that all classes are roots, because we
     * are going to be using the class statics scanner for each class.
     */
    CVMassert(options->allClassesAreRoots);
    CVMclassScanStatics(ee, cb, options->callback, options->data);
}

/*
 * This callback is used for keeping system classes live.
 */
static void
CVMclassScanSystemClassCallback(CVMExecEnv* ee, CVMClassBlock* cb, void* opts)
{
    CVMClassScanOptions* options = (CVMClassScanOptions*)opts;
    if (CVMcbClassLoader(cb) == NULL) {
	CVMassert(!options->allClassesAreRoots);
	/* This makes sure a class is not scanned in a cycle */
	CVMscanClassIfNeeded(ee, cb, options->callback, options->data);
    }
}

/*
 * Clear the scanned flag of a classblock
 */
static void
CVMclassClearScannedFlagCallback(CVMExecEnv* ee, CVMClassBlock* cb, void* opts)
{
    CVMtraceGcScan(("GC: Clearing marked bit of class %C (0x%x)\n",
		    cb, cb));
    CVMcbClearGcScanned(cb);
}

/*
 * Process special objects with liveness info from a particular GC
 * implementation. This covers special scans like string intern table,
 * weak references and monitor structures.
 */
void
CVMgcProcessSpecialWithLivenessInfo(CVMExecEnv* ee, CVMGCOptions* gcOpts,
				    CVMRefLivenessQueryFunc isLive,
				    void* isLiveData,
                                    CVMRefCallbackFunc transitiveScanner,
				    void* transitiveScannerData)
{
    /*
     * Handle weak reference structures, clearing out the right
     * referents. This can change the liveness of things, so it should
     * come first.
     */
    CVMweakrefProcessNonStrong(ee, isLive, isLiveData,
			       transitiveScanner, transitiveScannerData,
			       gcOpts);

    /*
     * Handle interned strings
     */
    CVMdeleteUnreferencedStrings(isLive, isLiveData,
				 transitiveScanner, transitiveScannerData);

    /*
     * Handle monitor structures
     */
    CVMdeleteUnreferencedMonitors(ee, isLive, isLiveData,
				  transitiveScanner, transitiveScannerData);
    
    /*
     * Purge unused entries for cb iff !CVMcbGcScanned(cb). NOTE: by
     * default, all ROM classes are assumed to be scanned, even though
     * CVMcbGcScanned(cbInRom) == FALSE.
     */
#ifdef CVM_CLASSLOADING
    CVMclassDoClassUnloadingPass1(ee, isLive, isLiveData,
				  transitiveScanner, transitiveScannerData,
				  gcOpts);
#endif
}

static CVMBool
ensureStackmapsForRootScans(CVMExecEnv* ee)
{
    CVM_WALK_ALL_THREADS_EXCEPT_NOHEAP(ee, currentEE, {
	CVMStack* iStack = &currentEE->interpreterStack;
	CVMstackDoRoots(iStack, {
	    if (CVMframeIsJava(frame)) {
		if (!CVMjavaFrameEnsureStackmaps(currentEE, frame)) {
		    return CVM_FALSE;
		}
	    }
	});
    });
    return CVM_TRUE;
}

/*
 * Start a GC cycle.
 */
void
CVMgcStartGC(CVMExecEnv* ee)
{
}

void
CVMgcClearClassMarks(CVMExecEnv* ee, CVMGCOptions* gcOpts)
{
    /*
     * Clear the scanned bits of all classes. Each class will be scanned
     * by the object scan as necessary. This includes the class
     * global roots.
     *
     * Of course if we are not doing a full GC, we don't have to do this.
     * All classes will be scanned anyway, so the GC_SCANNED bit is
     * irrelevant.
     */
    CVMclassIterateDynamicallyLoadedClasses(
        ee, CVMclassClearScannedFlagCallback, NULL);
}

/*
 * End a GC cycle.
 */
void
CVMgcEndGC(CVMExecEnv* ee)
{
#ifdef CVM_CLASSLOADING
    CVMclassDoClassUnloadingPass2(ee);
#endif
}

/*
 * Scan and optionally update special objects. This covers special
 * scans like string intern table, weak references and monitor
 * structures.
 */
void
CVMgcScanSpecial(CVMExecEnv* ee, CVMGCOptions* gcOpts,
		 CVMRefCallbackFunc refUpdate, void* refUpdateData)
{
    CVMtraceGcCollect(("GC_COMMON: "
		       "Scanning special objects (%s)\n",
		       gcOpts->fullGC ? "full" : "partial"));
    /*
     * Handle interned strings
     */
    CVMscanInternedStrings(refUpdate, refUpdateData);

    /*
     * Handle monitor structures
     */
    CVMscanMonitors(ee, refUpdate, refUpdateData);

    /*
     * Handle weak reference structures
     */
    CVMweakrefUpdate(ee, refUpdate, refUpdateData, gcOpts);

    /*
     * Update all live class loader roots
     */
    CVMstackScanRoots(ee, &CVMglobals.classLoaderGlobalRoots,
		      refUpdate, refUpdateData, gcOpts);
}

/*
 * Scan the root set of collection
 */
void
CVMgcScanRoots(CVMExecEnv *ee, CVMGCOptions* gcOpts,
	       CVMRefCallbackFunc callback, void* data)
{
    /*
     * Assert that we are holding the right locks before scanning
     * the data structures protected by those locks.
     */
    CVMassert(CVMreentrantMutexIAmOwner(
        ee, CVMsysMutexGetReentrantMutex(&CVMglobals.threadLock)));

    /*
     * Scan global roots
     */
    CVMstackScanRoots(ee, &CVMglobals.globalRoots, callback, data, gcOpts);

    /*
     * Do the preloader statics
     */
    {
	CVMUint32    numRefStatics = CVMpreloaderNumberOfRefStatics();
	CVMInt32*    refStatics    = CVMpreloaderGetRefStatics();
	CVMwalkContiguousRefs(refStatics, numRefStatics, callback, data);
    }

    if (gcOpts->allClassesAreRoots) {
	/*
	 * Do dynamically created class state if all classes are to be
	 * treated as roots.  
	 */
	CVMClassScanOptions opt;
	opt.callback = callback;
	opt.data = data;
	opt.allClassesAreRoots = CVM_TRUE;

	/*
	 * Mark all classes
	 */
	CVMclassIterateDynamicallyLoadedClasses(
            ee, CVMclassScanCallback, &opt);
    } else {
	CVMClassScanOptions opt;
	/*
	 * Now make sure to keep all system classes live. These cannot
	 * be unloaded.
	 */
	opt.callback = callback;
	opt.data = data;
	opt.allClassesAreRoots = CVM_FALSE;
	CVMclassIterateDynamicallyLoadedClasses(
            ee, CVMclassScanSystemClassCallback, &opt);
    }

    /*
     * Scan the local and interpreter roots of each thread. The previous
     * class table iterations already cleared the mark bits of classes
     * if they are to be scanned based on reachability.
     */
    CVM_WALK_ALL_THREADS_EXCEPT_NOHEAP(ee, currentEE, {
	CVMstackScanRoots(ee, &currentEE->localRootsStack,
			  callback, data, gcOpts);
	CVMstackScanRoots(ee, &currentEE->interpreterStack,
			  callback, data, gcOpts);
    });
}

/*
 * Initiate a GC. Acquire all GC locks, stop all threads, and then
 * call back to the particular GC to do the work. When the particular
 * GC is done, resume.
 *
 * If we can't execute the GC action successfully, return CVM_FALSE.
 * Otherwise, return CVM_TRUE.
 */
static CVMBool
CVMgcStopTheWorldAndGCSafe(CVMExecEnv* ee, CVMUint32 numBytes)
{
    CVMObjectICell* referenceLockCell =
	&CVMglobals.java_lang_ref_Reference_lock->r;
    CVMBool weakrefsInitialized;
    CVMBool gcOccurred;
    CVMassert(CVMD_isgcSafe(ee));

    weakrefsInitialized = !CVMID_icellIsNull(referenceLockCell);
    /*
     * First lock the reference lock.
     * %comment: fresko004
     */
    if (weakrefsInitialized) {
	CVMBool success;
	CVMD_gcUnsafeExec(ee, {
            /* %comment lamm002 */
            success = 
		CVMobjectTryLock(ee, CVMID_icellDirect(ee, referenceLockCell));
	    if (!success) {
                success = CVMobjectLock(ee, referenceLockCell);
	    }
	});
	if (!success) {
	    return CVM_FALSE;
	} else {
	    CVMglobals.referenceWorkTODO = CVM_FALSE;
	}
    }

    /*
     * And get the rest of the GC locks.
     */
    CVMlocksForGCAcquire(ee);

    /* Roll all threads to GC-safe points. */
    CVMD_gcBecomeSafeAll(ee);

    /* 
     * Do pre-GC actions
     */
    CVMgcStartGC(ee);

    if (ensureStackmapsForRootScans(ee)) {
	/*
	 * Call out to the particular GC implementation
	 * only if stackmaps are created to scan roots.
	 */
	CVMgcimplDoGC(ee, numBytes);
	gcOccurred = CVM_TRUE;
    } else {
	gcOccurred = CVM_FALSE;
    }

    /* Allow threads to become GC-unsafe again. */
    CVMD_gcAllowUnsafeAll(ee);

    CVMlocksForGCRelease(ee);

    /* 
     * Do post-GC actions
     */
    CVMgcEndGC(ee);

    /*
     * Now notify the reference lock waiters, and unlock.
     */
    if (weakrefsInitialized) {
	CVMD_gcUnsafeExec(ee, {
	    /*
	     * Wake the reference handler thread so it enqueues its weak
	     * references in the right ReferenceQueue's.
	     * 
	     * We're still holding the reference lock here, so another thread
	     * could not have changed CVMglobals.referenceWorkTODO from
	     * under us.
	     */
	    if (CVMglobals.referenceWorkTODO) {
		CVMobjectNotifyAll(ee, referenceLockCell);
		CVMglobals.referenceWorkTODO = CVM_FALSE;
	    }
            /* %comment lamm003 */
            if (!CVMobjectTryUnlock(
                    ee, CVMID_icellDirect(ee, referenceLockCell))) {
                if (!CVMobjectUnlock(ee, referenceLockCell)) {
		    CVMassert(CVM_FALSE); /* should never happen */
		}
            }
	});
    }
    return gcOccurred;
}

CVMBool
CVMgcStopTheWorldAndGC(CVMExecEnv* ee, CVMUint32 numBytes)
{
    CVMBool retVal;
    CVMD_gcSafeExec(ee, {
	retVal = CVMgcStopTheWorldAndGCSafe(ee, numBytes);
    });
    return retVal;
}

void
CVMgcRunGC(CVMExecEnv* ee)
{
    CVMgcPrivateLockHeapSafe(ee);
    (void)CVMgcStopTheWorldAndGCSafe(ee, ~0); /* Do a full-scale GC */
    CVMgcPrivateUnlockHeap(ee);
}

/*
 * Return the number of bytes free in the heap. 
 */
CVMJavaLong
CVMgcFreeMemory(CVMExecEnv* ee)
{
    CVMUint32 freeMem = CVMgcimplFreeMemory(ee);
    /* %comment fresko005 */
    return CVMint2Long(freeMem);
}

/*
 * Return the amount of total memory in the heap, in bytes.
 */
CVMJavaLong
CVMgcTotalMemory(CVMExecEnv* ee)
{
    CVMUint32 totalMem = CVMgcimplTotalMemory(ee);
    /* %comment fresko005 */
    return CVMint2Long(totalMem);
}

/*
 * Destroy heap
 */
CVMBool 
CVMgcDestroyHeap()
{
    if (CVMglobals.numGCOptions >= 1) {
	CVMassert(CVMglobals.gcOptions != NULL);
	CVMassert(CVMglobals.gcOptions[0] != NULL);
	/* The first option points to the beginning of the strdup()'ed
	   GC options string. Free it. */
	free(CVMglobals.gcOptions[0]);
	/* And now the set of gcOptions we allocated */
	free(CVMglobals.gcOptions);
    }
    return CVMgcimplDestroyHeap(&CVMglobals.gc);
}

#ifdef CVM_DEBUG
/*
 * Heap dump related routines
 */

/*
 * Scan objects in contiguous range, and do per-object callback in support
 * of heap dump.
 */
void
CVMgcScanObjectRange(CVMExecEnv* ee, CVMUint32* base, CVMUint32* top,
		     CVMObjectCallbackFunc callback, void* callbackData)
{
    CVMUint32* curr = base;

    while (curr < top) {
	CVMObject* currObj = (CVMObject*)curr;
	CVMClassBlock* currCb = CVMobjectGetClass(currObj);
	CVMUint32  objSize    = CVMobjectSizeGivenClass(currObj, currCb);

	(*callback)(currObj, currCb, objSize, callbackData);
	/* iterate */
	curr += objSize / 4;
    }
    CVMassert(curr == top); /* This had better be exact */
}

/*
 * Locks heap and dumps contents given a 'dumpFunc'
 */
static void
dumper(CVMExecEnv* ee, void (*dumpFunc)(CVMExecEnv* ee))
{
    CVMSysMutex* heapLock = &CVMglobals.heapLock;

    /*
     * Try to enter the heap lock without blocking. Refuse to dump heap
     * if recursively entering lock, or if another thread has already locked
     * the lock.
     */
    if (!CVMsysMutexTryLock(ee, heapLock)) {
	CVMconsolePrintf("Cannot lock heap without blocking -- "
			 "another thread already owns the heap lock!\n");
	return;
    } else {
	/*
	 * We did obtain the lock without blocking. Now just make sure
	 * it's not a recursive entry
	 */
	if (CVMsysMutexCount(heapLock) > 1) {
	    CVMconsolePrintf("Cannot dump heap -- "
			     "this thread already owns the heap lock!\n");
	    CVMsysMutexUnlock(ee, heapLock);
	    return;
	}
    }

    (*dumpFunc)(ee);

    CVMsysMutexUnlock(ee, heapLock);
}

/*
 * A dumper that counts objects on the heap
 */
static void
simpleDumpCallback(CVMObject* obj, CVMClassBlock* cb, CVMUint32 size,
		   void* data)
{
    CVMUint32* count = (CVMUint32*)data;
    *count += 1;
}

void
CVMgcSimpleHeapDump(CVMExecEnv* ee)
{
    CVMUint32 objCount;
    objCount = 0;
    CVMgcimplIterateHeap(ee, simpleDumpCallback, &objCount);
    CVMconsolePrintf("Counted %d objects\n", objCount);
}

/*
 * A dumper that dumps every object on the heap
 */
static void
verboseDumpCallback(CVMObject* obj, CVMClassBlock* cb, CVMUint32 size,
		   void* data)
{
    CVMconsolePrintf("Addr: 0x%x Size: %d  \tClass: %C\n", obj, size, cb);
}

void
CVMgcVerboseHeapDump(CVMExecEnv* ee)
{
    CVMgcimplIterateHeap(ee, verboseDumpCallback, NULL);
}

/*
 * A dumper that is used to output stats on heap objects
 */

/*
 * A per-object callback to determine the maximum TypeID of an object
 * on the heap.
 */
static void
findMaxTypeIDCallback(CVMObject* obj, CVMClassBlock* cb, CVMUint32 size,
		   void* data)
{
    CVMUint32* currMaxID = (CVMUint32*)data;
    if (CVMcbClassName(cb) > *currMaxID) {
	*currMaxID = CVMcbClassName(cb);
    }
}

/* 
 * The following table is used to hold instance stats. Currently it is
 * one large array, indexed by typeID: [0 .. maxTypeID). A first pass
 * optimization would remove the huge hole in the middle that is the
 * non-array -> array skip. Another pass would turn this into a
 * hashtable keyed by typeID. Left as an exercise to the reader, since
 * this is debugging-only code.  
 */
typedef struct {
    CVMUint32 totalSize; /* Total size consumed for this class */
    CVMUint32 numInstances; /* Total number of instances of this class */
    CVMClassTypeID typeID; /* We will sort this array, so record typeID here */
} TableEntry;

typedef struct {
    int  tableSize; /* Number of elements in the tables below */
    TableEntry* entries;
} Stats;

static void
statsCollectCallback(CVMObject* obj, CVMClassBlock* cb, CVMUint32 size,
		   void* data)
{
    CVMClassTypeID typeID = CVMcbClassName(cb);
    Stats*         stats  = (Stats*)data;
    TableEntry*    entry  = &stats->entries[typeID];

    entry->typeID        = typeID;
    entry->totalSize    += size;
    entry->numInstances += 1;
}

/*
 * Comparator for class stats table. This one sorts in descending total
 * memory usage
 */
static int 
tableEntrySizeCompare(const void *p1, const void *p2)
{
    TableEntry* e1 = (TableEntry*)p1;
    TableEntry* e2 = (TableEntry*)p2;
    return e2->totalSize - e1->totalSize;
}

/*
 * And now the heap stats dumper
 */
void
CVMgcStatsHeapDump(CVMExecEnv* ee)
{
    CVMUint32 maxTypeID = 0;
    int i;
    Stats stats;

    /* First look at all objects in the heap, and figure out the max TypeID
       value. We will use this to allocate stats tables */
    CVMgcimplIterateHeap(ee, findMaxTypeIDCallback, &maxTypeID);
    stats.tableSize = maxTypeID + 1;
    /*
     * OK, now allocate the stats table
     */
    stats.entries = calloc(sizeof(TableEntry), stats.tableSize);
    if (stats.entries == NULL) {
	CVMconsolePrintf("Out of memory allocating totalSize array\n");
	return;
    }
    CVMgcimplIterateHeap(ee, statsCollectCallback, &stats);
    /*
     * OK, now sort the table in decreasing order of memory consumption
     */
    qsort(stats.entries, stats.tableSize, sizeof(TableEntry), 
	  tableEntrySizeCompare);
    /*
     * And dump
     */
    for (i = 0; i < stats.tableSize; i++) {
	TableEntry* e = &stats.entries[i];
	char buf[256]; /* 256 chars max for the class name */
	if (e->numInstances == 0) {
	    continue;
	}
	CVMclassname2String(e->typeID, buf, 256);
	printf("TS=%-8d NI=%-8d CL=%s\n",
	       e->totalSize, e->numInstances, buf);
    }
    free(stats.entries);
}

/*
 * Dumping heap contents. These routines are designed to be called
 * from the debugger, so they don't take an ee as argument.
 */
void
CVMgcDumpHeapSimple()
{
    dumper(CVMgetEE(), CVMgcSimpleHeapDump);
}

void
CVMgcDumpHeapVerbose()
{
    dumper(CVMgetEE(), CVMgcVerboseHeapDump);
}

void
CVMgcDumpHeapStats()
{
    dumper(CVMgetEE(), CVMgcStatsHeapDump);
}

#endif /* CVM_DEBUG */
