/*
 * @(#)preloader.c	1.66 00/11/03
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/typeid.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/preloader.h"
#include "javavm/include/preloader_impl.h"
#include "javavm/include/stackmaps.h"
#include "javavm/include/utils.h"
#include "javavm/include/clib.h"

#ifdef CVM_DEBUG_ASSERTS

static void
CVMpreloaderVerifyGCMap(const CVMClassBlock* cb);

static void
CVMpreloaderVerifyGCMaps();
#endif

/*
 * Look up class by name. Only used from debuggers.
 */
#ifdef CVM_DEBUG
const CVMClassBlock* 
CVMpreloaderLookup(const char* className)
{
    CVMClassTypeID typeID = 
	CVMtypeidLookupClassID(NULL, className, strlen(className) );
    if (typeID == CVM_TYPEID_ERROR) {
	return NULL; /* not in type table, so don't bother looking. */
    }
    return CVMpreloaderLookupFromType(typeID);
}
#endif

CVMClassBlock*
CVMpreloaderLookupFromType(CVMClassTypeID typeID)
{
    int i, lowest, highbound;

    /* No CVMFieldTypeIDs allowed. Upper bits must never be set. */
    CVMassert(CVMtypeidGetType(typeID) == typeID);

    if (CVMtypeidIsPrimitive(typeID)) {
	/* NOTE: casting away const is safe here because we know that
           run-time flags are separated from ROMized classblocks */
	return (CVMClassBlock*) CVMterseTypeClassblocks[typeID];
    }

    /* NOTE: The Class instances of the Class and Array types in the
       CVM_ROMClasses[] are sorted in incremental order of typeid values.  The
       only exception is that deep array types (i.e. array types with depth >=
       7) will be sorted based only on the basetype field of their typeids
       (i.e. the depth field is ignored ... hence the masking operation in the
       computation of the index for the deep array case below).  This is what
       makes it possible to access Class instances for Class and Deep array
       types in CVM_ROMClasses[] by indexing. */
    if ( !CVMtypeidIsArray(typeID)) {
	CVMClassBlock * cb;
	
	i = typeID-CVMtypeidLastScalar-1;
	if ( (i >= CVM_firstROMVectorClass) || ( i < CVM_firstROMNonPrimitiveClass) ){
	    return NULL; /* not here! */
	}
	cb = CVM_ROMClasses[i].classBlockPointer;
	CVMassert( CVMcbClassName(cb) == typeID );
	return cb;
    }

    if (CVMtypeidIsBigArray(typeID)){
        CVMClassBlock * cb;

        i = (typeID & CVMtypeidBasetypeMask)-CVMtypeidLastScalar-1;
        if ( i >= CVM_firstROMVectorClass ){
                return NULL; /* not here! */
        }
        cb = CVM_ROMClasses[i].classBlockPointer;
        CVMassert( CVMcbClassName(cb) == typeID );
        return cb;
    }

    /*
     * It is either a one-dimensional array or a higher-dimensional
     * array. Choose the appropriate partition of the ROMClasses array
     * and binary search for it.
     */
    if ( CVMtypeidArrayDepth(typeID) == 1 ){
	lowest = CVM_firstROMVectorClass;
	highbound= CVM_lastROMVectorClass+1;
    } else {
	lowest = CVM_lastROMVectorClass+1;
	highbound= CVM_nROMClasses;
    }

    while ( lowest < highbound ){
	CVMClassBlock * cb;
	int candidateID;
	i = lowest + (highbound-lowest)/2;
	cb = CVM_ROMClasses[i].classBlockPointer;
	candidateID = CVMcbClassName(cb);
	if ( candidateID == typeID) return cb;
	if ( candidateID < typeID)
	    lowest = i+1;
	else
	    highbound = i;
    }

    /*
     * Fell out of while loop.
     * not in table.
     */
    return NULL;
}

/*
 * Only for lookup of primitive VM-defined types, by type name.
 */
CVMClassBlock*
CVMpreloaderLookupPrimitiveClassFromType(CVMClassTypeID classType){
    int i;
    CVMClassBlock* cb;

    i = classType-CVMtypeidLastScalar-1;
    if ( (i < 0) || ( i >= CVM_firstROMNonPrimitiveClass) ){
	return NULL; /* not here! */
    }
    cb = CVM_ROMClasses[i].classBlockPointer;
    CVMassert( CVMcbClassName(cb) == classType );
    return cb;
}


void
CVMpreloaderInit()
{
    int i, n;

    /* All romized classes with static initializers in their hierarchy have
     * all flags set except for the ERROR and INITIALIZED flags.
     */
    /* NOTE: The cast to CVMUint8 is a workaround for a bug in
       VxWorks' memset (more accurately, bfill) implementation, which
       behaves incorrectly for integer values having any of the three
       high bits set. Bug has been reported. */
    memset(CVMROMClassesRuntimeFlags, 
	       (CVMUint8) (~CVM_CLASS_ERROR & ~CVM_CLASS_INITIALIZED),
	       CVM_nROMClassesRuntimeFlags);

    /* All romized classes with no static initializers in their hierarchy have
     * a romIndex of 0, so we mark slot 0 as having all flags set except
     * for the ERROR flag.
     */
    CVMROMClassesRuntimeFlags[0] = ~CVM_CLASS_ERROR;

    /*
     * And finally, do a series of block initializations from read-only
     * master copies to BSS areas.
     */
    n = CVM_nPreloaderMapEntries;
    for ( i = 0; i < n; i++ ){
	struct CVM_preloaderInitTriple * t =
	    (struct CVM_preloaderInitTriple*)&CVM_preloaderInitMap[i];
	if ( t->from == NULL ){
	    memset( t->to, 0, t->bytecount );
	} else {
	    memcpy(t->to, t->from, t->bytecount);
	}
    }

#ifdef CVM_DEBUG_ASSERTS
    /*
     * Sanity check mb indexes
     */
    {
	int i, j;
	for (i = 0; i < CVM_nROMClasses; ++i) {
	    const CVMClassBlock* cb = CVM_ROMClasses[i].classBlockPointer;
	    if (CVMcbIs(cb, PRIMITIVE) || CVMcbMethodTablePtr(cb) == NULL) {
		continue;
	    }
	    for (j = 0; j < CVMcbMethodTableCount(cb); ++j) {
		CVMMethodBlock *mb = CVMcbMethodTableSlot(cb, j);
		const CVMClassBlock* mcb = CVMmbClassBlock(mb);
		CVMassert(CVMmbMethodTableIndex(mb) == j);
		if (CVMcbMethodCount(cb) <= 256) {
		    CVMassert(mb == 
			      CVMcbMethodSlot(mcb, CVMmbMethodIndex(mb)));
		}
	    }
	}
    }
#endif

#ifdef CVM_DEBUG_ASSERTS
    CVMpreloaderVerifyGCMaps();
#endif

}

CVMBool
CVMpreloaderDisambiguateAllMethods()
{
    int i;

    for (i = 0; i < CVM_nROMClasses; ++i) {
	const CVMClassBlock* cb = CVM_ROMClasses[i].classBlockPointer;
	int nmethods = CVMcbMethodCount(cb);
	int i;
	for (i = 0; i < nmethods; i++) {
	    CVMMethodBlock* mb = CVMcbMethodSlot(cb, i);
	    if (CVMmbInvokerIdx(mb) < CVM_INVOKE_CVM_METHOD) {
		if ( CVMjmdFlags(CVMmbJmd(mb)) & CVM_JMD_MAY_NEED_REWRITE ){
		    switch (CVMstackmapDisambiguate(mb, CVM_TRUE)) {
			case CVM_MAP_SUCCESS:
			    break;
			case CVM_MAP_OUT_OF_MEMORY:
			    CVMconsolePrintf(
			        "Ran out of memory while disambiguating %C.%M",
				cb, mb);
			    return CVM_FALSE;
			case CVM_MAP_EXCEEDED_LIMITS:
			    CVMconsolePrintf(
			        "Exceeded limits while disambiguating %C.%M",
				cb, mb);
			    return CVM_FALSE;
			default:
			    CVMassert(CVM_FALSE);
			    return CVM_FALSE;
		    }
		}
	    }
	}
    }
    return CVM_TRUE;
}

#ifdef CVM_DEBUG_ASSERTS

/*
 * For each romized class, verify that the romized gcmaps are the same
 * as the gcmap that the classloader would have generated if it had
 * dynamically loaded the class instead.
 */
static void
CVMpreloaderVerifyGCMaps()
{
    int i;
    for (i = 0; i < CVM_nROMClasses; ++i) {
	const CVMClassBlock* cb = CVM_ROMClasses[i].classBlockPointer;
	/* 
	 * We don't check java_lang_Object, array classes or
	 * primitive clases.
	 */
	if (cb != CVMsystemClass(java_lang_Object) &&
	    !CVMisArrayClass(cb) && !CVMcbIs(cb, PRIMITIVE)) {
	    CVMpreloaderVerifyGCMap(cb);
	}
   }
}

/*
 * This code is basically the same is CVMclassPrepareFields(), minus
 * a few things like setting up static field offsets. Also, instead
 * of setting cb fields like the CVMcbInstanceSize() and CVMcbGcMap(),
 * it instead verifies that the computed values are the same as the
 * romized values.
 */
static void
CVMpreloaderVerifyGCMap(const CVMClassBlock* cb)
{
    CVMUint16      i;
    CVMClassBlock* superCb = CVMcbSuperclass(cb);
    CVMGCBitMap    superGcMap = CVMcbGcMap(superCb);
    CVMGCBitMap    gcMap;
    CVMUint16      numFieldWords;
    CVMUint16      numSuperFieldWords;
    CVMUint32      instanceSize;
    CVMUint32      fieldOffset; /* word offset for next field */
    CVMUint32      currBigMapIndex = 0;
    CVMUint16      nextMapBit;
    CVMBool        hasShortMap;
    CVMUint32      map;  /* current map we are writing into. */

    numFieldWords = 
	(CVMcbInstanceSize(cb) - sizeof(CVMObjectHeader)) /
	 sizeof(CVMObject*);
    numSuperFieldWords = 
	(CVMcbInstanceSize(superCb) - sizeof(CVMObjectHeader)) /
	 sizeof(CVMObject*);

    /* Set instance size of the class. */
    instanceSize = 
	numFieldWords * sizeof(CVMObject*) + sizeof(CVMObjectHeader) ;

    /* Set fieldOffset to just past the last field of the superclass. */
    fieldOffset =
	numSuperFieldWords + sizeof(CVMObjectHeader) / sizeof(CVMObject*);

    /*
     * Determine if this class needs a long or a short gc map and copy
     * the superclass' gc map into this class' gc map.
     */
    hasShortMap = numFieldWords + CVM_GCMAP_NUM_FLAGS <= 32;
    if (hasShortMap) {
	/* 
	 * Both the current class and the superclass have short versions
	 * of the gcmap. Just copy the superclass map and turn off the flags
	 * for now.
	 */
	gcMap.map = superGcMap.map >>= CVM_GCMAP_NUM_FLAGS;
    } else {
	/*
	 * The current class needs a long version of the gc map.
	 */
	CVMUint32 superGcMapFlags = superGcMap.map & CVM_GCMAP_FLAGS_MASK;
	CVMUint32 maplen = (numFieldWords + 31) / 32;
	
	/* Allocate space for the current class' long map. */
	gcMap.bigmap = (CVMBigGCBitMap*)
	    calloc(1, CVMoffsetof(CVMBigGCBitMap, map) +
		       maplen * sizeof(CVMUint32));
	if (gcMap.bigmap == NULL) {
	    CVMassert(CVM_FALSE);
	}
	gcMap.bigmap->maplen = maplen;

	if (superGcMapFlags == CVM_GCMAP_SHORTGCMAP_FLAG) {
	    /*
	     * The superclass has a short gcmap and the current class
	     * needs a long one, so just copy the superclass's short
	     * map into the last word of the long map.
	     */
	    gcMap.bigmap->map[0] = (superGcMap.map >> CVM_GCMAP_NUM_FLAGS);
	} else {
	    /* 
	     * Both the superclass and the current class have long maps.
	     * Copy the superclass's long map into the current class'
	     * longmap.
	     */
	    CVMBigGCBitMap* superBigMap;
	    CVMUint16 i;
	    CVMassert(superGcMapFlags == CVM_GCMAP_LONGGCMAP_FLAG);
	    superBigMap = (CVMBigGCBitMap*)
		((CVMUint32)superGcMap.bigmap & ~CVM_GCMAP_FLAGS_MASK);
	    for (i = 0; i < superBigMap->maplen; i++) {
		gcMap.bigmap->map[i] = superBigMap->map[i];
	    }
	}
    }

    /* 
     * Setup the "map" variable that we will be setting bits in. Also
     * calculate the next bit in "map" that we will be working on.
     */
    nextMapBit = numSuperFieldWords % 32;
    if (hasShortMap) {
	map = gcMap.map;
    } else {
	currBigMapIndex = numSuperFieldWords / 32;
	map = gcMap.bigmap->map[currBigMapIndex];
    } 

    /* 
     * Iteterate over each field in the class and compute the gcmap for
     * the class.
     */
    for (i = 0; i < CVMcbFieldCount(cb) ; i++) {
	CVMFieldBlock* fb = CVMcbFieldSlot(cb, i);
	if (!CVMfbIs(fb, STATIC)) {
	    /* 
	     * Calculate the object offset of the next field and also set
	     * the gcmap bit if necessary.
	     */
	    if (CVMtypeidFieldIsDoubleword(CVMfbNameAndTypeID(fb))) {
		fieldOffset += 2;
		nextMapBit += 2;
	    } else {
		/* For ref types, we set the bit in the gc map. */
		if (CVMtypeidFieldIsRef(CVMfbNameAndTypeID(fb))) {
		    map |= 1<<nextMapBit;
		}
		fieldOffset += 1;
		nextMapBit += 1;
	    }

	    /* 
	     * If necessary, flush the current gc map into the big gcmap.
	     */
	    if (!hasShortMap) {
		/* Note that a 2-word field could cause nextMapBit == 33 */
		if (nextMapBit >= 32) {
		    nextMapBit %= 32;
		    gcMap.bigmap->map[currBigMapIndex] = map;
		    currBigMapIndex++;
		    map = gcMap.bigmap->map[currBigMapIndex];
		}
	    } else {
		CVMassert(nextMapBit + CVM_GCMAP_NUM_FLAGS <= 32);
	    }
	}
    }

    CVMassert(fieldOffset * sizeof(CVMObject*) == CVMcbInstanceSize(cb));

    /*
     * Store the calculated gc map into gcMap and set gc map flags.
     */
    if (hasShortMap) {
	gcMap.map =
	    (map << CVM_GCMAP_NUM_FLAGS) | CVM_GCMAP_SHORTGCMAP_FLAG;
	CVMassert(gcMap.map == CVMcbGcMap(cb).map);
    } else {
	int i;
	gcMap.bigmap->map[currBigMapIndex] = map;
	gcMap.map |= CVM_GCMAP_LONGGCMAP_FLAG;
	CVMassert((CVMcbGcMap(cb).map & CVM_GCMAP_LONGGCMAP_FLAG) != 0);
	CVMassert(CVMcbGcMap(cb).bigmap->maplen == gcMap.bigmap->maplen);
	for (i = 0; i < gcMap.bigmap->maplen; i++) {
	    CVMassert(CVMcbGcMap(cb).bigmap->map[i] = gcMap.bigmap->map[i]);
	}   
	free((char*)(gcMap.map & ~CVM_GCMAP_LONGGCMAP_FLAG));
    }
}

#endif

/*
 * Iterate over all preloaded classes,
 * and call 'callback' on each class.
 */
void
CVMpreloaderIterateAllClasses(CVMExecEnv* ee, 
			      CVMClassCallbackFunc callback,
			      void* data)
{
    int i;
    for (i = 0; i < CVM_nROMClasses; ++i) {
	callback(ee, CVM_ROMClasses[i].classBlockPointer, data);
    }

}

/*
 * Free allocated data structures associated with preloaded classes.
 * At the moment, this consists of stackmaps and any java method descriptors
 * re-allocated by the stackmap disambiguator.
 */
void
CVMpreloaderDestroy()
{
    int i;

    for (i = 0; i < CVM_nROMClasses; ++i) {
	CVMClassBlock * cb = CVM_ROMClasses[i].classBlockPointer;
	CVMtraceClassLoading(("CL: Destroying methods and maps for preloaded class %C (cb=0x%x)\n", cb, cb));
	CVMclassFreeJavaMethods( cb, CVM_TRUE );
    }
}
