/*
 * @(#)CVMInterfaceMethodTable.java	1.5 00/11/07
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

package runtime;
import vm.*;
import components.*;
import Const;
import java.util.Hashtable;

/*
 * This is the CVM-specific subclass of vm.InterfaceMethodTable
 * It knowns how to write out these data structures for the CVM
 * This is VERY similar to the CInterfaceMethodTable used for
 * JDK1.1-based systems. The only substantial difference is
 * the ordering requirement: rather than having two representations
 * of interfaces, CVM has only one. Those this class implements
 * directly come first, followed by those inhereted. Unfortunately,
 * this reduces, slightly, our capacity to share: for a class that
 * implements no interfaces directly, the totality of interfaces is
 * the same as that of the parent. But if the parent implemented some
 * directly, it will have a non-zero implementsCount, so we cannot
 * use its CVMInterfaces structure. Too bad.
 */

class CVMInterfaceMethodTable extends vm.InterfaceMethodTable {

    // names we generate
    public static final String vectorName = "CVMinterfaceVector";

    // private instance parts, parculiar to the CVM.
    // because the implementsCount is part of the CVMInterfaces
    // struct, children can not necessarily share with parents,
    // if these numbers are not the same. When implementsCount != 0,
    // we keep an almost-clone of this interaceMethodTable in the
    // defaultChild field.
    //
    int				implementsCount;
    CVMInterfaceMethodTable	defaultChild;

    boolean			generated;

    private static vm.InterfaceMethodFactory imf = 
        new vm.InterfaceMethodFactory() { 
	    public  InterfaceMethodTable 
	    newInterfaceMethodTable(ClassClass c, String n, 
				    InterfaceVector v[] ){
	        return new CVMInterfaceMethodTable( c, n, v );
	    }
        };
  
    public
    CVMInterfaceMethodTable( ClassClass c, String n, InterfaceVector v[] ){
	super( c, n, v );
	implementsCount = (c.ci.interfaces==null)?0:c.ci.interfaces.length;
	generated = false;
    }

    CVMInterfaceMethodTable getDefaultChild( ClassClass childClass ){
	if ( defaultChild == null ){
	    // create a default child. Attribute it to childClass
	    defaultChild = new CVMInterfaceMethodTable( childClass,
		childClass.ci.getGenericNativeName()+"_infMethodTable", iv );
	}
	childClass.inf = defaultChild;
	return defaultChild;
    }

    static void writeInterfaceTables(ClassClass[] classes, CCodeWriter out) {

        int n = classes.length;
	// generate the tables
	for ( int i = 0; i < n; i++ ){
	    ClassClass c = classes[i];
	    c.inf = vm.InterfaceMethodTable.generateInterfaceTable(c, imf );
	}

	//
	// print the vectors first.
	//
	out.println( "STATIC const CVMUint16 " + vectorName + "[] = {");
	int offset = 0;
	boolean emptyVector = true;
	for (int i = 0; i < n; i++) { 
	    ClassClass c = classes[i];
	    CVMInterfaceMethodTable simt = (CVMInterfaceMethodTable)c.inf;
	    if (simt == null || simt.parent != c) 
	        continue;
	    InterfaceVector iv[] = simt.iv;
	    for (int j = 0; j < iv.length; j++) { 
	        InterfaceVector vec = iv[j];
		if ( vec.generated )
		    continue; // done.
		vec.generated = true;
		vec.offset = offset;
		short vector[] = vec.v;
		if ( (vector==null) || (vector.length == 0) ){
		    continue;
		}
		int num = vector.length;
		emptyVector = false;
		out.comment(simt.parent.ci.className + 
			        " " + vec.intf.className );
		for (int k = 0, mod = 0; k < num; k++) { 
		    if (mod == 0) out.write('\t');
		    out.printHexInt( vector[k] );
		    out.write(',');
		    if (++mod == 8) { 
		        out.write('\n'); mod = 0;
		    }
		}
		if ((num % 8) != 0) 
		  out.write('\n');
		offset += num;
	    }
	}
	if ( emptyVector ){
	    out.write('0');// make it non-empty!
	}
	out.println("};\n");
	// end vectors.
	
	//
	// now write out the CVMInterfaces structures, which
	// roughly correspond to the InterfaceMethodTables.
	// But not quite.
	//
	for (int i = 0; i < n; i++) { 
	    ClassClass c = classes[i];
	    CVMInterfaceMethodTable simt = (CVMInterfaceMethodTable)c.inf;
	    if ( simt != null ){
		if ( simt.parent != c ){
		    // there seems to be sharing.
		    // this is probably not going to work under the CVM.
		    if ( simt.implementsCount != 0 )
			simt = simt.getDefaultChild(c);
		}
	        simt.write(out);
	    }
	}
    }

    private static boolean
    needVector( InterfaceVector vec ){
	if ( vec.v == null || vec.v.length == 0 ) return false; // don't bother.
	return true;
    }

    private static BitSet	interfaceSizes = new BitSet();
    private static final String		interfaceStructName = "CVMinterface";
	
    //struct CVMInterfaces {
    //    CVMUint16      interfaceCount;
    //    CVMUint16      implementsCount;
    //    union {
    //        CVMUint16* interfaces;
    //        struct {
    //            CVMClassBlock* interfacecb;
    //            CVMUint16*     index;
    //        } interfaceTable[1];
    //    } intfInfo;
    //}; 

    private static String declareInterfaceStruct( int n, CCodeWriter out ){
	String interfaceFlavor = "struct "+interfaceStructName+n;
	if ( ! interfaceSizes.get( n ) ){
	    out.println(interfaceFlavor+" { CVMUint16 interfaceCount; CVMUint16 implementsCount;");
	    if ( n > 0 ){
		out.println("    struct { const CVMClassBlock* interfacecb; const CVMUint16* index; } interfaceTable["+n+"];");
	    }
	    out.println("};");
	    interfaceSizes.set( n );
	}
	return interfaceFlavor;
    }

    public void
    write( CCodeWriter out ){
        if (this.generated) 
	    return;
	this.generated = true;
	String structType = declareInterfaceStruct( iv.length, out );
	out.print("STATIC const "+structType+" "+name);
	out.print(" = {\n\t");
	out.print( implementsCount );
	out.print(", ");
	out.print( iv.length );
	if ( iv.length != 0 ){
	    out.println(", {");
	    for( int i = 0; i < iv.length; i++ ){
		InterfaceVector entry = iv[i];
		ClassInfo intf = entry.intf;
		String intfName = intf.getGenericNativeName();
		out.print("\t{ ");
		out.print("(CVMClassBlock*)&"+intfName);
		out.print("_Classblock, ");
		if ( needVector( entry ) ){
		    out.print("&"+vectorName+"["+entry.offset+"]");
		} else {
		    out.write('0');
		}
		out.println(" },");
	    }
	    out.write('}');
	}
	out.println("};");
    }
}
